@@ -1,6 +1,8 @@
 # important core definitions
 
-using Core.Intrinsics
+using Core: Intrinsics, arraylen, arrayref, arrayset, arraysize,
+            tuplelen, tupleref, convert_default, kwcall,
+            typeassert, apply_type
 
 import Core.Array  # to add methods
 
@@ -8,8 +10,38 @@ const NonTupleType = Union(DataType,UnionType,TypeConstructor)
 
 typealias Callable Union(Function,DataType)
 
-convert(T, x)               = convert_default(T, x, convert)
-convert(T::Tuple, x::Tuple) = convert_tuple(T, x, convert)
+convert(T, x) = convert_default(T, x, convert)
+
+convert(::(), ::()) = ()
+convert(::Type{Tuple}, x::Tuple) = x
+
+argtail(x, rest...) = rest
+tupletail(x::Tuple) = argtail(x...)
+
+convert(T::(Any, Any...), x::(Any, Any...)) =
+    tuple(convert(T[1],x[1]), convert(tupletail(T), tupletail(x))...)
+
+convert{T}(::Type{(T...)}, x::Tuple) = cnvt_all(T, x...)
+cnvt_all(T) = ()
+cnvt_all(T, x, rest...) = tuple(convert(T,x), cnvt_all(T, rest...)...)
+
+
+ptr_arg_convert{T}(::Type{Ptr{T}}, x) = convert(T, x)
+ptr_arg_convert(::Type{Ptr{Void}}, x) = x
+
+# conversion used by ccall
+cconvert(T, x) = convert(T, x)
+# the following 3 definitions implement a 0.3 deprecation
+cconvert{T}(::Type{Ptr{T}}, x::Array{T}) = convert(Ptr{T}, x)
+cconvert(::Type{Ptr{None}}, x::Array) = convert(Ptr{None}, x)
+function cconvert{T}(::Type{Ptr{T}}, x::Array)
+    depwarn("ccall Ptr argument types must now match exactly, or be Ptr{Void}.", :cconvert)
+    convert(Ptr{T}, pointer(x))
+end
+# use the code in ccall.cpp to safely allocate temporary pointer arrays
+cconvert{T}(::Type{Ptr{Ptr{T}}}, a::Array) = a
+# convert strings to ByteString to pass as pointers
+cconvert{P<:Union(Int8,Uint8)}(::Type{Ptr{P}}, s::String) = bytestring(s)
 
 abstract IO
 
@@ -60,6 +92,10 @@ end
 
 type EOFError <: Exception end
 
+type DimensionMismatch <: Exception
+    name::ASCIIString
+end
+
 type WeakRef
     value
     WeakRef() = WeakRef(nothing)
@@ -79,12 +115,11 @@ type Colon
 end
 const (:) = Colon()
 
-hash(w::WeakRef) = hash(w.value)
-isequal(w::WeakRef, v::WeakRef) = isequal(w.value, v.value)
-isequal(w::WeakRef, v) = isequal(w.value, v)
-isequal(w, v::WeakRef) = isequal(w, v.value)
+==(w::WeakRef, v::WeakRef) = isequal(w.value, v.value)
+==(w::WeakRef, v) = isequal(w.value, v)
+==(w, v::WeakRef) = isequal(w, v.value)
 
-function finalizer(o::ANY, f::Function)
+function finalizer(o::ANY, f::Union(Function,Ptr))
     if isimmutable(o)
         error("objects of type ", typeof(o), " cannot be finalized")
     end
@@ -120,6 +155,39 @@ function append_any(xs...)
     out
 end
 
+# used by { } syntax
+function cell_1d(xs::ANY...)
+    n = length(xs)
+    a = Array(Any,n)
+    for i=1:n
+        arrayset(a,xs[i],i)
+    end
+    a
+end
+
+function cell_2d(nr, nc, xs::ANY...)
+    a = Array(Any,nr,nc)
+    for i=1:(nr*nc)
+        arrayset(a,xs[i],i)
+    end
+    a
+end
+
+# simple Array{Any} operations needed for bootstrap
+setindex!(A::Array{Any}, x::ANY, i::Real) = arrayset(A, x, to_index(i))
+
+function length_checked_equal(args...)
+    n = length(args[1])
+    for i=2:length(args)
+        if length(args[i]) != n
+            error("argument dimensions must match")
+        end
+    end
+    n
+end
+
+map(f::Callable, a::Array{Any,1}) = { f(a[i]) for i=1:length(a) }
+
 macro thunk(ex); :(()->$(esc(ex))); end
 macro L_str(s); s; end
 
@@ -129,18 +197,27 @@ function precompile(f, args::Tuple)
     end
 end
 
+esc(e::ANY) = Expr(:escape, e)
+
 macro boundscheck(yesno,blk)
-    quote
-        $(Expr(:boundscheck,yesno))
-        $(esc(blk))
-        $(Expr(:boundscheck,:pop))
-    end
+    # hack: use this syntax since it avoids introducing line numbers
+    :($(Expr(:boundscheck,yesno));
+      $(esc(blk));
+      $(Expr(:boundscheck,:pop)))
 end
 
 macro inbounds(blk)
     :(@boundscheck false $(esc(blk)))
 end
 
+macro label(name::Symbol)
+    Expr(:symboliclabel, name)
+end
+
+macro goto(name::Symbol)
+    Expr(:symbolicgoto, name)
+end
+
 # NOTE: Base shares Array with Core so we can add definitions to it
 
 Array{T,N}(::Type{T}, d::NTuple{N,Int}) =
