@@ -23,9 +23,10 @@ typemax(::Type{Bool}) = true
 (|)(x::Bool, y::Bool) = box(Bool,or_int(unbox(Bool,x),unbox(Bool,y)))
 ($)(x::Bool, y::Bool) = (x!=y)
 
-signbit(x::Bool) = 0
+signbit(x::Bool) = false
 sign(x::Bool) = x
 abs(x::Bool) = x
+abs2(x::Bool) = x
 
 <(x::Bool, y::Bool) = y&!x
 <=(x::Bool, y::Bool) = y|!x
@@ -35,11 +36,26 @@ abs(x::Bool) = x
 +(x::Bool) =  int(x)
 -(x::Bool) = -int(x)
 
-+(x::Bool, y::Bool) = int(x)+int(y)
--(x::Bool, y::Bool) = int(x)-int(y)
-*(x::Bool, y::Bool) = x&y
-^(x::Bool, y::Bool) = x|!y
-^(x::Integer, y::Bool) = y ? x : one(x)
++(x::Bool, y::Bool) = int(x) + int(y)
+-(x::Bool, y::Bool) = int(x) - int(y)
+*(x::Bool, y::Bool) = x & y
+^(x::Bool, y::Bool) = x | !y
+^(x::Integer, y::Bool) = ifelse(y, x, one(x))
+
+function +{T<:FloatingPoint}(x::Bool, y::T)
+    ifelse(x, one(promote_type(Bool,T)) + convert(promote_type(Bool,T),y),
+           convert(promote_type(Bool,T),y))
+end
++(y::FloatingPoint, x::Bool) = x + y
+
+function *{T<:Number}(x::Bool, y::T)
+    ifelse(x, convert(promote_type(Bool,T),y),
+           ifelse(signbit(y), -zero(promote_type(Bool,T)), zero(promote_type(Bool,T))))
+end
+function *{T<:Unsigned}(x::Bool, y::T)
+    ifelse(x, convert(promote_type(Bool,T),y), zero(promote_type(Bool,T)))
+end
+*(y::Number, x::Bool) = x * y
 
 div(x::Bool, y::Bool) = y ? x : throw(DivideError())
 fld(x::Bool, y::Bool) = div(x,y)
