@@ -1,10 +1,7 @@
 char(x) = convert(Char, x)
 char(x::FloatingPoint) = char(iround(x))
 
-is_valid_char(c) = !('\ud800' <= c <= '\udfff' || '\U10ffff' < c)
-
 integer(x::Char) = int(x)
-unsigned(x::Char) = uint(x)
 
 ## char promotions ##
 
@@ -22,9 +19,8 @@ promote_rule(::Type{Char}, ::Type{Uint128}) = Uint128
 ## character operations & comparisons ##
 
 # numeric operations
+# TODO: this should be removed, but needs to be here as long as Char <: Integer
 +(x::Char   , y::Char   ) = int(x)+int(y)
-*(x::Char   , y::Char   ) = int(x)*int(y)
-div(x::Char , y::Char   ) = div(int(x),int(y))
 
 # ordinal operations
 +(x::Char   , y::Integer) = char(int(x)+int(y))
@@ -54,4 +50,4 @@ sizeof(::Type{Char}) = 4
 ## printing & showing characters ##
 
 print(io::IO, c::Char) = (write(io,c); nothing)
-show(io::IO, c::Char) = (print(io,'\''); print_escaped(io,CharString(c),"'"); print(io,'\''))
+show(io::IO, c::Char) = (print(io,'\''); print_escaped(io,utf32(c),"'"); print(io,'\''))
