@@ -1,6 +1,8 @@
 ## client.jl - frontend handling command line options, environment setup,
 ##             and REPL
 
+const ARGS = UTF8String[]
+
 const text_colors = {
     :black   => "\033[1m\033[30m",
     :red     => "\033[1m\033[31m",
@@ -31,21 +33,20 @@ function input_color()
     return get(text_colors, c, default_color_input)
 end
 
-banner() = print(have_color ? banner_color : banner_plain)
-
 exit(n) = ccall(:jl_exit, Void, (Int32,), n)
 exit() = exit(0)
 quit() = exit()
 
-function repl_callback(ast::ANY, show_value)
-    global _repl_enough_stdin = true
-    stop_reading(STDIN) 
-    STDIN.readcb = false
-    put(repl_channel, (ast, show_value))
-end
-
 function repl_cmd(cmd)
     shell = shell_split(get(ENV,"JULIA_SHELL",get(ENV,"SHELL","/bin/sh")))
+    # Note that we can't support the fish shell due to its lack of subshells
+    #   See this for details: https://github.com/JuliaLang/julia/issues/4918
+    if Base.basename(shell[1]) == "fish"
+        warn_once("cannot use the fish shell, defaulting to /bin/sh\
+         set the JULIA_SHELL environment variable to silence this warning")
+        shell = "/bin/sh"
+    end
+
     if isempty(cmd.exec)
         error("no cmd to execute")
     elseif cmd.exec[1] == "cd"
@@ -59,7 +60,7 @@ function repl_cmd(cmd)
         end
         println(pwd())
     else
-        run(@windows? cmd : `$shell -i -c "($(shell_escape(cmd))) && true"`)
+        run(ignorestatus(@windows? cmd : (isa(STDIN, TTY) ? `$shell -i -c "($(shell_escape(cmd))) && true"` : `$shell -c "($(shell_escape(cmd))) && true"`)))
     end
     nothing
 end
@@ -69,22 +70,6 @@ function repl_hook(input::String)
          macroexpand(Expr(:macrocall,symbol("@cmd"),input)))
 end
 
-function repl_methods(input::String)
-    tokens = split(input, ".")
-    fn = Main
-    for token in tokens
-        sym = symbol(token)
-        isdefined(fn, sym) || return
-        fn = fn.(sym)
-    end
-    isgeneric(fn) || return
-
-    buf = IOBuffer()
-    show_method_table(buf, methods(fn), -1, false)
-    println(buf)
-    takebuf_string(buf)
-end
-
 display_error(er) = display_error(er, {})
 function display_error(er, bt)
     with_output_color(:red, STDERR) do io
@@ -136,60 +121,13 @@ function eval_user_input(ast::ANY, show_value)
     isa(STDIN,TTY) && println()
 end
 
-function read_buffer(stream::AsyncStream, nread)
-    global _repl_enough_stdin::Bool
-    while !_repl_enough_stdin && nb_available(stream.buffer) > 0
-        nread = int(search(stream.buffer,'\n')) # never more than one line or readline explodes :O
-        nread2 = int(search(stream.buffer,'\r'))
-        if nread == 0
-            if nread2 == 0
-                nread = nb_available(stream.buffer)
-            else
-                nread = nread2
-            end
-        else
-            if nread2 != 0 && nread2 < nread
-                nread = nread2
-            end
-        end
-        ptr = pointer(stream.buffer.data,stream.buffer.ptr)
-        skip(stream.buffer,nread)
-        #println(STDERR,stream.buffer.data[stream.buffer.ptr-nread:stream.buffer.ptr-1])
-        ccall(:jl_read_buffer,Void,(Ptr{Void},Cssize_t),ptr,nread)
-    end
-    return false
+function repl_callback(ast::ANY, show_value)
+    global _repl_enough_stdin = true
+    stop_reading(STDIN)
+    put!(repl_channel, (ast, show_value))
 end
 
-function run_repl()
-    global const repl_channel = RemoteRef()
-
-    ccall(:jl_init_repl, Void, (Cint,), _use_history)
-
-    # install Ctrl-C interrupt handler (InterruptException)
-    ccall(:jl_install_sigint_handler, Void, ())
-    STDIN.closecb = (x...)->put(repl_channel,(nothing,-1))
-
-    while true
-        if have_color
-            prompt_string = "\01\033[1m\033[32m\02julia> \01\033[0m"*input_color()*"\02"
-        else
-            prompt_string = "julia> "
-        end
-        ccall(:repl_callback_enable, Void, (Ptr{Uint8},), prompt_string)
-        global _repl_enough_stdin = false
-        start_reading(STDIN, read_buffer)
-        (ast, show_value) = take(repl_channel)
-        if show_value == -1
-            # exit flag
-            break
-        end
-        eval_user_input(ast, show_value!=0)
-    end
-
-    if have_color
-        print(color_normal)
-    end
-end
+_repl_start = Condition()
 
 function parse_input_line(s::String)
     # s = bytestring(s)
@@ -209,22 +147,56 @@ function parse_input_line(io::IO)
     while !eof(io)
         s = s*readline(io)
         e = parse_input_line(s)
-        if !(isa(e,Expr) && e.head === :continue)
+        if !(isa(e,Expr) && e.head === :incomplete)
             return e
         end
     end
 end
 
+# detect the reason which caused an :incomplete expression
+# from the error message
+# NOTE: the error messages are defined in src/julia-parser.scm
+incomplete_tag(ex) = :none
+function incomplete_tag(ex::Expr)
+    Meta.isexpr(ex, :incomplete) || return :none
+    msg = ex.args[1]
+    contains(msg, "string") && return :string
+    contains(msg, "comment") && return :comment
+    contains(msg, "requires end") && return :block
+    contains(msg, "\"`\"") && return :cmd
+    contains(msg, "character") && return :char
+    return :other
+end
+
 # try to include() a file, ignoring if not found
 try_include(path::String) = isfile(path) && include(path)
 
-function process_options(args::Array{Any,1})
-    global ARGS, bind_addr
+function init_bind_addr(args::Vector{UTF8String})
+    # Treat --bind-to in a position independent manner in ARGS since
+    # --worker, -n and --machinefile options are affected by it
+    btoidx = findfirst(args, "--bind-to")
+    if btoidx > 0
+        bind_addr = parseip(args[btoidx+1])
+    else
+        try
+            bind_addr = getipaddr()
+        catch
+            # All networking is unavailable, initialize bind_addr to the loopback address
+            # Will cause an exception to be raised only when used. 
+            bind_addr = ip"127.0.0.1"
+        end
+    end
+    global LPROC
+    LPROC.bind_addr = bind_addr
+end
+
+
+function process_options(args::Vector{UTF8String})
     quiet = false
     repl = true
     startup = true
     color_set = false
-    history = true
+    no_history_file = false
     i = 1
     while i <= length(args)
         if args[i]=="-q" || args[i]=="--quiet"
@@ -233,18 +205,17 @@ function process_options(args::Array{Any,1})
             start_worker()
             # doesn't return
         elseif args[i]=="--bind-to"
-            i += 1
-            bind_addr = args[i]
+            i+=1 # has already been processed
         elseif args[i]=="-e" || args[i]=="--eval"
             repl = false
             i+=1
-            ARGS = args[i+1:end]
+            splice!(ARGS, 1:length(ARGS), args[i+1:end])
             eval(Main,parse_input_line(args[i]))
             break
         elseif args[i]=="-E" || args[i]=="--print"
             repl = false
             i+=1
-            ARGS = args[i+1:end]
+            splice!(ARGS, 1:length(ARGS), args[i+1:end])
             show(eval(Main,parse_input_line(args[i])))
             println()
             break
@@ -271,19 +242,25 @@ function process_options(args::Array{Any,1})
             println("julia version ", VERSION)
             exit(0)
         elseif args[i]=="--no-history"
-            history = false
+            # deprecated in v0.3
+            warn("'--no-history' is deprecated; use '--no-history-file'")
+            no_history_file = true
+        elseif args[i] == "--no-history-file"
+            no_history_file = true
         elseif args[i] == "-f" || args[i] == "--no-startup"
             startup = false
         elseif args[i] == "-F"
             # load juliarc now before processing any more options
             load_juliarc()
             startup = false
+        elseif args[i] == "-i"
+            global is_interactive = true
         elseif beginswith(args[i], "--color")
             if args[i] == "--color"
                 color_set = true
                 global have_color = true
             elseif args[i][8] == '='
-                val = args[i][9:]
+                val = args[i][9:end]
                 if in(val, ("no","0","false"))
                     color_set = true
                     global have_color = false
@@ -296,10 +273,15 @@ function process_options(args::Array{Any,1})
                 error("invalid option: ", args[i])
             end
         elseif args[i][1]!='-'
+            if startup
+                load_juliarc()
+                startup = false
+            end
             # program
             repl = false
             # remove julia's arguments
-            ARGS = args[i+1:end]
+            splice!(ARGS, 1:length(ARGS), args[i+1:end])
+            ccall(:jl_exit_on_sigint, Void, (Cint,), 1)
             include(args[i])
             break
         else
@@ -307,7 +289,7 @@ function process_options(args::Array{Any,1})
         end
         i += 1
     end
-    return (quiet,repl,startup,color_set,history)
+    return (quiet,repl,startup,color_set,no_history_file)
 end
 
 const roottask = current_task()
@@ -315,94 +297,96 @@ const roottask = current_task()
 is_interactive = false
 isinteractive() = (is_interactive::Bool)
 
+const LOAD_PATH = ByteString[]
 function init_load_path()
     vers = "v$(VERSION.major).$(VERSION.minor)"
-    global const LOAD_PATH = ByteString[
-        abspath(JULIA_HOME,"..","local","share","julia","site",vers),
-        abspath(JULIA_HOME,"..","share","julia","site",vers)
-    ]
-end
-
-function init_sched()
-    global const Workqueue = Any[]
+    if haskey(ENV,"JULIA_LOAD_PATH")
+        prepend!(LOAD_PATH, split(ENV["JULIA_LOAD_PATH"], @windows? ';' : ':'))    
+    end
+    push!(LOAD_PATH,abspath(JULIA_HOME,"..","local","share","julia","site",vers))
+    push!(LOAD_PATH,abspath(JULIA_HOME,"..","share","julia","site",vers))
 end
 
 function init_head_sched()
     # start in "head node" mode
-    global const Scheduler = Task(()->event_loop(true), 1024*1024)
     global PGRP
     global LPROC
     LPROC.id = 1
     assert(length(PGRP.workers) == 0)
     register_worker(LPROC)
-    # make scheduler aware of current (root) task
-    unshift!(Workqueue, roottask)
-    yield()
-end
-
-function init_profiler()
-    # Use a max size of 1M profile samples, and fire timer evey 1ms
-    Profile.init(1_000_000, 0.001)
 end
 
 function load_juliarc()
-    # If the user built us with a specifi Base.SYSCONFDIR, check that location first for a juliarc.jl file
+    # If the user built us with a specific Base.SYSCONFDIR, check that location first for a juliarc.jl file
     #   If it is not found, then continue on to the relative path based on JULIA_HOME
-    if !isempty(Base.SYSCONFDIR) && isfile(joinpath(Base.SYSCONFDIR,"julia","juliarc.jl"))
-        include(abspath(Base.SYSCONFDIR,"julia","juliarc.jl"))
+    if !isempty(Base.SYSCONFDIR) && isfile(joinpath(JULIA_HOME,Base.SYSCONFDIR,"julia","juliarc.jl"))
+        include(abspath(JULIA_HOME,Base.SYSCONFDIR,"julia","juliarc.jl"))
     else
         try_include(abspath(JULIA_HOME,"..","etc","julia","juliarc.jl"))
     end
     try_include(abspath(homedir(),".juliarc.jl"))
 end
 
+function early_init()
+    Sys.init_sysinfo()
+    if CPU_CORES > 8 && !("OPENBLAS_NUM_THREADS" in keys(ENV)) && !("OMP_NUM_THREADS" in keys(ENV))
+        # Prevent openblas from stating to many threads, unless/until specifically requested
+        ENV["OPENBLAS_NUM_THREADS"] = 8
+    end
+end
+
+function init_parallel()
+    start_gc_msgs_task()
+    atexit(terminate_all_workers)
+end
+
+import .Terminals
+import .REPL
 
 function _start()
-    # set up standard streams
-    reinit_stdio()
-    fdwatcher_reinit()
-    # Initialize RNG
-    Random.librandom_init()
-    # Check that BLAS is correctly built
-    check_blas()
-    LinAlg.init()
-    Sys.init()
-    GMP.gmp_init()
-    global const CPU_CORES = Sys.CPU_CORES
-    init_profiler()
+    early_init()
 
-    @windows_only ENV["PATH"] = JULIA_HOME*";"*joinpath(JULIA_HOME,"..","Git","bin")*";"*ENV["PATH"]*
-        ";C:\\Program Files\\Git\\bin;C:\\Program Files (x86)\\Git\\bin"*
-        ";C:\\MinGW\\bin;C:\\MinGW\\msys\\1.0\\bin"*
-        ";C:\\Python27;C:\\Python26;C:\\Python25"
-    @windows_only haskey(ENV,"JULIA_EDITOR") || (ENV["JULIA_EDITOR"] = "start")
-    @windows_only begin
-        user_data_dir = abspath(ENV["AppData"],"julia")
-        isdir(user_data_dir) || mkdir(user_data_dir)
-    end
-
-    #atexit(()->flush(STDOUT))
     try
-        init_sched()
+        init_parallel()
+        init_bind_addr(ARGS)
         any(a->(a=="--worker"), ARGS) || init_head_sched()
         init_load_path()
-        (quiet,repl,startup,color_set,history) = process_options(ARGS)
-        global _use_history = history
-        repl && startup && load_juliarc()
+        (quiet,repl,startup,color_set,no_history_file) = process_options(copy(ARGS))
+
+        local term
+        global active_repl
+        if repl
+            if !isa(STDIN,TTY)
+                global is_interactive |= !isa(STDIN,Union(File,IOStream))
+                color_set || (global have_color = false)
+            else
+                term = Terminals.TTYTerminal(get(ENV,"TERM",@windows? "" : "dumb"),STDIN,STDOUT,STDERR)
+                global is_interactive = true
+                color_set || (global have_color = Terminals.hascolor(term))
+                quiet || REPL.banner(term,term)
+                if term.term_type == "dumb"
+                    active_repl = REPL.BasicREPL(term)
+                else
+                    active_repl = REPL.LineEditREPL(term, true)
+                    active_repl.no_history_file = no_history_file
+                    active_repl.hascolor = have_color
+                end
+                # Make sure any displays pushed in .juliarc.jl ends up above the
+                # REPLDisplay
+                pushdisplay(REPL.REPLDisplay(active_repl))
+            end
+        end
+
+        startup && load_juliarc()
 
         if repl
             if !isa(STDIN,TTY)
-                if !color_set
-                    global have_color = false
-                end
                 # note: currently IOStream is used for file STDIN
                 if isa(STDIN,File) || isa(STDIN,IOStream)
                     # reading from a file, behave like include
-                    global is_interactive = false
                     eval(parse_input_line(readall(STDIN)))
                 else
                     # otherwise behave repl-like
-                    global is_interactive = true
                     while !eof(STDIN)
                         eval_user_input(parse_input_line(STDIN), true)
                     end
@@ -412,21 +396,7 @@ function _start()
                 end
                 quit()
             end
-
-            if !color_set
-                @unix_only global have_color = (beginswith(get(ENV,"TERM",""),"xterm") || success(`tput setaf 0`))
-                @windows_only global have_color = true
-            end
-
-            global is_interactive = true
-            quiet || banner()
-
-            if haskey(ENV,"JL_ANSWER_COLOR")
-                warn("JL_ANSWER_COLOR is deprecated, use JULIA_ANSWER_COLOR instead.")
-                ENV["JULIA_ANSWER_COLOR"] = ENV["JL_ANSWER_COLOR"]
-            end
-
-            run_repl()
+            REPL.run_repl(active_repl)
         end
     catch err
         display_error(err,catch_backtrace())
