@@ -5,37 +5,45 @@ end
 Complex(x::Real, y::Real) = Complex(promote(x,y)...)
 Complex(x::Real) = Complex(x, zero(x))
 
+const im = Complex(false,true)
+
 typealias Complex128 Complex{Float64}
 typealias Complex64  Complex{Float32}
 typealias Complex32  Complex{Float16}
 
-sizeof(::Type{Complex128}) = 16
-sizeof(::Type{Complex64}) = 8
-sizeof(::Type{Complex32}) = 4
+sizeof{T<:Real}(::Type{Complex{T}}) = 2*sizeof(T)
 
-real(z::Complex) = z.re
-imag(z::Complex) = z.im
-real(x::Real) = x
-imag(x::Real) = zero(x)
+convert{T<:Real}(::Type{Complex{T}}, x::Real) = Complex{T}(x,0)
+convert{T<:Real}(::Type{Complex{T}}, z::Complex) = Complex{T}(real(z),imag(z))
+convert{T<:Real}(::Type{T}, z::Complex) =
+    isreal(z) ? convert(T,real(z)) : throw(InexactError())
 
-convert{T<:Real}(::Type{Complex{T}}, x::Real) =
-    Complex{T}(convert(T,x), convert(T,0))
-convert{T<:Real}(::Type{Complex{T}}, z::Complex{T}) = z
-convert{T<:Real}(::Type{Complex{T}}, z::Complex) =
-    Complex{T}(convert(T,real(z)),convert(T,imag(z)))
+convert(::Type{Complex}, z::Complex) = z
+convert(::Type{Complex}, x::Real) = Complex(x)
 
-convert{T<:Real}(::Type{T}, z::Complex) = (imag(z)==0 ? convert(T,real(z)) :
-                                           throw(InexactError()))
-
-promote_rule{T<:Real}(::Type{Complex{T}}, ::Type{T}) = Complex{T}
-promote_rule{T<:Real}(::Type{Complex}, ::Type{T}) = Complex
 promote_rule{T<:Real,S<:Real}(::Type{Complex{T}}, ::Type{S}) =
     Complex{promote_type(T,S)}
 promote_rule{T<:Real,S<:Real}(::Type{Complex{T}}, ::Type{Complex{S}}) =
     Complex{promote_type(T,S)}
 
-complex(x, y) = Complex(x, y)
-complex(x) = Complex(x)
+widen{T}(::Type{Complex{T}}) = Complex{widen(T)}
+
+real(z::Complex) = z.re
+imag(z::Complex) = z.im
+real(x::Real) = x
+imag(x::Real) = zero(x)
+reim(z) = (real(z), imag(z))
+
+isreal(x::Real) = true
+isreal(z::Complex) = imag(z) == 0
+isimag(z::Number) = real(z) == 0
+isinteger(z::Complex) = isreal(z) & isinteger(real(z))
+isfinite(z::Complex) = isfinite(real(z)) & isfinite(imag(z))
+isnan(z::Complex) = isnan(real(z)) | isnan(imag(z))
+isinf(z::Complex) = isinf(real(z)) | isinf(imag(z))
+
+complex(x::Real, y::Real) = Complex(x, y)
+complex(x::Real) = Complex(x)
 complex(z::Complex) = z
 
 complex128(r::Float64, i::Float64) = Complex{Float64}(r, i)
@@ -49,35 +57,26 @@ complex32(r::Real, i::Real) = complex32(float16(r),float16(i))
 complex32(z) = complex32(real(z), imag(z))
 
 for fn in _numeric_conversion_func_names
-    @eval $fn(z::Complex) = complex($fn(real(z)),$fn(imag(z)))
+    @eval $fn(z::Complex) = Complex($fn(real(z)),$fn(imag(z)))
 end
 
-isreal{T<:Real}(z::Complex{T}) = imag(z) == 0
-isinteger(z::Complex) = isreal(z) && isinteger(real(z))
-
-isfinite(z::Complex) = isfinite(real(z)) && isfinite(imag(z))
-reim(z) = (real(z), imag(z))
-
 function complex_show(io::IO, z::Complex, compact::Bool)
     r, i = reim(z)
-    if isnan(r) || isfinite(i)
-        compact ? showcompact(io,r) : show(io,r)
-        if signbit(i)==1 && !isnan(i)
-            i = -i
-            print(io, compact ? "-" : " - ")
-        else
-            print(io, compact ? "+" : " + ")
-        end
-        compact ? showcompact(io, i) : show(io, i)
-        if !(isa(i,Integer) || isa(i,Rational) ||
-             isa(i,FloatingPoint) && isfinite(i))
-            print(io, "*")
-        end
-        print(io, "im")
+    compact ? showcompact(io,r) : show(io,r)
+    if signbit(i) && !isnan(i)
+        i = -i
+        print(io, compact ? "-" : " - ")
     else
-        print(io, "complex(",r,",",i,")")
+        print(io, compact ? "+" : " + ")
     end
+    compact ? showcompact(io, i) : show(io, i)
+    if !(isa(i,Integer) && !isa(i,Bool) || isa(i,FloatingPoint) && isfinite(i))
+        print(io, "*")
+    end
+    print(io, "im")
 end
+complex_show(io::IO, z::Complex{Bool}, compact::Bool) =
+    print(io, z == im ? "im" : "Complex($(z.re),$(z.im))")
 show(io::IO, z::Complex) = complex_show(io, z, false)
 showcompact(io::IO, z::Complex) = complex_show(io, z, true)
 
@@ -92,66 +91,40 @@ function write(s::IO, z::Complex)
 end
 
 
-## singleton type for imaginary unit constant ##
-
-type ImaginaryUnit <: Number end
-const im = ImaginaryUnit()
-
-convert{T<:Real}(::Type{Complex{T}}, ::ImaginaryUnit) = Complex{T}(zero(T),one(T))
-convert(::Type{Complex}, ::ImaginaryUnit) = Complex(real(im),imag(im))
-
-real(::ImaginaryUnit) = int(0)
-imag(::ImaginaryUnit) = int(1)
-
-promote_rule{T<:Complex}(::Type{ImaginaryUnit}, ::Type{T}) = T
-promote_rule{T<:Real}(::Type{ImaginaryUnit}, ::Type{T}) = Complex{T}
-
-show(io::IO, ::ImaginaryUnit) = print(io, "im")
-
-
 ## generic functions of complex numbers ##
 
-convert(::Type{Complex}, z::Complex) = z
-convert(::Type{Complex}, x::Real) = complex(x)
-
 ==(z::Complex, w::Complex) = (real(z) == real(w)) & (imag(z) == imag(w))
 ==(z::Complex, x::Real) = isreal(z) && real(z) == x
 ==(x::Real, z::Complex) = isreal(z) && real(z) == x
 
 isequal(z::Complex, w::Complex) = isequal(real(z),real(w)) & isequal(imag(z),imag(w))
 
-hash(z::Complex) = bitmix(hash(real(z)),hash(imag(z)))
-
-conj(z::Complex) = complex(real(z),-imag(z))
+conj(z::Complex) = Complex(real(z),-imag(z))
 abs(z::Complex)  = hypot(real(z), imag(z))
 abs2(z::Complex) = real(z)*real(z) + imag(z)*imag(z)
 inv(z::Complex)  = conj(z)/abs2(z)
+inv{T<:Integer}(z::Complex{T}) = inv(float(z))
 sign(z::Complex) = z/abs(z)
 
-(-)(::ImaginaryUnit) = complex(0, -1)
--(z::Complex) = complex(-real(z), -imag(z))
-+(z::Complex, w::Complex) = complex(real(z) + real(w), imag(z) + imag(w))
--(z::Complex, w::Complex) = complex(real(z) - real(w), imag(z) - imag(w))
-*(z::Complex, w::Complex) = complex(real(z) * real(w) - imag(z) * imag(w),
+-(z::Complex) = Complex(-real(z), -imag(z))
++(z::Complex, w::Complex) = Complex(real(z) + real(w), imag(z) + imag(w))
+-(z::Complex, w::Complex) = Complex(real(z) - real(w), imag(z) - imag(w))
+*(z::Complex, w::Complex) = Complex(real(z) * real(w) - imag(z) * imag(w),
                                     real(z) * imag(w) + imag(z) * real(w))
 
 # adding or multiplying real & complex is common
-*(x::Real, z::Complex) = complex(x * real(z), x * imag(z))
-*(z::Complex, x::Real) = complex(x * real(z), x * imag(z))
-+(x::Real, z::Complex) = complex(x + real(z), imag(z))
-+(z::Complex, x::Real) = complex(x + real(z), imag(z))
--(x::Real, z::Complex) = complex(x - real(z), -imag(z))
--(z::Complex, x::Real) = complex(real(z) - x, imag(z))
-
-# multiplying by im is common
-*(z::ImaginaryUnit, w::ImaginaryUnit) = complex(-imag(z), real(z))
-*(z::ImaginaryUnit, x::Real)    = complex(zero(x), x)
-*(x::Real, z::ImaginaryUnit)    = complex(zero(x), x)
-*(z::ImaginaryUnit, w::Complex) = complex(-imag(w), real(w))
-*(w::Complex, z::ImaginaryUnit) = complex(-imag(w), real(w))
+*(x::Bool, z::Complex) = ifelse(x, z, zero(z))
+*(z::Complex, x::Bool) = ifelse(x, z, zero(z))
+*(x::Real, z::Complex) = Complex(x * real(z), x * imag(z))
+*(z::Complex, x::Real) = Complex(x * real(z), x * imag(z))
++(x::Real, z::Complex) = Complex(x + real(z), imag(z))
++(z::Complex, x::Real) = Complex(x + real(z), imag(z))
+-(x::Real, z::Complex) = Complex(x - real(z), -imag(z))
+-(z::Complex, x::Real) = Complex(real(z) - x, imag(z))
 
 /(z::Number, w::Complex) = z*inv(w)
-/(z::Complex, x::Real) = complex(real(z)/x, imag(z)/x)
+/(a::Real  , w::Complex) = a*inv(w)
+/(z::Complex, x::Real) = Complex(real(z)/x, imag(z)/x)
 
 function /(a::Complex, b::Complex)
     are = real(a); aim = imag(a); bre = real(b); bim = imag(b)
@@ -162,7 +135,7 @@ function /(a::Complex, b::Complex)
             r = bre / bim
         end
         den = bim + r*bre
-        complex((are*r + aim)/den, (aim*r - are)/den)
+        Complex((are*r + aim)/den, (aim*r - are)/den)
     else
         if isinf(bre) && isinf(bim)
             r = sign(bim)/sign(bre)
@@ -174,30 +147,78 @@ function /(a::Complex, b::Complex)
     end
 end
 
-/(a::Real, b::Complex) =
-    real(convert(promote_type(typeof(a),typeof(b)),a)) /
-    convert(promote_type(typeof(a),typeof(b)),b)
-function /{T<:Real}(a::T, b::Complex{T})
-    bre = real(b); bim = imag(b)
-    if abs(bre) <= abs(bim)
-        if isinf(bre) && isinf(bim)
-            r = sign(bre)/sign(bim)
-        else
-            r = bre / bim
-        end
-        den = bim + r*bre
-        complex(a*r/den, -a/den)
+inv{T<:Union(Float16,Float32)}(z::Complex{T}) =
+    oftype(z, conj(complex128(z))/abs2(complex128(z)))
+
+# robust complex division for double precision
+# the first step is to scale variables if appropriate ,then do calculations
+# in a way that avoids over/underflow (subfuncs 1 and 2), then undo the scaling.
+# scaling variable s and other techniques
+# based on arxiv.1210.4539 
+#             a + i*b
+#  p + i*q = ---------
+#             c + i*d
+function /(z::Complex128, w::Complex128)
+    a, b = reim(z); c, d = reim(w)
+    half = 0.5
+    two = 2.0
+    ab = max(abs(a), abs(b))
+    cd = max(abs(c), abs(d))
+    ov = realmax(a)
+    un = realmin(a)
+    ϵ = eps(Float64)
+    bs = two/(ϵ*ϵ)
+    s = 1.0
+    ab >= half*ov  && (a=half*a; b=half*b; s=two*s ) # scale down a,b
+    cd >= half*ov  && (c=half*c; d=half*d; s=s*half) # scale down c,d
+    ab <= un*two/ϵ && (a=a*bs; b=b*bs; s=s/bs      ) # scale up a,b
+    cd <= un*two/ϵ && (c=c*bs; d=d*bs; s=s*bs      ) # scale up c,d
+    abs(d)<=abs(c) ? ((p,q)=robust_cdiv1(a,b,c,d)  ) : ((p,q)=robust_cdiv1(b,a,d,c); q=-q)
+    return Complex128(p*s,q*s) # undo scaling
+end
+function robust_cdiv1(a::Float64, b::Float64, c::Float64, d::Float64)
+    r = d/c
+    t = 1.0/(c+d*r)
+    p = robust_cdiv2(a,b,c,d,r,t)
+    q = robust_cdiv2(b,-a,c,d,r,t)
+    return p,q
+end
+function robust_cdiv2(a::Float64, b::Float64, c::Float64, d::Float64, r::Float64, t::Float64)
+    if r != 0
+        br = b*r
+        return (br != 0 ? (a+br)*t : a*t + (b*t)*r)
     else
-        if isinf(bre) && isinf(bim)
-            r = sign(bim)/sign(bre)
-        else
-            r = bim / bre
-        end
-        den = bre + r*bim
-        complex(a/den, -a*r/den)
+        return (a + d*(b/c)) * t
     end
 end
 
+function inv(w::Complex128)
+    c, d = reim(w)
+    half = 0.5
+    two = 2.0
+    cd = max(abs(c), abs(d))
+    ov = realmax(c)
+    un = realmin(c)
+    ϵ = eps(Float64)
+    bs = two/(ϵ*ϵ)
+    s = 1.0
+    cd >= half*ov  && (c=half*c; d=half*d; s=s*half) # scale down c,d
+    cd <= un*two/ϵ && (c=c*bs; d=d*bs; s=s*bs      ) # scale up c,d
+    if abs(d)<=abs(c)
+        r = d/c
+        t = 1.0/(c+d*r)
+        p = t
+        q = -r * t
+    else
+        c, d = d, c
+        r = d/c
+        t = 1.0/(c+d*r)
+        p = r * t
+        q = -t
+    end
+    return Complex128(p*s,q*s) # undo scaling
+end
+
 function ssqs{T<:FloatingPoint}(x::T, y::T)
     k::Int = 0
     ρ = x*x + y*y
@@ -215,14 +236,14 @@ end
 function sqrt{T<:FloatingPoint}(z::Complex{T})
     x, y = reim(z)
     if x==y==0
-        return complex(zero(x),y)
+        return Complex(zero(x),y)
     end
     ρ, k::Int = ssqs(x, y)
     if isfinite(x) ρ=ldexp(abs(x),-k)+sqrt(ρ) end
     if isodd(k)
-        k = (k-1)/2
+        k = div(k-1,2)
     else
-        k = k/2-1
+        k = div(k,2)-1
         ρ += ρ
     end
     ρ = ldexp(sqrt(ρ),k) #sqrt((abs(z)+abs(x))/2) without over/underflow
@@ -235,7 +256,7 @@ function sqrt{T<:FloatingPoint}(z::Complex{T})
             η = copysign(ρ,y)
         end
     end
-    complex(ξ,η)
+    Complex(ξ,η)
 end
 sqrt(z::Complex) = sqrt(float(z))
 
@@ -244,27 +265,28 @@ sqrt(z::Complex) = sqrt(float(z))
 #     iz = float(imag(z))
 #     r = sqrt((hypot(rz,iz)+abs(rz))/2)
 #     if r == 0
-#         return complex(zero(iz), iz)
+#         return Complex(zero(iz), iz)
 #     end
 #     if rz >= 0
-#         return complex(r, iz/r/2)
+#         return Complex(r, iz/r/2)
 #     end
-#     return complex(abs(iz)/r/2, copysign(r,iz))
+#     return Complex(abs(iz)/r/2, copysign(r,iz))
 # end
 
-cis(theta::Real) = complex(cos(theta),sin(theta))
+# compute exp(im*theta)
+cis(theta::Real) = Complex(cos(theta),sin(theta))
 function cis(z::Complex)
-    v = 1/exp(imag(z))
-    complex(v*cos(real(z)), v*sin(real(z)))
+    v = exp(-imag(z))
+    Complex(v*cos(real(z)), v*sin(real(z)))
 end
+@vectorize_1arg Number cis
 
 angle(z::Complex) = atan2(imag(z), real(z))
 
 function log{T<:FloatingPoint}(z::Complex{T})
-    const T0::T  = 0.7071067811865475
     const T1::T  = 1.25
     const T2::T  = 3
-    const ln2::T = 0.6931471805599453
+    const ln2::T = log(oftype(T,2))  #0.6931471805599453
     x, y = reim(z)
     ρ, k = ssqs(x,y)
     ax = abs(x)
@@ -274,12 +296,12 @@ function log{T<:FloatingPoint}(z::Complex{T})
     else
         θ, β = ay, ax
     end
-    if k==0 && T0 < β && (β <= T1 || ρ < T2)
+    if k==0 && (0.5 < β*β) && (β <= T1 || ρ < T2)
         ρρ = log1p((β-1)*(β+1)+θ*θ)/2
     else
         ρρ = log(ρ)/2 + k*ln2
     end
-    complex(ρρ, angle(z))
+    Complex(ρρ, angle(z))
 end
 log(z::Complex) = log(float(z))
 
@@ -299,23 +321,80 @@ log(z::Complex) = log(float(z))
 #             re = log(ar) + log1p(r*r)/2
 #         end
 #     end
-#     complex(re, angle(z))
+#     Complex(re, angle(z))
 # end
 
-log10(z::Complex) = log(z)/oftype(real(z),2.302585092994046)
-log2(z::Complex) = log(z)/oftype(real(z),0.6931471805599453)
+function log10(z::Complex)
+    a = log(z)
+    a/log(oftype(real(a),10))
+end
+function log2(z::Complex)
+    a = log(z)
+    a/log(oftype(real(a),2))
+end
 
 function exp(z::Complex)
     zr, zi = reim(z)
-    if isfinite(zr) && !isfinite(zi) return complex(oftype(zr, NaN), oftype(zi, NaN)) end
-    if zr==Inf && zi==0 return complex(zr, zi) end
-    if zr==-Inf && !isfinite(zi) return complex(-zero(zr), copysign(zero(zi), zi)) end
-    if zr==Inf && !isfinite(zi) return complex(-zr, oftype(zr, NaN)) end
-    if isnan(zr) return complex(zr, zi==0 ? zi : zr) end
-    er = zr==Inf ? zr : exp(zr)
-    wr = er*(isfinite(zi) ? cos(zi) : zi)
-    wi = er*(isfinite(zi) ? sin(zi) : zi)
-    complex(wr, wi)
+    if isnan(zr)
+        Complex(zr, zi==0 ? zi : zr)
+    elseif !isfinite(zi)
+        if zr == inf(zr)
+            Complex(-zr, nan(zr))
+        elseif zr == -inf(zr)
+            Complex(-zero(zr), copysign(zero(zi), zi))
+        else
+            Complex(nan(zr), nan(zi))
+        end
+    else
+        er = exp(zr)
+        if zi == zero(zi)
+            Complex(er, zi)
+        else
+            Complex(er*cos(zi), er*sin(zi))
+        end
+    end
+end
+
+function expm1(z::Complex)
+    zr,zi = reim(z)
+    if isnan(zr)
+        Complex(zr, zi==0 ? zi : zr)
+    elseif !isfinite(zi)
+        if zr == inf(zr)
+            Complex(-zr, nan(zr))
+        elseif zr == -inf(zr)
+            Complex(-one(zr), copysign(zero(zi), zi))
+        else
+            Complex(nan(zr), nan(zi))
+        end
+    else
+        erm1 = expm1(zr)        
+        if zi == 0
+            Complex(erm1, zi)
+        else            
+            er = erm1+one(erm1)
+            wr = isfinite(er) ? erm1 - 2.0*er*(sin(0.5*zi))^2 : er*cos(zi)
+            Complex(wr, er*sin(zi))
+        end
+    end
+end
+
+function log1p{T}(z::Complex{T})
+    zr,zi = reim(z)
+    if isfinite(zr)
+        isinf(zi) && return log(z)
+        # This is based on a well-known trick for log1p of real z,
+        # allegedly due to Kahan, only modified to handle real(u) <= 0
+        # differently to avoid inaccuracy near z==-2 and for correct branch cut
+        u = float(one(T)) + z
+        u == 1 ? convert(typeof(u), z) : real(u) <= 0 ? log(u) : log(u)*z/(u-1)
+    elseif isnan(zr)
+        Complex(zr, zr)
+    elseif isfinite(zi)
+        Complex(inf(T), copysign(zr > 0 ? zero(T) : convert(T, pi), zi))
+    else
+        Complex(inf(T), nan(T))
+    end
 end
 
 function ^{T<:FloatingPoint}(z::Complex{T}, p::Complex{T})
@@ -334,7 +413,7 @@ function ^{T<:FloatingPoint}(z::Complex{T}, p::Complex{T})
         elseif isnan(y) && isinf(x)
             y = copysign(zero(T), y)
         end
-        complex(x,y)
+        Complex(x,y)
     elseif z!=0
         if p!=0 && isinteger(p)
             rp = real(p)
@@ -349,20 +428,20 @@ function ^{T<:FloatingPoint}(z::Complex{T}, p::Complex{T})
         zero(z) #CHECK SIGNS
     else #0^0
         zer = copysign(zero(T),real(p))*copysign(zero(T),imag(z))
-        complex(one(T), zer)
+        Complex(one(T), zer)
     end
 end
 
 function exp2{T}(z::Complex{T})
     er = exp2(real(z))
     theta = imag(z) * log(convert(T, 2))
-    complex(er*cos(theta), er*sin(theta))
+    Complex(er*cos(theta), er*sin(theta))
 end
 
 function exp10{T}(z::Complex{T})
     er = exp10(real(z))
     theta = imag(z) * log(convert(T, 10))
-    complex(er*cos(theta), er*sin(theta))
+    Complex(er*cos(theta), er*sin(theta))
 end
 
 function ^{T<:Complex}(z::T, p::T)
@@ -422,11 +501,11 @@ function ^{T<:Complex}(z::T, p::T)
         end
     end
 
-    complex(re, im)
+    Complex(re, im)
 end
 
 ^(z::Complex, n::Bool) = n ? z : one(z)
-^(z::Complex, n::Integer) = z^complex(n)
+^(z::Complex, n::Integer) = z^Complex(n)
 
 ^{T<:FloatingPoint}(z::Complex{T}, n::Bool) = n ? z : one(z)  # to resolve ambiguity
 ^{T<:Integer}(z::Complex{T}, n::Bool) = n ? z : one(z)        # to resolve ambiguity
@@ -437,97 +516,97 @@ end
 
 function sin(z::Complex)
     zr, zi = reim(z)
-    if !isfinite(zi) && zr == 0 return complex(zr, zi) end
-    if isnan(zr) && !isfinite(zi) return complex(zr, zi) end
-    if !isfinite(zr) && zi == 0 return complex(oftype(zr, NaN), zi) end
-    if !isfinite(zr) && isfinite(zi) return complex(oftype(zr, NaN), oftype(zi, NaN)) end
-    if !isfinite(zr) && !isfinite(zi) return complex(zr, oftype(zi, NaN)) end
-    complex(sin(zr)*cosh(zi), cos(zr)*sinh(zi))
+    if !isfinite(zi) && zr == 0 return Complex(zr, zi) end
+    if isnan(zr) && !isfinite(zi) return Complex(zr, zi) end
+    if !isfinite(zr) && zi == 0 return Complex(oftype(zr, NaN), zi) end
+    if !isfinite(zr) && isfinite(zi) return Complex(oftype(zr, NaN), oftype(zi, NaN)) end
+    if !isfinite(zr) && !isfinite(zi) return Complex(zr, oftype(zi, NaN)) end
+    Complex(sin(zr)*cosh(zi), cos(zr)*sinh(zi))
 end
 
 function cos(z::Complex)
     zr, zi = reim(z)
     if !isfinite(zi) && zr == 0
-        return complex(isnan(zi) ? zi : oftype(zi, Inf),
+        return Complex(isnan(zi) ? zi : oftype(zi, Inf),
                        isnan(zi) ? zr : zr*-sign(zi))
     end
     if !isfinite(zr) && isinf(zi)
-        return complex(oftype(zr, Inf), oftype(zi, NaN))
+        return Complex(oftype(zr, Inf), oftype(zi, NaN))
     end
     if isinf(zr)
-        return complex(oftype(zr, NaN), zi==0 ? -copysign(zi, zr) : oftype(zi, NaN))
+        return Complex(oftype(zr, NaN), zi==0 ? -copysign(zi, zr) : oftype(zi, NaN))
     end
-    if isnan(zr) && zi==0 return complex(zr, abs(zi)) end
-    complex(cos(zr)*cosh(zi), -sin(zr)*sinh(zi))
+    if isnan(zr) && zi==0 return Complex(zr, abs(zi)) end
+    Complex(cos(zr)*cosh(zi), -sin(zr)*sinh(zi))
 end
 
 function tan(z::Complex)
     zr, zi = reim(z)
-    w = tanh(complex(-zi, zr))
-    complex(imag(w), -real(w))
+    w = tanh(Complex(-zi, zr))
+    Complex(imag(w), -real(w))
 end
 
 function asin(z::Complex)
     zr, zi = reim(z)
     if isinf(zr) && isinf(zi)
-        return complex(copysign(pi/4, zr),zi)
+        return Complex(copysign(oftype(zr,pi)/4, zr),zi)
     elseif isnan(zi) && isinf(zr)
-        return complex(zi, oftype(zr, Inf))
+        return Complex(zi, oftype(zr, Inf))
     end
     ξ = zr == 0       ? zr :
-        !isfinite(zr) ? pi/2*sign(zr) :
+        !isfinite(zr) ? oftype(zr,pi)/2 * sign(zr) :
         atan2(zr, real(sqrt(1-z)*sqrt(1+z)))
     η = asinh(copysign(imag(sqrt(conj(1-z))*sqrt(1+z)), imag(z)))
-    complex(ξ,η)
+    Complex(ξ,η)
 end
 
 function acos{T<:FloatingPoint}(z::Complex{T})
     zr, zi = reim(z)
     if isnan(zr)
-        if isinf(zi) return complex(zr, -zi)
-        else         return complex(zr, zr) end
+        if isinf(zi) return Complex(zr, -zi)
+        else         return Complex(zr, zr) end
     elseif isnan(zi)
-        if isinf(zr) return complex(zi, abs(zr))
-        elseif zr==0 return complex(pi/2, zi)
-        else         return complex(zi, zi) end
+        if isinf(zr) return Complex(zi, abs(zr))
+        elseif zr==0 return Complex(oftype(zr,pi)/2, zi)
+        else         return Complex(zi, zi) end
     elseif zr==zi==0
-        return complex(pi/2, -zi)
+        return Complex(oftype(zr,pi)/2, -zi)
     elseif zr==Inf && zi===0.0
-        return complex(zi, -zr)
+        return Complex(zi, -zr)
     elseif zr==-Inf && zi===-0.0
-        return complex(oftype(zi, pi), -zr)
+        return Complex(oftype(zi,pi), -zr)
     end
     ξ = 2*atan2(real(sqrt(1-z)), real(sqrt(1+z)))
     η = asinh(imag(sqrt(conj(1+z))*sqrt(1-z)))
-    if isinf(zr) && isinf(zi) ξ -= pi/4 * sign(zr) end
-    complex(ξ,η)
+    if isinf(zr) && isinf(zi) ξ -= oftype(η,pi)/4 * sign(zr) end
+    Complex(ξ,η)
 end
 acos(z::Complex) = acos(float(z))
 
 function atan(z::Complex)
-    w = atanh(complex(-imag(z),real(z)))
-    complex(imag(w),-real(w))
+    w = atanh(Complex(-imag(z),real(z)))
+    Complex(imag(w),-real(w))
 end
 
 function sinh(z::Complex)
     zr, zi = reim(z)
-    if isinf(zr) && isinf(zi) return complex(zr, oftype(zi, NaN)) end
-    w = sin(complex(zi, zr))
-    complex(imag(w),real(w))
+    if isinf(zr) && isinf(zi) return Complex(zr, oftype(zi, NaN)) end
+    w = sin(Complex(zi, zr))
+    Complex(imag(w),real(w))
 end
 
 function cosh(z::Complex)
     zr, zi = reim(z)
-    if isnan(zr) && zi==0 return complex(zr, zi) end
-    cos(complex(-zi,zr))
+    if isnan(zr) && zi==0 return Complex(zr, zi) end
+    cos(Complex(-zi,zr))
 end
 
 function tanh{T<:FloatingPoint}(z::Complex{T})
     const Ω = prevfloat(typemax(T))
     ξ, η = reim(z)
-    if isnan(ξ) && η==0 return complex(ξ, η) end
+    if isnan(ξ) && η==0 return Complex(ξ, η) end
     if 4*abs(ξ) > asinh(Ω) #Overflow?
-        complex(copysign(one(T),ξ),
+        Complex(copysign(one(T),ξ),
                 copysign(zero(T),η*(isfinite(η) ? sin(2*abs(η)) : one(η))))
     else
         t = tan(η)
@@ -535,36 +614,36 @@ function tanh{T<:FloatingPoint}(z::Complex{T})
         s = sinh(ξ)
         ρ = sqrt(1 + s*s) #cosh(ξ)
         if isinf(t)
-            complex(ρ/s,1/t)
+            Complex(ρ/s,1/t)
         else
-            complex(β*ρ*s,t)/(1+β*s*s)
+            Complex(β*ρ*s,t)/(1+β*s*s)
         end
     end
 end
 tanh(z::Complex) = tanh(float(z))
 
 function asinh(z::Complex)
-    w = asin(complex(-imag(z),real(z)))
-    complex(imag(w),-real(w))
+    w = asin(Complex(-imag(z),real(z)))
+    Complex(imag(w),-real(w))
 end
 
 function acosh(z::Complex)
     zr, zi = reim(z)
     if isnan(zr) || isnan(zi)
         if isinf(zr) || isinf(zi)
-            return complex(oftype(zr, Inf), oftype(zi, NaN))
+            return Complex(oftype(zr, Inf), oftype(zi, NaN))
         else
-            return complex(oftype(zr, NaN), oftype(zi, NaN))
+            return Complex(oftype(zr, NaN), oftype(zi, NaN))
         end
     elseif zr==-Inf && zi===-0.0 #Edge case is wrong - WHY?
-        return complex(Inf, -pi)
+        return Complex(inf(zr), oftype(zi, -pi))
     end
     ξ = asinh(real(sqrt(conj(z-1))*sqrt(z+1)))
     η = 2atan2(imag(sqrt(z-1)),real(sqrt(z+1)))
     if isinf(zr) && isinf(zi)
-        η -= pi/4 * sign(zi) * sign(zr)
+        η -= oftype(η,pi)/4 * sign(zi) * sign(zr)
     end
-    complex(ξ, η)
+    Complex(ξ, η)
 end
 
 function atanh{T<:FloatingPoint}(z::Complex{T})
@@ -577,15 +656,15 @@ function atanh{T<:FloatingPoint}(z::Complex{T})
     if ax > θ || ay > θ #Prevent overflow
         if isnan(y)
             if isinf(x)
-                return complex(copysign(zero(x),x), y)
+                return Complex(copysign(zero(x),x), y)
             else
-                return complex(real(1/z), y)
+                return Complex(real(1/z), y)
             end
         end
         if isinf(y)
-            return complex(copysign(zero(x),x), copysign(pi/2, y))
+            return Complex(copysign(zero(x),x), copysign(oftype(y,pi)/2, y))
         end
-        return complex(real(1/z), copysign(pi/2, y))
+        return Complex(real(1/z), copysign(oftype(y,pi)/2, y))
     elseif ax==1
         if y == 0
             ξ = copysign(oftype(x,Inf),x)
@@ -593,7 +672,7 @@ function atanh{T<:FloatingPoint}(z::Complex{T})
         else
             ym = ay+ρ
             ξ = log(sqrt(sqrt(4+y*y))/sqrt(ym))
-            η = copysign(pi/2+atan(ym/2), y)/2
+            η = copysign(oftype(y,pi)/2 + atan(ym/2), y)/2
         end
     else #Normal case
         ysq = (ay+ρ)^2
@@ -602,8 +681,14 @@ function atanh{T<:FloatingPoint}(z::Complex{T})
         else
             ξ = log1p(4x/((1-x)^2 + ysq))/4
         end
-        η = angle(complex((1-x)*(1+x)-ysq, 2y))/2
+        η = angle(Complex((1-x)*(1+x)-ysq, 2y))/2
     end
-    complex(ξ, η)
+    Complex(ξ, η)
 end
 atanh(z::Complex) = atanh(float(z))
+
+function lexcmp(a::Complex, b::Complex)
+    c = cmp(real(a), real(b))
+    c == 0 || return c
+    cmp(imag(a), imag(b))
+end
