@@ -8,7 +8,12 @@ haskey(d::Associative, k) = in(k,keys(d))
 
 function in(p::(Any,Any), a::Associative)
     v = get(a,p[1],secret_table_token)
-    !is(v, secret_table_token) && isequal(v, p[2])
+    !is(v, secret_table_token) && (v == p[2])
+end
+
+function summary(t::Associative)
+    n = length(t)
+    string(typeof(t), " with ", n, (n==1 ? " entry" : " entries"))
 end
 
 function show{K,V}(io::IO, t::Associative{K,V})
@@ -22,7 +27,7 @@ function show{K,V}(io::IO, t::Associative{K,V})
         end
         print(io, delims[1])
         first = true
-        for (k, v) = t
+        for (k, v) in t
             first || print(io, ',')
             first = false
             show(io, k)
@@ -33,6 +38,73 @@ function show{K,V}(io::IO, t::Associative{K,V})
     end
 end
 
+function _truncate_at_width_or_chars(str, width, chars="", truncmark="…")
+    truncwidth = strwidth(truncmark)
+    (width <= 0 || width < truncwidth) && return ""
+
+    wid = truncidx = lastidx = 0
+    idx = start(str)
+    while !done(str, idx)
+        lastidx = idx
+        c, idx = next(str, idx)
+        wid += charwidth(c)
+        wid >= width - truncwidth && truncidx == 0 && (truncidx = lastidx)
+        (wid >= width || c in chars) && break
+    end
+
+    lastidx != 0 && str[lastidx] in chars && (lastidx = prevind(str, lastidx))
+    truncidx == 0 && (truncidx = lastidx)
+    if lastidx < sizeof(str)
+        return bytestring(SubString(str, 1, truncidx) * truncmark)
+    else
+        return bytestring(str)
+    end
+end
+
+showdict(t::Associative; kw...) = showdict(STDOUT, t; kw...)
+function showdict{K,V}(io::IO, t::Associative{K,V}; limit::Bool = false,
+                       sz=(s = tty_size(); (s[1]-3, s[2])))
+    rows, cols = sz
+    print(io, summary(t))
+    isempty(t) && return
+    print(io, ":")
+
+    if limit
+        rows < 2   && (print(io, " …"); return)
+        cols < 12  && (cols = 12) # Minimum widths of 2 for key, 4 for value
+        cols -= 6 # Subtract the widths of prefix "  " separator " => "
+        rows -= 2 # Subtract the summary and final ⋮ continuation lines
+
+        # determine max key width to align the output, caching the strings
+        ks = Array(String, min(rows, length(t)))
+        keylen = 0
+        for (i, k) in enumerate(keys(t))
+            i > rows && break
+            ks[i] = sprint(show, k)
+            keylen = clamp(length(ks[i]), keylen, div(cols, 3))
+        end
+    end
+
+    for (i, (k, v)) in enumerate(t)
+        print(io, "\n  ")
+        limit && i > rows && (print(io, rpad("⋮", keylen), " => ⋮"); break)
+
+        if limit
+            key = rpad(_truncate_at_width_or_chars(ks[i], keylen, "\r\n"), keylen)
+        else
+            key = sprint(show, k)
+        end
+        print(io, key)
+        print(io, " => ")
+
+        val = sprint(show, v)
+        if limit
+            val = _truncate_at_width_or_chars(val, cols - keylen, "\r\n")
+        end
+        print(io, val)
+    end
+end
+
 immutable KeyIterator{T<:Associative}
     dict::T
 end
@@ -40,6 +112,35 @@ immutable ValueIterator{T<:Associative}
     dict::T
 end
 
+summary{T<:Union(KeyIterator,ValueIterator)}(iter::T) =
+    string(T.name, " for a ", summary(iter.dict))
+
+show(io::IO, iter::Union(KeyIterator,ValueIterator)) = show(io, collect(iter))
+
+showkv(iter::Union(KeyIterator,ValueIterator); kw...) = showkv(STDOUT, iter; kw...)
+function showkv{T<:Union(KeyIterator,ValueIterator)}(io::IO, iter::T; limit::Bool = false,
+                                                     sz=(s = tty_size(); (s[1]-3, s[2])))
+    rows, cols = sz
+    print(io, summary(iter))
+    isempty(iter) && return
+    print(io, ". ", T<:KeyIterator ? "Keys" : "Values", ":")
+    if limit
+        rows < 2 && (print(io, " …"); return)
+        cols < 4 && (cols = 4)
+        cols -= 2 # For prefix "  "
+        rows -= 2 # For summary and final ⋮ continuation lines
+    end
+
+    for (i, v) in enumerate(iter)
+        print(io, "\n  ")
+        limit && i >= rows && (print(io, "⋮"); break)
+
+        str = sprint(show, v)
+        limit && (str = _truncate_at_width_or_chars(str, cols, "\r\n"))
+        print(io, str)
+    end
+end
+
 length(v::Union(KeyIterator,ValueIterator)) = length(v.dict)
 isempty(v::Union(KeyIterator,ValueIterator)) = isempty(v.dict)
 eltype(v::KeyIterator) = eltype(v.dict)[1]
@@ -94,14 +195,6 @@ filter(f::Function, d::Associative) = filter!(f,copy(d))
 
 eltype{K,V}(a::Associative{K,V}) = (K,V)
 
-function hash(d::Associative)
-    h::Uint = 0
-    for (k,v) in d
-        h $= bitmix(hash(k),~hash(v))
-    end
-    h
-end
-
 function isequal(l::Associative, r::Associative)
     if isa(l,ObjectIdDict) != isa(r,ObjectIdDict)
         return false
@@ -140,6 +233,11 @@ function getindex(t::Associative, key)
     return v
 end
 
+# t[k1,k2,ks...] is syntactic sugar for t[(k1,k2,ks...)].  (Note
+# that we need to avoid dispatch loops if setindex!(t,v,k) is not defined.)
+getindex(t::Associative, k1, k2, ks...) = getindex(t, tuple(k1,k2,ks...))
+setindex!(t::Associative, v, k1, k2, ks...) = setindex!(t, v, tuple(k1,k2,ks...))
+
 push!(t::Associative, key, v) = setindex!(t, v, key)
 
 # hashing objects by identity
@@ -147,6 +245,22 @@ push!(t::Associative, key, v) = setindex!(t, v, key)
 type ObjectIdDict <: Associative{Any,Any}
     ht::Array{Any,1}
     ObjectIdDict() = new(cell(32))
+
+    function ObjectIdDict(itr)
+        d = ObjectIdDict()
+        for (k,v) in itr
+            d[k] = v
+        end
+        d
+    end
+
+    function ObjectIdDict(o::ObjectIdDict)
+        N = length(o.ht)
+        ht = cell(N)
+        ccall(:memcpy, Ptr{Void}, (Ptr{Void}, Ptr{Void}, Uint),
+              ht, o.ht, N*sizeof(Ptr))
+        new(ht)
+    end
 end
 
 similar(d::ObjectIdDict) = ObjectIdDict()
@@ -174,11 +288,11 @@ end
 
 empty!(t::ObjectIdDict) = (t.ht = cell(length(t.ht)); t)
 
-start(t::ObjectIdDict) = 0
-done(t::ObjectIdDict, i) = is(next(t,i),())
-next(t::ObjectIdDict, i) = ccall(:jl_eqtable_next, Any, (Any, Uint32), t.ht, i)
+_oidd_nextind(a, i) = int(ccall(:jl_eqtable_nextind, Csize_t, (Any, Csize_t), a, i))
 
-isempty(t::ObjectIdDict) = is(next(t,0),())
+start(t::ObjectIdDict) = _oidd_nextind(t.ht, 0)
+done(t::ObjectIdDict, i) = (i == -1)
+next(t::ObjectIdDict, i) = ((t.ht[i+1],t.ht[i+2]), _oidd_nextind(t.ht, i+2))
 
 function length(d::ObjectIdDict)
     n = 0
@@ -188,87 +302,7 @@ function length(d::ObjectIdDict)
     n
 end
 
-# hashing
-
-bitmix(a::Union(Int32,Uint32), b::Union(Int32,Uint32)) =
-    ccall(:int64to32hash, Uint32, (Uint64,),
-          or_int(shl_int(zext_int(Uint64,unbox(Uint32,a)), 32), zext_int(Uint64,unbox(Uint32,b))))
-
-bitmix(a::Union(Int64,Uint64), b::Union(Int64, Uint64)) =
-    ccall(:int64hash, Uint64, (Uint64,),
-          xor_int(unbox(Uint64,a), or_int(lshr_int(unbox(Uint64,b), 32),
-                                           shl_int(unbox(Uint64,b), 32))))
-
-if WORD_SIZE == 64
-    hash64(x::Float64) =
-        ccall(:int64hash, Uint64, (Uint64,), box(Uint64,unbox(Float64,x)))
-    hash64(x::Union(Int64,Uint64)) =
-        ccall(:int64hash, Uint64, (Uint64,), x)
-else
-    hash64(x::Float64) =
-        ccall(:int64to32hash, Uint32, (Uint64,), box(Uint64,unbox(Float64,x)))
-    hash64(x::Union(Int64,Uint64)) =
-        ccall(:int64to32hash, Uint32, (Uint64,), x)
-end
-
-hash(x::Union(Bool,Char,Int8,Uint8,Int16,Uint16,Int32,Uint32,Int64,Uint64)) =
-    hash64(uint64(x))
-
-function hash(x::Integer)
-    h::Uint = hash(uint64(x&0xffffffffffffffff))
-    if typemin(Int64) <= x <= typemax(Uint64)
-        return h
-    end
-    x >>>= 64
-    while x != 0 && x != -1
-        h = bitmix(h, hash(uint64(x&0xffffffffffffffff)))
-        x >>>= 64
-    end
-    return h
-end
-
-hash(x::Float32) = hash(reinterpret(Uint32, isnan(x) ? NaN32 : x))
-hash(x::Float64) = hash(reinterpret(Uint64, isnan(x) ? NaN   : x))
-
-function hash(t::Tuple)
-    h::Uint = 0
-    for i=1:length(t)
-        h = bitmix(h,int(hash(t[i]))+42)
-    end
-    return h
-end
-
-function hash(a::AbstractArray)
-    h::Uint = hash(size(a))+1
-    for i=1:length(a)
-        h = bitmix(h,int(hash(a[i])))
-    end
-    return h
-end
-
-# make sure Array{Bool} and BitArray can be equivalent
-hash(a::AbstractArray{Bool}) = hash(bitpack(a))
-
-hash(x::ANY) = object_id(x)
-
-if WORD_SIZE == 64
-    hash(s::ByteString) =
-        ccall(:memhash, Uint64, (Ptr{Void}, Int), s.data, length(s.data))
-    hash(s::ByteString, seed::Union(Int,Uint)) =
-        ccall(:memhash_seed, Uint64, (Ptr{Void}, Int, Uint32),
-              s.data, length(s.data), uint32(seed))
-else
-    hash(s::ByteString) =
-        ccall(:memhash32, Uint32, (Ptr{Void}, Int), s.data, length(s.data))
-    hash(s::ByteString, seed::Union(Int,Uint)) =
-        ccall(:memhash32_seed, Uint32, (Ptr{Void}, Int, Uint32),
-              s.data, length(s.data), uint32(seed))
-end
-
-hash(s::String) = hash(bytestring(s))
-
-hash(x::Expr) = bitmix(hash(x.head),hash(x.args)+43)
-
+copy(o::ObjectIdDict) = ObjectIdDict(o)
 
 # dict
 
@@ -285,24 +319,50 @@ type Dict{K,V} <: Associative{K,V}
         new(zeros(Uint8,n), Array(K,n), Array(V,n), 0, 0, identity)
     end
     function Dict(ks, vs)
-        n = length(ks)
+        # TODO: eventually replace with a call to Dict(zip(ks,vs))
+        n = min(length(ks), length(vs))
         h = Dict{K,V}()
         for i=1:n
             h[ks[i]] = vs[i]
         end
         return h
     end
+    function Dict(kv)
+        h = Dict{K,V}()
+        for (k,v) in kv
+            h[k] = v
+        end
+        return h
+    end
 end
 Dict() = Dict{Any,Any}()
 
 Dict{K,V}(ks::AbstractArray{K}, vs::AbstractArray{V}) = Dict{K,V}(ks,vs)
 Dict(ks, vs) = Dict{Any,Any}(ks, vs)
 
+# conversion between Dict types
+function convert{K,V}(::Type{Dict{K,V}},d::Dict)
+    h = Dict{K,V}()
+    for (k,v) in d
+        ck = convert(K,k)
+        if !haskey(h,ck)
+            h[ck] = convert(V,v)
+        else
+            error("key collision during dictionary conversion")
+        end
+    end
+    return h
+end
+convert{K,V}(::Type{Dict{K,V}},d::Dict{K,V}) = d
+
 # syntax entry points
 Dict{K,V}(ks::(K...), vs::(V...)) = Dict{K  ,V  }(ks, vs)
 Dict{K  }(ks::(K...), vs::Tuple ) = Dict{K  ,Any}(ks, vs)
 Dict{V  }(ks::Tuple , vs::(V...)) = Dict{Any,V  }(ks, vs)
 
+Dict{K,V}(kv::AbstractArray{(K,V)}) = Dict{K,V}(kv)
+Dict{K,V}(kv::Associative{K,V}) = Dict{K,V}(kv)
+
 similar{K,V}(d::Dict{K,V}) = (K=>V)[]
 
 function serialize(s, t::Dict)
@@ -397,84 +457,21 @@ end
 function empty!{K,V}(h::Dict{K,V})
     fill!(h.slots, 0x0)
     sz = length(h.slots)
-    h.keys = Array(K, sz)
-    h.vals = Array(V, sz)
+    empty!(h.keys)
+    empty!(h.vals)
+    resize!(h.keys, sz)
+    resize!(h.vals, sz)
     h.ndel = 0
     h.count = 0
     return h
 end
 
-function setindex!{K,V}(h::Dict{K,V}, v0, key0)
-    key = convert(K,key0)
-    if !isequal(key,key0)
-        error(key0, " is not a valid key for type ", K)
-    end
-    v   = convert(V,  v0)
-
-    sz = length(h.keys)
-
-    if h.ndel >= ((3*sz)>>2) || h.count*3 > sz*2
-        # > 3/4 deleted or > 2/3 full
-        rehash(h, h.count > 64000 ? h.count*2 : h.count*4)
-        sz = length(h.keys)  # rehash may resize the table at this point!
-    end
-
-    iter = 0
-    maxprobe = max(16, sz>>6)
-    index = hashindex(key, sz)
-    orig = index
-    avail = -1  # an available slot
-    keys = h.keys; vals = h.vals
-
-    while true
-        if isslotempty(h,index)
-            if avail > 0; index = avail; end
-            h.slots[index] = 0x1
-            h.keys[index] = key
-            h.vals[index] = v
-            h.count += 1
-            return h
-        end
-
-        if isslotmissing(h,index)
-            if avail<0
-                # found an available slot, but need to keep scanning
-                # in case "key" already exists in a later collided slot.
-                avail = index
-            end
-        elseif isequal(key, keys[index])
-            vals[index] = v
-            return h
-        end
-
-        index = (index & (sz-1)) + 1
-        iter+=1
-        if iter > maxprobe || index==orig
-            break
-        end
-    end
-
-    if avail>0
-        index = avail
-        h.slots[index] = 0x1
-        h.keys[index] = key
-        h.vals[index] = v
-        h.count += 1
-        return h
-    end
-
-    rehash(h, h.count > 64000 ? sz*2 : sz*4)
-
-    setindex!(h, v, key)
-end
-
 # get the index where a key is stored, or -1 if not present
 function ht_keyindex{K,V}(h::Dict{K,V}, key)
     sz = length(h.keys)
     iter = 0
     maxprobe = max(16, sz>>6)
     index = hashindex(key, sz)
-    orig = index
     keys = h.keys
 
     while true
@@ -487,39 +484,160 @@ function ht_keyindex{K,V}(h::Dict{K,V}, key)
 
         index = (index & (sz-1)) + 1
         iter+=1
-        if iter > maxprobe || index==orig
-            break
-        end
+        iter > maxprobe && break
     end
 
     return -1
 end
 
+# get the index where a key is stored, or -pos if not present
+# and the key would be inserted at pos
+# This version is for use by setindex! and get!
+function ht_keyindex2{K,V}(h::Dict{K,V}, key)
+    sz = length(h.keys)
+    iter = 0
+    maxprobe = max(16, sz>>6)
+    index = hashindex(key, sz)
+    avail = 0
+    keys = h.keys
+
+    while true
+        if isslotempty(h,index)
+            avail < 0 && return avail
+            return -index
+        end
+
+        if isslotmissing(h,index)
+            if avail == 0
+                # found an available slot, but need to keep scanning
+                # in case "key" already exists in a later collided slot.
+                avail = -index
+            end
+        elseif isequal(key, keys[index])
+            return index
+        end
+
+        index = (index & (sz-1)) + 1
+        iter+=1
+        iter > maxprobe && break
+    end
+
+    avail < 0 && return avail
+
+    rehash(h, h.count > 64000 ? sz*2 : sz*4)
+
+    return ht_keyindex2(h, key)
+end
+
+function _setindex!(h::Dict, v, key, index)
+    h.slots[index] = 0x1
+    h.keys[index] = key
+    h.vals[index] = v
+    h.count += 1
+
+    sz = length(h.keys)
+    # Rehash now if necessary
+    if h.ndel >= ((3*sz)>>2) || h.count*3 > sz*2
+        # > 3/4 deleted or > 2/3 full
+        rehash(h, h.count > 64000 ? h.count*2 : h.count*4)
+    end
+end
+
+function setindex!{K,V}(h::Dict{K,V}, v0, key0)
+    key = convert(K,key0)
+    if !isequal(key,key0)
+        error(key0, " is not a valid key for type ", K)
+    end
+    v = convert(V,  v0)
+
+    index = ht_keyindex2(h, key)
+
+    if index > 0
+        h.keys[index] = key
+        h.vals[index] = v
+    else
+        _setindex!(h, v, key, -index)
+    end
+
+    return h
+end
+
+function get!{K,V}(h::Dict{K,V}, key0, default)
+    key = convert(K,key0)
+    if !isequal(key,key0)
+        error(key0, " is not a valid key for type ", K)
+    end
+
+    index = ht_keyindex2(h, key)
+
+    index > 0 && return h.vals[index]
+
+    v = convert(V,  default)
+    _setindex!(h, v, key, -index)
+    return v
+end
+
+function get!{K,V}(default::Callable, h::Dict{K,V}, key0)
+    key = convert(K,key0)
+    if !isequal(key,key0)
+        error(key0, " is not a valid key for type ", K)
+    end
+
+    index = ht_keyindex2(h, key)
+
+    index > 0 && return h.vals[index]
+
+    v = convert(V,  default())
+    _setindex!(h, v, key, -index)
+    return v
+end
+
+# NOTE: this macro is specific to Dict, not Associative, and should
+#       therefore not be exported as-is: it's for internal use only.
+macro get!(h, key0, default)
+    quote
+        K, V = eltype($(esc(h)))
+        key = convert(K, $(esc(key0)))
+        isequal(key, $(esc(key0))) || error($(esc(key0)), " is not a valid key for type ", K)
+        idx = ht_keyindex2($(esc(h)), key)
+        if idx < 0
+            idx = -idx
+            v = convert(V, $(esc(default)))
+            _setindex!($(esc(h)), v, key, idx)
+        else
+            @inbounds v = $(esc(h)).vals[idx]
+        end
+        v
+    end
+end
+
+
 function getindex{K,V}(h::Dict{K,V}, key)
     index = ht_keyindex(h, key)
     return (index<0) ? throw(KeyError(key)) : h.vals[index]::V
 end
 
-function get{K,V}(h::Dict{K,V}, key, deflt)
+function get{K,V}(h::Dict{K,V}, key, default)
     index = ht_keyindex(h, key)
-    return (index<0) ? deflt : h.vals[index]::V
+    return (index<0) ? default : h.vals[index]::V
+end
+
+function get{K,V}(default::Callable, h::Dict{K,V}, key)
+    index = ht_keyindex(h, key)
+    return (index<0) ? default() : h.vals[index]::V
 end
 
 haskey(h::Dict, key) = (ht_keyindex(h, key) >= 0)
 in{T<:Dict}(key, v::KeyIterator{T}) = (ht_keyindex(v.dict, key) >= 0)
 
-function getkey{K,V}(h::Dict{K,V}, key, deflt)
+function getkey{K,V}(h::Dict{K,V}, key, default)
     index = ht_keyindex(h, key)
-    return (index<0) ? deflt : h.keys[index]::K
+    return (index<0) ? default : h.keys[index]::K
 end
 
 function _pop!(h::Dict, index)
     val = h.vals[index]
-    h.slots[index] = 0x2
-    ccall(:jl_arrayunset, Void, (Any, Uint), h.keys, index-1)
-    ccall(:jl_arrayunset, Void, (Any, Uint), h.vals, index-1)
-    h.ndel += 1
-    h.count -= 1
+    _delete!(h, index)
     return val
 end
 
@@ -534,7 +652,6 @@ function pop!(h::Dict, key, default)
 end
 
 function _delete!(h::Dict, index)
-    val = h.vals[index]
     h.slots[index] = 0x2
     ccall(:jl_arrayunset, Void, (Any, Uint), h.keys, index-1)
     ccall(:jl_arrayunset, Void, (Any, Uint), h.vals, index-1)
@@ -612,17 +729,20 @@ WeakKeyDict() = WeakKeyDict{Any,Any}()
 
 setindex!{K}(wkh::WeakKeyDict{K}, v, key) = add_weak_key(wkh.ht, convert(K,key), v)
 
-function getkey{K}(wkh::WeakKeyDict{K}, kk, deflt)
+function getkey{K}(wkh::WeakKeyDict{K}, kk, default)
     k = getkey(wkh.ht, kk, secret_table_token)
     if is(k, secret_table_token)
-        return deflt
+        return default
     end
     return k.value::K
 end
 
-get{K}(wkh::WeakKeyDict{K}, key, def) = get(wkh.ht, key, def)
+get{K}(wkh::WeakKeyDict{K}, key, default) = get(wkh.ht, key, default)
+get{K}(default::Callable, wkh::WeakKeyDict{K}, key) = get(default, wkh.ht, key)
+get!{K}(wkh::WeakKeyDict{K}, key, default) = get!(wkh.ht, key, default)
+get!{K}(default::Callable, wkh::WeakKeyDict{K}, key) = get!(default, wkh.ht, key)
 pop!{K}(wkh::WeakKeyDict{K}, key) = pop!(wkh.ht, key)
-pop!{K}(wkh::WeakKeyDict{K}, key, def) = pop!(wkh.ht, key, def)
+pop!{K}(wkh::WeakKeyDict{K}, key, default) = pop!(wkh.ht, key, default)
 delete!{K}(wkh::WeakKeyDict{K}, key) = delete!(wkh.ht, key)
 empty!(wkh::WeakKeyDict)  = (empty!(wkh.ht); wkh)
 haskey{K}(wkh::WeakKeyDict{K}, key) = haskey(wkh.ht, key)
