@@ -2,68 +2,90 @@ module DSP
 
 importall Base.FFTW
 import Base.FFTW.normalization
+import Base.trailingsize
 
-export FFTW, filt, deconv, conv, conv2, xcorr, fftshift, ifftshift,
+export FFTW, filt, filt!, deconv, conv, conv2, xcorr, fftshift, ifftshift,
        dct, idct, dct!, idct!, plan_dct, plan_idct, plan_dct!, plan_idct!,
        # the rest are defined imported from FFTW:
        fft, bfft, ifft, rfft, brfft, irfft,
        plan_fft, plan_bfft, plan_ifft, plan_rfft, plan_brfft, plan_irfft,
        fft!, bfft!, ifft!, plan_fft!, plan_bfft!, plan_ifft!
 
-function filt{T<:Number}(b::Union(AbstractVector{T}, T),a::Union(AbstractVector{T}, T),x::AbstractVector{T})
-    if isempty(b); error("filt: b is empty"); end
-    if isempty(a); error("filt: a is empty"); end
-    if a[1]==0; error("filt: a[1] must be nonzero"); end
+_zerosi(b,a,T) = zeros(promote_type(eltype(b), eltype(a), T), max(length(a), length(b))-1)
+
+function filt{T,S}(b::Union(AbstractVector, Number), a::Union(AbstractVector, Number),
+                   x::AbstractArray{T}, si::AbstractArray{S}=_zerosi(b,a,T))
+    filt!(Array(promote_type(eltype(b), eltype(a), T, S), size(x)), b, a, x, si)
+end
+
+# in-place filtering: returns results in the out argument, which may shadow x
+# (and does so by default)
+function filt!{T,S,N}(out::AbstractArray, b::Union(AbstractVector, Number), a::Union(AbstractVector, Number),
+                      x::AbstractArray{T}, si::AbstractArray{S,N}=_zerosi(b,a,T))
+    isempty(b) && error("b must be non-empty")
+    isempty(a) && error("a must be non-empty")
+    a[1] == 0  && error("a[1] must be nonzero")
+    size(x) != size(out) && error("out size must match x")
 
     as = length(a)
     bs = length(b)
     sz = max(as, bs)
+    silen = sz - 1
+    ncols = trailingsize(x,2)
 
-    if sz == 1
-        return (b[1]/a[1]).*x
-    end
+    size(si, 1) != silen && error("si must have max(length(a),length(b))-1 rows")
+    N > 1 && trailingsize(si,2) != ncols && error("si must either be a vector or have the same number of columns as x")
 
-    if bs<sz
-        newb = zeros(T,sz)
-        newb[1:bs] = b
-        b = newb
-    end
-
-    xs = size(x,1)
-    y = Array(T, xs)
-    silen = sz-1
-    si = zeros(T, silen)
+    size(x,1) == 0 && return out
+    sz == 1 && return scale!(out, x, b[1]/a[1]) # Simple scaling without memory
 
+    # Filter coefficient normalization
     if a[1] != 1
         norml = a[1]
         a ./= norml
         b ./= norml
     end
+    # Pad the coefficients with zeros if needed
+    bs<sz   && (b = copy!(zeros(eltype(b), sz), b))
+    1<as<sz && (a = copy!(zeros(eltype(a), sz), a))
 
-    if as > 1
-        if as<sz
-            newa = zeros(T,sz)
-            newa[1:as] = a
-            a = newa
-        end
-
-        for i=1:xs
-            y[i] = si[1] + b[1]*x[i]
-            for j=1:(silen-1)
-                si[j] = si[j+1] + b[j+1]*x[i] - a[j+1]*y[i]
-            end
-            si[silen] = b[silen+1]*x[i] - a[silen+1]*y[i]
-        end
-    else
-        for i=1:xs
-            y[i] = si[1] + b[1]*x[i]
-            for j=1:(silen-1)
-                si[j] = si[j+1] + b[j+1]*x[i]
-            end
-            si[silen] = b[silen+1]*x[i]
+    initial_si = si
+    for col = 1:ncols
+        # Reset the filter state
+        si = initial_si[:, N > 1 ? col : 1]
+        if as > 1
+            _filt_iir!(out, b, a, x, si, col)
+        else
+            _filt_fir!(out, b, x, si, col)
         end
     end
-    return y
+    return out
+end
+
+function _filt_iir!(out, b, a, x, si, col)
+    silen = length(si)
+    @inbounds for i=1:size(x, 1)
+        xi = x[i,col]
+        val = si[1] + b[1]*xi
+        for j=1:(silen-1)
+            si[j] = si[j+1] + b[j+1]*xi - a[j+1]*val
+        end
+        si[silen] = b[silen+1]*xi - a[silen+1]*val
+        out[i,col] = val
+    end
+end
+
+function _filt_fir!(out, b, x, si, col)
+    silen = length(si)
+    @inbounds for i=1:size(x, 1)
+        xi = x[i,col]
+        val = si[1] + b[1]*xi
+        for j=1:(silen-1)
+            si[j] = si[j+1] + b[j+1]*xi
+        end
+        si[silen] = b[silen+1]*xi
+        out[i,col] = val
+    end
 end
 
 function deconv{T}(b::StridedVector{T}, a::StridedVector{T})
@@ -94,7 +116,7 @@ function conv{T<:Base.LinAlg.BlasFloat}(u::StridedVector{T}, v::StridedVector{T}
     end
     return y[1:n]
 end
-conv{T<:Integer}(u::StridedVector{T}, v::StridedVector{T}) = conv(float(u), float(v))
+conv{T<:Integer}(u::StridedVector{T}, v::StridedVector{T}) = int(conv(float(u), float(v)))
 conv{T<:Integer, S<:Base.LinAlg.BlasFloat}(u::StridedVector{T}, v::StridedVector{S}) = conv(float(u), v)
 conv{T<:Integer, S<:Base.LinAlg.BlasFloat}(u::StridedVector{S}, v::StridedVector{T}) = conv(u, float(v))
 
@@ -125,6 +147,8 @@ function conv2{T}(A::StridedMatrix{T}, B::StridedMatrix{T})
     end
     return C
 end
+conv2{T<:Integer}(A::StridedMatrix{T}, B::StridedMatrix{T}) = int(conv2(float(A), float(B)))
+conv2{T<:Integer}(u::StridedVector{T}, v::StridedVector{T}, A::StridedMatrix{T}) = int(conv2(float(u), float(v), float(A)))
 
 function xcorr(u, v)
     su = size(u,1); sv = size(v,1)
@@ -247,7 +271,7 @@ for (f, fr2r, Y, Tx) in ((:dct, :r2r, :Y, :Number),
 end
 
 # DCT of scalar is just the identity:
-dct(x::Number, dims) = length(dims) == 0 || dims[1] == 1 ? x : throw(BoundsError())x
+dct(x::Number, dims) = length(dims) == 0 || dims[1] == 1 ? x : throw(BoundsError())
 idct(x::Number, dims) = dct(x, dims)
 dct(x::Number) = x
 idct(x::Number) = x
