@@ -32,7 +32,8 @@ catch_backtrace() = ccall(:jl_get_backtrace, Array{Ptr{Void},1}, ())
 
 ## system error handling ##
 
-errno() = ccall(:jl_errno, Int32, ())
+errno() = ccall(:jl_errno, Cint, ())
+errno(e::Integer) = ccall(:jl_set_errno, Void, (Cint,), e)
 strerror(e::Integer) = bytestring(ccall(:strerror, Ptr{Uint8}, (Int32,), e))
 strerror() = strerror(errno())
 systemerror(p, b::Bool) = b ? throw(SystemError(string(p))) : nothing
@@ -40,9 +41,12 @@ systemerror(p, b::Bool) = b ? throw(SystemError(string(p))) : nothing
 ## assertion functions and macros ##
 
 assert(x) = x ? nothing : error("assertion failed")
-macro assert(ex,msg...)
-    msg = isempty(msg) ? ex : msg[1]
-    if isdefined(Base,:string)
+macro assert(ex,msgs...)
+    msg = isempty(msgs) ? ex : msgs[1]
+    if !isempty(msgs) && isa(msg, Expr)
+        # message is an expression needing evaluating
+        msg = :(string("assertion failed: ", $(esc(msg))))
+    elseif isdefined(Base,:string)
         msg = string("assertion failed: ", msg)
     else
         # string() might not be defined during bootstrap
