@@ -14,8 +14,11 @@ export
 
 # Types
     AbstractMatrix,
+    AbstractSparseArray,
     AbstractSparseMatrix,
+    AbstractSparseVector,
     AbstractVector,
+    AbstractVecOrMat,
     Array,
     Associative,
     Bidiagonal,
@@ -24,54 +27,46 @@ export
     BitArray,
     BitMatrix,
     BitVector,
-    BunchKaufman,
-    CharString,
-    Cholesky,
-    CholeskyPivoted,
+    CFILE,
     Cmd,
     Colon,
     Complex,
     Complex128,
     Complex64,
+    Complex32,
     DArray,
+    DenseMatrix,
+    DenseVecOrMat,
+    DenseVector,
     DevNull,
     Diagonal,
     Dict,
     Dims,
     EachLine,
-    Eigen,
     Enumerate,
     Factorization,
     FileMonitor,
     FileOffset,
     Filter,
-    GeneralizedSchur,
-    GeneralizedSVD,
+    FloatRange,
     Hermitian,
-    Hessenberg,
-    ImaginaryUnit,
+    UniformScaling,
     InsertionSort,
     IntSet,
     IO,
     IOBuffer,
     IOStream,
-    LDLTTridiagonal,
     LocalProcess,
-    LU,
-    LUTridiagonal,
     MathConst,
     Matrix,
     MergeSort,
     ObjectIdDict,
+    OrdinalRange,
     PollingFileWatcher,
     ProcessGroup,
-    QR,
-    QRPivoted,
     QuickSort,
     Range,
-    Range1,
     RangeIndex,
-    Ranges,
     Rational,
     Regex,
     RegexMatch,
@@ -85,10 +80,13 @@ export
     RoundNearest,
     RoundToZero,
     RoundUp,
-    Schur,
     Set,
+    SharedArray,
+    SharedMatrix,
+    SharedVector,
     SparseMatrixCSC,
-    Stat,
+    StatStruct,
+    StepRange,
     StridedArray,
     StridedMatrix,
     StridedVecOrMat,
@@ -97,19 +95,22 @@ export
     SubDArray,
     SubOrDArray,
     SubString,
-    SVD,
     Symmetric,
     SymTridiagonal,
     Timer,
     TmStruct,
     Triangular,
     Tridiagonal,
+    UnitRange,
+    UTF16String,
+    UTF32String,
     VecOrMat,
     Vector,
     VersionNumber,
     WeakKeyDict,
     WeakRef,
     Woodbury,
+    WString,
     Zip,
 
 # Ccall types
@@ -130,11 +131,10 @@ export
     Culonglong,
     Cushort,
     Cwchar_t,
-    #Ccomplex_float,
-    #Ccomplex_double,
 
 # Exceptions
     ArgumentError,
+    DimensionMismatch,
     EOFError,
     ErrorException,
     KeyError,
@@ -184,11 +184,15 @@ export
     γ, eulergamma,
     catalan,
     φ, golden,
+    I,
 
 # Operators
     !,
     !=,
+    ≠,
     !==,
+    ≡,
+    ≢, 
     $,
     %,
     &,
@@ -196,26 +200,32 @@ export
     +,
     -,
     .!=,
+    .≠,
     .+,
     .-,
     .*,
     ./,
     .<,
     .<=,
+    .≤,
     .==,
     .>,
     .>=,
+    .≥,
     .\,
     .^,
     /,
     //,
+    .//,
     <,
     <:,
     <<,
     <=,
+    ≤,
     ==,
     >,
     >=,
+    ≥,
     >>,
     #.>>,
     #.<<,
@@ -225,10 +235,10 @@ export
     |,
     ~,
     :,
+    ÷,
     A_ldiv_B!,
     A_ldiv_Bc,
     A_ldiv_Bt,
-    A_mul_B,
     A_mul_B!,
     A_mul_Bc,
     A_mul_Bc!,
@@ -238,7 +248,6 @@ export
     A_rdiv_Bt,
     Ac_ldiv_B,
     Ac_ldiv_Bc,
-    #Ac_mul_b_RFP,
     Ac_mul_B,
     Ac_mul_B!,
     Ac_mul_Bc,
@@ -255,6 +264,7 @@ export
     At_rdiv_Bt,
 
 # scalar math
+    @evalpoly,
     abs,
     abs2,
     acos,
@@ -279,7 +289,6 @@ export
     atanh,
     big,
     binomial,
-    bitmix,
     bool,
     bswap,
     cbrt,
@@ -291,6 +300,7 @@ export
     complex,
     complex128,
     complex64,
+    complex32,
     conj,
     copysign,
     cos,
@@ -307,7 +317,7 @@ export
     cscd,
     csch,
     dawson,
-    degrees2radians,
+    deg2rad,
     den,
     digamma,
     div,
@@ -382,6 +392,7 @@ export
     mod,
     mod1,
     modf,
+    mod2pi,
     nan,
     nextfloat,
     nextpow,
@@ -396,7 +407,7 @@ export
     prevpow2,
     prevprod,
     primes,
-    radians2degrees,
+    rad2deg,
     rationalize,
     real,
     realmax,
@@ -436,7 +447,10 @@ export
     uint64,
     uint8,
     unsigned,
+    widemul,
     zero,
+    √,
+    ∛,
 
 # specfun
     airy,
@@ -445,24 +459,32 @@ export
     airybi,
     airybiprime,
     airyprime,
+    airyx,
     besselh,
     besseli,
+    besselix,
     besselj,
     besselj0,
     besselj1,
+    besseljx,
     besselk,
+    besselkx,
     bessely,
     bessely0,
     bessely1,
+    besselyx,
     beta,
     eta,
     hankelh1,
+    hankelh1x,
     hankelh2,
+    hankelh2x,
     lbeta,
     polygamma,
     zeta,
 
 # arrays
+    bitbroadcast,
     broadcast!,
     broadcast!_function,
     broadcast,
@@ -480,8 +502,11 @@ export
     cummax,
     cummin,
     cumprod,
+    cumprod!,
     cumsum,
+    cumsum!,
     cumsum_kbn,
+    extrema,
     fill!,
     fill,
     find,
@@ -503,26 +528,33 @@ export
     indexin,
     indmax,
     indmin,
-    infs,
     invperm,
     ipermute!,
     ipermutedims,
     isperm,
     issorted,
     last,
+    linrange,
     linspace,
     logspace,
     mapslices,
     max,
+    maxabs,
+    maxabs!,
+    maximum!,
     maximum,
     min,
+    minabs,
+    minabs!,
+    minimum!,
     minimum,
-    nans,
+    minmax,
     ndims,
     nnz,
     nonzeros,
     nthperm!,
     nthperm,
+    countnz,
     ones,
     parent,
     parentindexes,
@@ -530,10 +562,15 @@ export
     permutations,
     permute!,
     permutedims,
+    permutedims!,
+    prod!,
     prod,
     promote_shape,
     randcycle,
     randperm,
+    randsubseq!,
+    randsubseq,
+    range,
     reducedim,
     repmat,
     reshape,
@@ -563,7 +600,12 @@ export
     strides,
     sub,
     sub2ind,
+    sum!,
     sum,
+    sumabs!,
+    sumabs,
+    sumabs2!,
+    sumabs2,
     sum_kbn,
     vcat,
     vec,
@@ -573,12 +615,12 @@ export
     bkfact!,
     bkfact,
     blas_set_num_threads,
+    blkdiag,
     chol,
     cholfact!,
     cholfact,
-    cholpfact!,
-    cholpfact,
     cond,
+    condskeel,
     cross,
     ctranspose,
     det,
@@ -597,8 +639,8 @@ export
     eigvecs,
     expm,
     eye,
-    factorize!,
     factorize,
+    givens,
     hessfact!,
     hessfact,
     ishermitian,
@@ -608,23 +650,21 @@ export
     istril,
     istriu,
     kron,
+    ldltfact,
+    ldltfact!,
     linreg,
     logdet,
     lu,
     lufact!,
     lufact,
+    lyap,
     norm,
-    normfro,
     null,
     peakflops,
     pinv,
     qr,
     qrfact!,
     qrfact,
-    qrp,
-    qrpfact!,
-    qrpfact,
-    randsym,
     rank,
     rref,
     scale!,
@@ -638,16 +678,18 @@ export
     svdfact,
     svdvals!,
     svdvals,
-    symmetrize!,
+    sylvester,
     trace,
     transpose,
     tril!,
     tril,
     triu!,
     triu,
+    vecnorm,
+    ⋅,
+    ×,
 
 # sparse
-    dense,
     etree,
     full,
     issparse,
@@ -682,7 +724,9 @@ export
     unshift!,
 
 # collections
+    all!,
     all,
+    any!,
     any,
     collect,
     complement!,
@@ -690,11 +734,14 @@ export
     contains,
     count,
     delete!,
+    deleteat!,
     eltype,
     empty!,
     endof,
     filter!,
     filter,
+    foldl,
+    foldr,
     get,
     get!,
     getindex,
@@ -727,6 +774,15 @@ export
     union,
     unique,
     values,
+    ∈,
+    ∉,
+    ∋,
+    ∌,
+    ⊆,
+    ⊈,
+    ⊊,
+    ∩,
+    ∪,
 
 # strings and text output
     ascii,
@@ -755,9 +811,11 @@ export
     hex2bytes,
     ind2chr,
     info,
+    is_assigned_char,
     is_valid_ascii,
     is_valid_char,
     is_valid_utf8,
+    is_valid_utf16,
     isalnum,
     isalpha,
     isascii,
@@ -782,6 +840,7 @@ export
     matchall,
     ndigits,
     nextind,
+    normalize_string,
     oct,
     parsefloat,
     parseint,
@@ -818,7 +877,10 @@ export
     unescape_string,
     uppercase,
     utf8,
+    utf16,
+    utf32,
     warn,
+    wstring,
     xdump,
 
 # random numbers
@@ -837,9 +899,6 @@ export
     get_bigfloat_precision,
     set_bigfloat_precision,
     with_bigfloat_precision,
-    get_bigfloat_rounding,
-    set_bigfloat_rounding,
-    with_bigfloat_rounding,
     get_rounding,
     set_rounding,
     with_rounding,
@@ -850,9 +909,11 @@ export
     hist,
     hist2d,
     histrange,
+    mean!,
     mean,
     median!,
     median,
+    middle,
     midpoints,
     quantile!,
     quantile,
@@ -874,6 +935,7 @@ export
     fft,
     fftshift,
     filt,
+    filt!,
     idct!,
     idct,
     ifft!,
@@ -962,6 +1024,7 @@ export
     super,
     typeintersect,
     typejoin,
+    widen,
 
 # syntax
     esc,
@@ -995,6 +1058,7 @@ export
     versioninfo,
     which,
     whos,
+    workspace,
 
 # loading source files
     evalfile,
@@ -1009,6 +1073,7 @@ export
     gc_disable,
     gc_enable,
     precompile,
+    clear_malloc_data,
 
 # misc
     atexit,
@@ -1025,6 +1090,7 @@ export
 
 # I/O and events
     accept,
+    bind,
     close,
     connect,
     countlines,
@@ -1040,11 +1106,13 @@ export
     getipaddr,
     htol,
     hton,
+    ismarked,
     isopen,
     isreadonly,
     listen,
     listenany,
     ltoh,
+    mark,
     mmap,
     mmap_array,
     mmap_bitarray,
@@ -1058,6 +1126,7 @@ export
     position,
     RawFD,
     read,
+    read!,
     readall,
     readavailable,
     readbytes!,
@@ -1072,9 +1141,12 @@ export
     redirect_stderr,
     redirect_stdin,
     redirect_stdout,
+    recv,
+    reset,
     seek,
     seekend,
     seekstart,
+    send,
     serialize,
     skip,
     skipchars,
@@ -1086,10 +1158,12 @@ export
     takebuf_array,
     takebuf_string,
     truncate,
+    unmark,
     watch_file,
     write,
     writecsv,
     writedlm,
+    UdpSocket,
 
 # multiprocessing
     addprocs,
@@ -1102,12 +1176,12 @@ export
     nworkers,
     pmap,
     procs,
-    put,
+    put!,
     remotecall,
     remotecall_fetch,
     remotecall_wait,
     rmprocs,
-    take,
+    take!,
     timedwait,
     wait,
     workers,
@@ -1137,9 +1211,13 @@ export
     drandn,
     dzeros,
     localpart,
-    myindexes,
+    localindexes,
     procs,
 
+# shared arrays
+    sdata,
+    indexpids,
+    
 # paths and file names
     abspath,
     basename,
@@ -1187,8 +1265,8 @@ export
     operm,
     pwd,
     rm,
-    rmdir,
     stat,
+    symlink,
     tempdir,
     tempname,
     touch,
@@ -1202,16 +1280,16 @@ export
     process_exited,
     process_running,
     readandwrite,
-    readsfrom,
     run,
     setenv,
     spawn,
     success,
-    writesto,
 
 # C interface
     c_free,
     c_malloc,
+    c_calloc,
+    c_realloc,
     cfunction,
     cglobal,
     disable_sigint,
@@ -1236,7 +1314,11 @@ export
     @__FILE__,
     @b_str,
     @r_str,
+    @r_mstr,
     @v_str,
+    @int128_str,
+    @uint128_str,
+    @bigint_str,
     @mstr,
     @unexpected,
     @assert,
@@ -1247,6 +1329,12 @@ export
     @allocated,
     @profile,
     @which,
+    @edit,
+    @less,
+    @code_typed,
+    @code_lowered,
+    @code_llvm,
+    @code_native,
     @windows,
     @unix,
     @osx,
@@ -1255,11 +1343,12 @@ export
     @unix_only,
     @osx_only,
     @linux_only,
+    @windowsxp_only,
+    @non_windowsxp_only,
     @schedule,
     @sync,
     @async,
     @spawn,
-    @spawnlocal,  # deprecated
     @spawnat,
     @fetch,
     @fetchfrom,
@@ -1276,4 +1365,7 @@ export
     @sprintf,
     @deprecate,
     @boundscheck,
-    @inbounds
+    @inbounds,
+    @simd,
+    @label,
+    @goto
