@@ -13,15 +13,8 @@ export fft, bfft, ifft, rfft, brfft, irfft,
 
 ## FFT: Implement fft by calling fftw.
 
-@unix_only begin
-const libfftw = "libfftw3_threads"
-const libfftwf = "libfftw3f_threads"
-end
-
-@windows_only begin
-const libfftw = "libfftw3"
-const libfftwf = "libfftw3f"
-end
+const libfftw = Base.libfftw_name
+const libfftwf = Base.libfftwf_name
 
 ## Direction of FFT
 
@@ -78,23 +71,17 @@ typealias fftwTypeSingle Union(Type{Float32},Type{Complex64})
 # FFTW's api/import-wisdom-from-file.c file].
 
 function export_wisdom(fname::String)
-    f = ccall(:fopen, Ptr{Void}, (Ptr{Uint8},Ptr{Uint8}),
-              bytestring(fname), bytestring("w"))
-    if f == C_NULL
-        error("could not open wisdom file $fname")
-    end
+    f = ccall(:fopen, Ptr{Void}, (Ptr{Uint8},Ptr{Uint8}), fname, "w")
+    systemerror("could not open wisdom file $fname for writing", f == C_NULL)
     ccall((:fftw_export_wisdom_to_file,libfftw), Void, (Ptr{Void},), f)
-    ccall(:fputs, Int32, (Ptr{Uint8},Ptr{Void}), bytestring(" "^256), f)
+    ccall(:fputs, Int32, (Ptr{Uint8},Ptr{Void}), " "^256, f)
     ccall((:fftwf_export_wisdom_to_file,libfftwf), Void, (Ptr{Void},), f)
     ccall(:fclose, Void, (Ptr{Void},), f)
 end
 
 function import_wisdom(fname::String)
-    f = ccall(:fopen, Ptr{Void}, (Ptr{Uint8},Ptr{Uint8}),
-              bytestring(fname), bytestring("r"))
-    if f == C_NULL
-        error("could not open wisdom file $fname")
-    end
+    f = ccall(:fopen, Ptr{Void}, (Ptr{Uint8},Ptr{Uint8}), fname, "r")
+    systemerror("could not open wisdom file $fname for reading", f == C_NULL)
     if ccall((:fftw_import_wisdom_from_file,libfftw),Int32,(Ptr{Void},),f)==0||
        ccall((:fftwf_import_wisdom_from_file,libfftwf),Int32,(Ptr{Void},),f)==0
         error("failed to import wisdom from $fname")
@@ -205,11 +192,23 @@ set_timelimit(precision::fftwTypeSingle,seconds) =
 #   Use the undocumented routine fftw_alignment_of to determine the 
 #   alignment of a given pointer modulo whatever FFTW needs.
 
-alignment_of{T<:fftwDouble}(A::StridedArray{T}) = 
-   ccall((:fftw_alignment_of, libfftw), Int32, (Ptr{T},), A)
+if Base.libfftw_name == "libmkl_rt"
 
-alignment_of{T<:fftwSingle}(A::StridedArray{T}) = 
-   ccall((:fftwf_alignment_of, libfftwf), Int32, (Ptr{T},), A)
+    alignment_of{T<:fftwDouble}(A::StridedArray{T}) =
+        convert(Int32, convert(Int64, pointer(A)) % 16)
+
+    alignment_of{T<:fftwSingle}(A::StridedArray{T}) =
+        convert(Int32, convert(Int64, pointer(A)) % 16)
+
+else
+
+    alignment_of{T<:fftwDouble}(A::StridedArray{T}) =
+        ccall((:fftw_alignment_of, libfftw), Int32, (Ptr{Void},), A)
+
+    alignment_of{T<:fftwSingle}(A::StridedArray{T}) =
+        ccall((:fftwf_alignment_of, libfftwf), Int32, (Ptr{Void},), A)
+
+end
 
 # Plan (low-level)
 
