@@ -37,15 +37,15 @@ convert(::Type{Float32}, x::Float64) = box(Float32,fptrunc(Float32,x))
 convert(::Type{Float64}, x::Float16) = convert(Float64, convert(Float32,x))
 convert(::Type{Float64}, x::Float32) = box(Float64,fpext(Float64,x))
 
-convert(::Type{FloatingPoint}, x::Bool)    = convert(Float32, x)
-convert(::Type{FloatingPoint}, x::Char)    = convert(Float32, x)
-convert(::Type{FloatingPoint}, x::Int8)    = convert(Float32, x)
-convert(::Type{FloatingPoint}, x::Int16)   = convert(Float32, x)
+convert(::Type{FloatingPoint}, x::Bool)    = convert(Float64, x)
+convert(::Type{FloatingPoint}, x::Char)    = convert(Float64, x)
+convert(::Type{FloatingPoint}, x::Int8)    = convert(Float64, x)
+convert(::Type{FloatingPoint}, x::Int16)   = convert(Float64, x)
 convert(::Type{FloatingPoint}, x::Int32)   = convert(Float64, x)
 convert(::Type{FloatingPoint}, x::Int64)   = convert(Float64, x) # LOSSY
 convert(::Type{FloatingPoint}, x::Int128)  = convert(Float64, x) # LOSSY
-convert(::Type{FloatingPoint}, x::Uint8)   = convert(Float32, x)
-convert(::Type{FloatingPoint}, x::Uint16)  = convert(Float32, x)
+convert(::Type{FloatingPoint}, x::Uint8)   = convert(Float64, x)
+convert(::Type{FloatingPoint}, x::Uint16)  = convert(Float64, x)
 convert(::Type{FloatingPoint}, x::Uint32)  = convert(Float64, x)
 convert(::Type{FloatingPoint}, x::Uint64)  = convert(Float64, x) # LOSSY
 convert(::Type{FloatingPoint}, x::Uint128) = convert(Float64, x) # LOSSY
@@ -55,9 +55,6 @@ float32(x) = convert(Float32, x)
 float64(x) = convert(Float64, x)
 float(x)   = convert(FloatingPoint, x)
 
-# possibly a hack, but useful for `f(x::Real) = f(float(x))` fallbacks
-float(x::Float16) = float32(x)
-
 ## conversions from floating-point ##
 
 # fallbacks using only convert, trunc, ceil, floor, round
@@ -71,35 +68,30 @@ iceil {T<:Integer}(::Type{T}, x::FloatingPoint) = convert(T,ceil(x))
 ifloor{T<:Integer}(::Type{T}, x::FloatingPoint) = convert(T,floor(x))
 iround{T<:Integer}(::Type{T}, x::FloatingPoint) = convert(T,round(x))
 
-## fast specific type converions ##
+## fast specific type conversions ##
 
-if WORD_SIZE == 64
-    iround(x::Float32) = iround(float64(x))
-    itrunc(x::Float32) = itrunc(float64(x))
-    iround(x::Float64) = box(Int64,fpsiround(unbox(Float64,x)))
-    itrunc(x::Float64) = box(Int64,fptosi(unbox(Float64,x)))
-else
-    iround(x::Float32) = box(Int32,fpsiround(unbox(Float32,x)))
-    itrunc(x::Float32) = box(Int32,fptosi(unbox(Float32,x)))
-    iround(x::Float64) = int32(box(Int64,fpsiround(unbox(Float64,x))))
-    itrunc(x::Float64) = int32(box(Int64,fptosi(unbox(Float64,x))))
-end
+iround(x::Float32) = iround(Int, x)
+iround(x::Float64) = iround(Int, x)
+itrunc(x::Float32) = itrunc(Int, x)
+itrunc(x::Float64) = itrunc(Int, x)
 
-for to in (Int8, Uint8, Int16, Uint16)
+for to in (Int8, Int16, Int32, Int64)
     @eval begin
-        iround(::Type{$to}, x::Float32) = box($to,trunc_int($to,fpsiround(unbox(Float32,x))))
-        iround(::Type{$to}, x::Float64) = box($to,trunc_int($to,fpsiround(unbox(Float64,x))))
+        iround(::Type{$to}, x::Float32) = box($to,fpsiround($to,unbox(Float32,x)))
+        iround(::Type{$to}, x::Float64) = box($to,fpsiround($to,unbox(Float64,x)))
+        itrunc(::Type{$to}, x::Float32) = box($to,fptosi($to,unbox(Float32,x)))
+        itrunc(::Type{$to}, x::Float64) = box($to,fptosi($to,unbox(Float64,x)))
     end
 end
 
-iround(::Type{Int32}, x::Float32) = box(Int32,fpsiround(unbox(Float32,x)))
-iround(::Type{Int32}, x::Float64) = box(Int32,trunc_int(Int32,fpsiround(unbox(Float64,x))))
-iround(::Type{Uint32}, x::Float32) = box(Uint32,fpuiround(unbox(Float32,x)))
-iround(::Type{Uint32}, x::Float64) = box(Uint32,trunc_int(Uint32,fpuiround(unbox(Float64,x))))
-iround(::Type{Int64}, x::Float32) = box(Int64,fpsiround(float64(x)))
-iround(::Type{Int64}, x::Float64) = box(Int64,fpsiround(unbox(Float64,x)))
-iround(::Type{Uint64}, x::Float32) = box(Uint64,fpuiround(float64(x)))
-iround(::Type{Uint64}, x::Float64) = box(Uint64,fpuiround(unbox(Float64,x)))
+for to in (Uint8, Uint16, Uint32, Uint64)
+    @eval begin
+        iround(::Type{$to}, x::Float32) = box($to,fpuiround($to,unbox(Float32,x)))
+        iround(::Type{$to}, x::Float64) = box($to,fpuiround($to,unbox(Float64,x)))
+        itrunc(::Type{$to}, x::Float32) = box($to,fptoui($to,unbox(Float32,x)))
+        itrunc(::Type{$to}, x::Float64) = box($to,fptoui($to,unbox(Float64,x)))
+    end
+end
 
 iround(::Type{Int128}, x::Float32) = convert(Int128,round(x))
 iround(::Type{Int128}, x::Float64) = convert(Int128,round(x))
@@ -116,18 +108,14 @@ promote_rule(::Type{Float32}, ::Type{Float16}) = Float32
 promote_rule(::Type{Float64}, ::Type{Float16}) = Float64
 promote_rule(::Type{Float64}, ::Type{Float32}) = Float64
 
-morebits(::Type{Float16}) = Float32
-morebits(::Type{Float32}) = Float64
+widen(::Type{Float16}) = Float32
+widen(::Type{Float32}) = Float64
 
 ## floating point arithmetic ##
 
--(x::Float16) = reinterpret(Float16, reinterpret(Uint16,x) $ 0x8000)
 -(x::Float32) = box(Float32,neg_float(unbox(Float32,x)))
 -(x::Float64) = box(Float64,neg_float(unbox(Float64,x)))
 
-for op in (:+,:-,:*,:/)
-    @eval ($op)(a::Float16, b::Float16) = ($op)(float32(a), float32(b))
-end
 +(x::Float32, y::Float32) = box(Float32,add_float(unbox(Float32,x),unbox(Float32,y)))
 +(x::Float64, y::Float64) = box(Float64,add_float(unbox(Float64,x),unbox(Float64,y)))
 -(x::Float32, y::Float32) = box(Float32,sub_float(unbox(Float32,x),unbox(Float32,y)))
@@ -157,26 +145,35 @@ mod{T<:FloatingPoint}(x::T, y::T) = rem(y+rem(x,y),y)
 <=(x::Float32, y::Float32) = le_float(unbox(Float32,x),unbox(Float32,y))
 <=(x::Float64, y::Float64) = le_float(unbox(Float64,x),unbox(Float64,y))
 
-isequal{T<:FloatingPoint}(x::T, y::T) =
-    ((x==y) & (signbit(x)==signbit(y))) | (isnan(x)&isnan(y))
-
 isequal(x::Float32, y::Float32) = fpiseq(unbox(Float32,x),unbox(Float32,y))
 isequal(x::Float64, y::Float64) = fpiseq(unbox(Float64,x),unbox(Float64,y))
 isless (x::Float32, y::Float32) = fpislt(unbox(Float32,x),unbox(Float32,y))
 isless (x::Float64, y::Float64) = fpislt(unbox(Float64,x),unbox(Float64,y))
 
-isless (a::Integer, b::FloatingPoint) = (a<b) | isless(float(a),b)
-isless (a::FloatingPoint, b::Integer) = (a<b) | isless(a,float(b))
+function cmp(x::FloatingPoint, y::FloatingPoint)
+    (isnan(x) || isnan(y)) && throw(DomainError())
+    ifelse(x<y, -1, ifelse(x>y, 1, 0))
+end
+
+function cmp(x::Real, y::FloatingPoint)
+    isnan(y) && throw(DomainError())
+    ifelse(x<y, -1, ifelse(x>y, 1, 0))
+end
+
+function cmp(x::FloatingPoint, y::Real)
+    isnan(x) && throw(DomainError())
+    ifelse(x<y, -1, ifelse(x>y, 1, 0))
+end
 
 ==(x::Float64, y::Int64  ) = eqfsi64(unbox(Float64,x),unbox(Int64,y))
 ==(x::Float64, y::Uint64 ) = eqfui64(unbox(Float64,x),unbox(Uint64,y))
 ==(x::Int64  , y::Float64) = eqfsi64(unbox(Float64,y),unbox(Int64,x))
 ==(x::Uint64 , y::Float64) = eqfui64(unbox(Float64,y),unbox(Uint64,x))
 
-==(x::Float32, y::Int64  ) = eqfsi64(unbox(Float64,float64(x)),unbox(Int64,y))
-==(x::Float32, y::Uint64 ) = eqfui64(unbox(Float64,float64(x)),unbox(Uint64,y))
-==(x::Int64  , y::Float32) = eqfsi64(unbox(Float64,float64(y)),unbox(Int64,x))
-==(x::Uint64 , y::Float32) = eqfui64(unbox(Float64,float64(y)),unbox(Uint64,x))
+==(x::Float32, y::Int64  ) = eqfsi64(unbox(Float32,x),unbox(Int64,y))
+==(x::Float32, y::Uint64 ) = eqfui64(unbox(Float32,x),unbox(Uint64,y))
+==(x::Int64  , y::Float32) = eqfsi64(unbox(Float32,y),unbox(Int64,x))
+==(x::Uint64 , y::Float32) = eqfui64(unbox(Float32,y),unbox(Uint64,x))
 
 < (x::Float64, y::Int64  ) = ltfsi64(unbox(Float64,x),unbox(Int64,y))
 < (x::Float64, y::Uint64 ) = ltfui64(unbox(Float64,x),unbox(Uint64,y))
@@ -207,6 +204,18 @@ isless (a::FloatingPoint, b::Integer) = (a<b) | isless(a,float(b))
 <=(x::Float32, y::Union(Int32,Uint32)) = float64(x)<=float64(y)
 <=(x::Union(Int32,Uint32), y::Float32) = float64(x)<=float64(y)
 
+abs(x::Float64) = box(Float64,abs_float(unbox(Float64,x)))
+abs(x::Float32) = box(Float32,abs_float(unbox(Float32,x)))
+
+isnan(x::FloatingPoint) = x != x
+isnan(x::Real) = false
+
+isfinite(x::FloatingPoint) = x - x == 0
+isfinite(x::Real) = decompose(x)[3] != 0
+isfinite(x::Integer) = true
+
+isinf(x::Real) = !isnan(x) & !isfinite(x)
+
 ## floating point traits ##
 
 const Inf16 = box(Float16,unbox(Uint16,0x7c00))
@@ -216,6 +225,31 @@ const NaN32 = box(Float32,unbox(Uint32,0x7fc00000))
 const Inf = box(Float64,unbox(Uint64,0x7ff0000000000000))
 const NaN = box(Float64,unbox(Uint64,0x7ff8000000000000))
 
+## precision, as defined by the effective number of bits in the mantissa ##
+precision(::Float16) = 11
+precision(::Float32) = 24
+precision(::Float64) = 53
+
+function float_lex_order(f::Integer, delta::Integer)
+    # convert from signed magnitude to 2's complement and back
+    neg = f < 0
+    if neg
+        f = oftype(f, -(f & typemax(f)))
+    end
+    f = oftype(f, f + delta)
+    neg && f == 0 && return typemin(f)  # nextfloat(-5e-324) === -0.0
+    f < 0 ? oftype(f, -(f & typemax(f))) : f
+end
+
+nextfloat(x::Float16, i::Integer) =
+    (isinf(x)&&sign(x)==sign(i)) ? x : reinterpret(Float16,float_lex_order(reinterpret(Int16,x), i))
+nextfloat(x::Float32, i::Integer) =
+    (isinf(x)&&sign(x)==sign(i)) ? x : reinterpret(Float32,float_lex_order(reinterpret(Int32,x), i))
+nextfloat(x::Float64, i::Integer) =
+    (isinf(x)&&sign(x)==sign(i)) ? x : reinterpret(Float64,float_lex_order(reinterpret(Int64,x), i))
+nextfloat(x::FloatingPoint) = nextfloat(x,1)
+prevfloat(x::FloatingPoint) = nextfloat(x,-1)
+
 @eval begin
     inf(::Type{Float16}) = $Inf16
     nan(::Type{Float16}) = $NaN16
@@ -249,13 +283,7 @@ const NaN = box(Float64,unbox(Uint64,0x7ff8000000000000))
     realmin() = realmin(Float64)
     realmax() = realmax(Float64)
 
-    nextfloat(x::Float16, i::Integer) = box(Float16,add_int(reinterpret(Int16,x),unbox(Int16,int16(i))))
-    nextfloat(x::Float32, i::Integer) = box(Float32,add_int(unbox(Float32,x),unbox(Int32,int32(i))))
-    nextfloat(x::Float64, i::Integer) = box(Float64,add_int(unbox(Float64,x),unbox(Int64,int64(i))))
-    nextfloat(x::FloatingPoint) = nextfloat(x,1)
-    prevfloat(x::FloatingPoint) = nextfloat(x,-1)
-
-    eps(x::FloatingPoint) = isfinite(x) ? abs(nextfloat(x)-x) : nan(x)
+    eps(x::FloatingPoint) = isfinite(x) ? abs(x) >= realmin(x) ? ldexp(eps(typeof(x)),exponent(x)) : nextfloat(zero(x)) : nan(x)
     eps(::Type{Float16}) = $(box(Float16,unbox(Uint16,0x1400)))
     eps(::Type{Float32}) = $(box(Float32,unbox(Uint32,0x34000000)))
     eps(::Type{Float64}) = $(box(Float64,unbox(Uint64,0x3cb0000000000000)))
