@@ -1,26 +1,17 @@
 ## floating-point functions ##
 
-abs(x::Float64) = box(Float64,abs_float(unbox(Float64,x)))
-abs(x::Float32) = box(Float32,abs_float(unbox(Float32,x)))
-
-isnan(x::FloatingPoint) = (x != x)
-isnan(x::Real) = isnan(float(x))
-isnan(x::Integer) = false
-
-isinf(x::FloatingPoint) = (abs(x) == Inf)
-isinf(x::Real) = isinf(float(x))
-isinf(x::Integer) = false
-
-isfinite(x::FloatingPoint) = (x-x == 0)
-isfinite(x::Real) = isfinite(float(x))
-isfinite(x::Integer) = true
-
 copysign(x::Float64, y::Float64) = box(Float64,copysign_float(unbox(Float64,x),unbox(Float64,y)))
 copysign(x::Float32, y::Float32) = box(Float32,copysign_float(unbox(Float32,x),unbox(Float32,y)))
 copysign(x::Float32, y::Real) = copysign(x, float32(y))
 copysign(x::Float64, y::Real) = copysign(x, float64(y))
 @vectorize_2arg Real copysign
 
+flipsign(x::Float64, y::Float64) = box(Float64,xor_int(unbox(Float64,x),and_int(unbox(Float64,y),0x8000000000000000)))
+flipsign(x::Float32, y::Float32) = box(Float32,xor_int(unbox(Float32,x),and_int(unbox(Float32,y),0x80000000)))
+flipsign(x::Float32, y::Real) = flipsign(x, float32(y))
+flipsign(x::Float64, y::Real) = flipsign(x, float64(y))
+@vectorize_2arg Real flipsign
+
 signbit(x::Float64) = signbit(reinterpret(Int64,x))
 signbit(x::Float32) = signbit(reinterpret(Int32,x))
 signbit(x::Float16) = signbit(reinterpret(Int16,x))
@@ -33,11 +24,6 @@ maxintfloat() = maxintfloat(Float64)
 
 isinteger(x::FloatingPoint) = (trunc(x)==x)&isfinite(x)
 
-## precision, as defined by the effective number of bits in the mantissa ##
-precision(::Float16) = 11
-precision(::Float32) = 24
-precision(::Float64) = 53
-
 num2hex(x::Float16) = hex(reinterpret(Uint16,x), 4)
 num2hex(x::Float32) = hex(box(Uint32,unbox(Float32,x)),8)
 num2hex(x::Float64) = hex(box(Uint64,unbox(Float64,x)),16)
@@ -58,9 +44,9 @@ end
 @vectorize_1arg Number abs2
 @vectorize_1arg Number angle
 
-@vectorize_1arg Real isnan
-@vectorize_1arg Real isinf
-@vectorize_1arg Real isfinite
+@vectorize_1arg Number isnan
+@vectorize_1arg Number isinf
+@vectorize_1arg Number isfinite
 
 iround{T<:Integer,R<:Real}(::Type{T}, x::AbstractArray{R,1}) = [ iround(T, x[i]) for i = 1:length(x) ]
 iround{T<:Integer,R<:Real}(::Type{T}, x::AbstractArray{R,2}) = [ iround(T, x[i,j]) for i = 1:size(x,1), j = 1:size(x,2) ]
