@@ -21,14 +21,6 @@ end
 -{T<:SmallUnsigned}(x::T, y::T) = uint(x) - uint(y)
 *{T<:SmallUnsigned}(x::T, y::T) = uint(x) * uint(y)
 
-div{T<:SmallSigned}(x::T, y::T) = div(int(x),int(y))
-rem{T<:SmallSigned}(x::T, y::T) = rem(int(x),int(y))
-mod{T<:SmallSigned}(x::T, y::T) = mod(int(x),int(y))
-
-div{T<:SmallUnsigned}(x::T, y::T) = div(uint(x),uint(y))
-rem{T<:SmallUnsigned}(x::T, y::T) = rem(uint(x),uint(y))
-mod{T<:Unsigned}(x::T, y::T) = rem(x,y)
-
 -(x::Int)     = box(Int,neg_int(unbox(Int,x)))
 -(x::Uint)    = box(Uint,neg_int(unbox(Uint,x)))
 -(x::Int64)   = box(Int64,neg_int(unbox(Int64,x)))
@@ -58,6 +50,32 @@ mod{T<:Unsigned}(x::T, y::T) = rem(x,y)
 /(x::Integer, y::Integer) = float(x)/float(y)
 inv(x::Integer) = float(one(x))/float(x)
 
+isodd(n::Integer) = bool(rem(n,2))
+iseven(n::Integer) = !isodd(n)
+
+signbit(x::Integer) = x < 0
+signbit(x::Unsigned) = false
+
+flipsign(x::Int,    y::Int)    = box(Int,flipsign_int(unbox(Int,x),unbox(Int,y)))
+flipsign(x::Int64,  y::Int64)  = box(Int64,flipsign_int(unbox(Int64,x),unbox(Int64,y)))
+flipsign(x::Int128, y::Int128) = box(Int128,flipsign_int(unbox(Int128,x),unbox(Int128,y)))
+
+flipsign{T<:Signed}(x::T,y::T)  = flipsign(int(x),int(y))
+flipsign(x::Signed, y::Signed)  = flipsign(promote(x,y)...)
+flipsign(x::Signed, y::Float32) = flipsign(x, reinterpret(Int32,y))
+flipsign(x::Signed, y::Float64) = flipsign(x, reinterpret(Int64,y))
+flipsign(x::Signed, y::Real)    = flipsign(x, -oftype(x,signbit(y)))
+
+copysign(x::Signed, y::Signed)  = flipsign(x, x$y)
+copysign(x::Signed, y::Float32) = copysign(x, reinterpret(Int32,y))
+copysign(x::Signed, y::Float64) = copysign(x, reinterpret(Int64,y))
+copysign(x::Signed, y::Real)    = copysign(x, -oftype(x,signbit(y)))
+
+abs(x::Unsigned) = x
+abs(x::Signed) = flipsign(x,x)
+
+~(n::Integer) = -n-1
+
 div(x::Signed, y::Unsigned) = flipsign(signed(div(unsigned(abs(x)),y)),x)
 div(x::Unsigned, y::Signed) = unsigned(flipsign(signed(div(x,unsigned(abs(y)))),y))
 
@@ -70,22 +88,23 @@ fld(x::Unsigned, y::Signed) = div(x,y)-(signbit(y)&(rem(x,y)!=0))
 mod(x::Signed, y::Unsigned) = rem(y+unsigned(rem(x,y)),y)
 mod(x::Unsigned, y::Signed) = rem(y+signed(rem(x,y)),y)
 
-div(x::Int,     y::Int)     = box(Int,sdiv_int(unbox(Int,x),unbox(Int,y)))
-div(x::Uint,    y::Uint)    = box(Uint,udiv_int(unbox(Uint,x),unbox(Uint,y)))
-div(x::Int64,   y::Int64)   = box(Int64,sdiv_int(unbox(Int64,x),unbox(Int64,y)))
-div(x::Uint64,  y::Uint64)  = box(Uint64,udiv_int(unbox(Uint64,x),unbox(Uint64,y)))
+# Don't promote integers for div/rem/mod since there no danger of overflow,
+# while there is a substantial performance penalty to 64-bit promotion.
+typealias Signed64 Union(Int8,Int16,Int32,Int64)
+typealias Unsigned64 Union(Uint8,Uint16,Uint32,Uint64)
+typealias Integer64 Union(Signed64,Unsigned64)
 
-rem(x::Int,     y::Int)     = box(Int,srem_int(unbox(Int,x),unbox(Int,y)))
-rem(x::Uint,    y::Uint)    = box(Uint,urem_int(unbox(Uint,x),unbox(Uint,y)))
-rem(x::Int64,   y::Int64)   = box(Int64,srem_int(unbox(Int64,x),unbox(Int64,y)))
-rem(x::Uint64,  y::Uint64)  = box(Uint64,urem_int(unbox(Uint64,x),unbox(Uint64,y)))
+div{T<:Signed64}  (x::T, y::T) = box(T,sdiv_int(unbox(T,x),unbox(T,y)))
+div{T<:Unsigned64}(x::T, y::T) = box(T,udiv_int(unbox(T,x),unbox(T,y)))
+rem{T<:Signed64}  (x::T, y::T) = box(T,srem_int(unbox(T,x),unbox(T,y)))
+rem{T<:Unsigned64}(x::T, y::T) = box(T,urem_int(unbox(T,x),unbox(T,y)))
+mod{T<:Signed64}  (x::T, y::T) = box(T,smod_int(unbox(T,x),unbox(T,y)))
+
+mod{T<:Unsigned}(x::T, y::T) = rem(x,y)
 
 fld{T<:Unsigned}(x::T, y::T) = div(x,y)
 fld{T<:Integer }(x::T, y::T) = div(x,y)-(signbit(x$y)&(rem(x,y)!=0))
 
-mod(x::Int,    y::Int)    = box(Int,smod_int(unbox(Int,x),unbox(Int,y)))
-mod(x::Int64,  y::Int64)  = box(Int64,smod_int(unbox(Int64,x),unbox(Int64,y)))
-
 ## integer bitwise operations ##
 
 ~(x::Int8 )  = box(Int8,not_int(unbox(Int8,x)))
@@ -316,6 +335,7 @@ convert(::Type{Signed}, x::Uint64 ) = convert(Int64,x)
 convert(::Type{Signed}, x::Uint128) = convert(Int128,x)
 convert(::Type{Signed}, x::Float32) = convert(Int,x)
 convert(::Type{Signed}, x::Float64) = convert(Int,x)
+convert(::Type{Signed}, x::Char)    = convert(Int,x)
 
 convert(::Type{Unsigned}, x::Int8   ) = convert(Uint,x)
 convert(::Type{Unsigned}, x::Int16  ) = convert(Uint,x)
@@ -324,6 +344,7 @@ convert(::Type{Unsigned}, x::Int64  ) = convert(Uint64,x)
 convert(::Type{Unsigned}, x::Int128 ) = convert(Uint128,x)
 convert(::Type{Unsigned}, x::Float32) = convert(Uint,x)
 convert(::Type{Unsigned}, x::Float64) = convert(Uint,x)
+convert(::Type{Unsigned}, x::Char)    = convert(Uint,x)
 
 convert(::Type{Integer}, x::Float32) = convert(Int,x)
 convert(::Type{Integer}, x::Float64) = convert(Int,x)
@@ -355,6 +376,20 @@ itrunc(x::Integer) = x
 ifloor(x::Integer) = x
  iceil(x::Integer) = x
 
+## integer construction ##
+
+macro int128_str(str)
+    int128(str)
+end
+
+macro uint128_str(str)
+    uint128(str)
+end
+
+macro bigint_str(str)
+    BigInt(str)
+end
+
 ## system word size ##
 
 const WORD_SIZE = int(Int.size)*8
@@ -455,40 +490,29 @@ sizeof(::Type{Uint64})  = 8
 sizeof(::Type{Int128})  = 16
 sizeof(::Type{Uint128}) = 16
 
-morebits(::Type{Int8}) = Int16
-morebits(::Type{Int16}) = Int32
-morebits(::Type{Int32}) = Int64
-morebits(::Type{Int64}) = Int128
-morebits(::Type{Uint8}) = Uint16
-morebits(::Type{Uint16}) = Uint32
-morebits(::Type{Uint32}) = Uint64
-morebits(::Type{Uint64}) = Uint128
+widen(::Type{Int8}) = Int
+widen(::Type{Int16}) = Int
+widen(::Type{Int32}) = Int64
+widen(::Type{Int64}) = Int128
+widen(::Type{Uint8}) = Uint
+widen(::Type{Uint16}) = Uint
+widen(::Type{Uint32}) = Uint64
+widen(::Type{Uint64}) = Uint128
 
 ## float to integer coercion ##
 
 # requires int arithmetic defined, for the loops to work
 
-for f in (:int, :int8, :int16, :int32, :signed, :integer)
-    @eval ($f)(x::FloatingPoint) = ($f)(iround(x))
-end
-
-for (f,t) in ((:uint8,:Uint8), (:uint16,:Uint16), (:uint32,:Uint32),
-              (:int64,:Int64), (:uint64,:Uint64),
+for (f,t) in ((:uint8,:Uint8), (:uint16,:Uint16), (:uint32,:Uint32), (:uint64,:Uint64),
+              (:int8,:Int8),   (:int16,:Int16),   (:int32,:Int32),   (:int64,:Int64),
               (:int128,:Int128), (:uint128,:Uint128),
-              (:unsigned,:Uint), (:uint,:Uint))
+              (:signed,:Int), (:unsigned,:Uint), (:integer,:Int),
+              (:int,:Int), (:uint,:Uint))
     @eval ($f)(x::FloatingPoint) = iround($t,x)
 end
 
 ## wide multiplication, Int128 multiply and divide ##
 
-widemul(x::Union(Int8,Uint8,Int16), y::Union(Int8,Uint8,Int16)) = int32(x)*int32(y)
-widemul(x::Uint16, y::Uint16) = uint32(x)*uint32(y)
-
-widemul(x::Int32, y::Int32) = int64(x)*int64(y)
-widemul(x::Uint32, y::Uint32) = uint64(x)*uint64(y)
-
-widemul(x::Integer, y::Integer) = widemul(promote(x,y)...)
-
 if WORD_SIZE==32
     function widemul(u::Int64, v::Int64)
         local u0::Uint64, v0::Uint64, w0::Uint64
@@ -559,9 +583,6 @@ if WORD_SIZE==32
     >>>(x::Int128,  y::Int32) = y == 0 ? x : box(Int128,lshr_int(unbox(Int128,x),unbox(Int32,y)))
     >>>(x::Uint128, y::Int32) = y == 0 ? x : box(Uint128,lshr_int(unbox(Uint128,x),unbox(Int32,y)))
 else
-    widemul(u::Int64, v::Int64) = int128(u)*int128(v)
-    widemul(u::Uint64, v::Uint64) = uint128(u)*uint128(v)
-
     *(x::Int128,  y::Int128)  = box(Int128,mul_int(unbox(Int128,x),unbox(Int128,y)))
     *(x::Uint128, y::Uint128) = box(Uint128,mul_int(unbox(Uint128,x),unbox(Uint128,y)))
 
@@ -573,3 +594,63 @@ else
 
     mod(x::Int128, y::Int128) = box(Int128,smod_int(unbox(Int128,x),unbox(Int128,y)))
 end
+
+## checked +, - and *
+
+for T in (Int8,Int16,Int32,Int64)#,Int128) ## FIXME: #4905
+    @eval begin
+        checked_add(x::$T, y::$T) = box($T,checked_sadd(unbox($T,x),unbox($T,y)))
+        checked_sub(x::$T, y::$T) = box($T,checked_ssub(unbox($T,x),unbox($T,y)))
+    end
+end
+for T in (Int16,Int32)
+    @eval begin
+        checked_mul(x::$T, y::$T) = box($T,checked_smul(unbox($T,x),unbox($T,y)))
+    end
+end
+for T in (Uint8,Uint16,Uint32,Uint64)#,Uint128) ## FIXME: #4905
+    @eval begin
+        checked_add(x::$T, y::$T) = box($T,checked_uadd(unbox($T,x),unbox($T,y)))
+        checked_sub(x::$T, y::$T) = box($T,checked_usub(unbox($T,x),unbox($T,y)))
+    end
+end
+for T in (Uint16,Uint32)
+    @eval begin
+        checked_mul(x::$T, y::$T) = box($T,checked_umul(unbox($T,x),unbox($T,y)))
+    end
+end
+
+# checked mul is broken for 8-bit types (LLVM bug?) ## FIXME: #4905
+
+for T in (Int8,Uint8)
+    @eval function checked_mul(x::$T, y::$T)
+        xy = x*y
+        xy8 = convert($T,xy)
+        xy == xy8 || throw(OverflowError())
+        return xy8
+    end
+end
+
+if WORD_SIZE == 32
+for T in (Int64,Uint64)
+    @eval function checked_mul(x::$T, y::$T)
+        xy = int128(x)*int128(y)
+        xy64 = convert($T,xy)
+        xy == xy64 || throw(OverflowError())
+        return xy64
+    end
+end
+else
+    checked_mul(x::Int64, y::Int64)   = box(Int64,checked_smul(unbox(Int64,x),unbox(Int64,y)))
+    checked_mul(x::Uint64, y::Uint64) = box(Uint64,checked_umul(unbox(Uint64,x),unbox(Uint64,y)))
+end
+
+# checked ops are broken for 128-bit types (LLVM bug) ## FIXME: #4905
+
+checked_add(x::Int128, y::Int128) = x + y
+checked_sub(x::Int128, y::Int128) = x - y
+checked_mul(x::Int128, y::Int128) = x * y
+
+checked_add(x::Uint128, y::Uint128) = x + y
+checked_sub(x::Uint128, y::Uint128) = x - y
+checked_mul(x::Uint128, y::Uint128) = x * y
