@@ -1,35 +1,3 @@
-## integer functions ##
-
-isodd(n::Integer) = bool(rem(n,2))
-iseven(n::Integer) = !isodd(n)
-
-signbit(x::Unsigned) = 0
-signbit(x::Int8) = int(x>>>7)
-signbit(x::Int16) = int(x>>>15)
-signbit(x::Int32) = int(x>>>31)
-signbit(x::Int64) = int(x>>>63)
-signbit(x::Int128) = int(x>>>127)
-
-flipsign(x::Int,    y::Int)    = box(Int,flipsign_int(unbox(Int,x),unbox(Int,y)))
-flipsign(x::Int64,  y::Int64)  = box(Int64,flipsign_int(unbox(Int64,x),unbox(Int64,y)))
-flipsign(x::Int128, y::Int128) = box(Int128,flipsign_int(unbox(Int128,x),unbox(Int128,y)))
-
-flipsign{T<:Signed}(x::T,y::T)  = flipsign(int(x),int(y))
-flipsign(x::Signed, y::Signed)  = flipsign(promote(x,y)...)
-flipsign(x::Signed, y::Float32) = flipsign(x, reinterpret(Int32,y))
-flipsign(x::Signed, y::Float64) = flipsign(x, reinterpret(Int64,y))
-flipsign(x::Signed, y::Real)    = flipsign(x, -oftype(x,signbit(y)))
-
-copysign(x::Signed, y::Signed)  = flipsign(x, x$y)
-copysign(x::Signed, y::Float32) = copysign(x, reinterpret(Int32,y))
-copysign(x::Signed, y::Float64) = copysign(x, reinterpret(Int64,y))
-copysign(x::Signed, y::Real)    = copysign(x, -oftype(x,signbit(y)))
-
-abs(x::Unsigned) = x
-abs(x::Signed) = flipsign(x,x)
-
-~(n::Integer) = -n-1
-
 ## number-theoretic functions ##
 
 function gcd{T<:Integer}(a::T, b::T)
@@ -51,6 +19,9 @@ lcm(a::Integer, b::Integer) = lcm(promote(a,b)...)
 gcd(a::Integer, b::Integer...) = gcd(a, gcd(b...))
 lcm(a::Integer, b::Integer...) = lcm(a, lcm(b...))
 
+gcd{T<:Integer}(abc::AbstractArray{T}) = reduce(gcd,abc)
+lcm{T<:Integer}(abc::AbstractArray{T}) = reduce(lcm,abc)
+
 # return (gcd(a,b),x,y) such that ax+by == gcd(a,b)
 function gcdx{T<:Integer}(a::T, b::T)
     s0, s1 = one(T), zero(T)
@@ -88,27 +59,26 @@ function power_by_squaring(x, p::Integer)
     elseif p < 0
         throw(DomainError())
     end
-    t = 1
-    while t <= p
-        t *= 2
+    t = trailing_zeros(p) + 1
+    p >>= t
+    while (t -= 1) > 0
+        x *= x
     end
-    t = div(t,2)
-    p -= t
-    a = x
-    while true
-        t = div(t,2)
-        if t > 0
-            x = x*x
-        else
-            break
-        end
-
-        if p >= t
-            x = x*a
-            p -= t
+    y = x
+    while p > 0
+        t = trailing_zeros(p) + 1
+        p >>= t
+        while (t -= 1) >= 0
+            x *= x
         end
+        y *= x
     end
-    return x
+    return y
+end
+power_by_squaring(x::Bool, p::Unsigned) = ((p==0) | x)
+function power_by_squaring(x::Bool, p::Integer)
+    p < 0 && throw(DomainError())
+    return (p==0) | x
 end
 
 ^{T<:Integer}(x::T, p::T) = power_by_squaring(x,p)
@@ -144,17 +114,23 @@ nextpow2(x::Integer) = oftype(x,x < 0 ? -nextpow2(unsigned(-x)) : nextpow2(unsig
 prevpow2(x::Unsigned) = (one(x)>>(x==0)) << ((sizeof(x)<<3)-leading_zeros(x)-1)
 prevpow2(x::Integer) = oftype(x,x < 0 ? -prevpow2(unsigned(-x)) : prevpow2(unsigned(x)))
 
-ispow2(x::Integer) = ((x<=0) == (x&(x-1)))
+ispow2(x::Integer) = count_ones(x)==1
 
-# smallest integer n for which a^n >= x
-function nextpow(a, x)
-    n = iceil(log(x) ./ log(a))
-    return n - int(a.^(n-1) .>= x) # guard against roundoff error, e.g., with a=5 and x=125
+# smallest a^n >= x, with integer n
+function nextpow(a::Real, x::Real)
+    (a <= 1 || x <= 0) && throw(DomainError())
+    x <= 1 && return one(a)
+    n = iceil(log(a, x))
+    p = a^(n-1)
+    # guard against roundoff error, e.g., with a=5 and x=125
+    p >= x ? p : a^n
 end
-# largest integer n for which a^n <= x
-function prevpow(a, x)
-    n = ifloor(log(x) ./ log(a))
-    return n + int(a.^(n+1) .<= x)
+# largest a^n <= x, with integer n
+function prevpow(a::Real, x::Real)
+    (a <= 1 || x < 1) && throw(DomainError())
+    n = ifloor(log(a, x))
+    p = a^(n+1)
+    p <= x ? p : a^n
 end
 
 # decimal digits in an unsigned integer
@@ -199,12 +175,12 @@ function ndigits0z(n::Unsigned, b::Int)
     end
     return d
 end
-ndigits0z(x::Integer, b::Integer) = ndigits0z(unsigned(abs(x)),b)
+ndigits0z(x::Integer, b::Integer) = ndigits0z(unsigned(abs(x)),int(b))
 
-ndigits(x::Unsigned, b::Integer) = x==0 ? 1 : ndigits0z(x,b)
+ndigits(x::Unsigned, b::Integer) = x==0 ? 1 : ndigits0z(x,int(b))
 ndigits(x::Unsigned)             = x==0 ? 1 : ndigits0z(x)
 
-ndigits(x::Integer, b::Integer) = ndigits(unsigned(abs(x)),b)
+ndigits(x::Integer, b::Integer) = ndigits(unsigned(abs(x)),int(b))
 ndigits(x::Integer) = ndigits(unsigned(abs(x)))
 
 ## integer to string functions ##
@@ -245,13 +221,12 @@ function dec(x::Unsigned, pad::Int, neg::Bool)
     ASCIIString(a)
 end
 
-digit(x::Integer) = '0'+x+39*(x>9)
-
 function hex(x::Unsigned, pad::Int, neg::Bool)
     i = neg + max(pad,(sizeof(x)<<1)-(leading_zeros(x)>>2))
     a = Array(Uint8,i)
     while i > neg
-        a[i] = digit(x&0xf)
+        d = x & 0xf
+        a[i] = '0'+d+39*(d>9)
         x >>= 4
         i -= 1
     end
@@ -261,25 +236,30 @@ end
 
 num2hex(n::Integer) = hex(n, sizeof(n)*2)
 
+const base36digits = ['0':'9','a':'z']
+const base62digits = ['0':'9','A':'Z','a':'z']
+
 function base(b::Int, x::Unsigned, pad::Int, neg::Bool)
     if !(2 <= b <= 62) error("invalid base: $b") end
+    digits = b <= 36 ? base36digits : base62digits
     i = neg + max(pad,ndigits0z(x,b))
     a = Array(Uint8,i)
     while i > neg
-        a[i] = digit(rem(x,b))
+        a[i] = digits[1+rem(x,b)]
         x = div(x,b)
         i -= 1
     end
     if neg; a[1]='-'; end
     ASCIIString(a)
 end
-base(b::Integer, n::Integer, pad::Integer=1) =
-    base(int(b), unsigned(abs(n)), pad, n<0)
+base(b::Integer, n::Integer, pad::Integer=1) = base(int(b), unsigned(abs(n)), pad, n<0)
 
 for sym in (:bin, :oct, :dec, :hex)
     @eval begin
         ($sym)(x::Unsigned, p::Int) = ($sym)(x,p,false)
         ($sym)(x::Unsigned)         = ($sym)(x,1,false)
+        ($sym)(x::Char, p::Int)     = ($sym)(unsigned(x),p,false)
+        ($sym)(x::Char)             = ($sym)(unsigned(x),1,false)
         ($sym)(x::Integer, p::Int)  = ($sym)(unsigned(abs(x)),p,x<0)
         ($sym)(x::Integer)          = ($sym)(unsigned(abs(x)),1,x<0)
     end
@@ -303,3 +283,12 @@ function digits{T<:Integer}(n::Integer, base::T=10, pad::Int=1)
 end
 
 isqrt(x::Integer) = oftype(x, trunc(sqrt(x)))
+
+function isqrt(x::Union(Int64,Uint64,Int128,Uint128))
+    x==0 && return x
+    s = oftype(x, trunc(sqrt(x)))
+    # fix with a Newton iteration, since conversion to float discards
+    # too many bits.
+    s = (s + div(x,s)) >> 1
+    s*s > x ? s-1 : s
+end
