@@ -6,7 +6,7 @@ type IntSet
     IntSet() = new(zeros(Uint32,256>>>5), 256, false)
 end
 
-IntSet(args...) = (s=IntSet(); for a in args; push!(s,a); end; s)
+IntSet(itr) = (s=IntSet(); for a in itr; push!(s,a); end; s)
 
 similar(s::IntSet) = IntSet()
 
@@ -36,6 +36,8 @@ function push!(s::IntSet, n::Integer)
             lim = int(n + div(n,2))
             sizehint(s, lim)
         end
+    elseif n < 0
+        throw(ArgumentError("IntSet elements cannot be negative"))
     end
     s.bits[n>>5 + 1] |= (uint32(1)<<(n&31))
     return s
@@ -88,7 +90,7 @@ function setdiff!(s::IntSet, ns)
 end
 
 setdiff(a::IntSet, b::IntSet) = setdiff!(copy(a),b)
-symdiff(a::IntSet, b::IntSet) =
+symdiff(s1::IntSet, s2::IntSet) =
     (s1.limit >= s2.limit ? symdiff!(copy(s1), s2) : symdiff!(copy(s2), s1))
 
 function empty!(s::IntSet)
@@ -100,6 +102,8 @@ function symdiff!(s::IntSet, n::Integer)
     if n >= s.limit
         lim = int(n + dim(n,2))
         sizehint(s, lim)
+    elseif n < 0
+        throw(ArgumentError("IntSet elements cannot be negative"))
     end
     s.bits[n>>5 + 1] $= (uint32(1)<<(n&31))
     return s
@@ -122,6 +126,8 @@ function in(n::Integer, s::IntSet)
         # max IntSet length is typemax(Int), so highest possible element is
         # typemax(Int)-1
         s.fill1s && n >= 0 && n < typemax(Int)
+    elseif n < 0
+        return false
     else
         (s.bits[n>>5 + 1] & (uint32(1)<<(n&31))) != 0
     end
@@ -144,7 +150,7 @@ isempty(s::IntSet) =
 function first(s::IntSet)
     n = next(s,0)[1]
     if n >= s.limit
-        error("first: set is empty")
+        error("set must be non-empty")
     end
     return n
 end
@@ -160,14 +166,14 @@ function last(s::IntSet)
             end
         end
     end
-    error("last: set has no last element")
+    error("set has no last element")
 end
 
 length(s::IntSet) = int(ccall(:bitvector_count, Uint64, (Ptr{Uint32}, Uint64, Uint64), s.bits, 0, s.limit)) +
     (s.fill1s ? typemax(Int) - s.limit : 0)
 
 function show(io::IO, s::IntSet)
-    print(io, "IntSet(")
+    print(io, "IntSet([")
     first = true
     for n in s
         if n > s.limit
@@ -182,7 +188,7 @@ function show(io::IO, s::IntSet)
     if s.fill1s
         print(io, ", ..., ", typemax(Int)-1, ")")
     else
-        print(io, ")")
+        print(io, "])")
     end
 end
 
@@ -258,7 +264,7 @@ function symdiff!(s::IntSet, s2::IntSet)
     s
 end
 
-function isequal(s1::IntSet, s2::IntSet)
+function ==(s1::IntSet, s2::IntSet)
     if s1.fill1s != s2.fill1s
         return false
     end
@@ -287,3 +293,5 @@ function isequal(s1::IntSet, s2::IntSet)
 end
 
 issubset(a::IntSet, b::IntSet) = isequal(a, intersect(a,b))
+<(a::IntSet, b::IntSet) = (a<=b) && !isequal(a,b)
+<=(a::IntSet, b::IntSet) = issubset(a, b)
