@@ -34,8 +34,8 @@ type TmStruct
 end
 
 strftime(t) = strftime("%c", t)
-strftime(fmt::ByteString, t::Real) = strftime(fmt, TmStruct(t))
-function strftime(fmt::ByteString, tm::TmStruct)
+strftime(fmt::String, t::Real) = strftime(fmt, TmStruct(t))
+function strftime(fmt::String, tm::TmStruct)
     timestr = Array(Uint8, 128)
     n = ccall(:strftime, Int, (Ptr{Uint8}, Int, Ptr{Uint8}, Ptr{Void}),
               timestr, length(timestr), fmt, &tm)
@@ -45,8 +45,8 @@ function strftime(fmt::ByteString, tm::TmStruct)
     bytestring(convert(Ptr{Uint8},timestr))
 end
 
-strptime(timestr::ByteString) = strptime("%c", timestr)
-function strptime(fmt::ByteString, timestr::ByteString)
+strptime(timestr::String) = strptime("%c", timestr)
+function strptime(fmt::String, timestr::String)
     tm = TmStruct()
     r = ccall(:strptime, Ptr{Uint8}, (Ptr{Uint8}, Ptr{Uint8}, Ptr{Void}),
               timestr, fmt, &tm)
@@ -55,7 +55,14 @@ function strptime(fmt::ByteString, timestr::ByteString)
     # exposed in the API.
     # tm.isdst = -1
     if r == C_NULL
-        error("strptime: invalid arguments")
+        error("invalid arguments")
+    end
+    @osx_only begin
+        # if we didn't explicitly parse the weekday or year day, use mktime
+        # to fill them in automatically.
+        if !ismatch(r"([^%]|^)%(a|A|j|w|Ow)", fmt)
+            ccall(:mktime, Int, (Ptr{Void},), &tm)
+        end
     end
     tm
 end
@@ -72,9 +79,7 @@ function gethostname()
     hn = Array(Uint8, 256)
     @unix_only err=ccall(:gethostname, Int32, (Ptr{Uint8}, Uint), hn, length(hn))
     @windows_only err=ccall(:gethostname, stdcall, Int32, (Ptr{Uint8}, Uint32), hn, length(hn))
-    if err != 0
-        error("gethostname")
-    end
+    systemerror("gethostname", err != 0)
     bytestring(convert(Ptr{Uint8},hn))
 end
 
@@ -82,3 +87,5 @@ end
 
 c_free(p::Ptr) = ccall(:free, Void, (Ptr{Void},), p)
 c_malloc(size::Integer) = ccall(:malloc, Ptr{Void}, (Csize_t,), size)
+c_realloc(p::Ptr, size::Integer) = ccall(:realloc, Ptr{Void}, (Ptr{Void}, Csize_t), p, size)
+c_calloc(num::Integer, size::Integer) = ccall(:calloc, Ptr{Void}, (Csize_t, Csize_t), num, size)
