@@ -1,7 +1,7 @@
 module LinAlg
 
 importall Base
-import Base.USE_BLAS64, Base.size, Base.copy, Base.copy_transpose!, Base.power_by_squaring, Base.print_matrix
+import Base: USE_BLAS64, size, copy, copy_transpose!, power_by_squaring, print_matrix, transpose!
 
 export 
 # Modules
@@ -9,7 +9,6 @@ export
     BLAS,
 
 # Types
-    BunchKaufman,
     SymTridiagonal,
     Tridiagonal,
     Bidiagonal,
@@ -19,12 +18,12 @@ export
     Cholesky,
     CholeskyPivoted,
     Eigen,
+    GeneralizedEigen,
     GeneralizedSVD,
     GeneralizedSchur,
     Hessenberg,
     LU,
-    LUTridiagonal,
-    LDLTTridiagonal,
+    LDLt,
     QR,
     QRPivoted,
     Schur,
@@ -33,17 +32,17 @@ export
     Symmetric,
     Triangular,
     Diagonal,
+    UniformScaling,
 
 # Functions
+    axpy!,
     bkfact,
     bkfact!,
-    check_blas,
     chol,
     cholfact,
     cholfact!,
-    cholpfact,
-    cholpfact!,
     cond,
+    condskeel,
     copy!,
     cross,
     ctranspose,
@@ -65,7 +64,7 @@ export
     sqrtm,
     eye,
     factorize,
-    factorize!,
+    givens,
     gradient,
     hessfact,
     hessfact!,
@@ -76,25 +75,21 @@ export
     istril,
     istriu,
     kron,
-    ldltd!,
-    ldltd,
+    ldltfact!,
+    ldltfact,
     linreg,
     logdet,
     lu,
     lufact,
     lufact!,
+    lyap,
     norm,
-    normfro,
     null,
     peakflops,
     pinv,
     qr,
     qrfact!,
     qrfact,
-    qrp,
-    qrpfact!,
-    qrpfact,
-    randsym,
     rank,
     rref,
     scale,
@@ -102,19 +97,19 @@ export
     schur,
     schurfact!,
     schurfact,
-    solve,
     svd,
     svdfact!,
     svdfact,
     svdvals!,
     svdvals,
-    symmetrize!,
+    sylvester,
     trace,
     transpose,
     tril,
     triu,
     tril!,
     triu!,
+    vecnorm,
 
 # Operators
     \,
@@ -122,7 +117,6 @@ export
     A_ldiv_B!,
     A_ldiv_Bc,
     A_ldiv_Bt,
-    A_mul_B,
     A_mul_B!,
     A_mul_Bc,
     A_mul_Bc!,
@@ -132,7 +126,6 @@ export
     A_rdiv_Bt,
     Ac_ldiv_B,
     Ac_ldiv_Bc,
-    Ac_mul_b_RFP,
     Ac_mul_B,
     Ac_mul_B!,
     Ac_mul_Bc,
@@ -146,7 +139,10 @@ export
     At_mul_Bt,
     At_mul_Bt!,
     At_rdiv_B,
-    At_rdiv_Bt
+    At_rdiv_Bt,
+
+# Constants
+    I
 
 typealias BlasFloat Union(Float64,Float32,Complex128,Complex64)
 typealias BlasReal Union(Float64,Float32)
@@ -161,6 +157,36 @@ else
     blas_int(x) = int32(x)
 end
 
+#Check that stride of matrix/vector is 1
+function chkstride1(A::StridedVecOrMat...)
+    for a in A 
+        stride(a,1)== 1 || error("Matrix does not have contiguous columns")
+    end  
+end
+
+#Check that matrix is square
+function chksquare(A::AbstractMatrix)
+    m,n = size(A)
+    m == n || throw(DimensionMismatch("Matrix is not square"))
+    m
+end
+
+function chksquare(A...)
+    sizes=Int[]
+    for a in A 
+        size(a,1)==size(a,2) || throw(DimensionMismatch("Matrix is not square: dimensions are $(size(a))"))
+        push!(sizes, size(a,1))
+    end
+    length(A)==1 ? sizes[1] : sizes
+end
+
+#Check that upper/lower (for special matrices) is correctly specified
+macro chkuplo()
+   :((uplo=='U' || uplo=='L') || throw(ArgumentError("""invalid uplo = $uplo
+            
+Valid choices are 'U' (upper) or 'L' (lower).""")))
+end
+
 include("linalg/exceptions.jl")
 include("linalg/generic.jl")
 
@@ -169,19 +195,22 @@ include("linalg/matmul.jl")
 include("linalg/lapack.jl")
 
 include("linalg/dense.jl")
+include("linalg/tridiag.jl")
+include("linalg/triangular.jl")
 include("linalg/factorization.jl")
+include("linalg/lu.jl")
 
 include("linalg/bunchkaufman.jl")
-include("linalg/triangular.jl")
-include("linalg/hermitian.jl")
 include("linalg/symmetric.jl")
 include("linalg/woodbury.jl")
-include("linalg/tridiag.jl")
-include("linalg/bidiag.jl")
 include("linalg/diagonal.jl")
+include("linalg/bidiag.jl")
+include("linalg/uniformscaling.jl")
 include("linalg/rectfullpacked.jl")
-
+include("linalg/givens.jl")
+include("linalg/special.jl")
 include("linalg/bitarray.jl")
+include("linalg/ldlt.jl")
 
 include("linalg/sparse.jl")
 include("linalg/umfpack.jl")
@@ -190,7 +219,8 @@ include("linalg/cholmod.jl")
 include("linalg/arpack.jl")
 include("linalg/arnoldi.jl")
 
-function init()
+function __init__()
+    Base.check_blas()
     if Base.blas_vendor() == :mkl
         ccall((:MKL_Set_Interface_Layer, Base.libblas_name), Void, (Cint,), USE_BLAS64 ? 1 : 0)
     end
