@@ -2,8 +2,7 @@ export LAPACKException,
        ARPACKException,
        SingularException,
        PosDefException,
-       RankDeficientException,
-       DimensionMismatch
+       RankDeficientException
 
 type LAPACKException <: Exception
     info::BlasInt
@@ -24,7 +23,3 @@ end
 type RankDeficientException <: Exception
     info::BlasInt
 end
-
-type DimensionMismatch <: Exception
-    name::ASCIIString
-end
