@@ -24,6 +24,14 @@ end
 find_in_node1_path(name) = myid()==1 ?
     find_in_path(name) : remotecall_fetch(1, find_in_path, name)
 
+function find_source_file(file)
+    (isabspath(file) || isfile(file)) && return file
+    file2 = find_in_path(file)
+    file2 != nothing && return file2
+    file2 = joinpath(JULIA_HOME, DATAROOTDIR, "julia", "base", file)
+    isfile(file2) ? file2 : nothing
+end
+
 # Store list of files and their load time
 package_list = (ByteString=>Float64)[]
 # to synchronize multiple tasks trying to require something
@@ -55,8 +63,11 @@ function _require(path)
     else
         last = toplevel_load
         toplevel_load = false
-        reload_path(path)
-        toplevel_load = last
+        try 
+            reload_path(path)
+        finally 
+            toplevel_load = last
+        end
     end
 end
 
@@ -70,8 +81,11 @@ function reload(name::String)
     end
     last = toplevel_load
     toplevel_load = false
-    reload_path(path)
-    toplevel_load = last
+    try 
+        reload_path(path)
+    finally 
+        toplevel_load = last
+    end
     if refs !== nothing
         for r in refs; wait(r); end
     end
@@ -80,10 +94,10 @@ end
 
 # remote/parallel load
 
-include_string(txt::ByteString, fname::ByteString) =
+include_string(txt::String, fname::String) =
     ccall(:jl_load_file_string, Any, (Ptr{Uint8},Ptr{Uint8}), txt, fname)
 
-include_string(txt::ByteString) = include_string(txt, "string")
+include_string(txt::String) = include_string(txt, "string")
 
 function source_path(default::Union(String,Nothing)="")
     t = current_task()
@@ -114,9 +128,7 @@ function include_from_node1(path::String)
             result = Core.include(path)
             nprocs()>1 && sleep(0.005)
         else
-            include_string(remotecall_fetch(1, readall, path), path)
-            # don't bother sending last value for remote include
-            result = nothing
+            result = include_string(remotecall_fetch(1, readall, path), path)
         end
     finally
         if prev == nothing
@@ -147,12 +159,17 @@ function reload_path(path::String)
         end
     end
     if !isready(package_locks[path])
-        put(package_locks[path],nothing)
+        put!(package_locks[path],nothing)
     end
     nothing
 end
 
-function evalfile(path::String, args::Array = {})
+function evalfile(path::String, args::Vector{UTF8String}=UTF8String[])
     return eval(Module(:__anon__),
-                Expr(:toplevel,:(ARGS=$args),:(include($path))))
+                Expr(:toplevel,
+                     :(const ARGS = $args),
+                     :(eval(x) = Core.eval(__anon__,x)),
+                     :(eval(m,x) = Core.eval(m,x)),
+                     :(include($path))))
 end
+evalfile(path::String, args::Vector) = evalfile(path, UTF8String[args...])
