@@ -1,15 +1,15 @@
 ### Generic interface ###
 
 # Arrays
-mmap_array{T,N}(::Type{T}, dims::NTuple{N,Int}, s::IO) = mmap_array(T, dims, s, position(s))
+mmap_array{T,N}(::Type{T}, dims::NTuple{N,Integer}, s::IO) = mmap_array(T, dims, s, position(s))
 
 msync{T}(A::Array{T}) = msync(pointer(A), length(A)*sizeof(T))
 
 # BitArrays
-mmap_bitarray{N}(::Type{Bool}, dims::NTuple{N,Int}, s::IOStream, offset::FileOffset) =
+mmap_bitarray{N}(::Type{Bool}, dims::NTuple{N,Integer}, s::IOStream, offset::FileOffset) =
     mmap_bitarray(dims, s, offset)
-mmap_bitarray{N}(::Type{Bool}, dims::NTuple{N,Int}, s::IOStream) = mmap_bitarray(dims, s, position(s))
-mmap_bitarray{N}(dims::NTuple{N,Int}, s::IOStream) = mmap_bitarray(dims, s, position(s))
+mmap_bitarray{N}(::Type{Bool}, dims::NTuple{N,Integer}, s::IOStream) = mmap_bitarray(dims, s, position(s))
+mmap_bitarray{N}(dims::NTuple{N,Integer}, s::IOStream) = mmap_bitarray(dims, s, position(s))
 
 msync(B::BitArray) = msync(pointer(B.chunks), length(B.chunks)*sizeof(Uint64))
 
@@ -18,20 +18,21 @@ msync(B::BitArray) = msync(pointer(B.chunks), length(B.chunks)*sizeof(Uint64))
 @unix_only begin
 # Low-level routines
 # These are needed for things like MAP_ANONYMOUS
-function mmap(len::Integer, prot::Integer, flags::Integer, fd::Integer, offset::FileOffset)
+function mmap(len::Integer, prot::Integer, flags::Integer, fd, offset::Integer)
     const pagesize::Int = ccall(:jl_getpagesize, Clong, ())
     # Check that none of the computations will overflow
+    if len < 0
+        error("requested size is negative")
+    end
     if len > typemax(Int)-pagesize
-        error("Cannot map such a large buffer")
+        error("requested size is too large")
     end
     # Set the offset to a page boundary
     offset_page::FileOffset = ifloor(offset/pagesize)*pagesize
     len_page::Int = (offset-offset_page) + len
     # Mmap the file
     p = ccall(:jl_mmap, Ptr{Void}, (Ptr{Void}, Csize_t, Cint, Cint, Cint, FileOffset), C_NULL, len_page, prot, flags, fd, offset_page)
-    if int(p) == -1
-        error("Memory mapping failed", strerror())
-    end
+    systemerror("memory mapping failed", int(p) == -1)
     # Also return a pointer that compensates for any adjustment in the offset
     return p, int(offset-offset_page)
 end
@@ -47,38 +48,26 @@ function mmap_grow(len::Integer, prot::Integer, flags::Integer, fd::Integer, off
     const SEEK_END::Cint = 2
     # Save current file position so we can restore it later
     cpos = ccall(:jl_lseek, FileOffset, (Cint, FileOffset, Cint), fd, 0, SEEK_CUR)
-    if cpos < 0
-        error(strerror())
-    end
+    systemerror("lseek", cpos < 0)
     filelen = ccall(:jl_lseek, FileOffset, (Cint, FileOffset, Cint), fd, 0, SEEK_END)
-    if filelen < 0
-        error(strerror())
-    end
+    systemerror("lseek", filelen < 0)
     if (filelen < offset + len)
-        n = ccall(:jl_pwrite, Cssize_t, (Cint, Ptr{Void}, Uint, FileOffset), fd, int8([0]), 1, offset + len - 1)
-        if (n < 1)
-            error(strerror())
-        end
+        systemerror("pwrite", ccall(:jl_pwrite, Cssize_t, (Cint, Ptr{Void}, Uint, FileOffset), fd, int8([0]), 1, offset + len - 1) < 1)
     end
     cpos = ccall(:jl_lseek, FileOffset, (Cint, FileOffset, Cint), fd, cpos, SEEK_SET)
+    systemerror("lseek", cpos < 0)
     return mmap(len, prot, flags, fd, offset)
 end
 
 function munmap(p::Ptr,len::Integer)
-    ret = ccall(:munmap,Cint,(Ptr{Void},Int),p,len)
-    if ret != 0
-        error(strerror())
-    end
+    systemerror("munmap", ccall(:munmap,Cint,(Ptr{Void},Int),p,len) != 0)
 end
 
 const MS_ASYNC = 1
 const MS_INVALIDATE = 2
 const MS_SYNC = 4
 function msync(p::Ptr, len::Integer, flags::Integer)
-    ret = ccall(:msync, Cint, (Ptr{Void}, Csize_t, Cint), p, len, flags)
-    if ret != 0
-        error(strerror())
-    end
+    systemerror("msync", ccall(:msync, Cint, (Ptr{Void}, Csize_t, Cint), p, len, flags) != 0)
 end
 msync(p::Ptr, len::Integer) = msync(p, len, MS_SYNC)
 
@@ -92,6 +81,7 @@ function mmap_stream_settings(s::IO)
     const MAP_SHARED::Cint = 1
     const F_GETFL::Cint = 3
     mode = ccall(:fcntl,Cint,(Cint,Cint),fd(s),F_GETFL)
+    systemerror("fcntl F_GETFL", mode == -1)
     mode = mode & 3
     if mode == 0
         prot = PROT_READ
@@ -101,25 +91,25 @@ function mmap_stream_settings(s::IO)
         prot = PROT_READ | PROT_WRITE
     end
     if prot & PROT_READ == 0
-        error("For mmap, you need read permissions on the file (choose r+)")
+        error("mmap requires read permissions on the file (choose r+)")
     end
     flags = MAP_SHARED
     return prot, flags, (prot & PROT_WRITE) > 0
 end
 
 # Mmapped-array constructor
-function mmap_array{T,N,TInt<:Integer}(::Type{T}, dims::NTuple{N,TInt}, s::IO, offset::FileOffset)
+function mmap_array{T,N}(::Type{T}, dims::NTuple{N,Integer}, s::IO, offset::FileOffset; grow::Bool=true)
     prot, flags, iswrite = mmap_stream_settings(s)
     len = prod(dims)*sizeof(T)
     if len > typemax(Int)
-        error("File is too large to memory-map on this platform")
+        error("file is too large to memory-map on this platform")
     end
-    if iswrite
+    if iswrite && grow
         pmap, delta = mmap_grow(len, prot, flags, fd(s), offset)
     else
         pmap, delta = mmap(len, prot, flags, fd(s), offset)
     end
-    A = pointer_to_array(pointer(T, uint(pmap)+delta), dims)
+    A = pointer_to_array(convert(Ptr{T}, uint(pmap)+delta), dims)
     finalizer(A,x->munmap(pmap,len+delta))
     return A
 end
@@ -129,29 +119,37 @@ end
 ### Windows implementation ###
 @windows_only begin
 # Mmapped-array constructor
-function mmap_array{T,N,TInt<:Integer}(::Type{T}, dims::NTuple{N,TInt}, s::IO, offset::FileOffset)
+function mmap_array{T,N}(::Type{T}, dims::NTuple{N,Integer}, s::IO, offset::FileOffset)
     shandle = _get_osfhandle(RawFD(fd(s)))
     if int(shandle.handle) == -1
-        error("Could not get handle")
+        error("could not get handle for file to map: $(FormatMessage())")
     end
     ro = isreadonly(s)
     flprotect = ro ? 0x02 : 0x04
     len = prod(dims)*sizeof(T)
-    if len > typemax(Int)
-        error("File is too large to memory-map on this platform")
+    const granularity::Int = ccall(:jl_getallocationgranularity, Clong, ())
+    if len < 0
+        error("requested size is negative")
     end
-    szarray = convert(Csize_t, len)
-    szfile = szarray + convert(Csize_t, offset)
-    mmaphandle = ccall(:CreateFileMappingA, stdcall, Ptr{Void}, (Ptr{Void}, Ptr{Void}, Cint, Cint, Cint, Ptr{Void}), shandle.handle, C_NULL, flprotect, szfile>>32, szfile&0xffffffff, C_NULL)
+    if len > typemax(Int)-granularity
+        error("file is too large to memory-map on this platform")
+    end
+    # Set the offset to a page boundary
+    offset_page::FileOffset = div(offset, granularity)*granularity
+    szfile = convert(Csize_t, len + offset)
+    szarray = szfile - convert(Csize_t, offset_page)    
+    mmaphandle = ccall(:CreateFileMappingW, stdcall, Ptr{Void}, (Ptr{Void}, Ptr{Void}, Cint, Cint, Cint, Ptr{Uint16}),
+        shandle.handle, C_NULL, flprotect, szfile>>32, szfile&typemax(Uint32), C_NULL)
     if mmaphandle == C_NULL
-        error("Could not create file mapping")
+        error("could not create file mapping: $(FormatMessage())")
     end
     access = ro ? 4 : 2
-    viewhandle = ccall(:MapViewOfFile, stdcall, Ptr{Void}, (Ptr{Void}, Cint, Cint, Cint, Csize_t), mmaphandle, access, offset>>32, offset&0xffffffff, szarray)
+    viewhandle = ccall(:MapViewOfFile, stdcall, Ptr{Void}, (Ptr{Void}, Cint, Cint, Cint, Csize_t),
+        mmaphandle, access, offset_page>>32, offset_page&typemax(Uint32), szarray)
     if viewhandle == C_NULL
-        error("Could not create mapping view")
+        error("could not create mapping view: $(FormatMessage())")
     end
-    A = pointer_to_array(pointer(T, viewhandle), dims)
+    A = pointer_to_array(convert(Ptr{T}, viewhandle+offset-offset_page), dims)
     finalizer(A, x->munmap(viewhandle, mmaphandle))
     return A
 end
@@ -160,21 +158,21 @@ function munmap(viewhandle::Ptr, mmaphandle::Ptr)
     status = bool(ccall(:UnmapViewOfFile, stdcall, Cint, (Ptr{Void},), viewhandle))
     status |= bool(ccall(:CloseHandle, stdcall, Cint, (Ptr{Void},), mmaphandle))
     if !status
-        error("Could not unmap view")
+        error("could not unmap view: $(FormatMessage())")
     end
 end
 
 function msync(p::Ptr, len::Integer)
     status = bool(ccall(:FlushViewOfFile, stdcall, Cint, (Ptr{Void}, Csize_t), p, len))
     if !status
-        error("Could not msync")
+        error("could not msync: $(FormatMessage())")
     end
 end
 
 end
 
 # Mmapped-bitarray constructor
-function mmap_bitarray{N,TInt<:Integer}(dims::NTuple{N,TInt}, s::IOStream, offset::FileOffset)
+function mmap_bitarray{N}(dims::NTuple{N,Integer}, s::IOStream, offset::FileOffset)
     iswrite = !isreadonly(s)
     n = 1
     for d in dims
@@ -185,14 +183,14 @@ function mmap_bitarray{N,TInt<:Integer}(dims::NTuple{N,TInt}, s::IOStream, offse
     end
     nc = num_bit_chunks(n)
     if nc > typemax(Int)
-        error("File is too large to memory-map on this platform")
+        error("file is too large to memory-map on this platform")
     end
     chunks = mmap_array(Uint64, (nc,), s, offset)
     if iswrite
         chunks[end] &= @_msk_end n
     else
         if chunks[end] != chunks[end] & @_msk_end n
-            error("The given file does not contain a valid BitArray of size ", join(dims, 'x'), " (open with \"r+\" mode to override)")
+            error("the given file does not contain a valid BitArray of size ", join(dims, 'x'), " (open with \"r+\" mode to override)")
         end
     end
     B = BitArray{N}(ntuple(N,i->0)...)
