@@ -4,22 +4,23 @@ export
     BigFloat,
     get_bigfloat_precision,
     set_bigfloat_precision,
-    with_bigfloat_precision,
-    set_bigfloat_rounding,
-    get_bigfloat_rounding,
-    with_bigfloat_rounding
+    with_bigfloat_precision
 
 import
     Base: (*), +, -, /, <, <=, ==, >, >=, ^, besselj, besselj0, besselj1, bessely,
-        bessely0, bessely1, ceil, cmp, convert, copysign, degrees2radians,
+        bessely0, bessely1, ceil, cmp, convert, copysign, deg2rad,
         exp, exp2, exponent, factorial, floor, hypot, isinteger, iround,
         isfinite, isinf, isnan, ldexp, log, log2, log10, max, min, mod, modf,
-        nextfloat, prevfloat, promote_rule, radians2degrees, rem, round, show,
+        nextfloat, prevfloat, promote_rule, rad2deg, rem, round, show,
         showcompact, sum, sqrt, string, print, trunc, precision, exp10, expm1,
-        gamma, lgamma, digamma, erf, erfc, zeta, log1p, airyai, iceil, ifloor,
+        gamma, lgamma, digamma, erf, erfc, zeta, eta, log1p, airyai, iceil, ifloor,
         itrunc, eps, signbit, sin, cos, tan, sec, csc, cot, acos, asin, atan,
         cosh, sinh, tanh, sech, csch, coth, acosh, asinh, atanh, atan2,
-        serialize, deserialize, inf, nan, hash, cbrt
+        serialize, deserialize, inf, nan, cbrt, typemax, typemin,
+        realmin, realmax, get_rounding, set_rounding, maxintfloat, widen,
+        significand, frexp
+
+import Base.GMP: ClongMax, CulongMax, CdoubleMax
 
 import Base.Math.lgamma_r
 
@@ -32,12 +33,12 @@ type BigFloat <: FloatingPoint
     prec::Clong
     sign::Cint
     exp::Clong
-    d::Ptr{Void}
+    d::Ptr{Culong}
     function BigFloat()
         N = get_bigfloat_precision()
         z = new(zero(Clong), zero(Cint), zero(Clong), C_NULL)
         ccall((:mpfr_init2,:libmpfr), Void, (Ptr{BigFloat}, Clong), &z, N)
-        finalizer(z, MPFR_clear)
+        finalizer(z, Base.GMP._mpfr_clear_func)
         return z
     end
     # Not recommended for general use
@@ -45,7 +46,9 @@ type BigFloat <: FloatingPoint
         new(prec, sign, exp, d)
     end
 end
-MPFR_clear(mpfr::BigFloat) = ccall((:mpfr_clear, :libmpfr), Void, (Ptr{BigFloat},), &mpfr)
+
+widen(::Type{Float64}) = BigFloat
+widen(::Type{BigFloat}) = BigFloat
 
 BigFloat(x::BigFloat) = x
 
@@ -68,7 +71,7 @@ end
 function BigFloat(x::String, base::Int)
     z = BigFloat()
     err = ccall((:mpfr_set_str, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{Uint8}, Int32, Int32), &z, x, base, ROUNDING_MODE[end])
-    if err != 0; error("Invalid input"); end
+    if err != 0; error("incorrectly formatted number"); end
     return z
 end
 BigFloat(x::String) = BigFloat(x, 10)
@@ -78,11 +81,11 @@ BigFloat(x::Integer) = BigFloat(BigInt(x))
 BigFloat(x::Union(Bool,Int8,Int16,Int32)) = BigFloat(convert(Clong,x))
 BigFloat(x::Union(Uint8,Uint16,Uint32)) = BigFloat(convert(Culong,x))
 
-BigFloat(x::Float32) = BigFloat(float64(x))
+BigFloat(x::Union(Float16,Float32)) = BigFloat(float64(x))
 BigFloat(x::Rational) = BigFloat(num(x)) / BigFloat(den(x))
 
 convert(::Type{Rational}, x::BigFloat) = convert(Rational{BigInt}, x)
-convert(::Type{BigFloat}, x::Rational) = BigFloat(x) # to resolve ambiguity
+convert{S}(::Type{BigFloat}, x::Rational{S}) = BigFloat(x) # to resolve ambiguity
 convert(::Type{BigFloat}, x::Real) = BigFloat(x)
 convert(::Type{FloatingPoint}, x::BigInt) = BigFloat(x)
 
@@ -106,12 +109,9 @@ for to in (Uint8, Uint16, Uint32, Uint64)
     end
 end
 
-function convert(::Type{BigInt}, x::BigFloat)
-    if isinteger(x)
-        return itrunc(x)
-    else
-        throw(InexactError())
-    end
+function Base.BigInt(x::BigFloat)
+    !isinteger(x) && throw(InexactError())
+    return itrunc(x)
 end
 convert(::Type{Float64}, x::BigFloat) =
     ccall((:mpfr_get_d,:libmpfr), Float64, (Ptr{BigFloat},Int32), &x, ROUNDING_MODE[end])
@@ -124,7 +124,13 @@ promote_rule{T<:Real}(::Type{BigFloat}, ::Type{T}) = BigFloat
 promote_rule{T<:FloatingPoint}(::Type{BigInt},::Type{T}) = BigFloat
 promote_rule{T<:FloatingPoint}(::Type{BigFloat},::Type{T}) = BigFloat
 
-rationalize(x::BigFloat; tol::Real=eps(x)) = rationalize(BigInt, x, tol=tol)
+function convert(::Type{Rational{BigInt}}, x::FloatingPoint)
+    if isnan(x); return zero(BigInt)//zero(BigInt); end
+    if isinf(x); return copysign(one(BigInt),x)//zero(BigInt); end
+    if x == 0;   return zero(BigInt) // one(BigInt); end
+    s = max(precision(x) - exponent(x), 0)
+    BigInt(ldexp(x,s)) // (BigInt(1) << s)
+end
 
 # serialization
 
@@ -136,195 +142,111 @@ end
 deserialize(s, ::Type{BigFloat}) = BigFloat(deserialize(s))
 
 # Basic arithmetic without promotion
-# Unsigned addition
-function +(x::BigFloat, c::Culong)
-    z = BigFloat()
-    ccall((:mpfr_add_ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-+(c::Culong, x::BigFloat) = x + c
-+(c::Unsigned, x::BigFloat) = x + convert(Culong, c)
-+(x::BigFloat, c::Unsigned) = x + convert(Culong, c)
+for (fJ, fC) in ((:+,:add), (:*,:mul))
+    @eval begin 
+        # BigFloat
+        function ($fJ)(x::BigFloat, y::BigFloat)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC)),:libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, &y, ROUNDING_MODE[end])
+            return z
+        end
 
-# Signed addition
-function +(x::BigFloat, c::Clong)
-    z = BigFloat()
-    ccall((:mpfr_add_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-+(c::Clong, x::BigFloat) = x + c
-+(x::BigFloat, c::Signed) = x + convert(Clong, c)
-+(c::Signed, x::BigFloat) = x + convert(Clong, c)
+        # Unsigned Integer
+        function ($fJ)(x::BigFloat, c::CulongMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_ui)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        ($fJ)(c::CulongMax, x::BigFloat) = ($fJ)(x,c)
 
-# Float64 addition
-function +(x::BigFloat, c::Float64)
-    z = BigFloat()
-    ccall((:mpfr_add_d, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Float64, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-+(c::Float64, x::BigFloat) = x + c
-+(c::Float32, x::BigFloat) = x + convert(Float64, c)
-+(x::BigFloat, c::Float32) = x + convert(Float64, c)
+        # Signed Integer
+        function ($fJ)(x::BigFloat, c::ClongMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_si)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        ($fJ)(c::ClongMax, x::BigFloat) = ($fJ)(x,c)
 
-# BigInt addition
-function +(x::BigFloat, c::BigInt)
-    z = BigFloat()
-    ccall((:mpfr_add_z, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
-    return z
-end
-+(c::BigInt, x::BigFloat) = x + c
+        # Float32/Float64
+        function ($fJ)(x::BigFloat, c::CdoubleMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_d)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Cdouble, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        ($fJ)(c::CdoubleMax, x::BigFloat) = ($fJ)(x,c)
 
-# Unsigned subtraction
-function -(x::BigFloat, c::Culong)
-    z = BigFloat()
-    ccall((:mpfr_sub_ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
+        # BigInt
+        function ($fJ)(x::BigFloat, c::BigInt)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_z)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
+            return z
+        end
+        ($fJ)(c::BigInt, x::BigFloat) = ($fJ)(x,c)
+    end
 end
-function -(c::Culong, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_ui_sub, :libmpfr), Int32, (Ptr{BigFloat}, Culong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
--(x::BigFloat, c::Unsigned) = -(x, convert(Culong, c))
--(c::Unsigned, x::BigFloat) = -(convert(Culong, c), x)
 
-# Signed subtraction
-function -(x::BigFloat, c::Clong)
-    z = BigFloat()
-    ccall((:mpfr_sub_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-function -(c::Clong, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_si_sub, :libmpfr), Int32, (Ptr{BigFloat}, Clong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
--(x::BigFloat, c::Signed) = -(x, convert(Clong, c))
--(c::Signed, x::BigFloat) = -(convert(Clong, c), x)
+for (fJ, fC) in ((:-,:sub), (:/,:div))
+    @eval begin 
+        # BigFloat
+        function ($fJ)(x::BigFloat, y::BigFloat)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC)),:libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, &y, ROUNDING_MODE[end])
+            return z
+        end
 
-# Float64 subtraction
-function -(x::BigFloat, c::Float64)
-    z = BigFloat()
-    ccall((:mpfr_sub_d, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Float64, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-function -(c::Float64, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_d_sub, :libmpfr), Int32, (Ptr{BigFloat}, Float64, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
--(x::BigFloat, c::Float32) = -(x, convert(Float64, c))
--(c::Float32, x::BigFloat) = -(convert(Float64, c), x)
+        # Unsigned Int
+        function ($fJ)(x::BigFloat, c::CulongMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_ui)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        function ($fJ)(c::CulongMax, x::BigFloat)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,:ui_,fC)), :libmpfr), Int32, (Ptr{BigFloat}, Culong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
+            return z
+        end
+        
+        # Signed Integer
+        function ($fJ)(x::BigFloat, c::ClongMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_si)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        function ($fJ)(c::ClongMax, x::BigFloat)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,:si_,fC)), :libmpfr), Int32, (Ptr{BigFloat}, Clong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
+            return z
+        end
 
-# BigInt subtraction
-function -(x::BigFloat, c::BigInt)
-    z = BigFloat()
-    ccall((:mpfr_sub_z, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
-    return z
+        # Float32/Float64
+        function ($fJ)(x::BigFloat, c::CdoubleMax)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_d)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Cdouble, Int32), &z, &x, c, ROUNDING_MODE[end])
+            return z
+        end
+        function ($fJ)(c::CdoubleMax, x::BigFloat)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,:d_,fC)), :libmpfr), Int32, (Ptr{BigFloat}, Cdouble, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
+            return z
+        end
+
+        # BigInt
+        function ($fJ)(x::BigFloat, c::BigInt)
+            z = BigFloat()
+            ccall(($(string(:mpfr_,fC,:_z)), :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
+            return z
+        end
+        # no :mpfr_z_div function
+    end
 end
+
 function -(c::BigInt, x::BigFloat)
     z = BigFloat()
     ccall((:mpfr_z_sub, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigInt}, Ptr{BigFloat}, Int32), &z, &c, &x, ROUNDING_MODE[end])
     return z
 end
 
-# Unsigned multiplication
-function *(x::BigFloat, c::Culong)
-    z = BigFloat()
-    ccall((:mpfr_mul_ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-*(c::Culong, x::BigFloat) = x * c
-*(c::Unsigned, x::BigFloat) = x * convert(Culong, c)
-*(x::BigFloat, c::Unsigned) = x * convert(Culong, c)
 
-# Signed multiplication
-function *(x::BigFloat, c::Clong)
-    z = BigFloat()
-    ccall((:mpfr_mul_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-*(c::Clong, x::BigFloat) = x * c
-*(x::BigFloat, c::Signed) = x * convert(Clong, c)
-
-# Float64 multiplication
-function *(x::BigFloat, c::Float64)
-    z = BigFloat()
-    ccall((:mpfr_mul_d, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Float64, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-*(c::Float64, x::BigFloat) = x * c
-*(c::Float32, x::BigFloat) = x * convert(Float64, c)
-*(x::BigFloat, c::Float32) = x * convert(Float64, c)
-
-# BigInt multiplication
-*(c::Signed, x::BigFloat) = x * convert(Clong, c)
-function *(x::BigFloat, c::BigInt)
-    z = BigFloat()
-    ccall((:mpfr_mul_z, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
-    return z
-end
-*(c::BigInt, x::BigFloat) = x * c
-
-# Unsigned division
-function /(x::BigFloat, c::Culong)
-    z = BigFloat()
-    ccall((:mpfr_div_ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-function /(c::Culong, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_ui_div, :libmpfr), Int32, (Ptr{BigFloat}, Culong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
-/(x::BigFloat, c::Unsigned) = /(x, convert(Culong, c))
-/(c::Unsigned, x::BigFloat) = /(convert(Culong, c), x)
-
-# Signed division
-function /(x::BigFloat, c::Clong)
-    z = BigFloat()
-    ccall((:mpfr_div_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-function /(c::Clong, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_si_div, :libmpfr), Int32, (Ptr{BigFloat}, Clong, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
-/(x::BigFloat, c::Signed) = /(x, convert(Clong, c))
-/(c::Signed, x::BigFloat) = /(convert(Clong, c), x)
-
-# Float64 division
-function /(x::BigFloat, c::Float64)
-    z = BigFloat()
-    ccall((:mpfr_div_d, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Float64, Int32), &z, &x, c, ROUNDING_MODE[end])
-    return z
-end
-function /(c::Float64, x::BigFloat)
-    z = BigFloat()
-    ccall((:mpfr_d_div, :libmpfr), Int32, (Ptr{BigFloat}, Float64, Ptr{BigFloat}, Int32), &z, c, &x, ROUNDING_MODE[end])
-    return z
-end
-/(x::BigFloat, c::Float32) = /(x, convert(Float64, c))
-/(c::Float32, x::BigFloat) = /(convert(Float64, c), x)
-
-# BigInt division
-function /(x::BigFloat, c::BigInt)
-    z = BigFloat()
-    ccall((:mpfr_div_z, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigInt}, Int32), &z, &x, &c, ROUNDING_MODE[end])
-    return z
-end
-
-# Basic operations
-for (fJ, fC) in ((:+,:add), (:-,:sub), (:*,:mul), (:/,:div), (:^, :pow))
-    @eval begin 
-        function ($fJ)(x::BigFloat, y::BigFloat)
-            z = BigFloat()
-            ccall(($(string(:mpfr_,fC)),:libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, &y, ROUNDING_MODE[end])
-            return z
-        end
-    end
-end
 
 # More efficient commutative operations
 for (fJ, fC, fI) in ((:+, :add, 0), (:*, :mul, 1))
@@ -359,10 +281,6 @@ function -(x::BigFloat)
     return z
 end
 
-function cmp(x::BigFloat, y::BigFloat)
-    ccall((:mpfr_cmp, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}), &x, &y)
-end
-
 function sqrt(x::BigFloat)
     z = BigFloat()
     ccall((:mpfr_sqrt, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, ROUNDING_MODE[end])
@@ -374,17 +292,22 @@ end
 
 sqrt(x::BigInt) = sqrt(BigFloat(x))
 
-radians2degrees(z::BigFloat) = 180/big(pi)*z
-degrees2radians(z::BigFloat) = big(pi)/180*z
+rad2deg(z::BigFloat) = 180/big(pi)*z
+deg2rad(z::BigFloat) = big(pi)/180*z
 
+function ^(x::BigFloat, y::BigFloat)
+    z = BigFloat()
+    ccall((:mpfr_pow, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, &y, ROUNDING_MODE[end])
+    return z
+end
 
-function ^(x::BigFloat, y::Unsigned)
+function ^(x::BigFloat, y::CulongMax)
     z = BigFloat()
     ccall((:mpfr_pow_ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, y, ROUNDING_MODE[end])
     return z
 end
 
-function ^(x::BigFloat, y::Signed)
+function ^(x::BigFloat, y::ClongMax)
     z = BigFloat()
     ccall((:mpfr_pow_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &x, y, ROUNDING_MODE[end])
     return z
@@ -405,6 +328,19 @@ for f in (:exp, :exp2, :exp10, :expm1, :digamma, :erf, :erfc, :zeta,
     end
 end
 
+# return log(2)
+function big_ln2()
+    c = BigFloat()
+    ccall((:mpfr_const_log2, :libmpfr), Cint, (Ptr{BigFloat}, Int32),
+          &c, MPFR.ROUNDING_MODE[end])
+    return c
+end
+
+function eta(x::BigFloat)
+    x == 1 && return big_ln2()
+    return -zeta(x) * expm1(big_ln2()*(1-x))
+end
+
 function airyai(x::BigFloat)
     z = BigFloat()
     ccall((:mpfr_ai, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Int32), &z, &x, ROUNDING_MODE[end])
@@ -421,8 +357,9 @@ function ldexp(x::BigFloat, n::Culong)
     ccall((:mpfr_mul_2ui, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Culong, Int32), &z, &x, n, ROUNDING_MODE[end])
     return z
 end
-ldexp(x::BigFloat, n::Signed) = ldexp(x, convert(Clong, n))
-ldexp(x::BigFloat, n::Unsigned) = ldexp(x, convert(Culong, n))
+ldexp(x::BigFloat, n::ClongMax) = ldexp(x, convert(Clong, n))
+ldexp(x::BigFloat, n::CulongMax) = ldexp(x, convert(Culong, n))
+ldexp(x::BigFloat, n::Integer) = x*exp2(BigFloat(n))
 
 function besselj0(x::BigFloat)
     z = BigFloat()
@@ -584,8 +521,43 @@ end
 <(x::BigFloat, y::BigFloat) = ccall((:mpfr_less_p, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}), &x, &y) != 0
 >(x::BigFloat, y::BigFloat) = ccall((:mpfr_greater_p, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}), &x, &y) != 0
 
-signbit(x::BigFloat) =
-    int(ccall((:mpfr_signbit, :libmpfr), Int32, (Ptr{BigFloat},), &x)!=0)
+function cmp(x::BigFloat, y::BigInt)
+    isnan(x) && throw(DomainError())
+    ccall((:mpfr_cmp_z, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigInt}), &x, &y)
+end
+function cmp(x::BigFloat, y::ClongMax)
+    isnan(x) && throw(DomainError())
+    ccall((:mpfr_cmp_si, :libmpfr), Int32, (Ptr{BigFloat}, Clong), &x, y)
+end
+function cmp(x::BigFloat, y::CulongMax)
+    isnan(x) && throw(DomainError())
+    ccall((:mpfr_cmp_ui, :libmpfr), Int32, (Ptr{BigFloat}, Culong), &x, y)
+end
+cmp(x::BigFloat, y::Integer) = cmp(x,big(y))
+cmp(x::Integer, y::BigFloat) = -cmp(y,x)
+
+function cmp(x::BigFloat, y::CdoubleMax)
+    (isnan(x) || isnan(y)) && throw(DomainError())
+    ccall((:mpfr_cmp_d, :libmpfr), Int32, (Ptr{BigFloat}, Cdouble), &x, y)
+end
+cmp(x::CdoubleMax, y::BigFloat) = -cmp(y,x)
+
+==(x::BigFloat, y::Integer)   = !isnan(x) && cmp(x,y) == 0
+==(x::Integer, y::BigFloat)   = y == x
+==(x::BigFloat, y::CdoubleMax) = !isnan(x) && !isnan(y) && cmp(x,y) == 0
+==(x::CdoubleMax, y::BigFloat) = y == x
+
+<(x::BigFloat, y::Integer)   = !isnan(x) && cmp(x,y) < 0
+<(x::Integer, y::BigFloat)   = !isnan(y) && cmp(y,x) > 0
+<(x::BigFloat, y::CdoubleMax) = !isnan(x) && !isnan(y) && cmp(x,y) < 0
+<(x::CdoubleMax, y::BigFloat) = !isnan(x) && !isnan(y) && cmp(y,x) > 0
+
+<=(x::BigFloat, y::Integer)   = !isnan(x) && cmp(x,y) <= 0
+<=(x::Integer, y::BigFloat)   = !isnan(y) && cmp(y,x) >= 0
+<=(x::BigFloat, y::CdoubleMax) = !isnan(x) && !isnan(y) && cmp(x,y) <= 0
+<=(x::CdoubleMax, y::BigFloat) = !isnan(x) && !isnan(y) && cmp(y,x) >= 0
+
+signbit(x::BigFloat) = ccall((:mpfr_signbit, :libmpfr), Int32, (Ptr{BigFloat},), &x) != 0
 
 function precision(x::BigFloat)
     return ccall((:mpfr_get_prec, :libmpfr), Clong, (Ptr{BigFloat},), &x)
@@ -599,6 +571,9 @@ function set_bigfloat_precision(x::Int)
     DEFAULT_PRECISION[end] = x
 end
 
+maxintfloat(x::BigFloat) = BigFloat(2)^precision(x)
+maxintfloat(::Type{BigFloat}) = BigFloat(2)^get_bigfloat_precision()
+
 function to_mpfr(r::RoundingMode)
     c = r.code
     if !(0 <= c <= 4)
@@ -614,8 +589,8 @@ function from_mpfr(c::Integer)
     RoundingMode(c)
 end
 
-get_bigfloat_rounding() = from_mpfr(ROUNDING_MODE[end])
-set_bigfloat_rounding(r::RoundingMode) = ROUNDING_MODE[end] = to_mpfr(r)
+get_rounding(::Type{BigFloat}) = from_mpfr(ROUNDING_MODE[end])
+set_rounding(::Type{BigFloat},r::RoundingMode) = ROUNDING_MODE[end] = to_mpfr(r)
 
 function copysign(x::BigFloat, y::BigFloat)
     z = BigFloat()
@@ -631,6 +606,22 @@ function exponent(x::BigFloat)
     return ccall((:mpfr_get_exp, :libmpfr), Clong, (Ptr{BigFloat},), &x) - 1
 end
 
+function frexp(x::BigFloat)
+    z = BigFloat()
+    c = Clong[0]
+    ccall((:mpfr_frexp, :libmpfr), Int32, (Ptr{Clong}, Ptr{BigFloat}, Ptr{BigFloat}, Cint), c, &z, &x, ROUNDING_MODE[end])
+    return (z, c[1])
+end
+
+function significand(x::BigFloat)
+    z = BigFloat()
+    c = Clong[0]
+    ccall((:mpfr_frexp, :libmpfr), Int32, (Ptr{Clong}, Ptr{BigFloat}, Ptr{BigFloat}, Cint), c, &z, &x, ROUNDING_MODE[end])
+    # Double the significand to make it work as Base.significand
+    ccall((:mpfr_mul_si, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Clong, Int32), &z, &z, 2, ROUNDING_MODE[end])
+    return z
+end
+
 function isinteger(x::BigFloat)
     return ccall((:mpfr_integer_p, :libmpfr), Int32, (Ptr{BigFloat},), &x) != 0
 end
@@ -666,6 +657,9 @@ isfinite(x::BigFloat) = !isinf(x) && !isnan(x)
 @eval inf(::Type{BigFloat}) = $(BigFloat(Inf))
 @eval nan(::Type{BigFloat}) = $(BigFloat(NaN))
 
+typemax(::Type{BigFloat}) = inf(BigFloat)
+@eval typemin(::Type{BigFloat}) = $(BigFloat(-Inf))
+
 function nextfloat(x::BigFloat)
     z = BigFloat()
     ccall((:mpfr_set, :libmpfr), Int32, (Ptr{BigFloat}, Ptr{BigFloat}, Int32),
@@ -684,6 +678,9 @@ end
 
 eps(::Type{BigFloat}) = nextfloat(BigFloat(1)) - BigFloat(1)
 
+realmin(::Type{BigFloat}) = nextfloat(zero(BigFloat))
+realmax(::Type{BigFloat}) = prevfloat(inf(BigFloat))
+
 function with_bigfloat_precision(f::Function, precision::Integer)
     old_precision = get_bigfloat_precision()
     set_bigfloat_precision(precision)
@@ -694,16 +691,6 @@ function with_bigfloat_precision(f::Function, precision::Integer)
     end
 end
 
-function with_bigfloat_rounding(f::Function, rounding::RoundingMode)
-    old_rounding = get_bigfloat_rounding()
-    set_bigfloat_rounding(rounding)
-    try
-        return f()
-    finally
-        set_bigfloat_rounding(old_rounding)
-    end
-end
-
 function string(x::BigFloat)
     lng = 128
     for i = 1:2
@@ -719,24 +706,4 @@ print(io::IO, b::BigFloat) = print(io, string(b))
 show(io::IO, b::BigFloat) = print(io, string(b), " with $(precision(b)) bits of precision")
 showcompact(io::IO, b::BigFloat) = print(io, string(b))
 
-function hash(x::BigFloat)
-    if isnan(x)
-        return hash(NaN)
-    end
-    if isinf(x)
-        return hash(float64(x))
-    end
-    n = ceil(precision(x)/53)
-    e = exponent(x)
-    h::Uint = signbit(x)
-    h = h<<30 + e
-    x = ldexp(x, -e)
-    for i=1:n
-        f64 = float64(x)
-        h = bitmix(h, hash(f64)$11111)
-        x -= f64
-    end
-    h
-end
-
 end #module
