@@ -24,7 +24,7 @@
 ##
 ## RemoteRef(p) - ...or on a particular processor
 ##
-## put(r, val) - store a value to an uninitialized RemoteRef
+## put!(r, val) - store a value to an uninitialized RemoteRef
 ##
 ## @spawn expr -
 ##     evaluate expr somewhere. returns a RemoteRef. all variables in expr
@@ -42,8 +42,8 @@
 
 # todo:
 # - more indexing
-# * take() to empty a Ref (full/empty variables)
-# * have put() wait on non-empty Refs
+# * take!() to empty a Ref (full/empty variables)
+# * have put!() wait on non-empty Refs
 # - removing nodes
 # - more dynamic scheduling
 # * fetch/wait latency seems to be excessive
@@ -96,20 +96,41 @@ type Worker
     add_msgs::Array{Any,1}
     id::Int
     gcflag::Bool
-    privhost::ByteString
+    bind_addr::IpAddr
     manage::Function
     config::Dict
     
     Worker(host::String, port::Integer, sock::TcpSocket, id::Int) =
-        new(bytestring(host), uint16(port), sock, IOBuffer(), {}, {}, id, false, "")
+        new(bytestring(host), uint16(port), sock, IOBuffer(), {}, {}, id, false)
 end
 Worker(host::String, port::Integer, sock::TcpSocket) =
     Worker(host, port, sock, 0)
-Worker(host::String, port::Integer) =
-    Worker(host, port, connect(host,uint16(port)))
-Worker(host::String, privhost::String, port::Integer, tunnel_user::String, sshflags) =
-    Worker(host, port, connect("localhost",
-                               ssh_tunnel(tunnel_user, host, privhost, uint16(port), sshflags)))
+function Worker(host::String, port::Integer) 
+    # Connect to the loopback port if requested host has the same ipaddress as self.
+    if host == string(LPROC.bind_addr)
+        w = Worker(host, port, connect("127.0.0.1", uint16(port)))
+    else
+        w = Worker(host, port, connect(host, uint16(port)))
+    end
+    # Avoid calling getaddrinfo if possible - involves a DNS lookup
+    # host may be a stringified ipv4 / ipv6 address or a dns name
+    if host == "localhost"
+        w.bind_addr = parseip("127.0.0.1")
+    else
+        try 
+            w.bind_addr = parseip(host)
+        catch
+            w.bind_addr = getaddrinfo(host) 
+        end
+    end
+    w
+end
+function Worker(host::String, bind_addr::String, port::Integer, tunnel_user::String, sshflags) 
+    w = Worker(host, port, connect("localhost",
+                               ssh_tunnel(tunnel_user, host, bind_addr, uint16(port), sshflags)))
+    w.bind_addr = parseip(bind_addr)
+    w
+end
 
 
 function send_msg_now(w::Worker, kind, args...)
@@ -173,9 +194,11 @@ end
 
 type LocalProcess
     id::Int
+    bind_addr::IpAddr
+    LocalProcess() = new(1)
 end
 
-const LPROC = LocalProcess(0)
+const LPROC = LocalProcess()
 
 const map_pid_wrkr = Dict{Int, Union(Worker, LocalProcess)}()
 const map_sock_wrkr = ObjectIdDict()
@@ -201,24 +224,47 @@ type ProcessGroup
 end
 const PGRP = ProcessGroup({})
 
+get_bind_addr(pid::Integer) = get_bind_addr(worker_from_id(pid))
+function get_bind_addr(w::Union(Worker, LocalProcess)) 
+    if !isdefined(w, :bind_addr) 
+        if w.id != myid()
+            w.bind_addr = remotecall_fetch(w.id, get_bind_addr, w.id)
+        else
+            error("LPROC.bind_addr not defined") # Should never happend since LPROC.bind_addr 
+                                           # is defined early on during process init.
+        end
+    end
+    w.bind_addr
+end 
+
 function add_workers(pg::ProcessGroup, ws::Array{Any,1})
     # NOTE: currently only node 1 can add new nodes, since nobody else
     # has the full list of address:port
     assert(LPROC.id == 1)
-    for w in ws
+    rr_join = similar(ws, RemoteRef)
+    for (i, w) in enumerate(ws)
         w.id = get_next_pid()
         register_worker(w)
-        create_message_handler_loop(w.socket) 
+        rr_join[i] = RemoteRef()
+        create_message_handler_loop(w.socket; ntfy_join_complete=rr_join[i]) 
     end
-    all_locs = map(x -> isa(x, Worker) ? (x.privhost, x.port, x.id) : ("", 0, x.id), pg.workers)
+
+    all_locs = map(x -> isa(x, Worker) ? (string(x.bind_addr), x.port, x.id, x.manage == manage_local_worker) : ("", 0, x.id, true), pg.workers)
+    
     for w in ws
-        send_msg_now(w, :join_pgrp, w.id, all_locs)
+        send_msg_now(w, :join_pgrp, w.id, all_locs, w.manage == manage_local_worker)
     end
     for w in ws
         @schedule begin
             w.manage(w.id, w.config, :register)
         end
     end 
+    
+    # Wait till the all nodes have connected to each other
+    for rr in rr_join
+        wait(rr)
+    end
+    
     [w.id for w in ws]
 end
 
@@ -231,6 +277,18 @@ function nworkers()
 end
 
 procs() = Int[x.id for x in PGRP.workers]
+function procs(pid::Integer)
+    if myid() == 1
+        if (pid == 1) || (map_pid_wrkr[pid].manage == manage_local_worker)
+            Int[x.id for x in filter(w -> (w.id==1) || (w.manage == manage_local_worker), PGRP.workers)]
+        else
+            ipatpid = get_bind_addr(pid)
+            Int[x.id for x in filter(w -> get_bind_addr(w) == ipatpid, PGRP.workers)]
+        end
+    else
+        remotecall_fetch(1, procs, pid)
+    end
+end
 
 function workers()
     allp = procs()
@@ -288,7 +346,7 @@ function worker_from_id(pg::ProcessGroup, i)
     end
     start = time()
     while (!haskey(map_pid_wrkr, i) && ((time() - start) < 60.0))
-        sleep(0.1)		
+        sleep(0.1)
         yield()
     end
     map_pid_wrkr[i]
@@ -394,30 +452,12 @@ type RemoteRef
     RemoteRef(w::Worker) = RemoteRef(w.id)
     RemoteRef() = RemoteRef(myid())
 
-    global WeakRemoteRef
-    function WeakRemoteRef(w, wh, id)
-        return new(w, wh, id)
-    end
-
-    function WeakRemoteRef(pid::Integer)
-        rr = WeakRemoteRef(pid, myid(), REQ_ID)
-        REQ_ID += 1
-        if mod(REQ_ID,200) == 0
-            gc()
-        end
-        rr
-    end
-
-    WeakRemoteRef(w::LocalProcess) = WeakRemoteRef(myid())
-    WeakRemoteRef(w::Worker) = WeakRemoteRef(w.id)
-    WeakRemoteRef() = WeakRemoteRef(myid())
-
     global next_id
     next_id() = (id=(myid(),REQ_ID); REQ_ID+=1; id)
 end
 
-hash(r::RemoteRef) = hash(r.whence)+3*hash(r.id)
-isequal(r::RemoteRef, s::RemoteRef) = (r.whence==s.whence && r.id==s.id)
+hash(r::RemoteRef, h::Uint) = hash(r.whence, hash(r.id, h))
+==(r::RemoteRef, s::RemoteRef) = (r.whence==s.whence && r.id==s.id)
 
 rr2id(r::RemoteRef) = (r.whence, r.id)
 
@@ -467,7 +507,13 @@ function del_clients(pairs::(Any,Any)...)
     end
 end
 
-any_gc_flag = false
+any_gc_flag = Condition()
+function start_gc_msgs_task()
+    @schedule while true
+        wait(any_gc_flag)
+        flush_gc_msgs()
+    end
+end
 
 function send_del_client(rr::RemoteRef)
     if rr.where == myid()
@@ -480,7 +526,7 @@ function send_del_client(rr::RemoteRef)
         w = worker_from_id(rr.where)
         push!(w.del_msgs, (rr2id(rr), myid()))
         w.gcflag = true
-        global any_gc_flag = true
+        notify(any_gc_flag)
     end
 end
 
@@ -508,7 +554,7 @@ function send_add_client(rr::RemoteRef, i)
         #println("$(myid()) adding $((rr2id(rr), i)) for $(rr.where)")
         push!(w.add_msgs, (rr2id(rr), i))
         w.gcflag = true
-        global any_gc_flag = true
+        notify(any_gc_flag)
     end
 end
 
@@ -566,7 +612,7 @@ function wait_empty(rv::RemoteValue)
     return nothing
 end
 
-## core messages: do, call, fetch, wait, ref, put ##
+## core messages: do, call, fetch, wait, ref, put! ##
 
 
 function run_work_thunk(thunk)
@@ -581,14 +627,15 @@ function run_work_thunk(thunk)
     result
 end
 function run_work_thunk(rv::RemoteValue, thunk)
-    put(rv, run_work_thunk(thunk))
+    put!(rv, run_work_thunk(thunk))
+    nothing
 end
 
 function schedule_call(rid, thunk)
     rv = RemoteValue()
     (PGRP::ProcessGroup).refs[rid] = rv
     push!(rv.clientset, rid[1])
-    enq_work(@task(run_work_thunk(rv,thunk)))
+    schedule(@task(run_work_thunk(rv,thunk)))
     rv
 end
 
@@ -684,7 +731,7 @@ function remote_do(w::LocalProcess, f, args...)
     # does when it gets a :do message.
     # same for other messages on LocalProcess.
     thk = local_remotecall_thunk(f, args)
-    enq_work(Task(thk))
+    schedule(Task(thk))
     nothing
 end
 
@@ -713,17 +760,18 @@ fetch(r::RemoteRef) = call_on_owner(fetch_ref, r)
 fetch(x::ANY) = x
 
 # storing a value to a Ref
-function put(rv::RemoteValue, val::ANY)
+function put!(rv::RemoteValue, val::ANY)
     wait_empty(rv)
     rv.result = val
     rv.done = true
     notify_full(rv)
+    rv
 end
 
-put_ref(rid, v) = put(lookup_ref(rid), v)
-put(rr::RemoteRef, val::ANY) = (call_on_owner(put_ref, rr, val); val)
+put_ref(rid, v) = put!(lookup_ref(rid), v)
+put!(rr::RemoteRef, val::ANY) = (call_on_owner(put_ref, rr, val); rr)
 
-function take(rv::RemoteValue)
+function take!(rv::RemoteValue)
     wait_full(rv)
     val = rv.result
     rv.done = false
@@ -732,37 +780,8 @@ function take(rv::RemoteValue)
     val
 end
 
-take_ref(rid) = take(lookup_ref(rid))
-take(rr::RemoteRef) = call_on_owner(take_ref, rr)
-
-## work queue ##
-
-function enq_work(t::Task)
-    ccall(:uv_stop,Void,(Ptr{Void},),eventloop())
-    unshift!(Workqueue, t)
-end
-
-function perform_work()
-    perform_work(pop!(Workqueue))
-end
-
-function perform_work(t::Task)
-    if !istaskstarted(t)
-        # starting new task
-        yieldto(t)
-    else
-        # continuing interrupted work item
-        arg = t.result
-        t.result = nothing
-        t.runnable = true
-        yieldto(t, arg)
-    end
-    t = current_task().last
-    if t.runnable
-        # still runnable; return to queue
-        enq_work(t)
-    end
-end
+take_ref(rid) = take!(lookup_ref(rid))
+take!(rr::RemoteRef) = call_on_owner(take_ref, rr)
 
 function deliver_result(sock::IO, msg, oid, value)
     #print("$(myid()) sending result $oid\n")
@@ -799,18 +818,20 @@ notify_empty(rv::RemoteValue) = notify(rv.empty)
 # activity on accept fd
 function accept_handler(server::TcpServer, status::Int32)
     if status == -1
-        error("An error occured during the creation of the server")
+        error("an error occured during the creation of the server")
     end
     client = accept_nonblock(server)
     create_message_handler_loop(client)
 end
 
-function create_message_handler_loop(sock::AsyncStream) #returns immediately
-    enq_work(@task begin
+function create_message_handler_loop(sock::AsyncStream; ntfy_join_complete=nothing) #returns immediately
+    schedule(@task begin
         global PGRP
         #println("message_handler_loop")
+        disable_nagle(sock)
         start_reading(sock)
         wait_connected(sock)
+        
         try
             while true
                 msg = deserialize(sock)
@@ -856,7 +877,7 @@ function create_message_handler_loop(sock::AsyncStream) #returns immediately
                     oid = deserialize(sock)
                     #print("$(myid()) got $msg $oid\n")
                     val = deserialize(sock)
-                    put(lookup_ref(oid), val)
+                    put!(lookup_ref(oid), val)
                 elseif is(msg, :identify_socket)
                     otherid = deserialize(sock)
                     register_worker(Worker("", 0, sock, otherid))
@@ -864,16 +885,25 @@ function create_message_handler_loop(sock::AsyncStream) #returns immediately
                     # first connection; get process group info from client
                     self_pid = LPROC.id = deserialize(sock)
                     locs = deserialize(sock)
+                    self_is_local = deserialize(sock)
                     #print("\nLocation: ",locs,"\nId:",myid(),"\n")
                     # joining existing process group
                     
-                    register_worker(Worker("", 0, sock, 1))
+                    controller = Worker("", 0, sock, 1)
+                    register_worker(controller)
                     register_worker(LPROC)
                     
-                    for (rhost, rport, rpid) in locs
+                    for (rhost, rport, rpid, r_is_local) in locs
                         if (rpid < self_pid) && (!(rpid == 1))
                             # Connect to them
-                            w = Worker(rhost, rport)
+                            if self_is_local && r_is_local
+                                # If on localhost, use the loopback address - this addresses
+                                # the special case of system suspend wherein the local ip 
+                                # may be changed upon system awake.
+                                w = Worker("127.0.0.1", rport)
+                            else
+                                w = Worker(rhost, rport)
+                            end
                             w.id = rpid
                             register_worker(w)
                             create_message_handler_loop(w.socket)
@@ -883,6 +913,12 @@ function create_message_handler_loop(sock::AsyncStream) #returns immediately
                             continue
                         end
                     end
+                    
+                    send_msg_now(controller, :join_complete)
+                    
+                elseif is(msg, :join_complete)
+                    put!(ntfy_join_complete, :join_complete)
+                    ntfy_join_complete = nothing    # so that it gets gc'ed
                 end
                 
             end # end of while
@@ -929,8 +965,6 @@ end
 # argument is descriptor to write listening port # to.
 start_worker() = start_worker(STDOUT)
 function start_worker(out::IO)
-    global bind_addr
-
     # we only explicitly monitor worker STDOUT on the console, so redirect
     # stderr to stdout so we can see the output.
     # at some point we might want some or all worker output to go to log
@@ -939,28 +973,22 @@ function start_worker(out::IO)
     # exit when process 1 shut down. Don't yet know why.
     #redirect_stderr(STDOUT)
 
-    if !isdefined(Base,:bind_addr)
-        bind_addr = getipaddr()
-    end
     (actual_port,sock) = listenany(uint16(9009))
     sock.ccb = accept_handler
     print(out, "julia_worker:")  # print header
     print(out, "$(dec(actual_port))#") # print port
-    print(out, bind_addr)      #TODO: print hostname
+    print(out, LPROC.bind_addr)
     print(out, '\n')
     flush(out)
     # close STDIN; workers will not use it
     #close(STDIN)
 
     disable_threaded_libs()
-
-    ccall(:jl_install_sigint_handler, Void, ())
-
-    global const Scheduler = current_task()
+    disable_nagle(sock)
 
     try
         check_master_connect(60.0)
-        event_loop(false)
+        while true; wait(); end
     catch err
         print(STDERR, "unhandled exception on $(myid()): $(err)\nexiting.\n")
     end
@@ -984,20 +1012,20 @@ function start_cluster_workers(np::Integer, config::Dict, cman::ClusterManager)
     elseif  insttype == :io_host
         read_cb_response(inst) = 
         begin
-            (priv_hostname, port) = read_worker_host_port(inst[1])
-            inst[1], priv_hostname, port, inst[2], inst[3]
+            (bind_addr, port) = read_worker_host_port(inst[1])
+            inst[1], bind_addr, port, inst[2], inst[3]
         end
     elseif insttype == :io_host_port
         read_cb_response(inst) = (inst[1], inst[2], inst[3], inst[2], inst[4])
     elseif insttype == :host_port
         read_cb_response(inst) = (nothing, inst[1], inst[2], inst[1], inst[3])
     else
-        error("Unsupported format from Cluster Manager callback")
+        error("unsupported format from ClusterManager callback")
     end
     
     for i=1:np
-        (io, privhost, port, pubhost, wconfig) = read_cb_response(instances[i])
-        ws[i] = create_worker(privhost, port, pubhost, io, wconfig, cman.manage)
+        (io, bind_addr, port, pubhost, wconfig) = read_cb_response(instances[i])
+        ws[i] = create_worker(bind_addr, port, pubhost, io, wconfig, cman.manage)
     end
     ws
 end
@@ -1006,14 +1034,14 @@ function read_worker_host_port(io::IO)
     io.line_buffered = true
     while true
         conninfo = readline(io)
-        private_hostname, port = parse_connection_info(conninfo)
-        if private_hostname != ""
-            return private_hostname, port
+        bind_addr, port = parse_connection_info(conninfo)
+        if bind_addr != ""
+            return bind_addr, port
         end
     end
 end
 
-function create_worker(privhost, port, pubhost, stream, config, manage)
+function create_worker(bind_addr, port, pubhost, stream, config, manage)
     tunnel = config[:tunnel]
     
     s = split(pubhost,'@')
@@ -1025,41 +1053,29 @@ function create_worker(privhost, port, pubhost, stream, config, manage)
         if haskey(ENV, "USER")
             user = ENV["USER"]
         elseif tunnel
-            error("USER must be specified either in the environment or as part of the hostname when tunnel option is used.")
+            error("USER must be specified either in the environment or as part of the hostname when tunnel option is used")
         end
     end
     
     if tunnel
         sshflags = config[:sshflags]
-        w = Worker(pubhost, privhost, port, user, sshflags)
+        w = Worker(pubhost, bind_addr, port, user, sshflags)
     else
-        w = Worker(pubhost, port)
+        w = Worker(bind_addr, port)
     end
     
-    w.privhost = privhost
     w.config = config
     w.manage = manage
     
     if isa(stream, AsyncStream)
-        stream.line_buffered = true
         let wrker = w
             # redirect console output from workers to the client's stdout:
-            start_reading(stream,function(stream::AsyncStream,nread::Int)
-                if nread>0
-                    try
-                        line = readbytes(stream.buffer, nread)
-                        if length(line) < nread
-                            println(STDERR,"\tTruncated reply from worker $(wrker.id)")
-                            return false
-                        end
-                        print("\tFrom worker $(wrker.id):\t",bytestring(line))
-                    catch err
-                        println(STDERR,"\tError parsing reply from worker $(wrker.id):\t",err)
-                        return false
-                    end
+            @async begin
+                while !eof(stream)
+                    line = readline(stream)
+                    print("\tFrom worker $(wrker.id):\t$line")
                 end
-                true
-            end)
+            end
         end
     end
 
@@ -1082,15 +1098,15 @@ end
 tunnel_port = 9201
 # establish an SSH tunnel to a remote worker
 # returns P such that localhost:P connects to host:port
-function ssh_tunnel(user, host, privhost, port, sshflags)
+function ssh_tunnel(user, host, bind_addr, port, sshflags)
     global tunnel_port
     localp = tunnel_port::Int
-    while !success(detach(`ssh -f -o ExitOnForwardFailure=yes $sshflags $(user)@$host -L $localp:$privhost:$(int(port)) sleep 60`)) && localp < 10000
+    while !success(detach(`ssh -T -a -x -o ExitOnForwardFailure=yes -f $sshflags $(user)@$host -L $localp:$bind_addr:$(int(port)) sleep 60`)) && localp < 10000
         localp += 1
     end
     
     if localp >= 10000
-        error("Unable to assign a local tunnel port between 9201 and 10000")
+        error("unable to assign a local tunnel port between 9201 and 10000")
     end
     
     tunnel_port = localp+1
@@ -1117,7 +1133,7 @@ function launch_local_workers(cman::LocalManager, np::Integer, config::Dict)
 
     # start the processes first...
     for i in 1:np
-        io, pobj = readsfrom(detach(`$(dir)/$(exename) --bind-to 127.0.0.1 $exeflags`))
+        io, pobj = open(detach(`$(dir)/$(exename) $exeflags --bind-to $(LPROC.bind_addr)`), "r")
         io_objs[i] = io
         configs[i] = merge(config, {:process => pobj})
     end
@@ -1145,7 +1161,7 @@ show(io::IO, cman::SSHManager) = println("SSHManager(machines=", cman.machines,
 function launch_ssh_workers(cman::SSHManager, np::Integer, config::Dict)
     dir = config[:dir]
     exename = config[:exename]
-    exeflags = config[:exeflags]
+    exeflags_base = config[:exeflags]
 
     io_objs = cell(np)
     configs = cell(np)
@@ -1153,22 +1169,38 @@ function launch_ssh_workers(cman::SSHManager, np::Integer, config::Dict)
     # start the processes first...
 
     for i in 1:np
-        # machine could be of the format [user@]host[:port]
-        machine_def = split(cman.machines[i], ':')
+        thisconfig = copy(config) # config for this worker
+
+        # machine could be of the format [user@]host[:port] bind_addr
+        machine_bind = split(cman.machines[i])
+        if length(machine_bind) > 1
+            exeflags = `--bind-to $(machine_bind[2]) $exeflags_base`
+        else
+            exeflags = exeflags_base
+        end
+        machine_def = machine_bind[1]
+        
+        machine_def = split(machine_def, ':')
         portopt = length(machine_def) == 2 ? ` -p $(machine_def[2]) ` : ``
-        config[:sshflags] = `$(config[:sshflags]) $portopt`
+        sshflags = `$(config[:sshflags]) $portopt`
+        thisconfig[:sshflags] = sshflags
         
-        sshflags = config[:sshflags]
-        cman.machines[i] = machine_def[1]
+        host = cman.machines[i] = machine_def[1]
         
-        io, pobj = readsfrom(detach(`ssh -n $sshflags $(machine_def[1]) "sh -l -c \"cd $dir && $exename $exeflags\""`))
+        # Build up the ssh command
+        cmd = `cd $dir && $exename $exeflags` # launch julia
+        cmd = `sh -l -c $(shell_escape(cmd))` # shell to launch under
+        cmd = `ssh -T -a -x -o ClearAllForwardings=yes -n $sshflags $host $(shell_escape(cmd))` # use ssh to remote launch
+        
+        io, pobj = open(detach(cmd), "r")
         io_objs[i] = io
-        configs[i] = merge(config, {:machine => cman.machines[i]})
+        thisconfig[:machine] = cman.machines[i]
+        configs[i] = thisconfig
     end
 
     # ...and then read the host:port info. This optimizes overall start times.
     # For ssh, the tunnel connection, if any, has to be with the specified machine name.
-    # but the port needs to be forwarded to the private hostname/ip-address
+    # but the port needs to be forwarded to the bound hostname/ip-address
     return (:io_host, collect(zip(io_objs, cman.machines, configs)))
 end
 
@@ -1176,7 +1208,7 @@ function manage_ssh_worker(id::Integer, config::Dict, op::Symbol)
     if op == :interrupt
         if haskey(config, :ospid)
             machine = config[:machine]
-            if !success(`ssh -n $(config[:sshflags]) $machine "kill -2 $(config[:ospid])"`)
+            if !success(`ssh -T -a -x -o ClearAllForwardings=yes -n $(config[:sshflags]) $machine "kill -2 $(config[:ospid])"`)
                 println("Error sending a Ctrl-C to julia worker $id on $machine")
             end
         else
@@ -1194,7 +1226,7 @@ end
 # to be mutually reachable without a tunnel, as is often the case in a cluster.
 function addprocs_internal(np::Integer;
                   tunnel=false, dir=JULIA_HOME,
-                  exename=(ccall(:jl_is_debugbuild,Cint,())==0?"./julia-basic":"./julia-debug-basic"),
+                  exename=(ccall(:jl_is_debugbuild,Cint,())==0?"./julia":"./julia-debug"),
                   sshflags::Cmd=``, cman=LocalManager(), exeflags=``)
                   
     config={:dir=>dir, :exename=>exename, :exeflags=>`$exeflags --worker`, :tunnel=>tunnel, :sshflags=>sshflags}
@@ -1216,38 +1248,7 @@ end
 
 ## higher-level functions: spawn, pmap, pfor, etc. ##
 
-sync_begin() = task_local_storage(:SPAWNS, ({}, get(task_local_storage(), :SPAWNS, ())))
-
-function sync_end()
-    spawns = get(task_local_storage(), :SPAWNS, ())
-    if is(spawns,())
-        error("sync_end() without sync_begin()")
-    end
-    refs = spawns[1]
-    task_local_storage(:SPAWNS, spawns[2])
-    for r in refs
-        wait(r)
-    end
-end
-
-macro sync(block)
-    quote
-        sync_begin()
-        v = $(esc(block))
-        sync_end()
-        v
-    end
-end
-
-function sync_add(r)
-    spawns = get(task_local_storage(), :SPAWNS, ())
-    if !is(spawns,())
-        push!(spawns[1], r)
-    end
-    r
-end
-
-let nextidx = 1
+let nextidx = 0
     global chooseproc
     function chooseproc(thunk::Function)
         p = -1
@@ -1297,23 +1298,6 @@ macro fetchfrom(p, expr)
     :(remotecall_fetch($(esc(p)), $(esc(expr))))
 end
 
-function spawnlocal(thunk)
-    t = Task(thunk)
-    sync_add(t)
-    enq_work(t)
-    t
-end
-
-macro async(expr)
-    expr = localize_vars(:(()->($expr)), false)
-    :(spawnlocal($(esc(expr))))
-end
-
-macro spawnlocal(expr)
-    warn_once("@spawnlocal is deprecated, use @async instead.")
-    :(@async $(esc(expr)))
-end
-
 function at_each(f, args...)
     for w in PGRP.workers
         sync_add(remotecall(w.id, f, args...))
@@ -1359,12 +1343,11 @@ function pmap(f, lsts...; err_retry=true, err_stop=false)
     function getnext_tasklet()
         if is_task_in_error() && err_stop
             return nothing
-        elseif all([!done(lsts[idx],states[idx]) for idx in 1:len])
+        elseif !any(idx->done(lsts[idx],states[idx]), 1:len)
             nxts = [next(lsts[idx],states[idx]) for idx in 1:len]
-            map(idx->states[idx]=nxts[idx][2], 1:len)
+            for idx in 1:len; states[idx] = nxts[idx][2]; end
             nxtvals = [x[1] for x in nxts]
             return (getnextidx(), nxtvals)
-            
         elseif !isempty(retryqueue)
             return shift!(retryqueue)
         else    
@@ -1412,7 +1395,7 @@ function splitrange(N::Int, np::Int)
     each = div(N,np)
     extras = rem(N,np)
     nchunks = each > 0 ? np : extras
-    chunks = Array(Range1{Int}, nchunks)
+    chunks = Array(UnitRange{Int}, nchunks)
     lo = 1
     for i in 1:nchunks
         hi = lo + each - 1
@@ -1480,12 +1463,12 @@ macro parallel(args...)
             ex = loop.args[1]
             loop.args[1] = esc(ex)
             nd = length(loop.args)-1
-            ranges = map(e->esc(e.args[2]), loop.args[2:])
+            ranges = map(e->esc(e.args[2]), loop.args[2:end])
             for i=1:nd
                 var = loop.args[1+i].args[1]
                 loop.args[1+i] = :( $(esc(var)) = ($(ranges[i]))[I[$i]] )
             end
-            return :( DArray((I::(Range1{Int}...))->($loop),
+            return :( DArray((I::(UnitRange{Int}...))->($loop),
                              tuple($(map(r->:(length($r)),ranges)...))) )
         end
     elseif na==2
@@ -1531,75 +1514,6 @@ end
 #     2/(nc/niter)
 # end
 
-## event processing, I/O and work scheduling ##
-
-function yield(args...)
-    ct = current_task()
-    # preserve Task.last across calls to the scheduler
-    prev = ct.last
-    v = yieldto(Scheduler, args...)
-    ct.last = prev
-    return v
-end
-
-function pause()
-    @unix_only    ccall(:pause, Void, ())
-    @windows_only ccall(:Sleep,stdcall, Void, (Uint32,), 0xffffffff)
-end
-
-function event_loop(isclient)
-    iserr, lasterr, bt = false, nothing, {}
-    while true
-        try
-            if iserr
-                display_error(lasterr, bt)
-                println(STDERR)
-                iserr, lasterr, bt = false, nothing, {}
-            else
-                while true
-                    if isempty(Workqueue)
-                        if any_gc_flag
-                            flush_gc_msgs()
-                        end
-                        c = process_events(true)
-                        if c==0 && eventloop()!=C_NULL && isempty(Workqueue) && !any_gc_flag
-                            # if there are no active handles and no runnable tasks, just
-                            # wait for signals.
-                            pause()
-                        end
-                    else
-                        perform_work()
-                        process_events(false)
-                    end
-                end
-            end
-        catch err
-            if iserr
-                ccall(:jl_, Void, (Any,), (
-                    "\n!!!An ERROR occurred while printing the last error!!!\n",
-                    lasterr,
-                    bt
-                ))
-            end
-            iserr, lasterr = true, err
-            bt = catch_backtrace()
-            if isclient && isa(err,InterruptException)
-                # root task is waiting for something on client. allow C-C
-                # to interrupt.
-                interrupt_waiting_task(roottask,err)
-                iserr, lasterr = false, nothing
-            end
-        end
-    end
-end
-
-# force a task to stop waiting with an exception
-function interrupt_waiting_task(t::Task, err)
-    if !t.runnable
-        t.exception = err
-        enq_work(t)
-    end
-end
 
 function check_master_connect(timeout)
     # If we do not have at least process 1 connect to us within timeout
@@ -1621,17 +1535,15 @@ end
 function timedwait(testcb::Function, secs::Float64; pollint::Float64=0.1)
     start = time()
     done = RemoteRef()
-    timercb(aw, status) = begin
+    timercb(aw) = begin
         try
             if testcb()
-                put(done, :ok)
+                put!(done, :ok)
             elseif (time() - start) > secs
-                put(done, :timed_out)
-            elseif status != 0
-                put(done, :error)
+                put!(done, :timed_out)
             end
         catch e
-            put(done, :error)
+            put!(done, :error)
         finally
             isready(done) && stop_timer(aw)
         end
@@ -1665,3 +1577,49 @@ function interrupt(pids::AbstractVector=workers())
         end
     end
 end
+
+function disable_nagle(sock)
+    # disable nagle on all OSes
+    ccall(:uv_tcp_nodelay, Cint, (Ptr{Void}, Cint), sock.handle, 1)
+    @linux_only begin
+        # tcp_quickack is a linux only option
+        if ccall(:jl_tcp_quickack, Cint, (Ptr{Void}, Cint), sock.handle, 1) < 0 
+            warn_once("Parallel networking unoptimized ( Error enabling TCP_QUICKACK : ", strerror(errno()), " )")
+        end
+    end
+end
+
+function check_same_host(pids)
+    if myid() != 1
+        return remotecall_fetch(1, check_same_host, pids)
+    else
+        # We checkfirst if all test pids have been started using the local manager,
+        # else we check for the same bind_to addr. This handles the special case
+        # where the local ip address may change - as during a system sleep/awake
+        if all(p -> (p==1) || (map_pid_wrkr[p].manage == manage_local_worker), pids)
+            return true
+        else
+            first_bind_addr = map_pid_wrkr[pids[1]].bind_addr
+            return all(p -> (p != 1) && (map_pid_wrkr[p].bind_addr == first_bind_addr), pids[2:end])
+        end
+    end
+end
+
+function terminate_all_workers()
+    if myid() != 1
+        return
+    end
+    
+    if nprocs() > 1
+        ret = rmprocs(workers(); waitfor=0.5)
+        if ret != :ok
+            warn("Forcibly interrupting busy workers")
+            # Might be computation bound, interrupt them and try again
+            interrupt(workers())
+            ret = rmprocs(workers(); waitfor=0.5)
+            if ret != :ok
+                warn("Unable to terminate all workers")
+            end
+        end
+     end   
+end
