@@ -12,7 +12,7 @@ end
 @windows_only begin
     const path_separator    = "\\"
     const path_separator_re = r"[/\\]+"
-    const path_absolute_re  = r"^(?:\w+:)[/\\]"
+    const path_absolute_re  = r"^(?:\w+:)?[/\\]"
     const path_directory_re = r"(?:^|[/\\])\.{0,2}$"
     const path_dir_splitter = r"^(.*?)([/\\]+)([^/\\]*)$"
     const path_ext_splitter = r"^((?:.*[/\\])?(?:\.|[^/\\\.])[^/\\]*?)(\.[^/\\\.]*|)$"
@@ -21,7 +21,7 @@ end
         m = match(r"^(\w+:|\\\\\w+\\\w+|\\\\\?\\UNC\\\w+\\\w+|\\\\\?\\\w+:|)(.*)$", path)
         bytestring(m.captures[1]), bytestring(m.captures[2])
     end
-    homedir() = get(ENV,"HOME",joinpath(ENV["HOMEDRIVE"],ENV["HOMEPATH"]))
+    homedir() = get(ENV,"HOME",string(ENV["HOMEDRIVE"],ENV["HOMEPATH"]))
 end
 
 isabspath(path::String) = ismatch(path_absolute_re, path)
@@ -78,7 +78,7 @@ function normpath(path::String)
         clean = true
         for j = 1:length(parts)-1
             if parts[j] != ".." && parts[j+1] == ".."
-                splice!(parts, j:j+1)
+                deleteat!(parts, j:j+1)
                 clean = false
                 break
             end
@@ -106,20 +106,21 @@ normpath(a::String, b::String...) = normpath(joinpath(a,b...))
 abspath(a::String) = normpath(isabspath(a) ? a : joinpath(pwd(),a))
 abspath(a::String, b::String...) = abspath(joinpath(a,b...))
 
-@windows_only function realpath(path::String)
-    buflength = length(path)+1
-    buf = zeros(Uint8,buflength)
-    p = ccall((:GetFullPathNameA, "Kernel32"), stdcall, 
-        Uint32, (Ptr{Uint8}, Uint32, Ptr{Uint8}, Ptr{Void}), 
-        path, buflength, buf, C_NULL)
-    if p > buflength
-        buf = zeros(Uint8,p)
-        p = ccall((:GetFullPathNameA, "Kernel32"), stdcall, 
-            Uint32, (Ptr{Uint8}, Uint32, Ptr{Uint8}, Ptr{Void}), 
-            path, p, buf, C_NULL)
+@windows_only realpath(path::String) = realpath(utf16(path))
+@windows_only function realpath(path::UTF16String)
+    p = uint32((sizeof(path)>>2) + 1)
+    while true
+        buflength = p
+        buf = zeros(Uint16,buflength)
+        p = ccall((:GetFullPathNameW, "Kernel32"), stdcall, 
+            Uint32, (Ptr{Uint16}, Uint32, Ptr{Uint16}, Ptr{Void}), 
+            path, buflength, buf, C_NULL)
+        systemerror(:realpath, p == 0)
+        if (p < buflength)
+            resize!(buf, p+1)
+            return utf8(UTF16String(buf))
+        end
     end
-    systemerror(:realpath, p == 0)
-    return bytestring(buf)[1:end-1]
 end
 
 @unix_only function realpath(path::String)
