@@ -46,7 +46,7 @@ const EXECUTE_MASK      =
 const OPTIONS_MASK = COMPILE_MASK | EXECUTE_MASK
 
 function info{T}(
-    regex::Union(Ptr{Void},Vector{Uint8}),
+    regex::Ptr{Void},
     extra::Ptr{Void}, what::Integer, ::Type{T}
 )
     buf = Array(Uint8,sizeof(T))
@@ -54,8 +54,7 @@ function info{T}(
                 (Ptr{Void}, Ptr{Void}, Int32, Ptr{Uint8}),
                 regex, extra, what, buf)
     if ret != 0
-        error("info: ",
-              ret == ERROR_NULL      ? "NULL regex object" :
+        error(ret == ERROR_NULL      ? "NULL regex object" :
               ret == ERROR_BADMAGIC  ? "invalid regex object" :
               ret == ERROR_BADOPTION ? "invalid option flags" :
                                        "unknown error $ret")
@@ -70,7 +69,7 @@ function config{T}(what::Integer, ::Type{T})
                 what, buf)
 
     if ret != 0
-        error("config: error $n")
+        error("error $n")
     end
     reinterpret(T,buf)[1]
 end
@@ -78,56 +77,63 @@ end
 function compile(pattern::String, options::Integer)
     errstr = Array(Ptr{Uint8},1)
     erroff = Array(Int32,1)
-    re_ptr = (()->ccall((:pcre_compile, :libpcre), Ptr{Void},
-                        (Ptr{Uint8}, Int32, Ptr{Ptr{Uint8}}, Ptr{Int32}, Ptr{Uint8}),
-                        pattern, options, errstr, erroff, C_NULL))()
+    re_ptr = ccall((:pcre_compile, :libpcre), Ptr{Void},
+                    (Ptr{Uint8}, Int32, Ptr{Ptr{Uint8}}, Ptr{Int32}, Ptr{Uint8}),
+                    pattern, options, errstr, erroff, C_NULL)
     if re_ptr == C_NULL
-        error("compile: $(bytestring(errstr[1]))",
+        error("$(bytestring(errstr[1]))",
               " at position $(erroff[1]+1)",
               " in $(repr(pattern))")
     end
-    size = info(re_ptr, C_NULL, INFO_SIZE, Int32)
-    regex = Array(Uint8,size)
-    ccall(:memcpy, Ptr{Void}, (Ptr{Void}, Ptr{Void}, Uint), regex, re_ptr, size)
-    regex
+    re_ptr
 end
 
-function study(regex::Array{Uint8}, options::Integer)
+function study(regex::Ptr{Void}, options::Integer)
     # NOTE: options should always be zero in current PCRE
     errstr = Array(Ptr{Uint8},1)
     extra = ccall((:pcre_study, :libpcre), Ptr{Void},
                   (Ptr{Void}, Int32, Ptr{Ptr{Uint8}}),
                   regex, options, errstr)
     if errstr[1] != C_NULL
-        error("study: $(bytestring(errstr[1]))")
+        error("$(bytestring(errstr[1]))")
     end
     extra
 end
-study(re::Array{Uint8}) = study(re, int32(0))
+
+study(re::Ptr{Void}) = study(re, int32(0))
 
 free_study(extra::Ptr{Void}) =
     ccall((:pcre_free_study, :libpcre), Void, (Ptr{Void},), extra)
+free(regex::Ptr{Void}) =
+    ccall(unsafe_load(cglobal((:pcre_free, :libpcre),Ptr{Void})), Void, (Ptr{Void},), regex)
 
-exec(regex::Array{Uint8}, extra::Ptr{Void}, str::SubString, offset::Integer, options::Integer, cap::Bool) =
-    exec(regex, extra, str.string, str.offset, offset, sizeof(str), options, cap)
-exec(regex::Array{Uint8}, extra::Ptr{Void}, str::ByteString, offset::Integer, options::Integer, cap::Bool) =
-    exec(regex, extra, str, 0, offset, sizeof(str), options, cap)
-function exec(regex::Array{Uint8}, extra::Ptr{Void},
-              str::ByteString, shift::Integer, offset::Integer, len::Integer, options::Integer, cap::Bool)
+function exec(regex::Ptr{Void}, extra::Ptr{Void}, str::SubString, offset::Integer,
+              options::Integer, ovec::Vector{Int32})
+    return exec(regex, extra, str.string, str.offset, offset, sizeof(str),
+                options, ovec)
+end
+
+function exec(regex::Ptr{Void}, extra::Ptr{Void}, str::ByteString, offset::Integer,
+              options::Integer, ovec::Vector{Int32})
+    return exec(regex, extra, str, 0, offset, sizeof(str), options, ovec)
+end
+
+function exec(regex::Ptr{Void}, extra::Ptr{Void},
+              str::ByteString, shift::Integer, offset::Integer,
+              len::Integer, options::Integer,
+              ovec::Vector{Int32})
     if offset < 0 || len < offset || len+shift > sizeof(str)
         error(BoundsError)
     end
-    ncap = info(regex, extra, INFO_CAPTURECOUNT, Int32)
-    ovec = Array(Int32, 3(ncap+1))
     n = ccall((:pcre_exec, :libpcre), Int32,
               (Ptr{Void}, Ptr{Void}, Ptr{Uint8}, Int32,
                Int32, Int32, Ptr{Int32}, Int32),
               regex, extra, pointer(str.data,shift+1), len,
               offset, options, ovec, length(ovec))
     if n < -1
-        error("exec: error $n")
+        error("error $n")
     end
-    cap ? ((n > -1 ? ovec[1:2(ncap+1)] : Array(Int32,0)), ncap) : n > -1
+    return n > -1
 end
 
 end # module
