@@ -1,7 +1,7 @@
 module Dir
 
-import Base.Git
 import ..Pkg: DEFAULT_META, META_BRANCH
+import ..Git
 
 const DIR_NAME = ".julia"
 
@@ -45,7 +45,7 @@ function init(meta::String=DEFAULT_META, branch::String=META_BRANCH)
             run(`touch REQUIRE`)
         end
     catch e
-        run(`rm -rf $dir`)
+        rm(dir, recursive=true)
         rethrow(e)
     end
 end
