@@ -1,8 +1,8 @@
 module Entry
 
+import Base: thispatch, nextpatch, nextminor, nextmajor, check_new_version
+import ..Git, ..Reqs, ..Read, ..Query, ..Resolve, ..Cache, ..Write, ..GitHub
 using ..Types
-import ..Reqs, ..Read, ..Query, ..Resolve, ..Cache, ..Write
-import Base: Git, thispatch, nextpatch, nextminor, nextmajor, check_new_version
 
 macro recover(ex)
     quote
@@ -23,7 +23,7 @@ function edit(f::Function, pkg::String, args...)
     reqsʹ = Reqs.parse(rʹ)
     reqsʹ != reqs && resolve(reqsʹ,avail)
     Reqs.write("REQUIRE",rʹ)
-    info("REQUIRE updated.")
+    info("Package database updated")
     return true
 end
 
@@ -35,21 +35,44 @@ function edit()
     reqs = Reqs.parse("REQUIRE")
     run(`$editor REQUIRE`)
     reqsʹ = Reqs.parse("REQUIRE")
-    reqs == reqsʹ && return info("Nothing to be done.")
+    reqs == reqsʹ && return info("Nothing to be done")
     info("Computing changes...")
     resolve(reqsʹ)
 end
 
 function add(pkg::String, vers::VersionSet)
-    edit(Reqs.add,pkg,vers) && return
-    ispath(pkg) || error("unknown package $pkg")
-    info("Nothing to be done.")
+    outdated = :maybe
+    @sync begin
+        @async if !edit(Reqs.add,pkg,vers)
+            ispath(pkg) || error("unknown package $pkg")
+            info("Nothing to be done")
+        end
+        branch = Pkg.META_BRANCH
+        if Git.branch(dir="METADATA") == branch
+            if !Git.success(`diff --quiet origin/$branch`, dir="METADATA")
+                outdated = :yes
+            else
+                try
+                    run(Git.cmd(`fetch -q --all`, dir="METADATA") |>DevNull .>DevNull)
+                    outdated = Git.success(`diff --quiet origin/$branch`, dir="METADATA") ?
+                        (:no) : (:yes)
+                end
+            end
+        else
+            outdated = :no # user is doing something funky with METADATA
+        end
+    end
+    if outdated != :no
+        is = outdated == :yes ? "is" : "might be"
+        info("METADATA $is out-of-date — you may not have the latest version of $pkg")
+        info("Use `Pkg.update()` to get the latest versions of your packages")
+    end
 end
 add(pkg::String, vers::VersionNumber...) = add(pkg,VersionSet(vers...))
 
 function rm(pkg::String)
     edit(Reqs.rm,pkg) && return
-    ispath(pkg) || return info("Nothing to be done.")
+    ispath(pkg) || return info("Nothing to be done")
     info("Removing $pkg (unregistered)")
     Write.remove(pkg)
 end
@@ -84,7 +107,7 @@ function status(io::IO)
     instd = Read.installed()
     required = sort!([keys(reqs)...])
     if !isempty(required)
-        println(io, "Required packages:")
+        println(io, "$(length(required)) required packages:")
         for pkg in required
             ver,fix = pop!(instd,pkg)
             status(io,pkg,ver,fix)
@@ -92,14 +115,14 @@ function status(io::IO)
     end
     additional = sort!([keys(instd)...])
     if !isempty(additional)
-        println(io, "Additional packages:")
+        println(io, "$(length(additional)) additional packages:")
         for pkg in additional
             ver,fix = instd[pkg]
             status(io,pkg,ver,fix)
         end
     end
     if isempty(required) && isempty(additional)
-        println(io, "No packages installed.")
+        println(io, "No packages installed")
     end
 end
 # TODO: status(io::IO, pkg::String)
@@ -127,7 +150,7 @@ function clone(url::String, pkg::String)
         Git.run(`clone -q $url $pkg`)
         Git.set_remote_url(url, dir=pkg)
     catch
-        run(`rm -rf $pkg`)
+        Base.rm(pkg, recursive=true)
         rethrow()
     end
     isempty(Reqs.parse("$pkg/REQUIRE")) && return
@@ -143,17 +166,17 @@ function clone(url_or_pkg::String)
         # TODO: Cache.prefetch(pkg,url)
     else
         url = url_or_pkg
-        m = match(r"/(\w+?)(?:\.jl)?(?:\.git)?$", url)
+        m = match(r"(?:^|[/\\])(\w+?)(?:\.jl)?(?:\.git)?$", url)
         m != nothing || error("can't determine package name from URL: $url")
         pkg = m.captures[1]
     end
     clone(url,pkg)
 end
 
-function _checkout(pkg::String, what::String, merge::Bool=false, pull::Bool=false)
+function _checkout(pkg::String, what::String, merge::Bool=false, pull::Bool=false, branch::Bool=false)
     Git.transact(dir=pkg) do
         Git.dirty(dir=pkg) && error("$pkg is dirty, bailing")
-        Git.run(`checkout -q $what`, dir=pkg)
+        branch ? Git.run(`checkout -q -B $what -t origin/$what`, dir=pkg) : Git.run(`checkout -q $what`, dir=pkg)
         merge && Git.run(`merge -q --ff-only $what`, dir=pkg)
         if pull
             info("Pulling $pkg latest $what...")
@@ -166,7 +189,7 @@ end
 function checkout(pkg::String, branch::String, merge::Bool, pull::Bool)
     ispath(pkg,".git") || error("$pkg is not a git repo")
     info("Checking out $pkg $branch...")
-    _checkout(pkg,branch,merge,pull)
+    _checkout(pkg,branch,merge,pull,true)
 end
 
 function free(pkg::String)
@@ -251,14 +274,42 @@ function update(branch::String)
     end
     info("Computing changes...")
     resolve(Reqs.parse("REQUIRE"), avail, instd, fixed, free)
+    # Don't use instd here since it may have changed
+    updatehook(sort!([keys(installed())...]))
+end
+
+function pull_request(dir::String, commit::String="", url::String="")
+    commit = isempty(commit) ? Git.head(dir=dir) :
+        Git.readchomp(`rev-parse --verify $commit`, dir=dir)
+    isempty(url) && (url = Git.readchomp(`config remote.origin.url`, dir=dir))
+    m = match(Git.GITHUB_REGEX, url)
+    m == nothing && error("not a GitHub repo URL, can't make a pull request: $url")
+    owner, repo = m.captures[2:3]
+    user = GitHub.user()
+    info("Forking $owner/$repo to $user")
+    response = GitHub.fork(owner,repo)
+    fork = response["ssh_url"]
+    branch = "pull-request/$(commit[1:8])"
+    info("Pushing changes as branch $branch")
+    Git.run(`push -q $fork $commit:refs/heads/$branch`, dir=dir)
+    pr_url = "$(response["html_url"])/compare/$branch"
+    @osx? run(`open $pr_url`) : info("To create a pull-request, open:\n\n  $pr_url\n")
+end
+
+function submit(pkg::String, commit::String="")
+    urlpath = joinpath("METADATA",pkg,"url")
+    url = ispath(urlpath) ? readchomp(urlpath) : ""
+    pull_request(pkg, commit, url)
 end
 
 function publish(branch::String)
     Git.branch(dir="METADATA") == branch ||
         error("METADATA must be on $branch to publish changes")
-    Git.success(`push -q -n origin $branch`, dir="METADATA") ||
-        error("METADATA is behind origin/$branch – run Pkg.update() before publishing")
     Git.run(`fetch -q`, dir="METADATA")
+    ahead_remote, ahead_local = map(int,split(Git.readchomp(`rev-list --count --left-right origin/$branch...$branch`, dir="METADATA"),'\t'))
+    ahead_remote > 0 && error("METADATA is behind origin/$branch – run `Pkg.update()` before publishing")
+    ahead_local == 0 && error("There are no METADATA changes to publish")
+
     info("Validating METADATA")
     check_metadata()
     tags = Dict{ASCIIString,Vector{ASCIIString}}()
@@ -272,7 +323,7 @@ function publish(branch::String)
         sha1 = readchomp(joinpath("METADATA",path))
         if Git.success(`cat-file -e origin/$branch:$path`, dir="METADATA")
             old = Git.readchomp(`cat-file blob origin/$branch:$path`, dir="METADATA")
-            old == sha1 || error("$pkg v$ver SHA1 changed in METADATA – refusing to push")
+            old == sha1 || error("$pkg v$ver SHA1 changed in METADATA – refusing to publish")
         end
         any(split(Git.readall(`tag --contains $sha1`, dir=pkg))) do tag
             ver == convert(VersionNumber,tag) || return false
@@ -281,7 +332,7 @@ function publish(branch::String)
             return true
         end || error("$pkg v$ver is incorrectly tagged – $sha1 expected")
     end
-    isempty(tags) && info("No new package versions to publish.")
+    isempty(tags) && info("No new package versions to publish")
     @sync for pkg in sort!([keys(tags)...])
         @async begin
             forced = ASCIIString[]
@@ -302,8 +353,8 @@ function publish(branch::String)
             end
         end
     end
-    info("Pushing METADATA changes")
-    Git.run(`push -q origin $branch`, dir="METADATA")
+    info("Submitting METADATA changes")
+    pull_request("METADATA")
 end
 
 function resolve(
@@ -318,8 +369,12 @@ function resolve(
 
     for pkg in keys(reqs)
         if !haskey(deps,pkg)
-            error("$pkg's requirements can't be satisfied because of the following fixed packages: ",
+            if "julia" in conflicts[pkg]
+                error("$pkg can't be installed because it has no versions that support ", VERSION, " of julia")
+            else
+                error("$pkg's requirements can't be satisfied because of the following fixed packages: ",
                    join(conflicts[pkg], ", ", " and "))
+            end
         end
     end
 
@@ -330,7 +385,7 @@ function resolve(
 
     # compare what is installed with what should be
     changes = Query.diff(have, want, avail, fixed)
-    isempty(changes) && return info("No packages to install, update or remove.")
+    isempty(changes) && return info("No packages to install, update or remove")
 
     # prefetch phase isolates network activity, nothing to roll back
     missing = {}
@@ -545,11 +600,12 @@ function check_metadata()
 end
 
 function warnbanner(msg...; label="[ WARNING ]", prefix="")
-    warn(prefix="", Base.cpad(label,Base.tty_cols(),"="))
+    cols = Base.tty_size()[2]
+    warn(prefix="", Base.cpad(label,cols,"="))
     println(STDERR)
     warn(prefix=prefix, msg...)
     println(STDERR)
-    warn(prefix="", "="^Base.tty_cols())
+    warn(prefix="", "="^cols)
 end
 
 function build!(pkgs::Vector, errs::Dict, seen::Set=Set())
@@ -560,7 +616,7 @@ function build!(pkgs::Vector, errs::Dict, seen::Set=Set())
         isfile(path) || continue
         info("Building $pkg")
         cd(dirname(path)) do
-            try eval(Module(),Expr(:toplevel,:(ARGS={}),:(include($path))))
+            try evalfile(path)
             catch err
                 warnbanner(err, label="[ ERROR: $pkg ]")
                 errs[pkg] = err
@@ -584,4 +640,82 @@ function build(pkgs::Vector)
 end
 build() = build(sort!([keys(installed())...]))
 
+function updatehook!(pkgs::Vector, errs::Dict, seen::Set=Set())
+    for pkg in pkgs
+        pkg in seen && continue
+        updatehook!(Read.requires_list(pkg),errs,push!(seen,pkg))
+        path = abspath(pkg,"deps","update.jl")
+        isfile(path) || continue
+        info("Running update script for $pkg")
+        cd(dirname(path)) do
+            try evalfile(path)
+            catch err
+                warnbanner(err, label="[ ERROR: $pkg ]")
+                errs[pkg] = err
+            end
+        end
+    end
+end
+
+function updatehook(pkgs::Vector)
+    errs = Dict()
+    updatehook!(pkgs,errs)
+    isempty(errs) && return
+    println(STDERR)
+    warnbanner(label="[ UPDATE ERRORS ]", """
+    WARNING: $(join(map(x->x[1],errs),", "," and ")) had update errors.
+
+     - Unrelated packages are unaffected
+     - To retry, run Pkg.update() again
+    """)
+end
+
+function test!(pkg::String, errs::Vector{String}, notests::Vector{String}; coverage::Bool=false)
+    reqs_path = abspath(pkg,"test","REQUIRE")
+    if isfile(reqs_path)
+        tests_require = Reqs.parse(reqs_path)
+        if (!isempty(tests_require))
+            info("Computing test dependencies for $pkg...")
+            resolve(merge(Reqs.parse("REQUIRE"), tests_require))
+        end
+    end
+    test_path = abspath(pkg,"test","runtests.jl")
+    if isfile(test_path)
+        info("Testing $pkg")
+        cd(dirname(test_path)) do
+            try
+                if coverage
+                    cmd = `$JULIA_HOME/julia --code-coverage $test_path`
+                else
+                    cmd = `$JULIA_HOME/julia $test_path`
+                end
+                run(cmd)
+                info("$pkg tests passed")
+            catch err
+                warnbanner(err, label="[ ERROR: $pkg ]")
+                push!(errs,pkg)
+            end
+        end
+    else
+        push!(notests,pkg)
+    end
+    resolve()
+end
+
+function test(pkgs::Vector{String}; coverage::Bool=false)
+    errs = String[]
+    notests = String[]
+    for pkg in pkgs
+        test!(pkg,errs,notests; coverage=coverage)
+    end
+    if !isempty(errs) || !isempty(notests)
+        messages = String[]
+        isempty(errs) || push!(messages, "$(join(errs,", "," and ")) had test errors")
+        isempty(notests) || push!(messages, "$(join(notests,", "," and ")) did not provide a test/runtests.jl file")
+        error(join(messages, "and"))
+    end
+end
+
+test(;coverage::Bool=false) = test(sort!(String[keys(installed())...]); coverage=coverage)
+
 end # module
