@@ -1,6 +1,6 @@
 module Write
 
-using Base.Git, ..Cache, ..Read
+import ..Git, ..Cache, ..Read
 
 function prefetch(pkg::String, sha1::String)
     isempty(Cache.prefetch(pkg, Read.url(pkg), sha1)) && return
@@ -8,7 +8,7 @@ function prefetch(pkg::String, sha1::String)
 end
 
 function fetch(pkg::String, sha1::String)
-    refspec = "+refs/heads/*:refs/remotes/cache/*"
+    refspec = "+refs/*:refs/remotes/cache/*"
     Git.run(`fetch -q $(Cache.path(pkg)) $refspec`, dir=pkg)
     Git.iscommit(sha1, dir=pkg) && return
     f = Git.iscommit(sha1, dir=Cache.path(pkg)) ? "fetch" : "prefetch"
@@ -23,7 +23,7 @@ end
 function install(pkg::String, sha1::String)
     prefetch(pkg, sha1)
     if isdir(".trash/$pkg")
-        run(`mv .trash/$pkg ./`)
+        mv(".trash/$pkg", "./$pkg")
     else
         Git.run(`clone -q $(Cache.path(pkg)) $pkg`)
     end
@@ -39,8 +39,8 @@ end
 
 function remove(pkg::String)
     isdir(".trash") || mkdir(".trash")
-    ispath(".trash/$pkg") && run(`rm -rf .trash/$pkg`)
-    run(`mv $pkg .trash/`)
+    ispath(".trash/$pkg") && rm(".trash/$pkg", recursive=true)
+    mv(pkg, ".trash/$pkg")
 end
 
 end # module
