@@ -4,10 +4,13 @@ include("pcre.jl")
 
 const DEFAULT_OPTS = PCRE.JAVASCRIPT_COMPAT | PCRE.UTF8 | PCRE.NO_UTF8_CHECK
 
-immutable Regex
+type Regex
     pattern::ByteString
     options::Uint32
-    regex::Array{Uint8}
+    regex::Ptr{Void}
+    extra::Ptr{Void}
+    ovec::Vector{Int32}
+
 
     function Regex(pattern::String, options::Integer)
         pattern = bytestring(pattern)
@@ -15,8 +18,13 @@ immutable Regex
         if (options & ~PCRE.OPTIONS_MASK) != 0
             error("invalid regex options: $options")
         end
-        regex = PCRE.compile(pattern, options & PCRE.COMPILE_MASK)
-        new(pattern, options, regex)
+        re = compile(new(pattern, options, C_NULL, C_NULL, Array(Int32, 0)))
+        finalizer(re,
+            function(re::Regex)
+                re.extra != C_NULL && PCRE.free_study(re.extra)
+                re.regex != C_NULL && PCRE.free(re.regex)
+            end)
+        re
     end
 end
 
@@ -33,7 +41,19 @@ function Regex(pattern::String, flags::String)
 end
 Regex(pattern::String) = Regex(pattern, DEFAULT_OPTS)
 
+function compile(regex::Regex)
+    if regex.regex == C_NULL
+        regex.regex = PCRE.compile(regex.pattern, regex.options & PCRE.COMPILE_MASK)
+        regex.extra = PCRE.study(regex.regex, PCRE.STUDY_JIT_COMPILE)
+        ncap  = PCRE.info(regex.regex, regex.extra,
+                          PCRE.INFO_CAPTURECOUNT, Int32)
+        resize!(regex.ovec, 3(ncap+1))
+    end
+    regex
+end
+
 macro r_str(pattern, flags...) Regex(pattern, flags...) end
+macro r_mstr(pattern, flags...) Regex(pattern, flags...) end
 
 copy(r::Regex) = r
 
@@ -81,21 +101,30 @@ function show(io::IO, m::RegexMatch)
     print(io, ")")
 end
 
-ismatch(r::Regex, s::String, offset::Integer=0) =
-    PCRE.exec(r.regex, C_NULL, bytestring(s), offset, r.options & PCRE.EXECUTE_MASK, false)
-ismatch(r::Regex, s::SubString, offset::Integer=0) =
-    PCRE.exec(r.regex, C_NULL, s, offset, r.options & PCRE.EXECUTE_MASK, false)
+function ismatch(r::Regex, s::String, offset::Integer=0)
+    compile(r)
+    return PCRE.exec(r.regex, r.extra, bytestring(s), offset, r.options & PCRE.EXECUTE_MASK,
+                     r.ovec)
+end
 
-function match(re::Regex, str::UTF8String, idx::Integer, add_opts::Uint32=uint32(0),
-               extra::Ptr{Void}=C_NULL)
+function ismatch(r::Regex, s::SubString, offset::Integer=0)
+    compile(r)
+    return PCRE.exec(r.regex, r.extra, s, offset, r.options & PCRE.EXECUTE_MASK,
+                  r.ovec)
+end
+
+function match(re::Regex, str::UTF8String, idx::Integer, add_opts::Uint32=uint32(0))
     opts = re.options & PCRE.EXECUTE_MASK | add_opts
-    m, n = PCRE.exec(re.regex, extra, str, idx-1, opts, true)
-    if isempty(m); return nothing; end
-    mat = SubString(str, m[1]+1, m[2])
+    compile(re)
+    if !PCRE.exec(re.regex, re.extra, str, idx-1, opts, re.ovec)
+        return nothing
+    end
+    n = length(re.ovec)/3 - 1
+    mat = SubString(str, re.ovec[1]+1, re.ovec[2])
     cap = Union(Nothing,SubString{UTF8String})[
-            m[2i+1] < 0 ? nothing : SubString(str, m[2i+1]+1, m[2i+2]) for i=1:n ]
-    off = Int[ m[2i+1]::Int32+1 for i=1:n ]
-    RegexMatch(mat, cap, m[1]+1, off)
+            re.ovec[2i+1] < 0 ? nothing : SubString(str, re.ovec[2i+1]+1, re.ovec[2i+2]) for i=1:n ]
+    off = Int[ re.ovec[2i+1]::Int32+1 for i=1:n ]
+    RegexMatch(mat, cap, re.ovec[1]+1, off)
 end
 
 match(re::Regex, str::Union(ByteString,SubString), idx::Integer, add_opts::Uint32=uint32(0)) =
@@ -106,7 +135,8 @@ match(r::Regex, s::String, i::Integer) =
     error("regex matching is only available for bytestrings; use bytestring(s) to convert")
 
 function matchall(re::Regex, str::UTF8String, overlap::Bool=false)
-    extra = PCRE.study(re.regex, PCRE.STUDY_JIT_COMPILE)
+    regex = compile(re).regex
+    extra = re.extra
     n = length(str.data)
     matches = SubString{UTF8String}[]
     offset = int32(0)
@@ -118,7 +148,7 @@ function matchall(re::Regex, str::UTF8String, overlap::Bool=false)
         result = ccall((:pcre_exec, :libpcre), Int32,
                        (Ptr{Void}, Ptr{Void}, Ptr{Uint8}, Int32,
                        Int32, Int32, Ptr{Int32}, Int32),
-                       re.regex, extra, str, n,
+                       regex, extra, str, n,
                        offset, prevempty ? opts_nonempty : opts, ovec, 3)
 
         if result < 0
@@ -141,7 +171,6 @@ function matchall(re::Regex, str::UTF8String, overlap::Bool=false)
             offset = ovec[2]
         end
     end
-    PCRE.free_study(extra)
     matches
 end
 
@@ -153,8 +182,9 @@ function search(str::Union(ByteString,SubString), re::Regex, idx::Integer)
         throw(BoundsError())
     end
     opts = re.options & PCRE.EXECUTE_MASK
-    m, n = PCRE.exec(re.regex, C_NULL, str, idx-1, opts, true)
-    isempty(m) ? (0:-1) : ((m[1]+1):prevind(str,m[2]+1))
+    compile(re)
+    PCRE.exec(re.regex, re.extra, str, idx-1, opts, re.ovec) ?
+        ((re.ovec[1]+1):prevind(str,re.ovec[2]+1)) : (0:-1)
 end
 search(s::String, r::Regex, idx::Integer) =
     error("regex search is only available for bytestrings; use bytestring(s) to convert")
@@ -164,16 +194,14 @@ immutable RegexMatchIterator
     regex::Regex
     string::UTF8String
     overlap::Bool
-    extra::Ptr{Void}
 
     function RegexMatchIterator(regex::Regex, string::String, ovr::Bool=false)
-        extra = PCRE.study(regex.regex, PCRE.STUDY_JIT_COMPILE)
-        new(regex, string, ovr, extra)
+        new(regex, string, ovr)
     end
 end
-
+compile(itr::RegexMatchIterator) = (compile(itr.regex); itr)
 eltype(itr::RegexMatchIterator) = RegexMatch
-start(itr::RegexMatchIterator) = match(itr.regex, itr.string, 1, uint32(0), itr.extra)
+start(itr::RegexMatchIterator) = match(itr.regex, itr.string, 1, uint32(0))
 done(itr::RegexMatchIterator, prev_match) = (prev_match == nothing)
 
 # Assumes prev_match is not nothing
@@ -193,7 +221,7 @@ function next(itr::RegexMatchIterator, prev_match)
     opts_nonempty = uint32(PCRE.ANCHORED | PCRE.NOTEMPTY_ATSTART)
     while true
         mat = match(itr.regex, itr.string, offset,
-                    prevempty ? opts_nonempty : uint32(0), itr.extra)
+                    prevempty ? opts_nonempty : uint32(0))
 
         if mat === nothing
             if prevempty && offset <= length(itr.string.data)
@@ -207,8 +235,6 @@ function next(itr::RegexMatchIterator, prev_match)
             return (prev_match, mat)
         end
     end
-
-    PCRE.free_study(itr.extra)
     (prev_match, nothing)
 end
 
@@ -225,3 +251,17 @@ filter(r::Regex, v)  = filter(x->ismatch(r,x), v)
 
 filter!(r::Regex, d::Dict) = filter!((k,v)->ismatch(r,k),d)
 filter(r::Regex,  d::Dict) = filter!(r,copy(d))
+
+
+# Don't serialize the pointers
+function serialize(s, r::Regex)
+    serialize_type(s, typeof(r))
+    serialize(s, r.pattern) 
+    serialize(s, r.options) 
+end
+
+function deserialize(s, t::Type{Regex})
+    pattern = deserialize(s)
+    options = deserialize(s)
+    Regex(pattern, options)        
+end
