@@ -1,26 +1,46 @@
+
 show(x) = show(STDOUT::IO, x)
 
-function print(io::IO, s::Symbol)
-    pname = convert(Ptr{Uint8}, s)
-    write(io, pname, int(ccall(:strlen, Csize_t, (Ptr{Uint8},), pname)))
-end
+print(io::IO, s::Symbol) = (write(io,s);nothing)
 
 function show(io::IO, x::ANY)
     t = typeof(x)::DataType
     show(io, t)
     print(io, '(')
     if t.names !== () || t.size==0
-        n = length(t.names)
-        for i=1:n
-            f = t.names[i]
-            if !isdefined(x, f)
-                print(io, undef_ref_str)
+        recorded = false
+        oid = object_id(x)
+        shown_set = get(task_local_storage(), :SHOWNSET, nothing)
+        if shown_set == nothing 
+            shown_set = Set()
+            task_local_storage(:SHOWNSET, shown_set) 
+        end
+        
+        try
+            if oid in shown_set
+                print(io, "#= circular reference =#")
             else
-                show(io, x.(f))
-            end
-            if i < n
-                print(io, ',')
+                push!(shown_set, oid)
+                recorded = true
+            
+                n = length(t.names)
+                for i=1:n
+                    f = t.names[i]
+                    if !isdefined(x, f)
+                        print(io, undef_ref_str)
+                    else
+                        show(io, x.(f))
+                    end
+                    if i < n
+                        print(io, ',')
+                    end
+                end
             end
+        catch e
+            rethrow(e)
+        
+        finally
+            if recorded delete!(shown_set, oid) end
         end
     else
         nb = t.size
@@ -70,20 +90,12 @@ function show(io::IO, x::DataType)
     end
 end
 
-function showcompact{T<:Number}(io::IO, x::Vector{T})
-    print(io, "[")
-    if length(x) > 0
-        showcompact(io, x[1])
-        for j in 2:length(x)
-            print(io, ",")
-            showcompact(io, x[j])
-        end
-    end
-    print(io, "]")
-end
 showcompact(io::IO, x) = show(io, x)
 showcompact(x) = showcompact(STDOUT::IO, x)
 
+showcompact_lim(io, x) = _limit_output ? showcompact(io, x) : show(io, x)
+showcompact_lim(io, x::Number) = _limit_output ? showcompact(io, x) : print(io, x)
+
 macro show(exs...)
     blk = Expr(:block)
     for ex in exs
@@ -94,7 +106,6 @@ macro show(exs...)
     return blk
 end
 
-show(io::IO, s::Symbol) = show_indented(io, s)
 show(io::IO, tn::TypeName) = print(io, tn.name)
 show(io::IO, ::Nothing) = print(io, "nothing")
 show(io::IO, b::Bool) = print(io, b ? "true" : "false")
@@ -119,6 +130,41 @@ function show(io::IO, l::LambdaStaticData)
     print(io, ")")
 end
 
+function show_delim_array(io::IO, itr::AbstractArray, op, delim, cl, delim_one, compact=false)
+    print(io, op)
+    newline = true
+    first = true
+    i = 1
+    l = length(itr)
+    if l > 0
+        while true
+            if !isassigned(itr, i)
+                print(io, undef_ref_str)
+                multiline = false
+            else
+                x = itr[i]
+                multiline = isa(x,AbstractArray) && ndims(x)>1 && length(x)>0
+                newline && multiline && println(io)
+                compact ? showcompact_lim(io, x) : show(io, x)
+            end
+            i += 1
+            if i > l
+                delim_one && first && print(io, delim)
+                break
+            end
+            first = false
+            print(io, delim)
+            if multiline
+                println(io); println(io)
+                newline = false
+            else
+                newline = true
+            end
+        end
+    end
+    print(io, cl)
+end
+
 function show_delim_array(io::IO, itr, op, delim, cl, delim_one)
     print(io, op)
     state = start(itr)
@@ -126,28 +172,19 @@ function show_delim_array(io::IO, itr, op, delim, cl, delim_one)
     first = true
     if !done(itr,state)
         while true
-            if isa(itr,Array) && !isdefined(itr,state)
-                print(io, undef_ref_str)
-                state += 1
-                multiline = false
-            else
-                x, state = next(itr,state)
-                multiline = isa(x,AbstractArray) && ndims(x)>1 && length(x)>0
-                if newline
-                    if multiline; println(io); end
-                end
-                show(io, x)
-            end
+            x, state = next(itr,state)
+            multiline = isa(x,AbstractArray) && ndims(x)>1 && length(x)>0
+            newline && multiline && println(io)
+            show(io, x)
             if done(itr,state)
-                if delim_one && first
-                    print(io, delim)
-                end
+                delim_one && first && print(io, delim)
                 break
             end
             first = false
             print(io, delim)
             if multiline
-                println(io); println(io); newline=false
+                println(io); println(io)
+                newline = false
             else
                 newline = true
             end
@@ -159,8 +196,95 @@ end
 show_comma_array(io::IO, itr, o, c) = show_delim_array(io, itr, o, ',', c, false)
 show(io::IO, t::Tuple) = show_delim_array(io, t, '(', ',', ')', true)
 
+show(io::IO, s::Symbol) = show_unquoted(io, QuoteNode(s))
+show(io::IO, tn::TypeName) = print(io, tn.name)
+show(io::IO, ::Nothing) = print(io, "nothing")
+show(io::IO, b::Bool) = print(io, b ? "true" : "false")
+show(io::IO, n::Signed) = (write(io, dec(n)); nothing)
+show(io::IO, n::Unsigned) = print(io, "0x", hex(n,sizeof(n)<<1))
+print(io::IO, n::Unsigned) = print(io, dec(n))
+
+## Abstract Syntax Tree (AST) printing ##
+
+# Summary:
+#   print(io, ex) defers to show_unquoted(io, ex)
+#   show(io, ex) defers to show_unquoted(io, QuoteNode(ex))
+#   show_unquoted(io, ex) does the heavy lifting
+# 
+# AST printing should follow two rules:
+#   1. parse(string(ex)) == ex
+#   2. eval(parse(repr(ex))) == ex
+#
+# Rule 1 means that printing an expression should generate Julia code which
+# could be reparsed to obtain the original expression. This code should be
+# unambiguous and as readable as possible.
+#
+# Rule 2 means that showing an expression should generate a quoted version of
+# print’s output. Parsing and then evaling this output should return the
+# original expression.
+#
+# This is consistent with many other show methods, i.e.:
+#   show(Set([1,2,3]))                # ==> "Set{Int64}([2,3,1])"
+#   eval(parse("Set{Int64}([2,3,1])”) # ==> An actual set
+# While this isn’t true of ALL show methods, it is of all ASTs.
+
+typealias ExprNode Union(Expr, QuoteNode, SymbolNode, LineNumberNode,
+                         LabelNode, GotoNode, TopNode)
+print        (io::IO, ex::ExprNode)    = show_unquoted(io, ex)
+show         (io::IO, ex::ExprNode)    = show_unquoted(io, QuoteNode(ex))
+show_unquoted(io::IO, ex)              = show_unquoted(io, ex, 0, 0)
+show_unquoted(io::IO, ex, indent::Int) = show_unquoted(io, ex, indent, 0)
+show_unquoted(io::IO, ex, ::Int,::Int) = show(io, ex)
+
+## AST printing constants ##
+
+const indent_width = 4
+const quoted_syms = Set{Symbol}([:(:),:(::),:(:=),:(=),:(==),:(===),:(=>)])
+const uni_ops = Set{Symbol}([:(+), :(-), :(!), :(~), :(<:), :(>:)])
+const bin_ops_by_prec = [
+    "= := += -= *= /= //= .//= .*= ./= \\= .\\= ^= .^= %= .%= |= &= \$= => <<= >>= >>>= ~ .+= .-=",
+    "?",
+    "||",
+    "&&",
+    "-- -->",
+    "> < >= <= == === != !== .> .< .>= .<= .== .!= .= .! <: >:",
+    "|> <|",
+    ": ..",
+    "+ - .+ .- | \$",
+    "<< >> >>> .<< .>> .>>>",
+    "* / ./ % .% & .* \\ .\\",
+    "// .//",
+    "^ .^",
+    "::",
+    "."
+]
+const bin_op_precs = Dict{Symbol,Int}(merge([{symbol(op)=>i for op=split(bin_ops_by_prec[i])} for i=1:length(bin_ops_by_prec)]...))
+const bin_ops = Set{Symbol}(keys(bin_op_precs))
+const expr_infix_wide = Set([:(=), :(+=), :(-=), :(*=), :(/=), :(\=), :(&=),
+    :(|=), :($=), :(>>>=), :(>>=), :(<<=), :(&&), :(||)])
+const expr_infix = Set([:(:), :(<:), :(->), :(=>), symbol("::")])
+const expr_calls  = [:call =>('(',')'), :calldecl =>('(',')'), :ref =>('[',']'), :curly =>('{','}')]
+const expr_parens = [:tuple=>('(',')'), :vcat=>('[',']'), :cell1d=>('{','}'),
+                     :hcat =>('[',']'), :row =>('[',']')]
+
 ## AST decoding helpers ##
 
+is_id_start_char(c::Char) = ccall(:jl_id_start_char, Cint, (Uint32,), c) != 0
+is_id_char(c::Char) = ccall(:jl_id_char, Cint, (Uint32,), c) != 0
+function isidentifier(s::String)
+    i = start(s)
+    done(s, i) && return false
+    (c, i) = next(s, i)
+    is_id_start_char(c) || return false
+    while !done(s, i)
+        (c, i) = next(s, i)
+        is_id_char(c) || return false
+    end
+    return true
+end
+isoperator(s::ByteString) = ccall(:jl_is_operator, Cint, (Ptr{Uint8},), s) != 0
+isoperator(s::String) = isoperator(bytestring(s))
+
 is_expr(ex, head::Symbol)         = (isa(ex, Expr) && (ex.head == head))
 is_expr(ex, head::Symbol, n::Int) = is_expr(ex, head) && length(ex.args) == n
 
@@ -168,63 +292,13 @@ is_linenumber(ex::LineNumberNode) = true
 is_linenumber(ex::Expr)           = is(ex.head, :line)
 is_linenumber(ex)                 = false
 
+is_quoted(ex)            = false
 is_quoted(ex::QuoteNode) = true
 is_quoted(ex::Expr)      = is_expr(ex, :quote, 1)
-is_quoted(ex)            = false
 
 unquoted(ex::QuoteNode)  = ex.value
 unquoted(ex::Expr)       = ex.args[1]
 
-## AST printing ##
-
-# show(ex::Expr) delegates to show_indented(), 
-# which shows the contents using show_unquoted(),
-# which shows subexpressions using show_unquoted()
-# ==> AST:s are printed wrapped in a single quotation
-show_indented(x)                 = show_indented(STDOUT, x)
-show_indented(io::IO, x)         = show_indented(io, x, 0)
-show_indented(io::IO, x, indent) = show(io, x)
-show_unquoted(x)                 = show_unquoted(STDOUT, x)
-show_unquoted(io::IO, x)         = show_unquoted(io, x, 0)
-show_unquoted(io::IO, x, indent) = (print(io,"\$("); show(io,x); print(io,')'))
-
-## Quoted AST printing ##
-
-typealias ExprNode Union(SymbolNode, LineNumberNode, LabelNode, GotoNode,
-                         TopNode, QuoteNode)
-show(io::IO, ex::ExprNode) = show_indented(io, ex)
-
-function show_indented(io::IO, ex::ExprNode, indent::Int)
-    default_show_quoted(io, ex, indent)
-end
-function show_indented(io::IO, ex::QuoteNode, indent::Int)
-    show_indented(io, ex.value, indent)
-end
-
-show(io::IO, ex::Expr) = show_indented(io, ex)
-function show_indented(io::IO, ex::Expr, indent::Int)
-    if is(ex.head, :block) || is(ex.head, :body)
-        show_block(io, "quote", ex, indent); print(io, "end")
-    elseif in(ex.head, (:tuple, :vcat, :cell1))
-        print(io, ':'); show_unquoted(io, ex, indent + indent_width)        
-    else
-        default_show_quoted(io, ex, indent)
-    end
-end
-const paren_quoted_syms = Set{Symbol}(:(:),:(::),:(:=),:(=),:(==),:(===),:(=>))
-function show_indented(io::IO, sym::Symbol, indent::Int)
-    if in(sym, paren_quoted_syms)
-        print(io, ":($sym)")
-    else
-        print(io, ":$sym")
-    end
-end
-function default_show_quoted(io::IO, ex, indent::Int)
-    print(io, ":(")
-    show_unquoted(io, ex, indent + indent_width)
-    print(io, ")")
-end
-
 ## AST printing helpers ##
 
 const indent_width = 4
@@ -258,212 +332,326 @@ function show_block(io::IO, head, args::Vector, body, indent::Int)
     print(io, '\n', " "^indent)
 end
 show_block(io::IO,head,    block,i::Int) = show_block(io,head,{},   block,i)
-show_block(io::IO,head,arg,block,i::Int) = show_block(io,head,{arg},block,i)
+function show_block(io::IO, head, arg, block, i::Int)
+    if is_expr(arg, :block)
+        show_block(io, head, arg.args, block, i)
+    else
+        show_block(io, head, {arg}, block, i)
+    end
+end
 
 # show an indented list
-function show_list(io::IO, items, sep, indent::Int)
+function show_list(io::IO, items, sep, indent::Int, prec::Int=0)
     n = length(items)
     if n == 0; return end
     indent += indent_width
-    show_unquoted(io, items[1], indent)
+    show_unquoted(io, items[1], indent, prec)
     for item in items[2:end]
         print(io, sep)
-        show_unquoted(io, item, indent)        
+        show_unquoted(io, item, indent, prec)
     end
 end
 # show an indented list inside the parens (op, cl)
-function show_enclosed_list(io::IO, op, items, sep, cl, indent)
-    print(io, op); show_list(io, items, sep, indent); print(io, cl)
+function show_enclosed_list(io::IO, op, items, sep, cl, indent, prec=0)
+    print(io, op); show_list(io, items, sep, indent, prec); print(io, cl)
 end
 
-## Unquoted AST printing ##
+# show a normal (non-operator) function call, e.g. f(x,y) or A[z]
+function show_call(io::IO, head, func, func_args, indent)
+    op, cl = expr_calls[head]
+    if isa(func, Symbol) || (isa(func, Expr) && func.head == :.)
+        show_unquoted(io, func, indent)
+    else
+        print(io, '(')
+        show_unquoted(io, func, indent)
+        print(io, ')')
+    end
+    if !isempty(func_args) && isa(func_args[1], Expr) && func_args[1].head === :parameters
+        print(io, op)
+        show_list(io, func_args[2:end], ',', indent, 0)
+        print(io, "; ")
+        show_list(io, func_args[1].args, ',', indent, 0)
+        print(io, cl)
+    else
+        show_enclosed_list(io, op, func_args, ",", cl, indent)
+    end
+end
 
-show_unquoted(io::IO, sym::Symbol, indent::Int) = print(io, sym)
-show_unquoted(io::IO, x::Number, indent::Int)   = show(io, x)
-show_unquoted(io::IO, x::String, indent::Int)   = show(io, x)
-show_unquoted(io::IO, x::Char, indent::Int)     = show(io, x)
+## AST printing ##
 
-const _expr_infix_wide = Set(:(=), :(+=), :(-=), :(*=), :(/=), :(\=), :(&=), 
-    :(|=), :($=), :(>>>=), :(>>=), :(<<=), :(&&), :(||))
-const _expr_infix = Set(:(:), :(<:), :(->), :(=>), symbol("::"))
-const _expr_calls  = [:call =>('(',')'), :ref =>('[',']'), :curly =>('{','}')]
-const _expr_parens = [:tuple=>('(',')'), :vcat=>('[',']'), :cell1d=>('{','}')]
+show_unquoted(io::IO, sym::Symbol, ::Int, ::Int)        = print(io, sym)
+show_unquoted(io::IO, ex::LineNumberNode, ::Int, ::Int) = show_linenumber(io, ex.line)
+show_unquoted(io::IO, ex::LabelNode, ::Int, ::Int)      = print(io, ex.label, ": ")
+show_unquoted(io::IO, ex::GotoNode, ::Int, ::Int)       = print(io, "goto ", ex.label)
+show_unquoted(io::IO, ex::TopNode, ::Int, ::Int)        = print(io,"top(",ex.name,')')
 
-function show_unquoted(io::IO, ex::Expr, indent::Int)
+function show_unquoted(io::IO, ex::SymbolNode, ::Int, ::Int)
+    print(io, ex.name)
+    show_expr_type(io, ex.typ)
+end
+
+show_unquoted(io::IO, ex::QuoteNode, indent::Int, prec::Int) =
+    show_unquoted_quote_expr(io, ex.value, indent, prec)
+
+function show_unquoted_quote_expr(io::IO, value, indent::Int, prec::Int)
+    if isa(value, Symbol) && !(value in quoted_syms)
+        s = string(value)
+        if isidentifier(s) || isoperator(s)
+            print(io, ":")
+            print(io, value)
+        else
+            print(io, "symbol(\"", escape_string(s), "\")")
+        end
+    else
+        if isa(value,Expr) && value.head === :block
+            show_block(io, "quote", value, indent)
+            print(io, "end")
+        else
+            print(io, ":(")
+            show_unquoted(io, value, indent+indent_width, 0)
+            print(io, ")")
+        end
+    end
+end
+
+# TODO: implement interpolated strings
+function show_unquoted(io::IO, ex::Expr, indent::Int, prec::Int)
     head, args, nargs = ex.head, ex.args, length(ex.args)
 
+    # dot (i.e. "x.y")
     if is(head, :(.))
         show_unquoted(io, args[1], indent + indent_width)
         print(io, '.')
-        if is_quoted(args[2]) 
+        if is_quoted(args[2])
             show_unquoted(io, unquoted(args[2]), indent + indent_width)
         else
             print(io, '(')
             show_unquoted(io, args[2], indent + indent_width)
             print(io, ')')
-        end                  
-    elseif (in(head, _expr_infix) && nargs==2) || (is(head,:(:)) && nargs==3)
+        end
+
+    # infix (i.e. "x<:y" or "x = y")
+    elseif (head in expr_infix && nargs==2) || (is(head,:(:)) && nargs==3)
         show_list(io, args, head, indent)
-    elseif in(head, _expr_infix_wide) && nargs == 2
-        show_list(io, args, " $head ", indent)
-    elseif is(head, symbol("::")) && nargs == 1
-        print(io, "::")        
-        show_unquoted(io, args[1], indent)
-    elseif haskey(_expr_parens, head)               # :tuple/:vcat/:cell1d
-        op, cl = _expr_parens[head]
-        print(io, op)
-        show_list(io, args, ",", indent)
+
+    elseif head in expr_infix_wide && nargs == 2
+        func_prec = get(bin_op_precs, head, 0)
+        if func_prec < prec
+            show_enclosed_list(io, '(', args, " $head ", ')', indent, func_prec)
+        else
+            show_list(io, args, " $head ", indent, func_prec)
+        end
+
+    # list (i.e. "(1,2,3)" or "[1,2,3]")
+    elseif haskey(expr_parens, head)               # :tuple/:vcat/:cell1d
+        op, cl = expr_parens[head]
+        if head === :vcat && !isempty(args) && is_expr(args[1], :row)
+            sep = ";"
+        elseif head === :hcat || head === :row
+            sep = " "
+        else
+            sep = ","
+        end
+        head !== :row && print(io, op)
+        show_list(io, args, sep, indent)
         if is(head, :tuple) && nargs == 1; print(io, ','); end
-        print(io, cl)
-    elseif haskey(_expr_calls, head) && nargs >= 1  # :call/:ref/:curly
-        op, cl = _expr_calls[head]
+        head !== :row && print(io, cl)
+
+    # function declaration (like :call but always printed with parens)
+    # (:calldecl is a "fake" expr node created when we find a :function expr)
+    elseif head == :calldecl && nargs >= 1
+        show_call(io, head, args[1], args[2:end], indent)
+        
+    # function call
+    elseif haskey(expr_calls, head) && nargs >= 1  # :call/:ref/:curly
+        func = args[1]
+        func_prec = get(bin_op_precs, func, 0)
+        func_args = args[2:end]
+
+        # scalar multiplication (i.e. "100x")
+        if (func == :(*) && length(func_args)==2 &&
+            isa(func_args[1], Real) && isa(func_args[2], Symbol))
+            if func_prec <= prec
+                show_enclosed_list(io, '(', func_args, "", ')', indent, func_prec)
+            else
+                show_list(io, func_args, "", indent, func_prec)
+            end
+
+        # unary operator (i.e. "!z")
+        elseif func in uni_ops && length(func_args) == 1
+            show_unquoted(io, func, indent)
+            if isa(func_args[1], Expr) || length(func_args) > 1
+                show_enclosed_list(io, '(', func_args, ",", ')', indent, func_prec)
+            else
+                show_unquoted(io, func_args[1])
+            end
+
+        # binary operator (i.e. "x + y")
+        elseif func in bin_ops
+            if length(func_args) > 1
+                sep = func_prec >= bin_op_precs[:(^)] ? "$func" : " $func "
+                if func_prec <= prec
+                    show_enclosed_list(io, '(', func_args, sep, ')', indent, func_prec)
+                else
+                    show_list(io, func_args, sep, indent, func_prec)
+                end
+            else
+                # 1-argument call to normally-binary operator
+                op, cl = expr_calls[head]
+                print(io, "(")
+                show_unquoted(io, func, indent)
+                print(io, ")")
+                show_enclosed_list(io, op, func_args, ",", cl, indent)
+            end
+
+        # normal function (i.e. "f(x,y)" or "A[x,y]")
+        else
+            show_call(io, head, func, func_args, indent)
+        end
+
+    # typed comprehension
+    elseif is(head, :typed_comprehension) && length(args) == 3
         show_unquoted(io, args[1], indent)
-        show_enclosed_list(io, op, args[2:end], ",", cl, indent)
+        print(io, '[')
+        show_unquoted(io, args[2], indent)
+        print(io, " for ")
+        show_unquoted(io, args[3], indent)
+        print(io, ']')
+
+    elseif is(head, :ccall)
+        show_unquoted(io, :ccall, indent)
+        show_enclosed_list(io, '(', args, ",", ')', indent)
+
+    # comparison (i.e. "x < y < z")
     elseif is(head, :comparison) && nargs >= 3 && (nargs&1==1)
-        show_enclosed_list(io, '(', args, "", ')', indent)
+        comp_prec = minimum([get(bin_op_precs, comp, 0) for comp=args[2:2:end]])
+        if comp_prec <= prec
+            show_enclosed_list(io, '(', args, " ", ')', indent, comp_prec)
+        else
+            show_list(io, args, " ", indent, comp_prec)
+        end
+
+    # function calls need to transform the function from :call to :calldecl
+    # so that operators are printed correctly
+    elseif head == :function && nargs==2 && is_expr(args[1], :call)
+        show_block(io, head, Expr(:calldecl, args[1].args...), args[2], indent)
+        print(io, "end")
+
+    # block with argument
+    elseif head in (:for,:while,:function,:if,:module) && nargs==2
+        show_block(io, head, args[1], args[2], indent); print(io, "end")
+
+    # type declaration
+    elseif is(head, :type) && nargs==3
+        show_block(io, args[1] ? :type : :immutable, args[2], args[3], indent)
+        print(io, "end")
+
+    # empty return (i.e. "function f() return end")
+    elseif is(head, :return) && nargs == 1 && is(args[1], nothing)
+        print(io, head)
+
+    # type annotation (i.e. "::Int")
+    elseif is(head, symbol("::")) && nargs == 1
+        print(io, "::")
+        show_unquoted(io, args[1], indent)
+
+    # var-arg declaration or expansion
+    # (i.e. "function f(L...) end" or "f(B...)")
     elseif is(head, :(...)) && nargs == 1
         show_unquoted(io, args[1], indent)
         print(io, "...")
-    elseif (nargs == 1 && in(head, (:return, :abstract, :const))) ||
-                          in(head, (:local,  :global))
+
+    elseif (nargs == 1 && head in (:return, :abstract, :const)) ||
+                          head in (:local,  :global)
         print(io, head, ' ')
-        show_list(io, args, ",", indent)
+        show_list(io, args, ", ", indent)
+
     elseif is(head, :macrocall) && nargs >= 1
         show_list(io, args, ' ', indent)
+
     elseif is(head, :typealias) && nargs == 2
         print(io, "typealias ")
         show_list(io, args, ' ', indent)
+
     elseif is(head, :line) && 1 <= nargs <= 2
         show_linenumber(io, args...)
-    elseif is(head, :if) && nargs == 3           # if/else
+
+    elseif is(head, :if) && nargs == 3     # if/else
         show_block(io, "if",   args[1], args[2], indent)
         show_block(io, "else", args[3], indent)
         print(io, "end")
-    elseif is(head, :try) && nargs == 3
+
+    elseif is(head, :try) && 3 <= nargs <= 4
         show_block(io, "try", args[1], indent)
-        if !(is(args[2], false) && is_expr(args[3], :block, 0))
-            show_block(io, "catch", args[2], args[3], indent)
+        if is_expr(args[3], :block)
+            show_block(io, "catch", is(args[2], false) ? [] : args[2], args[3], indent)
+        end
+        if nargs >= 4 && is_expr(args[4], :block)
+            show_block(io, "finally", [], args[4], indent)
         end
         print(io, "end")
+
     elseif is(head, :let) && nargs >= 1
         show_block(io, "let", args[2:end], args[1], indent); print(io, "end")
+
     elseif is(head, :block) || is(head, :body)
         show_block(io, "begin", ex, indent); print(io, "end")
-    elseif in(head,(:for,:while,:function,:if,:type,:module)) && nargs==2
-        show_block(io, head, args[1], args[2], indent); print(io, "end")
+
     elseif is(head, :quote) && nargs == 1
-        show_indented(io, args[1], indent)
+        show_unquoted_quote_expr(io, args[1], indent, 0)
+
     elseif is(head, :gotoifnot) && nargs == 2
         print(io, "unless ")
         show_list(io, args, " goto ", indent)
+
     elseif is(head, :string) && nargs == 1 && isa(args[1], String)
         show(io, args[1])
+
     elseif is(head, :null)
         print(io, "nothing")
-    elseif is(head, :kw)
+
+    elseif is(head, :kw) && length(args)==2
         show_unquoted(io, args[1], indent+indent_width)
         print(io, '=')
         show_unquoted(io, args[2], indent+indent_width)
+
+    elseif is(head, :string)
+        a = map(args) do x
+            if !isa(x,String)
+                if isa(x,Symbol) && !(x in quoted_syms)
+                    string("\$", x)
+                else
+                    string("\$(", sprint(show_unquoted,x), ")")
+                end
+            else
+                sprint(print_escaped, x, "\"\$")
+            end
+        end
+        print(io, '"', a..., '"')
+
+    elseif is(head, :&) && length(args) == 1
+        print(io, '&')
+        show_unquoted(io, args[1])
+
+    # transpose
+    elseif is(head, symbol('\'')) && length(args) == 1
+        show_unquoted(io, args[1])
+        print(io, '\'')
+
+    # print anything else as "Expr(head, args...)"
     else
         print(io, "\$(Expr(")
-        show_indented(io, ex.head, indent)
+        show(io, ex.head)
         for arg in args
             print(io, ", ")
-            show_indented(io, arg, indent)
+            show(io, arg)
         end
         print(io, "))")
     end
+
     show_expr_type(io, ex.typ)
 end
 
-show_unquoted(io::IO, ex::ExprNode, indent::Int) = show_unquoted(io, ex)
-
-show_unquoted(io::IO, ex::LineNumberNode) = show_linenumber(io, ex.line)
-show_unquoted(io::IO, ex::LabelNode)      = print(io, ex.label, ": ")
-show_unquoted(io::IO, ex::GotoNode)       = print(io, "goto ", ex.label)
-show_unquoted(io::IO, ex::TopNode)        = print(io, "top(", ex.name, ')')
-show_unquoted(io::IO, ex::QuoteNode, ind::Int) = show_indented(io,ex.value,ind)
-function show_unquoted(io::IO, ex::SymbolNode) 
-    print(io, ex.name)
-    show_expr_type(io, ex.typ)
-end
-
-function clean_gensym(s::Symbol)
-    s = string(s)
-    i = search(s,'#')
-    if i > 0
-        return s[1:i-1]
-    end
-    return s
-end
-
-function argtype_decl_string(n, t)
-    if isa(n,Expr)
-        n = n.args[1]  # handle n::T in arg list
-    end
-    n = clean_gensym(n)
-    if t === Any && !isempty(n)
-        return n
-    end
-    if t <: Vararg && t !== None && t.parameters[1] === Any
-        return string(n, "...")
-    end
-    return string(n, "::", t)
-end
-
-function show(io::IO, m::Method)
-    tv = m.tvars
-    if !isa(tv,Tuple)
-        tv = (tv,)
-    end
-    if !isempty(tv)
-        show_delim_array(io, tv, '{', ',', '}', false)
-    end
-    li = m.func.code
-    e = uncompressed_ast(li)
-    argnames = e.args[1]
-    if length(argnames) != length(m.sig)
-        s = symbol("?")
-        decls = map(argtype_decl_string, { s for i=1:length(m.sig) }, {m.sig...})
-    else
-        decls = map(argtype_decl_string, argnames, {m.sig...})
-    end
-    print(io, "(")
-    print_joined(io, decls, ",", ",")
-    print(io, ")")
-    if li.line > 0
-        print(io, " at ", li.file, ":", li.line)
-    end
-end
-
-function show_method_table(io::IO, mt::MethodTable, max::Int=-1, header=true)
-    name = mt.name
-    n = length(mt)
-    if header
-        m = n==1 ? "method" : "methods"
-        print(io, "# $n $m for generic function \"$name\":")
-    end
-    d = mt.defs
-    n = rest = 0
-    while !is(d,())
-        if max==-1 || n<max || (rest==0 && n==max && d.next === ())
-            println(io)
-            print(io, name)
-            show(io, d)
-            n += 1
-        else
-            rest += 1
-        end
-        d = d.next
-    end
-    if rest > 0
-        println(io)
-        print(io,"... $rest methods not shown (use methods($name) to see them all)")
-    end
-end
-
-show(io::IO, mt::MethodTable) = show_method_table(io, mt)
-
 # dump & xdump - structured tree representation like R's str()
 # - dump is for the user-facing structure
 # - xdump is for the internal structure
@@ -509,14 +697,14 @@ function xdump_elts(fn::Function, io::IO, x::Array{Any}, n::Int, indent, i0, i1)
     for i in i0:i1
         print(io, indent, "  ", i, ": ")
         if !isdefined(x,i)
-            println(undef_ref_str)
+            println(io, undef_ref_str)
         else
             fn(io, x[i], n - 1, string(indent, "  "))
         end
     end
 end
 function xdump(fn::Function, io::IO, x::Array{Any}, n::Int, indent)
-    println("Array($(eltype(x)),$(size(x)))")
+    println(io, "Array($(eltype(x)),$(size(x)))")
     if n > 0
         xdump_elts(fn, io, x, n, indent, 1, (length(x) <= 10 ? length(x) : 5))
         if length(x) > 10
@@ -542,7 +730,7 @@ function xdump(fn::Function, io::IO, x::DataType, n::Int, indent)
                 if isa(x.types[idx], DataType)
                     xdump(fn, io, x.types[idx], n - 1, string(indent, "  "))
                 else
-                    println(x.types[idx])
+                    println(io, x.types[idx])
                 end
             end
         end
@@ -618,7 +806,7 @@ dump(io::IO, args...) = error("invalid arguments to dump")
 dump(args...) = with_output_limit(()->dump(STDOUT::IO, args...), true)
 
 function dump(io::IO, x::Dict, n::Int, indent)
-    println(typeof(x), " len ", length(x))
+    println(io, typeof(x), " len ", length(x))
     if n > 0
         i = 1
         for (k,v) in x
@@ -640,22 +828,22 @@ dump(io::IO, x::DataType) = dump(io, x, 5, "")
 dump(io::IO, x::TypeVar, n::Int, indent) = println(io, x.name)
 
 
-alignment(x::Any) = (0, length(sprint(showcompact, x)))
-alignment(x::Number) = (length(sprint(showcompact, x)), 0)
-alignment(x::Integer) = (length(sprint(showcompact, x)), 0)
+alignment(x::Any) = (0, length(sprint(showcompact_lim, x)))
+alignment(x::Number) = (length(sprint(showcompact_lim, x)), 0)
+alignment(x::Integer) = (length(sprint(showcompact_lim, x)), 0)
 function alignment(x::Real)
-    m = match(r"^(.*?)((?:[\.eE].*)?)$", sprint(showcompact, x))
-    m == nothing ? (length(sprint(showcompact, x)), 0) :
+    m = match(r"^(.*?)((?:[\.eE].*)?)$", sprint(showcompact_lim, x))
+    m == nothing ? (length(sprint(showcompact_lim, x)), 0) :
                    (length(m.captures[1]), length(m.captures[2]))
 end
 function alignment(x::Complex)
-    m = match(r"^(.*,)(.*)$", sprint(showcompact, x))
-    m == nothing ? (length(sprint(showcompact, x)), 0) :
+    m = match(r"^(.*[\+\-])(.*)$", sprint(showcompact_lim, x))
+    m == nothing ? (length(sprint(showcompact_lim, x)), 0) :
                    (length(m.captures[1]), length(m.captures[2]))
 end
 function alignment(x::Rational)
-    m = match(r"^(.*?/)(/.*)$", sprint(showcompact, x))
-    m == nothing ? (length(sprint(showcompact, x)), 0) :
+    m = match(r"^(.*?/)(/.*)$", sprint(showcompact_lim, x))
+    m == nothing ? (length(sprint(showcompact_lim, x)), 0) :
                    (length(m.captures[1]), length(m.captures[2]))
 end
 
@@ -663,7 +851,7 @@ const undef_ref_str = "#undef"
 const undef_ref_alignment = (3,3)
 
 function alignment(
-    X::Union(AbstractMatrix,AbstractVector),
+    X::AbstractVecOrMat,
     rows::AbstractVector, cols::AbstractVector,
     cols_if_complete::Integer, cols_otherwise::Integer, sep::Integer
 )
@@ -694,7 +882,7 @@ function alignment(
 end
 
 function print_matrix_row(io::IO,
-    X::Union(AbstractMatrix,AbstractVector), A::Vector,
+    X::AbstractVecOrMat, A::Vector,
     i::Integer, cols::AbstractVector, sep::String
 )
     for k = 1:length(A)
@@ -702,7 +890,7 @@ function print_matrix_row(io::IO,
         if isassigned(X,i,j)
             x = X[i,j]
             a = alignment(x)
-            sx = sprint(showcompact, x)
+            sx = sprint(showcompact_lim, x)
         else
             a = undef_ref_alignment
             sx = undef_ref_str
@@ -730,12 +918,16 @@ function print_matrix_vdots(io::IO,
     end
 end
 
-function print_matrix(io::IO,
-    X::Union(AbstractMatrix,AbstractVector), rows::Integer, cols::Integer,
-    pre::String, sep::String, post::String,
-    hdots::String, vdots::String, ddots::String,
-    hmod::Integer, vmod::Integer
-)
+function print_matrix(io::IO, X::AbstractVecOrMat,
+                      sz::(Integer, Integer) = (s = tty_size(); (s[1]-4, s[2])),
+                      pre::String = " ",
+                      sep::String = "  ",
+                      post::String = "",
+                      hdots::String = "  \u2026  ",
+                      vdots::String = "\u22ee",
+                      ddots::String = "  \u22f1  ",
+                      hmod::Integer = 5, vmod::Integer = 5)
+    rows, cols = sz
     cols -= length(pre) + length(post)
     presp = repeat(" ", length(pre))
     postsp = ""
@@ -805,13 +997,6 @@ function print_matrix(io::IO,
         end
     end
 end
-print_matrix(io::IO, X::Union(AbstractMatrix,AbstractVector),
-             rows::Integer, cols::Integer) =
-    print_matrix(io, X, rows, cols, " ", "  ", "",
-                 "  \u2026  ", "\u22ee", "  \u22f1  ", 5, 5)
-
-print_matrix(io::IO, X::Union(AbstractMatrix,AbstractVector)) =
-               print_matrix(io, X, tty_rows()-4, tty_cols())
 
 summary(x) = string(typeof(x))
 
@@ -822,11 +1007,11 @@ dims2string(d) = length(d) == 0 ? "0-dimensional" :
 summary(a::AbstractArray) =
     string(dims2string(size(a)), " ", typeof(a))
 
-function show_nd(io::IO, a::AbstractArray, limit, rows, cols)
+function show_nd(io::IO, a::AbstractArray, limit, print_matrix, label_slices)
     if isempty(a)
         return
     end
-    tail = size(a)[3:]
+    tail = size(a)[3:end]
     nd = ndims(a)-2
     function print_slice(idxs...)
         if limit
@@ -850,11 +1035,13 @@ function show_nd(io::IO, a::AbstractArray, limit, rows, cols)
                 end
             end
         end
-        print(io, "[:, :, ")
-        for i = 1:(nd-1); print(io, "$(idxs[i]), "); end
-        println(io, idxs[end], "] =")
+        if label_slices
+            print(io, "[:, :, ")
+            for i = 1:(nd-1); print(io, "$(idxs[i]), "); end
+            println(io, idxs[end], "] =")
+        end
         slice = sub(a, 1:size(a,1), 1:size(a,2), idxs...)
-        print_matrix(io, slice, rows, cols)
+        print_matrix(io, slice)
         print(io, idxs == tail ? "" : "\n\n")
     end
     cartesianmap(print_slice, tail)
@@ -872,47 +1059,75 @@ whos() = whos(r"")
 whos(m::Module) = whos(m, r"")
 whos(pat::Regex) = whos(current_module(), pat)
 
-function show{T}(io::IO, x::AbstractArray{T,0})
-    println(io, summary(x),":")
-    if isassigned(x)
-        sx = sprint(showcompact, x[])
-    else
-        sx = undef_ref_str
-    end
-    print(io, sx)
-end
-
 # global flag for limiting output
 # TODO: this should be replaced with a better mechanism. currently it is only
 # for internal use in showing arrays.
 _limit_output = false
 
+function print_matrix_repr(io, X::AbstractArray)
+    compact, prefix = array_eltype_show_how(eltype(X))
+    prefix *= "["
+    ind = " "^length(prefix)
+    print(io, prefix)
+    for i=1:size(X,1)
+        i > 1 && print(io, ind)
+        for j=1:size(X,2)
+            j > 1 && print(io, " ")
+            if !isassigned(X,i,j)
+                print(io, undef_ref_str)
+            else
+                el = X[i,j]
+                compact ? showcompact_lim(io, el) : show(io, el)
+            end
+        end
+        if i < size(X,1)
+            println(io)
+        else
+            print(io, "]")
+        end
+    end
+end
+
 # NOTE: this is a possible, so-far-unexported function, providing control of
 # array output. Not sure I want to do it this way.
 showarray(X::AbstractArray; kw...) = showarray(STDOUT, X; kw...)
 function showarray(io::IO, X::AbstractArray;
                    header::Bool=true, limit::Bool=_limit_output,
-                   rows = tty_rows()-4, cols = tty_cols())
+                   sz = (s = tty_size(); (s[1]-4, s[2])), repr=false)
+    rows, cols = sz
     header && print(io, summary(X))
     if !isempty(X)
         header && println(io, ":")
         if ndims(X) == 0
-            return showcompact(io, X[])
+            if isassigned(X)
+                return showcompact_lim(io, X[])
+            else
+                return print(io, undef_ref_str)
+            end
         end
         if !limit
             rows = cols = typemax(Int)
         end
-        if ndims(X)<=2
-            print_matrix(io, X, rows, cols)
+        if repr
+            if ndims(X)<=2
+                print_matrix_repr(io, X)
+            else
+                show_nd(io, X, limit, print_matrix_repr, false)
+            end
         else
-            show_nd(io, X, limit, rows, cols)
+            punct = (" ", "  ", "")
+            if ndims(X)<=2
+                print_matrix(io, X, sz, punct...)
+            else
+                show_nd(io, X, limit,
+                        (io,slice)->print_matrix(io,slice,sz,punct...),
+                        !repr)
+            end
         end
     end
 end
 
-show(io::IO, X::AbstractArray) = showarray(io, X)
-
-print(io::IO, X::AbstractArray) = writedlm(io, X)
+show(io::IO, X::AbstractArray) = showarray(io, X, header=_limit_output, repr=!_limit_output)
 
 function with_output_limit(thk, lim=true)
     global _limit_output
@@ -947,13 +1162,32 @@ function showlimited(io::IO, x)
     end
 end
 
-function show_vector(io::IO, v, opn, cls)
-    if _limit_output && length(v) > 20
-        show_delim_array(io, v[1:10], opn, ",", "", false)
-        print(io, "  \u2026  ")
-        show_delim_array(io, v[end-9:end], "", ",", cls, false)
+# returns compact, prefix
+function array_eltype_show_how(e)
+    leaf = isleaftype(e)
+    plain = e<:Number || e<:String
+    if isa(e,DataType) && e === e.name.primary
+        str = string(e.name)
     else
-        show_delim_array(io, v, opn, ",", cls, false)
+        str = string(e)
+    end
+    leaf&&plain, (e===Float64 || e===Int || (leaf && !plain) ? "" : str)
+end
+
+function show_vector(io::IO, v, opn, cls)
+    e = eltype(v)
+    compact = false
+    if e !== Any
+        compact, prefix = array_eltype_show_how(e)
+        print(io, prefix)
+    end
+    if _limit_output && length(v) > 20
+        show_delim_array(io, sub(v,1:10), opn, ",", "", false, compact)
+        print(io, "  \u2026  ")
+        n = length(v)
+        show_delim_array(io, sub(v,(n-9):n), "", ",", cls, false, compact)
+    else
+        show_delim_array(io, v, opn, ",", cls, false, compact)
     end
 end
 
