@@ -18,9 +18,9 @@ end
 
 function IPv4(host::Integer)
     if host < 0
-        error("IP address may not be negative")
+        error("IP address must not be negative")
     elseif typemax(typeof(host)) > typemax(Uint32) && host > typemax(Uint32) 
-        error("Number too large for IP Address")
+        error("IPv4 address must fit within 32 bits")
     else
         return IPv4(uint32(host))
     end
@@ -57,11 +57,11 @@ end
 
 function IPv6(host::Integer)
     if host < 0
-        error("IP address may not be negative")
-    # We allow passing bigger integer types, but need to be careful to avoid overflow
-    # Let's hope promotion rules are sensible
+        error("IP address must not be negative")
+        # We allow passing bigger integer types, but need to be careful to avoid overflow
+        # Let's hope promotion rules are sensible
     elseif typemax(typeof(host)) > typemax(Uint128) && host > typemax(Uint128) 
-        error("Number too large for IP Address")
+        error("IPv6 address must fit within 128 bits")
     else
         return IPv6(uint128(host))
     end
@@ -134,7 +134,7 @@ function parseipv4(str)
     ret = 0
     for f in fields 
         if length(f) == 0
-            error("Empty field in IPv4")
+            error("empty field in IPv4 address")
         end
         if f[1] == '0'
             if length(f) >= 2 && f[2] == 'x'
@@ -153,7 +153,7 @@ function parseipv4(str)
         end
         if i != length(fields)
             if r < 0 || r > 255
-                error("Invalid IPv4 field")
+                error("IPv4 field out of range (must be 0-255)")
             end
             ret |= uint32(r) << ((4-i)*8)
         else
@@ -169,7 +169,7 @@ end
 
 function parseipv6fields(fields,num_fields)
     if length(fields) > num_fields
-        error("Too many fields in IPv6 address")
+        error("too many fields in IPv6 address")
     end
     cf = 7
     ret = uint128(0)
@@ -192,7 +192,7 @@ parseipv6fields(fields) = parseipv6fields(fields,8)
 function parseipv6(str)
     fields = split(str,':')
     if length(fields) > 8
-        error("Too many fields in IPv6 address")
+        error("too many fields in IPv6 address")
     elseif length(fields) == 8
         return IPv6(parseipv6fields(fields))
     elseif in('.',fields[end])
@@ -261,13 +261,14 @@ type TcpSocket <: Socket
 end
 function TcpSocket()
     this = TcpSocket(c_malloc(_sizeof_uv_tcp))
-    associate_julia_struct(this.handle, this)
+    associate_julia_struct(this.handle,this)
+    finalizer(this,uvfinalize)
     err = ccall(:uv_tcp_init,Cint,(Ptr{Void},Ptr{Void}),
                   eventloop(),this.handle)
     if err != 0 
         c_free(this.handle)
         this.handle = C_NULL
-        error(UVError("Failed to create tcp socket",err))
+        error(UVError("failed to create tcp socket",err))
     end
     this.status = StatusInit
     this
@@ -289,29 +290,23 @@ end
 function TcpServer()
     this = TcpServer(c_malloc(_sizeof_uv_tcp))
     associate_julia_struct(this.handle, this)
+    finalizer(this,uvfinalize)
     err = ccall(:uv_tcp_init,Cint,(Ptr{Void},Ptr{Void}),
                   eventloop(),this.handle)
     if err != 0 
         c_free(this.handle)
         this.handle = C_NULL
-        error(UVError("Failed to create tcp server",err))
+        error(UVError("failed to create tcp server",err))
     end
     this.status = StatusInit
     this
 end
 
-#type UdpSocket <: Socket
-#    handle::Ptr{Void}
-#    open::Bool
-#    line_buffered::Bool
-#    buffer::IOBuffer
-#    readcb::Callback
-#    readnotify::Condition
-#    ccb::Callback
-#    connectnotify::Condition
-#    closecb::Callback
-#    closenotify::Condition
-#end
+function uvfinalize(uv)
+    close(uv)
+    disassociate_julia_struct(uv)
+    uv.handle = 0
+end
 
 isreadable(io::TcpSocket) = true
 iswritable(io::TcpSocket) = true
@@ -321,9 +316,6 @@ show(io::IO,sock::TcpSocket) = print(io,"TcpSocket(",uv_status_string(sock),", "
 
 show(io::IO,sock::TcpServer) = print(io,"TcpServer(",uv_status_string(sock),")")
 
-#show(io::IO,sock::UdpSocket) = print(io,"UdpSocket(",uv_status_string(sock),", ",
-#    nb_available(sock.buffer)," bytes waiting)")
-
 ## VARIOUS METHODS TO BE MOVED TO BETTER LOCATION
 
 _jl_connect_raw(sock::TcpSocket,sockaddr::Ptr{Void}) = 
@@ -339,12 +331,65 @@ accept(server::PipeServer) = accept(server, Pipe())
 ##
 
 bind(sock::TcpServer, addr::InetAddr) = bind(sock,addr.host,addr.port)
-bind(sock::TcpServer, host::IpAddr, port) = bind(sock, InetAddr(host,port))
 
-function bind(sock::TcpServer, host::IPv4, port::Uint16)
+_bind(sock::TcpServer, host::IPv4, port::Uint16) = ccall(:jl_tcp_bind, Int32, (Ptr{Void}, Uint16, Uint32, Cuint),
+            sock.handle, hton(port), hton(host.host), 0)
+
+_bind(sock::TcpServer, host::IPv6, port::Uint16) = ccall(:jl_tcp_bind6, Int32, (Ptr{Void}, Uint16, Ptr{Uint128}, Cuint),
+            sock.handle, hton(port), &hton(host.host), 0)
+
+# UDP 
+
+type UdpSocket <: Socket
+    handle::Ptr{Void}
+    status::Int
+    recvnotify::Condition
+    sendnotify::Condition
+    closenotify::Condition
+    UdpSocket(handle::Ptr) = new(handle, StatusUninit, Condition(), Condition(), Condition())
+end
+
+function UdpSocket()
+    this = UdpSocket(c_malloc(_sizeof_uv_udp))
+    associate_julia_struct(this.handle, this)
+    err = ccall(:uv_udp_init,Cint,(Ptr{Void},Ptr{Void}),
+                  eventloop(),this.handle)
+    finalizer(this, uvfinalize)
+    if err != 0 
+        c_free(this.handle)
+        this.handle = C_NULL
+        error(UVError("failed to create udp socket",err))
+    end
+    this.status = StatusInit
+    this
+end
+
+function uvfinalize(uv::Union(TTY,Pipe,PipeServer,TcpServer,TcpSocket,UdpSocket))
+    if (uv.status != StatusUninit && uv.status != StatusInit)
+        close(uv)
+    end
+    disassociate_julia_struct(uv)
+    uv.handle = 0
+end
+
+function _uv_hook_close(sock::UdpSocket)
+    sock.handle = 0
+    sock.status = StatusClosed
+    notify(sock.closenotify)
+    notify(sock.sendnotify)
+    notify_error(sock.recvnotify,EOFError())
+end
+
+# Disables dual stack mode. Only available when using ipv6 binf
+const UV_UDP_IPV6ONLY = 1
+
+# Indicates message was truncated because read buffer was too small. The
+# remainder was discarded by the OS. 
+const UV_UDP_PARTIAL = 2
+
+function bind(sock::Union(TcpServer,UdpSocket), host::IPv4, port::Integer)
     @assert sock.status == StatusInit
-    err = ccall(:jl_tcp_bind, Int32, (Ptr{Void}, Uint16, Uint32),
-	        sock.handle, hton(port), hton(host.host))
+    err = _bind(sock,host,uint16(port))
     if err < 0
         if err != UV_EADDRINUSE && err != UV_EACCES
             error(UVError("bind",err))
@@ -356,10 +401,15 @@ function bind(sock::TcpServer, host::IPv4, port::Uint16)
     true
 end
 
-function bind(sock::TcpServer, host::IPv6, port::Uint16)
+_bind(sock::UdpSocket, host::IPv4, port::Uint16) = ccall(:jl_udp_bind, Int32, (Ptr{Void}, Uint16, Uint32, Uint32),
+            sock.handle, hton(port), hton(host.host), 0)
+
+_bind(sock::UdpSocket, host::IPv6, port::Uint16, flags::Uint32 = uint32(0)) = ccall(:jl_udp_bind6, Int32, (Ptr{Void}, Uint16, Ptr{Uint128}, Uint32),
+            sock.handle, hton(port), &hton(host.host), flags)
+
+function bind(sock::UdpSocket, host::IPv6, port::Uint16; ipv6only = false)
     @assert sock.status == StatusInit
-    err = ccall(:jl_tcp_bind6, Int32, (Ptr{Void}, Uint16, Ptr{Uint128}),
-            sock.handle, hton(port), &hton(host.host))
+    err = _bind(sock,host,ipv6only ? UV_UDP_IPV6ONLY : 0)
     if err < 0
         if err != UV_EADDRINUSE && err != UV_EACCES
             error(UVError("bind",err))
@@ -371,6 +421,82 @@ function bind(sock::TcpServer, host::IPv6, port::Uint16)
     true
 end
 
+
+function setopt(sock::UdpSocket; multicast_loop = nothing, multicast_ttl=nothing, enable_broadcast=nothing, ttl=nothing)
+    if sock.status == StatusUninit 
+        error("Cannot set options on unitialized socket")
+    end
+    if multicast_loop !== nothing
+        uv_error("multicast_loop",ccall(:uv_udp_set_multicast_loop,Cint,(Ptr{Void},Cint),sock.handle,multicast_loop) < 0)
+    end
+    if multicast_ttl !== nothing
+        uv_error("multicast_ttl",ccall(:uv_udp_set_multicast_ttl,Cint,(Ptr{Void},Cint),sock.handle,multicast_ttl))
+    end
+    if enable_broadcast !== nothing
+        uv_error("enable_broadcast",ccall(:uv_udp_set_broadcast,Cint,(Ptr{Void},Cint),sock.handle,enable_broadcast))
+    end
+    if ttl !== nothing
+        uv_error("ttl",ccall(:uv_udp_set_ttl,Cint,(Ptr{Void},Cint),sock.handle,ttl))
+    end
+end
+
+_uv_hook_alloc_buf(sock::UdpSocket,size::Uint) = (c_malloc(size),size)
+
+function _recv_start(sock::UdpSocket)
+    if ccall(:uv_is_active,Cint,(Ptr{Void},),sock.handle) == 0
+        uv_error("recv_start",ccall(:uv_udp_recv_start,Cint,(Ptr{Void},Ptr{Void},Ptr{Void}),
+                                        sock.handle,cglobal(:jl_uv_alloc_buf),cglobal(:jl_uv_recvcb)))
+    end
+end
+
+_recv_stop(sock::UdpSocket) = uv_error("recv_stop",ccall(:uv_udp_recv_stop,Cint,(Ptr{Void},),sock.handle))
+
+function recv(sock::UdpSocket)
+    # If the socket has not been bound, it will be bound implicitly to ::0 and a random port
+    if sock.status != StatusInit && sock.status != StatusOpen 
+        error("Invalid socket state")
+    end
+    _recv_start(sock)
+    stream_wait(sock,sock.recvnotify)::Vector{Uint8}
+end
+
+function _uv_hook_recv(sock::UdpSocket, nread::Int, buf_addr::Ptr{Void}, buf_size::Uint, addr::Ptr{Void}, flags::Int32)
+    if flags & UV_UDP_PARTIAL > 0
+        # TODO: Decide what to do in this case. For now throw an error
+        c_free(buf_addr)
+        notify_error(sock.recvnotify,"Partial message received")
+    end
+    buf = pointer_to_array(convert(Ptr{Uint8},buf_addr),int(buf_size),true)
+    notify(sock.recvnotify,buf[1:nread])
+end
+
+function _send(sock::UdpSocket,ipaddr::IPv4,port::Uint16,buf) 
+    ccall(:jl_udp_send,Cint,(Ptr{Void},Uint16,Uint32,Ptr{Uint8},Csize_t),sock.handle,hton(port),hton(ipaddr.host),buf,sizeof(buf))
+end
+
+function _send(sock::UdpSocket,ipaddr::IPv6,port::Uint16,buf) 
+    ccall(:jl_udp_send6,Cint,(Ptr{Void},Uint16,Ptr{Uint128},Ptr{Uint8},Csize_t),sock.handle,hton(port),&hton(ipaddr.host),buf,sizeof(buf))
+end
+
+function send(sock::UdpSocket,ipaddr,port,msg)
+    # If the socket has not been bound, it will be bound implicitly to ::0 and a random port
+    if sock.status != StatusInit && sock.status != StatusOpen 
+        error("Invalid socket state")
+    end
+    uv_error("send",_send(sock,ipaddr,uint16(port),msg))
+    stream_wait(sock,sock.sendnotify)
+    nothing
+end
+
+function _uv_hook_send(sock::UdpSocket,status::Cint)
+    if status < 0
+        notify_error(sock.sendnotify,UVError("UDP send failed",status))
+    end
+    notify(sock.sendnotify)
+end
+
+##
+
 callback_dict = ObjectIdDict()
 
 function _uv_hook_getaddrinfo(cb::Function, addrinfo::Ptr{Void}, status::Int32)
@@ -419,6 +545,7 @@ const _sizeof_uv_interface_address = ccall(:jl_uv_sizeof_interface_address,Int32
 function getipaddr()
     addr = Array(Ptr{Uint8},1)
     count = Array(Int32,1)
+    lo_present = false
     err = ccall(:jl_uv_interface_addresses,Int32,(Ptr{Ptr{Uint8}},Ptr{Int32}),addr,count)
     addr, count = addr[1],count[1]
     if err != 0
@@ -428,6 +555,7 @@ function getipaddr()
     for i = 0:(count-1)
         current_addr = addr + i*_sizeof_uv_interface_address
         if 1 == ccall(:jl_uv_interface_address_is_internal,Int32,(Ptr{Uint8},),current_addr)
+            lo_present = true
             continue
         end
         sockaddr = ccall(:jl_uv_interface_address_sockaddr,Ptr{Void},(Ptr{Uint8},),current_addr)
@@ -441,7 +569,7 @@ function getipaddr()
         end
     end
     ccall(:uv_free_interface_addresses,Void,(Ptr{Uint8},Int32),addr,count)
-    error("No active external interfaces")
+    lo_present ? ip"127.0.0.1" : error("No networking interface available")
 end
 
 ##
@@ -492,7 +620,7 @@ listen(sock::UVServer; backlog::Integer=BACKLOG_DEFAULT) = (uv_error("listen",_l
 
 function listen(addr; backlog::Integer=BACKLOG_DEFAULT)
     sock = TcpServer()
-    !bind(sock,addr) && error("Cannot bind to port (may already be in use or access denied)")
+    !bind(sock,addr) && error("cannot bind to port; may already be in use or access denied")
     uv_error("listen",_listen(sock;backlog=backlog))
     sock
 end
@@ -528,7 +656,7 @@ function listenany(default_port)
             return (addr.port,sock)
         end
         close(sock)
-	addr.port += 1
+        addr.port += 1
         if addr.port==default_port
             error("no ports available")
         end
