@@ -81,9 +81,10 @@ function select!(v::AbstractVector, k::Int, lo::Int, hi::Int, o::Ordering)
     return v[lo]
 end
 
-function select!(v::AbstractVector, r::Range1, lo::Int, hi::Int, o::Ordering)
-    a, b = first(r), last(r)
-    lo <= a <= b <= hi || error("select index $k is out of range $lo:$hi")
+function select!(v::AbstractVector, r::OrdinalRange, lo::Int, hi::Int, o::Ordering)
+    isempty(r) && (return v[r])
+    a, b = extrema(r)
+    lo <= a <= b <= hi || error("selection $r is out of range $lo:$hi")
     @inbounds while true
         if lo == a && hi == b
             sort!(v, lo, hi, DEFAULT_UNSTABLE, o)
@@ -111,12 +112,12 @@ function select!(v::AbstractVector, r::Range1, lo::Int, hi::Int, o::Ordering)
     end
 end
 
-select!(v::AbstractVector, k::Union(Int,Range1), o::Ordering) = select!(v,k,1,length(v),o)
-select!(v::AbstractVector, k::Union(Int,Range1);
+select!(v::AbstractVector, k::Union(Int,OrdinalRange), o::Ordering) = select!(v,k,1,length(v),o)
+select!(v::AbstractVector, k::Union(Int,OrdinalRange);
     lt::Function=isless, by::Function=identity, rev::Bool=false, order::Ordering=Forward) =
     select!(v, k, ord(lt,by,rev,order))
 
-select(v::AbstractVector, k::Union(Int,Range1); kws...) = select!(copy(v), k; kws...)
+select(v::AbstractVector, k::Union(Int,OrdinalRange); kws...) = select!(copy(v), k; kws...)
 
 # reference on sorted binary search:
 #   http://www.tbray.org/ongoing/When/200x/2003/03/22/Binary
@@ -154,11 +155,11 @@ function searchsortedlast(v::AbstractVector, x, lo::Int, hi::Int, o::Ordering)
 end
 
 # returns the range of indices of v equal to x
-# if v does not contain x, returns a 0-length range 
+# if v does not contain x, returns a 0-length range
 # indicating the insertion point of x
-function searchsorted(v::AbstractVector, x, lo::Int, hi::Int, o::Ordering)
-    lo = lo-1
-    hi = hi+1
+function searchsorted(v::AbstractVector, x, ilo::Int, ihi::Int, o::Ordering)
+    lo = ilo-1
+    hi = ihi+1
     @inbounds while lo < hi-1
         m = (lo+hi)>>>1
         if lt(o, v[m], x)
@@ -166,15 +167,15 @@ function searchsorted(v::AbstractVector, x, lo::Int, hi::Int, o::Ordering)
         elseif lt(o, x, v[m])
             hi = m
         else
-            a = searchsortedfirst(v, x, max(lo,1), m, o)
-            b = searchsortedlast(v, x, m, min(hi,length(v)), o)
+            a = searchsortedfirst(v, x, max(lo,ilo), m, o)
+            b = searchsortedlast(v, x, m, min(hi,ihi), o)
             return a:b
         end
     end
     return lo+1:hi-1
 end
 
-function searchsortedlast{T<:Real}(a::Ranges{T}, x::Real, o::Ordering=Forward)
+function searchsortedlast{T<:Real}(a::Range{T}, x::Real, o::DirectOrdering)
     if step(a) == 0
         lt(o, x, first(a)) ? 0 : length(a)
     else
@@ -183,7 +184,7 @@ function searchsortedlast{T<:Real}(a::Ranges{T}, x::Real, o::Ordering=Forward)
     end
 end
 
-function searchsortedfirst{T<:Real}(a::Ranges{T}, x::Real, o::Ordering=Forward)
+function searchsortedfirst{T<:Real}(a::Range{T}, x::Real, o::DirectOrdering)
     if step(a) == 0
         lt(o, first(a), x) ? length(a)+1 : 1
     else
@@ -192,7 +193,7 @@ function searchsortedfirst{T<:Real}(a::Ranges{T}, x::Real, o::Ordering=Forward)
     end
 end
 
-function searchsortedlast{T<:Integer}(a::Ranges{T}, x::Real, o::Ordering=Forward)
+function searchsortedlast{T<:Integer}(a::Range{T}, x::Real, o::DirectOrdering)
     if step(a) == 0
         lt(o, x, first(a)) ? 0 : length(a)
     else
@@ -200,7 +201,7 @@ function searchsortedlast{T<:Integer}(a::Ranges{T}, x::Real, o::Ordering=Forward
     end
 end
 
-function searchsortedfirst{T<:Integer}(a::Ranges{T}, x::Real, o::Ordering=Forward)
+function searchsortedfirst{T<:Integer}(a::Range{T}, x::Real, o::DirectOrdering)
     if step(a) == 0
         lt(o, first(a), x) ? length(a)+1 : 1
     else
@@ -208,8 +209,8 @@ function searchsortedfirst{T<:Integer}(a::Ranges{T}, x::Real, o::Ordering=Forwar
     end
 end
 
-searchsorted{T<:Real}(a::Ranges{T}, x::Real; kws...) =
-    searchsortedfirst(a,x; kws...):searchsortedlast(a,x; kws...)
+searchsorted{T<:Real}(a::Range{T}, x::Real, o::DirectOrdering) =
+    searchsortedfirst(a,x,o):searchsortedlast(a,x,o)
 
 for s in {:searchsortedfirst, :searchsortedlast, :searchsorted}
     @eval begin
@@ -257,17 +258,29 @@ end
 function sort!(v::AbstractVector, lo::Int, hi::Int, a::QuickSortAlg, o::Ordering)
     @inbounds while lo < hi
         hi-lo <= SMALL_THRESHOLD && return sort!(v, lo, hi, SMALL_ALGORITHM, o)
-        pivot = v[(lo+hi)>>>1]
-        i, j = lo, hi
-        while true
-            while lt(o, v[i], pivot); i += 1; end
-            while lt(o, pivot, v[j]); j -= 1; end
-            i <= j || break
-            v[i], v[j] = v[j], v[i]
-            i += 1; j -= 1
+        mi = (lo+hi)>>>1
+        if lt(o, v[mi], v[lo])
+            v[mi], v[lo] = v[lo], v[mi]
         end
-        lo < j && sort!(v, lo, j, a, o)
-        lo = i
+        if lt(o, v[hi], v[mi])
+            v[hi], v[mi] = v[mi], v[hi]
+        end
+        if lt(o, v[mi], v[lo])
+            v[mi], v[lo] = v[lo], v[mi]
+        end
+        v[mi], v[lo] = v[lo], v[mi]
+        i, j = lo, hi
+        pivot = v[lo]
+        while true
+            i += 1; j -= 1;
+            while lt(o, v[i], pivot); i += 1; end;
+            while lt(o, pivot, v[j]); j -= 1; end;
+            i >= j && break
+            v[i], v[j] = v[j], v[i]
+        end
+        v[j], v[lo] = v[lo], v[j]
+        lo < (j-1) && sort!(v, lo, j-1, a, o)
+        lo = j+1
     end
     return v
 end
@@ -322,7 +335,7 @@ sort(v::AbstractVector; kws...) = sort!(copy(v); kws...)
 
 ## sortperm: the permutation to sort an array ##
 
-sortperm(v::AbstractVector; alg::Algorithm=DEFAULT_STABLE,
+sortperm(v::AbstractVector; alg::Algorithm=DEFAULT_UNSTABLE,
     lt::Function=isless, by::Function=identity, rev::Bool=false, order::Ordering=Forward) =
     sort!([1:length(v)], alg, Perm(ord(lt,by,rev,order),v))
 
@@ -362,8 +375,8 @@ immutable Right <: Ordering end
 left(::DirectOrdering) = Left()
 right(::DirectOrdering) = Right()
 
-left(o::Perm) = Perm(left(o.order),o.data)
-right(o::Perm) = Perm(right(o.order),o.data)
+left(o::Perm) = Perm(left(o.order), o.data)
+right(o::Perm) = Perm(right(o.order), o.data)
 
 lt{T<:Floats}(::Left, x::T, y::T) = slt_int(unbox(T,y),unbox(T,x))
 lt{T<:Floats}(::Right, x::T, y::T) = slt_int(unbox(T,x),unbox(T,y))
@@ -372,44 +385,32 @@ isnan(o::DirectOrdering, x::Floats) = (x!=x)
 isnan(o::Perm, i::Int) = isnan(o.order,o.data[i])
 
 function nans2left!(v::AbstractVector, o::Ordering, lo::Int=1, hi::Int=length(v))
-    hi < lo && return lo, hi
     i = lo
-    @inbounds while (i < hi) & isnan(o, v[i])
+    @inbounds while i <= hi && isnan(o,v[i])
         i += 1
     end
-    r = 0
-    @inbounds while true
-        if isnan(o, v[i])
-            i += 1
-        else
-            r += 1
-        end
-        j = i + r
-        j > hi && break
-        if r > 0
+    j = i + 1
+    @inbounds while j <= hi
+        if isnan(o,v[j])
             v[i], v[j] = v[j], v[i]
+            i += 1
         end
+        j += 1
     end
     return i, hi
 end
 function nans2right!(v::AbstractVector, o::Ordering, lo::Int=1, hi::Int=length(v))
-    hi < lo && return lo, hi
     i = hi
-    @inbounds while (i > lo) & isnan(o, v[i])
+    @inbounds while lo <= i && isnan(o,v[i])
         i -= 1
     end
-    r = 0
-    @inbounds while true
-        if isnan(o, v[i])
-            i -= 1
-        else
-            r += 1
-        end
-        j = i - r
-        j < lo && break
-        if r > 0
+    j = i - 1
+    @inbounds while lo <= j
+        if isnan(o,v[j])
             v[i], v[j] = v[j], v[i]
+            i -= 1
         end
+        j -= 1
     end
     return lo, i
 end
@@ -419,7 +420,8 @@ nans2end!(v::AbstractVector, o::ReverseOrdering) = nans2left!(v,o)
 nans2end!{O<:ForwardOrdering}(v::AbstractVector{Int}, o::Perm{O}) = nans2right!(v,o)
 nans2end!{O<:ReverseOrdering}(v::AbstractVector{Int}, o::Perm{O}) = nans2left!(v,o)
 
-issignleft(o::DirectOrdering, x::Floats) = lt(o, x, zero(x))
+issignleft(o::ForwardOrdering, x::Floats) = lt(o, x, zero(x))
+issignleft(o::ReverseOrdering, x::Floats) = lt(o, x, -zero(x))
 issignleft(o::Perm, i::Int) = issignleft(o.order, o.data[i])
 
 function fpsort!(v::AbstractVector, a::Algorithm, o::Ordering)
@@ -427,13 +429,9 @@ function fpsort!(v::AbstractVector, a::Algorithm, o::Ordering)
     @inbounds while true
         while i <= j &&  issignleft(o,v[i]); i += 1; end
         while i <= j && !issignleft(o,v[j]); j -= 1; end
-        if i <= j
-            v[i], v[j] = v[j], v[i]
-            i += 1
-            j -= 1
-        else
-            break
-        end
+        i <= j || break
+        v[i], v[j] = v[j], v[i]
+        i += 1; j -= 1
     end
     sort!(v, lo, j,  a, left(o))
     sort!(v, i,  hi, a, right(o))
