@@ -17,6 +17,7 @@ next(s::String, i::Integer) = next(s,int(i))
 ## conversion of general objects to strings ##
 
 function print_to_string(xs...)
+    # specialized for performance reasons
     s = IOBuffer(Array(Uint8,isa(xs[1],String) ? endof(xs[1]) : 0), true, true)
     truncate(s,0)
     for x in xs
@@ -47,24 +48,24 @@ convert(::Type{Array{Uint8,1}}, s::String) = bytestring(s).data
 convert(::Type{Array{Uint8}}, s::String) = bytestring(s).data
 convert(::Type{ByteString}, s::String) = bytestring(s)
 convert(::Type{Array{Char,1}}, s::String) = collect(s)
+convert(::Type{Symbol}, s::String) = symbol(s)
 
 ## generic supplied functions ##
 
 start(s::String) = 1
 done(s::String,i) = (i > endof(s))
-isempty(s::String) = done(s,start(s))
 getindex(s::String, i::Int) = next(s,i)[1]
 getindex(s::String, i::Integer) = s[int(i)]
 getindex(s::String, x::Real) = s[to_index(x)]
-getindex{T<:Integer}(s::String, r::Range1{T}) = s[int(first(r)):int(last(r))]
+getindex{T<:Integer}(s::String, r::UnitRange{T}) = s[int(first(r)):int(last(r))]
 # TODO: handle other ranges with stride ±1 specially?
 getindex(s::String, v::AbstractVector) =
     sprint(length(v), io->(for i in v write(io,s[i]) end))
 
 symbol(s::String) = symbol(bytestring(s))
 
-print(io::IO, s::String) = write(io, s)
-write(io::IO, s::String) = for c in s write(io, c) end
+print(io::IO, s::String) = (write(io, s); nothing)
+write(io::IO, s::String) = (len = 0; for c in s; len += write(io, c); end; len)
 show(io::IO, s::String) = print_quoted(io, s)
 
 sizeof(s::String) = error("type $(typeof(s)) has no canonical binary representation")
@@ -94,6 +95,8 @@ end
 
 isvalid(s::DirectIndexString, i::Integer) = (start(s) <= i <= endof(s))
 function isvalid(s::String, i::Integer)
+    i < 1 && return false
+    done(s,i) && return false
     try
         next(s,i)
         true
@@ -138,8 +141,13 @@ function nextind(s::String, i::Integer)
     next(s,e)[2] # out of range
 end
 
-ind2chr(s::DirectIndexString, i::Integer) = i
-chr2ind(s::DirectIndexString, i::Integer) = i
+checkbounds(s::String, i::Integer) = start(s) <= i <= endof(s) || throw(BoundsError())
+checkbounds(s::String, i::Real) = checkbounds(s, to_index(i))
+checkbounds{T<:Integer}(s::String, r::Range{T}) = isempty(r) || (minimum(r) >= start(s) && maximum(r) <= endof(s)) || throw(BoundsError())
+checkbounds{T<:Real}(s::String, I::AbstractArray{T}) = all(i -> checkbounds(s, i), I)
+
+ind2chr(s::DirectIndexString, i::Integer) = begin checkbounds(s,i); i end
+chr2ind(s::DirectIndexString, i::Integer) = begin checkbounds(s,i); i end
 
 function ind2chr(s::String, i::Integer)
     s[i] # throws error if invalid
@@ -156,9 +164,7 @@ function ind2chr(s::String, i::Integer)
 end
 
 function chr2ind(s::String, i::Integer)
-    if i < 1
-        return i
-    end
+    i < start(s) && throw(BoundsError())
     j = 1
     k = start(s)
     while true
@@ -183,7 +189,7 @@ function search(s::String, c::Chars, i::Integer)
     i = nextind(s,i-1)
     while !done(s,i)
         d, j = next(s,i)
-        if in(d,c)
+        if d in c
             return i
         end
         i = j
@@ -311,7 +317,7 @@ function search(s::Union(Array{Uint8,1},Array{Int8,1}),t::Union(Array{Uint8,1},A
     end
 end
 
-function search(s::String, t::String, i::Integer)
+function search(s::String, t::String, i::Integer=start(s))
     idx = searchindex(s,t,i)
     if isempty(t)
         idx:idx-1
@@ -320,12 +326,12 @@ function search(s::String, t::String, i::Integer)
     end
 end
 
-search(s::String, t::String) = search(s,t,start(s))
-
-rsearch(s::String, c::Chars, i::Integer) =
-    endof(s)-search(RevString(s), c, endof(s)-i+1)+1
-
-rsearch(s::String, c::Chars) = rsearch(s,c,endof(s))
+function rsearch(s::String, c::Chars, i::Integer=endof(s))
+    e = endof(s)
+    j = search(RevString(s), c, e-i+1)
+    j == 0 && return 0
+    e-j+1
+end
 
 function _rsearchindex(s, t, i)
     if isempty(t)
@@ -451,7 +457,7 @@ function rsearch(s::Union(Array{Uint8,1},Array{Int8,1}),t::Union(Array{Uint8,1},
     end
 end
 
-function rsearch(s::String, t::String, i::Integer)
+function rsearch(s::String, t::String, i::Integer=endof(s))
     idx = rsearchindex(s,t,i)
     if isempty(t)
         idx:idx-1
@@ -460,8 +466,6 @@ function rsearch(s::String, t::String, i::Integer)
     end
 end
 
-rsearch(s::String, t::String) = rsearch(s,t,endof(s))
-
 contains(a::String, b::String) = searchindex(a,b)!=0
 
 in(::String, ::String) = error("use contains(x,y) for string containment")
@@ -483,8 +487,8 @@ function cmp(a::String, b::String)
     !done(a,i) && done(b,j) ? +1 : 0
 end
 
-isequal(a::String, b::String) = cmp(a,b) == 0
-isless(a::String, b::String)  = cmp(a,b) <  0
+==(a::String, b::String) = cmp(a,b) == 0
+isless(a::String, b::String) = cmp(a,b) < 0
 
 # begins with and ends with predicates
 
@@ -498,7 +502,7 @@ function beginswith(a::String, b::String)
     end
     done(b,i)
 end
-beginswith(a::String, c::Char) = !isempty(a) && a[start(a)] == c
+beginswith(str::String, chars::Chars) = !isempty(str) && str[start(str)] in chars
 
 function endswith(a::String, b::String)
     i = endof(a)
@@ -514,21 +518,20 @@ function endswith(a::String, b::String)
     end
     j < b1
 end
-endswith(a::String, c::Char) = !isempty(a) && a[end] == c
+endswith(str::String, chars::Chars) = !isempty(str) && str[end] in chars
 
-# faster comparisons for byte strings
+# faster comparisons for byte strings and symbols
+
+cmp(a::ByteString, b::ByteString) = lexcmp(a.data, b.data)
+cmp(a::Symbol, b::Symbol) = int(sign(ccall(:strcmp, Int32, (Ptr{Uint8}, Ptr{Uint8}), a, b)))
+
+==(a::ByteString, b::ByteString) = endof(a) == endof(b) && cmp(a,b) == 0
+isless(a::Symbol, b::Symbol) = cmp(a,b) < 0
 
-cmp(a::ByteString, b::ByteString)     = lexcmp(a.data, b.data)
-isequal(a::ByteString, b::ByteString) = endof(a)==endof(b) && cmp(a,b)==0
 beginswith(a::ByteString, b::ByteString) = beginswith(a.data, b.data)
-
 beginswith(a::Array{Uint8,1}, b::Array{Uint8,1}) =
     (length(a) >= length(b) && ccall(:strncmp, Int32, (Ptr{Uint8}, Ptr{Uint8}, Uint), a, b, length(b)) == 0)
 
-cmp(a::Symbol, b::Symbol) =
-    int(sign(ccall(:strcmp, Int32, (Ptr{Uint8}, Ptr{Uint8}), a, b)))
-isless(a::Symbol, b::Symbol) = cmp(a,b)<0
-
 # TODO: fast endswith
 
 ## character column width function ##
@@ -563,25 +566,6 @@ end
 endof(s::GenericString) = endof(s.string)
 next(s::GenericString, i::Int) = next(s.string, i)
 
-## plain old character arrays ##
-
-immutable CharString <: DirectIndexString
-    chars::Array{Char,1}
-
-    CharString(a::Array{Char,1}) = new(a)
-    CharString(c::Char...) = new([ c[i] for i=1:length(c) ])
-end
-CharString(x...) = CharString(map(char,x)...)
-
-next(s::CharString, i::Int) = (s.chars[i], i+1)
-endof(s::CharString) = length(s.chars)
-length(s::CharString) = length(s.chars)
-
-convert(::Type{CharString}, s::String) = CharString(Char[c for c in s])
-convert{T<:String}(::Type{T}, v::Vector{Char}) = convert(T, CharString(v))
-
-reverse(s::CharString) = CharString(reverse(s.chars))
-
 ## substrings reference original strings ##
 
 immutable SubString{T<:String} <: String
@@ -591,7 +575,7 @@ immutable SubString{T<:String} <: String
 
     function SubString(s::T, i::Int, j::Int)
         if i > endof(s) || j<i
-            return new(s, i, 0)
+            return new(s, i-1, 0)
         else
             if !isvalid(s,i)
                 error("invalid SubString indexes")
@@ -617,6 +601,17 @@ print(io::IOBuffer, s::SubString) = write(io, s)
 
 sizeof{T<:ByteString}(s::SubString{T}) = s.endof==0 ? 0 : next(s,s.endof)[2]-1
 
+# TODO: length(s::SubString) = ??
+# default implementation will work but it's slow
+# can this be delegated efficiently somehow?
+# that may require additional string interfaces
+length{T<:DirectIndexString}(s::SubString{T}) = endof(s)
+
+function length(s::SubString{UTF8String})
+    return s.endof==0 ? 0 : int(ccall(:u8_charnum, Csize_t, (Ptr{Uint8}, Csize_t),
+                                      pointer(s), next(s,s.endof)[2]-1))
+end
+
 function next(s::SubString, i::Int)
     if i < 1 || i > s.endof
         error(BoundsError)
@@ -625,27 +620,37 @@ function next(s::SubString, i::Int)
     c, i-s.offset
 end
 
-getindex(s::SubString, i::Int) = getindex(s.string, i+s.offset)
-
-isempty(s::SubString) = s.endof == 0
+function getindex(s::SubString, i::Int)
+    if i < 1 || i > s.endof
+        error(BoundsError)
+    end
+    getindex(s.string, i+s.offset)
+end
 
 endof(s::SubString) = s.endof
-# TODO: length(s::SubString) = ??
-# default implementation will work but it's slow
-# can this be delegated efficiently somehow?
-# that may require additional string interfaces
+
+function isvalid(s::SubString, i::Integer)
+    return (start(s) <= i <= endof(s)) &&  isvalid(s.string, s.offset+i)
+end
+
+isvalid{T<:DirectIndexString}(s::SubString{T}, i::Integer) = (start(s) <= i <= endof(s))
+
+ind2chr{T<:DirectIndexString}(s::SubString{T}, i::Integer) = begin checkbounds(s,i); i end
+chr2ind{T<:DirectIndexString}(s::SubString{T}, i::Integer) = begin checkbounds(s,i); i end
 
 nextind(s::SubString, i::Integer) = nextind(s.string, i+s.offset)-s.offset
 prevind(s::SubString, i::Integer) = prevind(s.string, i+s.offset)-s.offset
 
 convert{T<:String}(::Type{SubString{T}}, s::T) = SubString(s, 1, endof(s))
 
+bytestring{T <: ByteString}(p::SubString{T}) = bytestring(pointer(p.string.data)+p.offset, nextind(p, p.endof)-1)
+
 function serialize{T}(s, ss::SubString{T})
     # avoid saving a copy of the parent string, keeping the type of ss
     invoke(serialize, (Any,Any), s, convert(SubString{T}, convert(T,ss)))
 end
 
-function getindex(s::String, r::Range1{Int})
+function getindex(s::String, r::UnitRange{Int})
     if first(r) < 1 || endof(s) < last(r)
         error(BoundsError)
     end
@@ -661,6 +666,16 @@ end
 
 isascii(s::SubString{ASCIIString}) = true
 
+## hashing strings ##
+
+const memhash = Uint == Uint64 ? :memhash_seed : :memhash32_seed
+
+function hash{T<:ByteString}(s::Union(T,SubString{T}), h::Uint)
+    h += uint(0x71e729fd56419c81)
+    ccall(memhash, Uint, (Ptr{Uint8}, Csize_t, Uint32), s, sizeof(s), h) + h
+end
+hash(s::String, h::Uint) = hash(bytestring(s), h)
+
 ## efficient representation of repeated strings ##
 
 immutable RepString <: String
@@ -668,15 +683,27 @@ immutable RepString <: String
     repeat::Integer
 end
 
-endof(s::RepString)  = endof(s.string)*s.repeat
+function endof(s::RepString)
+    e = endof(s.string)
+    (next(s.string,e)[2]-1) * (s.repeat-1) + e
+end
 length(s::RepString) = length(s.string)*s.repeat
 sizeof(s::RepString) = sizeof(s.string)*s.repeat
 
 function next(s::RepString, i::Int)
-    if i < 1 || i > endof(s)
-        error(BoundsError)
+    if i < 1
+        throw(BoundsError())
     end
-    j = mod1(i,length(s.string))
+    e = endof(s.string)
+    sz = next(s.string,e)[2]-1
+
+    r, j = divrem(i-1, sz)
+    j += 1
+
+    if r >= s.repeat || j > e
+        throw(BoundsError())
+    end
+
     c, k = next(s.string, j)
     c, k-j+i
 end
@@ -777,22 +804,29 @@ lowercase(c::Char) = convert(Char, ccall(:towlower, Cwchar_t, (Cwchar_t,), c))
 uppercase(s::String) = map(uppercase, s)
 lowercase(s::String) = map(lowercase, s)
 
-ucfirst(s::String) = isupper(s[1]) ? s : string(uppercase(s[1]),s[nextind(s,1):end])
-lcfirst(s::String) = islower(s[1]) ? s : string(lowercase(s[1]),s[nextind(s,1):end])
+function ucfirst(s::String)
+    isempty(s) || isupper(s[1]) ? s : string(uppercase(s[1]),s[nextind(s,1):end])
+end
+function lcfirst(s::String)
+    isempty(s) || islower(s[1]) ? s : string(lowercase(s[1]),s[nextind(s,1):end])
+end
 
 ## string map, filter, has ##
 
+map_result(s::String, a::Vector{Uint8}) = UTF8String(a)
+map_result(s::Union(ASCIIString,SubString{ASCIIString}), a::Vector{Uint8}) = bytestring(a)
+
 function map(f::Function, s::String)
     out = IOBuffer(Array(Uint8,endof(s)),true,true)
     truncate(out,0)
     for c in s
         c2 = f(c)
         if !isa(c2,Char)
-            error("map(f,s::String) requires f to return Char. Try map(f,collect(s)) or a comprehension instead.")
+            error("map(f,s::String) requires f to return Char; try map(f,collect(s)) or a comprehension instead")
         end
         write(out, c2::Char)
     end
-    takebuf_string(out)
+    map_result(s, takebuf_array(out))
 end
 
 function filter(f::Function, s::String)
@@ -808,10 +842,7 @@ end
 
 ## string promotion rules ##
 
-promote_rule(::Type{UTF8String} , ::Type{ASCIIString}) = UTF8String
-promote_rule(::Type{UTF8String} , ::Type{CharString} ) = UTF8String
-promote_rule(::Type{ASCIIString}, ::Type{CharString} ) = UTF8String
-promote_rule{T<:String}(::Type{RepString}, ::Type{T}) = RepString
+promote_rule{S<:String,T<:String}(::Type{S}, ::Type{T}) = UTF8String
 
 ## printing literal quoted string data ##
 
@@ -839,7 +870,7 @@ function print_escaped(io, s::String, esc::String)
         c == '\0'       ? print(io, escape_nul(s,j)) :
         c == '\e'       ? print(io, "\\e") :
         c == '\\'       ? print(io, "\\\\") :
-        in(c,esc)       ? print(io, '\\', c) :
+        c in esc        ? print(io, '\\', c) :
         7 <= c <= 13    ? print(io, '\\', "abtnvfr"[int(c-6)]) :
         isprint(c)      ? print(io, c) :
         c <= '\x7f'     ? print(io, "\\x", hex(c, 2)) :
@@ -859,13 +890,13 @@ end
 # bare minimum unescaping function unescapes only given characters
 
 function print_unescaped_chars(io, s::String, esc::String)
-    if !in('\\',esc)
+    if !('\\' in esc)
         esc = string("\\", esc)
     end
     i = start(s)
     while !done(s,i)
         c, i = next(s,i)
-        if c == '\\' && !done(s,i) && in(s[i],esc)
+        if c == '\\' && !done(s,i) && s[i] in esc
             c, i = next(s,i)
         end
         print(io, c)
@@ -966,6 +997,7 @@ function indentation(s::String)
 end
 
 function unindent(s::String, indent::Int)
+    indent == 0 && return s
     buf = IOBuffer(Array(Uint8,endof(s)), true, true)
     truncate(buf,0)
     a = i = start(s)
@@ -976,7 +1008,9 @@ function unindent(s::String, indent::Int)
         if cutting && isblank(c)
             a = i_
             cut += blank_width(c)
-            if cut > indent
+            if cut == indent
+                cutting = false
+            elseif cut > indent
                 cutting = false
                 for _ = (indent+1):cut write(buf, ' ') end
             end
@@ -1122,7 +1156,7 @@ function shell_parse(raw::String, interp::Bool)
     if in_single_quotes; error("unterminated single quote"); end
     if in_double_quotes; error("unterminated double quote"); end
 
-    update_arg(s[i:])
+    update_arg(s[i:end])
     append_arg()
 
     if !interp
@@ -1210,7 +1244,13 @@ end
 
 function parse(str::String; raise::Bool=true)
     ex, pos = parse(str, start(str), greedy=true, raise=raise)
-    done(str, pos) || error("syntax: extra token after end of expression")
+    if isa(ex,Expr) && ex.head === :error
+        return ex
+    end
+    if !done(str, pos)
+        raise && error("extra token after end of expression")
+        return Expr(:error, "extra token after end of expression")
+    end
     return ex
 end
 
@@ -1244,11 +1284,12 @@ lpad(s, n::Integer, p=" ") = lpad(string(s),n,string(p))
 rpad(s, n::Integer, p=" ") = rpad(string(s),n,string(p))
 cpad(s, n::Integer, p=" ") = rpad(lpad(s,div(n+strwidth(s),2),p),n,p)
 
+
 # splitter can be a Char, Vector{Char}, String, Regex, ...
 # any splitter that provides search(s::String, splitter)
-
-function split(str::String, splitter, limit::Integer, keep_empty::Bool)
-    strs = String[]
+split{T<:SubString}(str::T, splitter, limit::Integer, keep_empty::Bool) = _split(str, splitter, limit, keep_empty, T[])
+split{T<:String}(str::T, splitter, limit::Integer, keep_empty::Bool) = _split(str, splitter, limit, keep_empty, SubString{T}[])
+function _split{T<:String,U<:Array}(str::T, splitter, limit::Integer, keep_empty::Bool, strs::U)
     i = start(str)
     n = endof(str)
     r = search(str,splitter,i)
@@ -1256,7 +1297,7 @@ function split(str::String, splitter, limit::Integer, keep_empty::Bool)
     while 0 < j <= n && length(strs) != limit-1
         if i < k
             if keep_empty || i < j
-                push!(strs, str[i:prevind(str,j)])
+                push!(strs, SubString(str,i,prevind(str,j)))
             end
             i = k
         end
@@ -1265,7 +1306,7 @@ function split(str::String, splitter, limit::Integer, keep_empty::Bool)
         j, k = first(r), nextind(str,last(r))
     end
     if keep_empty || !done(str,i)
-        push!(strs, str[i:])
+        push!(strs, SubString(str,i))
     end
     return strs
 end
@@ -1275,10 +1316,12 @@ split(s::String, spl)             = split(s, spl, 0, true)
 
 # a bit oddball, but standard behavior in Perl, Ruby & Python:
 const _default_delims = [' ','\t','\n','\v','\f','\r']
-split(str::String) = split(str, _default_delims, 0, false)
+split(str::String)                = split(str, _default_delims, 0, false)
 
-function rsplit(str::String, splitter, limit::Integer, keep_empty::Bool)
-    strs = String[]
+
+rsplit{T<:SubString}(str::T, splitter, limit::Integer, keep_empty::Bool) = _rsplit(str, splitter, limit, keep_empty, T[])
+rsplit{T<:String}(str::T, splitter, limit::Integer, keep_empty::Bool) = _rsplit(str, splitter, limit, keep_empty, SubString{T}[])
+function _rsplit{T<:String,U<:Array}(str::T, splitter, limit::Integer, keep_empty::Bool, strs::U)
     i = start(str)
     n = endof(str)
     r = rsearch(str,splitter)
@@ -1286,7 +1329,7 @@ function rsplit(str::String, splitter, limit::Integer, keep_empty::Bool)
     k = last(r)
     while((0 <= j < n) && (length(strs) != limit-1))
         if i <= k
-            (keep_empty || (k < n)) && unshift!(strs, str[k+1:n])
+            (keep_empty || (k < n)) && unshift!(strs, SubString(str,k+1,n))
             n = j
         end
         (k <= j) && (j = prevind(str,j))
@@ -1294,7 +1337,7 @@ function rsplit(str::String, splitter, limit::Integer, keep_empty::Bool)
         j = first(r)-1
         k = last(r)
     end
-    (keep_empty || (n > 0)) && unshift!(strs, str[1:n])
+    (keep_empty || (n > 0)) && unshift!(strs, SubString(str,1,n))
     return strs
 end
 rsplit(s::String, spl, n::Integer) = rsplit(s, spl, n, true)
@@ -1302,7 +1345,6 @@ rsplit(s::String, spl, keep::Bool) = rsplit(s, spl, 0, keep)
 rsplit(s::String, spl)             = rsplit(s, spl, 0, true)
 #rsplit(str::String) = rsplit(str, _default_delims, 0, false)
 
-
 function replace(str::ByteString, pattern, repl::Function, limit::Integer)
     n = 1
     e = endof(str)
@@ -1391,7 +1433,7 @@ function lstrip(s::String, chars::Chars=_default_delims)
     i = start(s)
     while !done(s,i)
         c, j = next(s,i)
-        if !in(c, chars)
+        if !(c in chars)
             return s[i:end]
         end
         i = j
@@ -1404,7 +1446,7 @@ function rstrip(s::String, chars::Chars=_default_delims)
     i = start(r)
     while !done(r,i)
         c, j = next(r,i)
-        if !in(c, chars)
+        if !(c in chars)
             return s[1:end-i+1]
         end
         i = j
@@ -1417,75 +1459,100 @@ strip(s::String, chars::Chars) = lstrip(rstrip(s, chars), chars)
 
 ## string to integer functions ##
 
-function parseint{T<:Integer}(::Type{T}, s::String, base::Integer)
-    if !(2 <= base <= 36); error("invalid base: ",base); end
-    i = start(s)
-    while true
-        if done(s,i)
-            throw(ArgumentError(string(
-                "premature end of integer (in ", repr(s) ,")"
-            )))
-        end
-        c,i = next(s,i)
-        if !isspace(c)
-            break
-        end
+function parseint(c::Char, base::Integer=36, a::Int=(base <= 36 ? 10 : 36))
+    2 <= base <= 62 || error("invalid base: $base")
+    d = '0' <= c <= '9' ? c-'0'    :
+        'A' <= c <= 'Z' ? c-'A'+10 :
+        'a' <= c <= 'z' ? c-'a'+a  : error("invalid digit: $(repr(c))")
+    d < base || error("invalid base $base digit $(repr(c))")
+    d
+end
+parseint{T<:Integer}(::Type{T}, c::Char, base::Integer) = convert(T,parseint(c,base))
+parseint{T<:Integer}(::Type{T}, c::Char) = convert(T,parseint(c))
+
+function parseint_next(s::String, i::Int=start(s))
+    done(s,i) && error("premature end of integer: $(repr(s))")
+    j = i
+    c, i = next(s,i)
+    c, i, j
+end
+
+function parseint_preamble(signed::Bool, s::String, base::Int)
+    c, i, j = parseint_next(s)
+    while isspace(c)
+        c, i, j = parseint_next(s,i)
     end
     sgn = 1
-    if T <: Signed && c == '-'
-        sgn = -1
-        if done(s,i)
-            throw(ArgumentError(string(
-                "premature end of integer (in ", repr(s), ")"
-            )))
+    if signed
+        if c == '-' || c == '+'
+            (c == '-') && (sgn = -1)
+            c, i, j = parseint_next(s,i)
         end
-        c,i = next(s,i)
-    elseif c == '+'
-        if done(s,i)
-            throw(ArgumentError(string(
-                "premature end of integer (in ", repr(s), ")"
-            )))
-        end
-        c,i = next(s,i)
     end
-    base = convert(T,base)
-    n::T = 0
-    while true
-        d = '0' <= c <= '9' ? c-'0' :
-            'A' <= c <= 'Z' ? c-'A'+10 :
-            'a' <= c <= 'z' ? c-'a'+10 : typemax(Int)
-        if d >= base
-            if !isspace(c)
-                throw(ArgumentError(string(
-                    repr(c)," is not a valid digit (in ", repr(s), ")"
-                )))
-            end
-            while !done(s,i)
-                c,i = next(s,i)
-                if !isspace(c)
-                    throw(ArgumentError(string(
-                        "extra characters after whitespace (in ", repr(s), ")"
-                    )))
-                end
+    while isspace(c)
+        c, i, j = parseint_next(s,i)
+    end
+    if base == 0
+        if c == '0' && !done(s,i)
+            c, i = next(s,i)
+            base = c=='b' ? 2 : c=='o' ? 8 : c=='x' ? 16 : 10
+            if base != 10
+                c, i, j = parseint_next(s,i)
             end
         else
-            # TODO: overflow detection?
-            n = n*base + d
+            base = 10
         end
-        if done(s,i)
-            break
-        end
-        c,i = next(s,i)
     end
-    if T <: Signed
-        n = flipsign(n,sgn)
+    return sgn, base, j
+end
+
+function parseint_nocheck{T<:Integer}(::Type{T}, s::String, base::Int, a::Int)
+    sgn, base, i = parseint_preamble(T<:Signed,s,base)
+    c, i = parseint_next(s,i)
+    base = convert(T,base)
+    ## FIXME: remove 128-bit specific code once 128-bit div doesn't rely on BigInt
+    m::T = T===Uint128 || T===Int128 ? typemax(T) : div(typemax(T)-base+1,base)
+    n::T = 0
+    while n <= m
+        d::T = '0' <= c <= '9' ? c-'0'    :
+               'A' <= c <= 'Z' ? c-'A'+10 :
+               'a' <= c <= 'z' ? c-'a'+a  : base
+        d < base || error("invalid base $base digit $(repr(c)) in $(repr(s))")
+        n *= base
+        n += d
+        if done(s,i)
+            n *= sgn
+            return n
+        end
+        c, i = next(s,i)
+        isspace(c) && break
+    end
+    (T <: Signed) && (n *= sgn)
+    while !isspace(c)
+        d::T = '0' <= c <= '9' ? c-'0'    :
+               'A' <= c <= 'Z' ? c-'A'+10 :
+               'a' <= c <= 'z' ? c-'a'+a  : base
+        d < base || error("invalid base $base digit $(repr(c)) in $(repr(s))")
+        (T <: Signed) && (d *= sgn)
+        n = checked_mul(n,base)
+        n = checked_add(n,d)
+        done(s,i) && return n
+        c, i = next(s,i)
+    end
+    while !done(s,i)
+        c, i = next(s,i)
+        isspace(c) || error("extra characters after whitespace in $(repr(s))")
     end
     return n
 end
+parseint_nocheck{T<:Integer}(::Type{T}, s::String, base::Int) =
+    parseint_nocheck(T, s, base, base <= 36 ? 10 : 36)
 
+parseint{T<:Integer}(::Type{T}, s::String, base::Integer) =
+    2 <= base <= 62 ? parseint_nocheck(T,s,int(base)) : error("invalid base: $base")
+parseint{T<:Integer}(::Type{T}, s::String) = parseint_nocheck(T,s,0)
 parseint(s::String, base::Integer) = parseint(Int,s,base)
-parseint(T::Type, s::String)       = parseint(T,s,10)
-parseint(s::String)                = parseint(Int,s,10)
+parseint(s::String) = parseint_nocheck(Int,s,0)
 
 integer (s::String) = int(s)
 unsigned(s::String) = uint(s)
@@ -1514,9 +1581,9 @@ float32_isvalid(s::String, out::Array{Float32,1}) =
     ccall(:jl_strtof, Int32, (Ptr{Uint8},Ptr{Float32}), s, out) == 0
 
 float64_isvalid(s::SubString, out::Array{Float64,1}) =
-    ccall(:jl_substrtod, Int32, (Ptr{Uint8},Csize_t,Int,Ptr{Float64}), s.string, convert(Csize_t,s.offset), s.endof, out) == 0
+    ccall(:jl_substrtod, Int32, (Ptr{Uint8},Csize_t,Cint,Ptr{Float64}), s.string, s.offset, s.endof, out) == 0
 float32_isvalid(s::SubString, out::Array{Float32,1}) =
-    ccall(:jl_substrtof, Int32, (Ptr{Uint8},Csize_t,Int,Ptr{Float32}), s.string, convert(Csize_t,s.offset), s.endof, out) == 0
+    ccall(:jl_substrtof, Int32, (Ptr{Uint8},Csize_t,Cint,Ptr{Float32}), s.string, s.offset, s.endof, out) == 0
 
 begin
     local tmp::Array{Float64,1} = Array(Float64,1)
@@ -1594,12 +1661,9 @@ let
     randstring() = randstring(8)
 end
 
-
 function hex2bytes(s::ASCIIString)
     len = length(s)
-    if isodd(len)
-        error("Input string length should be even")
-    end
+    iseven(len) || error("string length must be even: $(repr(s))")
     arr = zeros(Uint8, div(len,2))
     i = j = 0
     while i < len
@@ -1607,20 +1671,40 @@ function hex2bytes(s::ASCIIString)
         c = s[i+=1]
         n = '0' <= c <= '9' ? c - '0' :
             'a' <= c <= 'f' ? c - 'a' + 10 :
-            'A' <= c <= 'F' ? c - 'A' + 10 : error("Input string isn't a hexadecimal string")
+            'A' <= c <= 'F' ? c - 'A' + 10 :
+                error("not a hexadecimal string: $(repr(s))")
         c = s[i+=1]
         n = '0' <= c <= '9' ? n << 4 + c - '0' :
             'a' <= c <= 'f' ? n << 4 + c - 'a' + 10 :
-            'A' <= c <= 'F' ? n << 4 + c - 'A' + 10 : error("Input string isn't a hexadecimal string")
+            'A' <= c <= 'F' ? n << 4 + c - 'A' + 10 :
+                error("not a hexadecimal string: $(repr(s))")
         arr[j+=1] = n
     end
     return arr
 end
 
-bytes2hex(arr::Array{Uint8,1}) = join([hex(i, 2) for i in arr])
+bytes2hex{T<:Uint8}(arr::Array{T,1}) = join([hex(i,2) for i in arr])
 
 function repr(x)
     s = IOBuffer()
     showall(s, x)
     takebuf_string(s)
 end
+
+if sizeof(Cwchar_t) == 2
+    const WString = UTF16String # const, not typealias, to get constructor
+    const wstring = utf16
+elseif sizeof(Cwchar_t) == 4
+    const WString = UTF32String # const, not typealias, to get constructor
+    const wstring = utf32
+end
+
+# pointer conversions of ASCII/UTF8/UTF16/UTF32 strings:
+pointer(x::Union(ByteString,UTF16String,UTF32String)) = pointer(x.data)
+pointer{T<:ByteString}(x::SubString{T}) = pointer(x.string.data) + x.offset
+pointer(x::ByteString, i::Integer) = pointer(x.data)+(i-1)
+pointer{T<:ByteString}(x::SubString{T}, i::Integer) = pointer(x.string.data) + x.offset + (i-1)
+pointer(x::Union(UTF16String,UTF32String), i::Integer) = pointer(x)+(i-1)*sizeof(eltype(x.data))
+pointer{T<:Union(UTF16String,UTF32String)}(x::SubString{T}) = pointer(x.string.data) + x.offset*sizeof(eltype(x.data))
+pointer{T<:Union(UTF16String,UTF32String)}(x::SubString{T}, i::Integer) = pointer(x.string.data) + (x.offset + (i-1))*sizeof(eltype(x.data))
+
