@@ -1,11 +1,12 @@
+
 ## subarrays ##
 
-typealias RangeIndex Union(Int, Range{Int}, Range1{Int})
+typealias RangeIndex Union(Int, Range{Int}, UnitRange{Int})
 
 type SubArray{T,N,A<:AbstractArray,I<:(RangeIndex...,)} <: AbstractArray{T,N}
     parent::A
     indexes::I
-    dims::Dims
+    dims::NTuple{N,Int}
     strides::Array{Int,1}  # for accessing parent with linear indexes
     first_index::Int
 
@@ -17,7 +18,7 @@ type SubArray{T,N,A<:AbstractArray,I<:(RangeIndex...,)} <: AbstractArray{T,N}
         end
     #linear indexing constructor (ranges)
     elseif N == 1 && length(I) == 1 && A <: Array
-        function SubArray(p::A, i::(Range1{Int},))
+        function SubArray(p::A, i::(UnitRange{Int},))
             new(p, i, (length(i[1]),), [1], first(i[1]))
         end
         function SubArray(p::A, i::(Range{Int},))
@@ -35,7 +36,7 @@ type SubArray{T,N,A<:AbstractArray,I<:(RangeIndex...,)} <: AbstractArray{T,N}
                 else
                     push!(newdims, length(i[j]))
                     #may want to return error if step(i[j]) <= 0
-                    push!(newstrides, isa(i[j],Range1) ? pstride :
+                    push!(newstrides, isa(i[j],UnitRange) ? pstride :
                          pstride * step(i[j]))
                     newfirst += (first(i[j])-1)*pstride
                 end
@@ -47,16 +48,31 @@ type SubArray{T,N,A<:AbstractArray,I<:(RangeIndex...,)} <: AbstractArray{T,N}
 end
 
 #linear indexing sub (may want to rename as slice)
-function sub{T,N}(A::Array{T,N}, i::(RangeIndex,))
-    SubArray{T,(isa(i[1], Int) ? 0 : 1),typeof(A),typeof(i)}(A, i)
+function sub{T,N}(A::Array{T,N}, i::(Union(Range{Int}, UnitRange{Int}),))
+    SubArray{T,1,typeof(A),typeof(i)}(A, i)
+end
+
+# if `I` were a vector, index_ranges would do the following:
+# j = length(I)
+# while j > 0 && isa(I[j], Int)
+#     j -= 1
+# end
+# for i = 1:j
+#     if isa(I[i], Int)
+#         I[i] = I[i]:I[i]
+#     end
+# end
+to_range(j::Int) = j:j
+to_range(j::RangeIndex) = j
+index_ranges(I::Int...) = I
+index_ranges(i, I...) = tuple(to_range(i), index_ranges(I...)...)
+
+function sub_internal{T,N,L}(A::AbstractArray{T,N}, i::NTuple{N,RangeIndex}, ::NTuple{L,Int})
+    SubArray{T,L,typeof(A),typeof(i)}(A, i)
 end
 
 function sub{T,N}(A::AbstractArray{T,N}, i::NTuple{N,RangeIndex})
-    L = length(i)
-    while L > 0 && isa(i[L], Int); L-=1; end
-    i0 = map(j -> isa(j, Int) ? (j:j) : j, i[1:L])
-    i = ntuple(length(i), k->(k<=L ? i0[k] : i[k]))
-    SubArray{T,L,typeof(A),typeof(i)}(A, i)
+    sub_internal(A, index_ranges(i...), index_shape(i...))
 end
 
 sub{N}(A::SubArray, i::NTuple{N,RangeIndex}) = sub(A, i...)
@@ -82,12 +98,15 @@ function sub(A::SubArray, i::RangeIndex...)
     SubArray{eltype(A),L,typeof(A.parent),typeof(ni)}(A.parent, ni)
 end
 
+# Drops all Ints from a tuple of RangeIndexes
+ranges_only(I::Int...) = ()
+ranges_only(i::Int, I...) = ranges_only(I...)
+ranges_only(i::Union(Range{Int}, UnitRange{Int}), I...) = tuple(i, ranges_only(I...)...)
 
-function slice{T,N}(A::AbstractArray{T,N}, i::NTuple{N,RangeIndex})
-    n = 0
-    for j = i; if !isa(j, Int); n += 1; end; end
-    SubArray{T,n,typeof(A),typeof(i)}(A, i)
+function slice_internal{T,N,L}(A::AbstractArray{T,N}, i::NTuple{N,RangeIndex}, ::NTuple{L,RangeIndex})
+    SubArray{T,L,typeof(A),typeof(i)}(A, i)
 end
+slice{T,N}(A::AbstractArray{T,N}, i::NTuple{N,RangeIndex}) = slice_internal(A, i, ranges_only(i...))
 
 # Throw error on slice dimension mismatch
 slice{T,N,M}(A::AbstractArray{T,N}, i::NTuple{M,RangeIndex}) = throw(BoundsError())
@@ -109,9 +128,9 @@ function slice(A::SubArray, i::RangeIndex...)
     slice(A.parent, tuple(newindexes...))
 end
 
-# Generic fallback for Colon translation
-sub(A::AbstractArray, I...) = sub(A, ntuple(length(I), i-> isa(I[i], Colon) ? (1:size(A,i)) : I[i])...)
-slice(A::AbstractArray, I...) = slice(A, ntuple(length(I), i-> isa(I[i], Colon) ? (1:size(A,i)) : I[i])...)
+# Colon translation
+sub(A::AbstractArray, I::Union(RangeIndex, Colon)...) = sub(A, ntuple(length(I), i-> isa(I[i], Colon) ? (1:size(A,i)) : I[i])...)
+slice(A::AbstractArray, I::Union(RangeIndex, Colon)...) = slice(A, ntuple(length(I), i-> isa(I[i], Colon) ? (1:size(A,i)) : I[i])...)
 
 
 ### rename the old slice function ###
@@ -129,7 +148,7 @@ slice(A::AbstractArray, I...) = slice(A, ntuple(length(I), i-> isa(I[i], Colon)
 #        for j in sdims
 #            if i == j
 #                if size(a, i) != 1
-#                    error("slice: dimension ", i, " has length greater than 1")
+#                    error("dimension ", i, " has length greater than 1")
 #                end
 #                next = 1
 #                break
@@ -146,7 +165,7 @@ slice(A::AbstractArray, I...) = slice(A, ntuple(length(I), i-> isa(I[i], Colon)
 #        for j in sdims
 #            if i == j
 #                if length(next) != 1
-#                    error("slice: dimension ", i," has length greater than 1")
+#                    error("dimension ", i," has length greater than 1")
 #                end
 #                next = isa(next, Int) ? next : first(next)
 #                break
@@ -170,7 +189,6 @@ parentindexes(a::AbstractArray) = ntuple(ndims(a), i->1:size(a,i))
 copy(s::SubArray) = copy!(similar(s.parent, size(s)), s)
 similar(s::SubArray, T, dims::Dims) = similar(s.parent, T, dims)
 
-getindex{T}(s::SubArray{T,0,AbstractArray{T,0}}) = s.parent[]
 getindex{T}(s::SubArray{T,0}) = s.parent[s.first_index]
 
 getindex{T}(s::SubArray{T,1}, i::Integer) =
@@ -197,61 +215,11 @@ getindex(s::SubArray, i0::Real, i1::Real, i2::Real, i3::Real, i4::Real) =
     getindex(s, to_index(i0), to_index(i1), to_index(i2), to_index(i3), to_index(i4))
 getindex(s::SubArray, i0::Real, i1::Real, i2::Real, i3::Real, i4::Real, i5::Real) =
     getindex(s, to_index(i0), to_index(i1), to_index(i2), to_index(i3), to_index(i4), to_index(i5))
-getindex(s::SubArray, i0::Real, i1::Real, i2::Real, i3::Real, i4::Real, i5::Real, is::Int...) =
-    getindex(s, to_index(i0), to_index(i1), to_index(i2), to_index(i3), to_index(i4), to_index(5), is...)
+getindex(s::SubArray, i0::Real, i1::Real, i2::Real, i3::Real, i4::Real, i5::Real, is::Real...) =
+    getindex(s, to_index(i0), to_index(i1), to_index(i2), to_index(i3), to_index(i4), to_index(i5), to_index(is)...)
 
 getindex(s::SubArray, i::Integer) = s[ind2sub(size(s), i)...]
 
-function getindex{T}(s::SubArray{T,2}, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    ind -= 1
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2]]
-end
-
-function getindex{T}(s::SubArray{T,3}, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    ind -= 1
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3]]
-end
-
-function getindex{T}(s::SubArray{T,4}, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    @inbounds strd4 = strd3*s.dims[3]
-    ind -= 1
-    i4 = div(ind,strd4)
-    ind -= i4*strd4
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3] + i4*s.strides[4]]
-end
-
-function getindex{T}(s::SubArray{T,5}, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    @inbounds strd4 = strd3*s.dims[3]
-    @inbounds strd5 = strd4*s.dims[4]
-    ind -= 1
-    i5 = div(ind,strd5)
-    ind -= i5*strd5
-    i4 = div(ind,strd4)
-    ind -= i4*strd4
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3] + i4*s.strides[4] + i5*s.strides[5]]
-end
-
 function getindex(s::SubArray, is::Integer...)
     index = s.first_index
     for i = 1:length(is)
@@ -263,7 +231,26 @@ function getindex(s::SubArray, is::Integer...)
     s.parent[index]
 end
 
-getindex{T}(s::SubArray{T,1}, I::Range1{Int}) =
+function getindex_bool_1d(S::SubArray, I::AbstractArray{Bool})
+    n = sum(I)
+    out = similar(S, n)
+    c = 1
+    for i = 1:length(I)
+        if I[i]
+            out[c] = S[i]
+            c += 1
+        end
+    end
+    out
+end
+
+getindex{T}(S::SubArray{T,1}, I::AbstractArray{Bool,1}) = getindex_bool_1d(S, I)
+getindex{T}(S::SubArray{T,2}, I::AbstractArray{Bool,2}) = getindex_bool_1d(S, I)
+getindex{T}(S::SubArray{T,3}, I::AbstractArray{Bool,3}) = getindex_bool_1d(S, I)
+getindex{T}(S::SubArray{T,4}, I::AbstractArray{Bool,4}) = getindex_bool_1d(S, I)
+getindex{T}(S::SubArray{T,5}, I::AbstractArray{Bool,5}) = getindex_bool_1d(S, I)
+
+getindex{T}(s::SubArray{T,1}, I::UnitRange{Int}) =
     getindex(s.parent, (s.first_index+(first(I)-1)*s.strides[1]):s.strides[1]:(s.first_index+(last(I)-1)*s.strides[1]))
 
 getindex{T}(s::SubArray{T,1}, I::Range{Int}) =
@@ -285,9 +272,8 @@ function translate_indexes(s::SubArray, I::Union(Real,AbstractArray)...)
     for i = 1:n-havelinear
         newindexes[pdims[i]] = s.indexes[pdims[i]][I[i]]
     end
-    lastdim = pdims[n]
     if havelinear
-        newindexes = newindexes[1:lastdim]
+        newindexes = newindexes[1:pdims[n]]
         newindexes[pdims[n]] = translate_linear_indexes(s, n, I[end], pdims)
     end
     newindexes
@@ -297,8 +283,8 @@ end
 # the parent array
 function translate_linear_indexes(s, n, I, pdims)
     idx = Array(Int, length(I))
-    ssztail = size(s)[n:]
-    indexestail = s.indexes[pdims[n:]]
+    ssztail = size(s)[n:end]
+    indexestail = s.indexes[pdims[n:end]]
     # The next gets the strides of dimensions listed in pdims[n:end], relative to the stride of pdims[n]
     pstrd = [1]
     j = n+1
@@ -359,60 +345,6 @@ end
 
 setindex!(s::SubArray, v, i::Integer) = setindex!(s, v, ind2sub(size(s), i)...)
 
-function setindex!{T}(s::SubArray{T,2}, v, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    ind -= 1
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2]] = v
-    s
-end
-
-function setindex!{T}(s::SubArray{T,3}, v, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    ind -= 1
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3]] = v
-    s
-end
-
-function setindex!{T}(s::SubArray{T,4}, v, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    @inbounds strd4 = strd3*s.dims[3]
-    ind -= 1
-    i4 = div(ind,strd4)
-    ind -= i4*strd4
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3] + i4*s.strides[4]] = v
-    s
-end
-
-function setindex!{T}(s::SubArray{T,5}, v, ind::Integer)
-    @inbounds strd2 = s.dims[1]
-    @inbounds strd3 = strd2*s.dims[2]
-    @inbounds strd4 = strd3*s.dims[3]
-    @inbounds strd5 = strd4*s.dims[4]
-    ind -= 1
-    i5 = div(ind,strd5)
-    ind -= i5*strd5
-    i4 = div(ind,strd4)
-    ind -= i4*strd4
-    i3 = div(ind,strd3)
-    ind -= i3*strd3
-    i2 = div(ind,strd2)
-    i1 = ind-i2*strd2
-    s.parent[s.first_index + i1*s.strides[1] + i2*s.strides[2] + i3*s.strides[3] + i4*s.strides[4] + i5*s.strides[5]] = v
-    s
-end
-
 function setindex!(s::SubArray, v, is::Integer...)
     index = s.first_index
     for i = 1:length(is)
@@ -422,8 +354,6 @@ function setindex!(s::SubArray, v, is::Integer...)
     return s
 end
 
-setindex!{T}(s::SubArray{T,0,AbstractArray{T,0}},v) = setindex!(s.parent, v)
-
 setindex!{T}(s::SubArray{T,0}, v) = setindex!(s.parent, v, s.first_index)
 
 
@@ -442,7 +372,7 @@ setindex!{T}(s::SubArray{T,4}, v, i::Integer, j::Integer, k::Integer, l::Integer
 setindex!{T}(s::SubArray{T,5}, v, i::Integer, j::Integer, k::Integer, l::Integer, m::Integer) =
     setindex!(s.parent, v, s.first_index +(i-1)*s.strides[1]+(j-1)*s.strides[2]+(k-1)*s.strides[3]+(l-1)*s.strides[4]+(m-1)*s.strides[5])
 
-setindex!{T}(s::SubArray{T,1}, v, I::Range1{Int}) =
+setindex!{T}(s::SubArray{T,1}, v, I::UnitRange{Int}) =
     setindex!(s.parent, v, (s.first_index+(first(I)-1)*s.strides[1]):s.strides[1]:(s.first_index+(last(I)-1)*s.strides[1]))
 
 setindex!{T}(s::SubArray{T,1}, v, I::Range{Int}) =
@@ -458,11 +388,11 @@ end
 
 # to avoid ambiguity warning
 function setindex!(s::SubArray, v, I::Real)
-    newindexes = translate_indexes(s, (I,))
+    newindexes = translate_indexes(s, (to_index(I),))
     setindex!(s.parent, v, newindexes...)
 end
 function setindex!(s::SubArray, v, I::Union(Real,AbstractArray)...)
-    newindexes = translate_indexes(s, I...)
+    newindexes = translate_indexes(s, to_index(I)...)
     setindex!(s.parent, v, newindexes...)
 end
 
@@ -470,6 +400,8 @@ stride(s::SubArray, i::Integer) = i <= length(s.strides) ? s.strides[i] : s.stri
 
 convert{T}(::Type{Ptr{T}}, x::SubArray{T}) =
     pointer(x.parent) + (x.first_index-1)*sizeof(T)
+convert{T}(::Type{Ptr{None}}, x::SubArray{T}) = convert(Ptr{None}, convert(Ptr{T},x))
+convert{T,S,N}(::Type{Array{T,N}},A::SubArray{S,N}) = copy!(Array(T,size(A)), A)
 
 pointer(s::SubArray, i::Int) = pointer(s, ind2sub(size(s), i))
 
