@@ -39,7 +39,7 @@ function endof(s::UTF8String)
 end
 length(s::UTF8String) = int(ccall(:u8_strlen, Csize_t, (Ptr{Uint8},), s.data))
 
-function getindex(s::UTF8String, i::Int)
+function next(s::UTF8String, i::Int)
     # potentially faster version
     # d = s.data
     # a::Uint32 = d[i]
@@ -62,11 +62,12 @@ function getindex(s::UTF8String, i::Int)
             # b is a continuation byte of a valid UTF-8 character
             error("invalid UTF-8 character index")
         end
-        return '\ufffd'
+        # move past 1 byte in case the data is actually Latin-1
+        return '\ufffd', i+1
     end
     trailing = utf8_trailing[b+1]
     if length(d) < i + trailing
-        return '\ufffd'
+        return '\ufffd', i+1
     end
     c::Uint32 = 0
     for j = 1:trailing+1
@@ -75,12 +76,9 @@ function getindex(s::UTF8String, i::Int)
         i += 1
     end
     c -= utf8_offset[trailing+1]
-    char(c)
+    char(c), i
 end
 
-# this is a trick to allow inlining and tuple elision
-next(s::UTF8String, i::Int) = (s[i], i+1+utf8_trailing[s.data[i]+1])
-
 function first_utf8_byte(c::Char)
     c < 0x80    ? uint8(c)            :
     c < 0x800   ? uint8((c>>6 )|0xc0) :
@@ -95,11 +93,20 @@ sizeof(s::UTF8String) = sizeof(s.data)
 isvalid(s::UTF8String, i::Integer) =
     (1 <= i <= endof(s.data)) && is_utf8_start(s.data[i])
 
-function getindex(s::UTF8String, r::Range1{Int})
-    a, b = first(r), last(r)
-    i = isvalid(s,a) ? a : nextind(s,a)
-    j = b < endof(s) ? nextind(s,b)-1 : endof(s.data)
-    UTF8String(s.data[i:j])
+const empty_utf8 = UTF8String(Uint8[])
+
+function getindex(s::UTF8String, r::UnitRange{Int})
+    isempty(r) && return empty_utf8
+    i, j = first(r), last(r)
+    d = s.data
+    if !is_utf8_start(d[i])
+        i = nextind(s,i)
+    end
+    if j > endof(s)
+        throw(BoundsError())
+    end
+    j = nextind(s,j)-1
+    UTF8String(d[i:j])
 end
 
 function search(s::UTF8String, c::Char, i::Integer)
@@ -122,6 +129,9 @@ function rsearch(s::UTF8String, c::Char, i::Integer)
 end
 
 function string(a::ByteString...)
+    if length(a) == 1
+        return a[1]::UTF8String
+    end
     # ^^ at least one must be UTF-8 or the ASCII-only method would get called
     data = Array(Uint8,0)
     for d in a
@@ -130,9 +140,6 @@ function string(a::ByteString...)
     UTF8String(data)
 end
 
-ucfirst(s::UTF8String) = string(uppercase(s[1]), s[2:])
-lcfirst(s::UTF8String) = string(lowercase(s[1]), s[2:])
-
 function reverse(s::UTF8String)
     out = similar(s.data)
     if ccall(:u8_reverse, Cint, (Ptr{Uint8}, Ptr{Uint8}, Csize_t),
