@@ -64,8 +64,8 @@ const VERSION_REGEX = r"^
     (\d+)                                   # major         (required)
     (?:\.(\d+))?                            # minor         (optional)
     (?:\.(\d+))?                            # patch         (optional)
-    (?:(-)|
-    (?:-((?:[0-9a-z-]+\.)*[0-9a-z-]+))?     # pre-release   (optional)
+    (?:(-)|                                 # pre-release   (optional)
+    ([a-z][0-9a-z-]*(?:\.[0-9a-z-]+)*|-(?:[0-9a-z-]+\.)*[0-9a-z-]+)?
     (?:(\+)|
     (?:\+((?:[0-9a-z-]+\.)*[0-9a-z-]+))?    # build         (optional)
     ))
@@ -86,6 +86,9 @@ function convert(::Type{VersionNumber}, v::String)
     major = int(major)
     minor = minor != nothing ? int(minor) : 0
     patch = patch != nothing ? int(patch) : 0
+    if prerl != nothing && !isempty(prerl) && prerl[1] == '-'
+        prerl = prerl[2:end] # strip leading '-'
+    end
     prerl = prerl != nothing ? split_idents(prerl) : minus == "-" ? ("",) : ()
     build = build != nothing ? split_idents(build) : plus  == "+" ? ("",) : ()
     VersionNumber(major, minor, patch, prerl, build)
@@ -115,7 +118,7 @@ function ident_cmp(A::(Union(Int,ASCIIString)...),
     !done(A,i) && done(B,j) ? +1 : 0
 end
 
-function isequal(a::VersionNumber, b::VersionNumber)
+function ==(a::VersionNumber, b::VersionNumber)
     (a.major != b.major) && return false
     (a.minor != b.minor) && return false
     (a.patch != b.patch) && return false
@@ -145,7 +148,14 @@ function isless(a::VersionNumber, b::VersionNumber)
     return false
 end
 
-hash(v::VersionNumber) = hash([v.(n) for n in VersionNumber.names])
+function hash(v::VersionNumber, h::Uint)
+    h += uint(0x8ff4ffdb75f9fede)
+    h = hash(v.major, h)
+    h = hash(v.minor, h)
+    h = hash(v.patch, h)
+    h = hash(v.prerelease, ~h)
+    h = hash(v.build, ~h)
+end
 
 lowerbound(v::VersionNumber) = VersionNumber(v.major, v.minor, v.patch, ("",), ())
 upperbound(v::VersionNumber) = VersionNumber(v.major, v.minor, v.patch, (), ("",))
@@ -180,58 +190,64 @@ end
 
 ## julia version info
 
-begin
 # Include build number if we've got at least some distance from a tag (e.g. a release)
-prerelease = BUILD_INFO.prerelease ? "-prerelease" : ""
-build_number = BUILD_INFO.build_number != 0 ? "+$(BUILD_INFO.build_number)" : ""
 try
-    global const VERSION = convert(VersionNumber, "$(BUILD_INFO.version_string)$(prerelease)$(build_number)")
+    build_number = GIT_VERSION_INFO.build_number != 0 ? "+$(GIT_VERSION_INFO.build_number)" : ""
+    global const VERSION = convert(VersionNumber, "$(VERSION_STRING)$(build_number)")
 catch e
     println("while creating Base.VERSION, ignoring error $e")
     global const VERSION = VersionNumber(0)
 end
 
-if BUILD_INFO.tagged_commit
-    const commit_string = BUILD_INFO.TAGGED_RELEASE_BANNER
-elseif BUILD_INFO.commit == ""
-    const commit_string = ""
-else
-    local days = int(floor((ccall(:clock_now, Float64, ()) - BUILD_INFO.fork_master_timestamp) / (60 * 60 * 24)))
-    if BUILD_INFO.fork_master_distance == 0
-        const commit_string = "Commit $(BUILD_INFO.commit_short) ($(days) days old master)"
+function banner(io::IO = STDOUT)
+    if GIT_VERSION_INFO.tagged_commit
+        commit_string = TAGGED_RELEASE_BANNER
+    elseif GIT_VERSION_INFO.commit == ""
+        commit_string = ""
     else
-        const commit_string = "$(BUILD_INFO.branch)/$(BUILD_INFO.commit_short) (fork: $(BUILD_INFO.fork_master_distance) commits, $(days) days)"
+        days = int(floor((ccall(:clock_now, Float64, ()) - GIT_VERSION_INFO.fork_master_timestamp) / (60 * 60 * 24)))
+        unit = days == 1 ? "day" : "days"
+        distance = GIT_VERSION_INFO.fork_master_distance
+        commit = GIT_VERSION_INFO.commit_short
+
+        if distance == 0
+            commit_string = "Commit $(commit) ($(days) $(unit) old master)"
+        else
+            branch = GIT_VERSION_INFO.branch
+            commit_string = "$(branch)/$(commit) (fork: $(distance) commits, $(days) $(unit))"
+        end
+    end
+    commit_date = GIT_VERSION_INFO.date_string != "" ? " ($(GIT_VERSION_INFO.date_string))": ""
+
+    if have_color
+        tx = "\033[0m\033[1m" # text
+        jl = "\033[0m\033[1m" # julia
+        d1 = "\033[34m" # first dot
+        d2 = "\033[31m" # second dot
+        d3 = "\033[32m" # third dot
+        d4 = "\033[35m" # fourth dot
+
+        print(io,"""\033[1m               $(d3)_
+           $(d1)_       $(jl)_$(tx) $(d2)_$(d3)(_)$(d4)_$(tx)     |  A fresh approach to technical computing
+          $(d1)(_)$(jl)     | $(d2)(_)$(tx) $(d4)(_)$(tx)    |  Documentation: http://docs.julialang.org
+           $(jl)_ _   _| |_  __ _$(tx)   |  Type \"help()\" for help.
+          $(jl)| | | | | | |/ _` |$(tx)  |
+          $(jl)| | |_| | | | (_| |$(tx)  |  Version $(VERSION)$(commit_date)
+         $(jl)_/ |\\__'_|_|_|\\__'_|$(tx)  |  $(commit_string)
+        $(jl)|__/$(tx)                   |  $(Sys.MACHINE)
+
+        \033[0m""")
+    else
+        print(io,"""
+                       _
+           _       _ _(_)_     |  A fresh approach to technical computing
+          (_)     | (_) (_)    |  Documentation: http://docs.julialang.org
+           _ _   _| |_  __ _   |  Type \"help()\" for help.
+          | | | | | | |/ _` |  |
+          | | |_| | | | (_| |  |  Version $(VERSION)$(commit_date)
+         _/ |\\__'_|_|_|\\__'_|  |  $(commit_string)
+        |__/                   |  $(Sys.MACHINE)
+
+        """)
     end
 end
-commit_date = BUILD_INFO.date_string != "" ? " ($(BUILD_INFO.date_string))": ""
-
-const banner_plain =
-"""
-               _
-   _       _ _(_)_     |  A fresh approach to technical computing
-  (_)     | (_) (_)    |  Documentation: http://docs.julialang.org
-   _ _   _| |_  __ _   |  Type \"help()\" to list help topics
-  | | | | | | |/ _` |  |
-  | | |_| | | | (_| |  |  Version $(VERSION)$(commit_date)
- _/ |\\__'_|_|_|\\__'_|  |  $(commit_string)
-|__/                   |  $(Sys.MACHINE)
-
-"""
-local tx = "\033[0m\033[1m" # text
-local jl = "\033[0m\033[1m" # julia
-local d1 = "\033[34m" # first dot
-local d2 = "\033[31m" # second dot
-local d3 = "\033[32m" # third dot
-local d4 = "\033[35m" # fourth dot
-const banner_color =
-"\033[1m               $(d3)_
-   $(d1)_       $(jl)_$(tx) $(d2)_$(d3)(_)$(d4)_$(tx)     |  A fresh approach to technical computing
-  $(d1)(_)$(jl)     | $(d2)(_)$(tx) $(d4)(_)$(tx)    |  Documentation: http://docs.julialang.org
-   $(jl)_ _   _| |_  __ _$(tx)   |  Type \"help()\" to list help topics
-  $(jl)| | | | | | |/ _` |$(tx)  |
-  $(jl)| | |_| | | | (_| |$(tx)  |  Version $(VERSION)$(commit_date)
- $(jl)_/ |\\__'_|_|_|\\__'_|$(tx)  |  $(commit_string)
-$(jl)|__/$(tx)                   |  $(Sys.MACHINE)
-
-\033[0m"
-end # begin
