@@ -1,7 +1,7 @@
 # Set up environment for Julia OSX binary distribution
 let
     ROOT = abspath(JULIA_HOME,"..")
-    ENV["PATH"]="$JULIA_HOME:$(ENV["PATH"])"
+    ENV["PATH"]="$JULIA_HOME:$(joinpath(ROOT, "libexec", "git-core")):$(ENV["PATH"])"
     ENV["FONTCONFIG_PATH"] = joinpath(ROOT, "etc", "fonts")
     ENV["GIT_EXEC_PATH"] = joinpath(ROOT, "libexec", "git-core")
     ENV["GIT_TEMPLATE_DIR"] = joinpath(ROOT, "share", "git-core")
