@@ -1,3 +1,4 @@
+module TypeTrees
 ##
 # Generate a text graphic of Julia modules type tree
 ##
@@ -81,7 +82,7 @@ function store_all_from(m::Module)
             t = eval(m,expr)
             isa(t, Type) && store_type(string(expr), t)
         #catch ex
-        #    println("Error adding ", string(expr), " ", t, " ", m, " (", ex, ")")
+        #    println("Error adding ", string(expr), m, " (", ex, ")")
         end
     end
 end
@@ -95,7 +96,8 @@ type_props(typ::DataType) = string("<<",
                                  " >>")
 
 function print_tree(subtypes::Dict{String, TTNode}, pfx::String="")
-    for (n,v) in subtypes
+    for n in sort!([keys(subtypes)...])
+        v = subtypes[n]
         if(n == string(v.typ))
             println(pfx, "+- ", n, " ", type_props(v.typ))
         else
@@ -111,9 +113,10 @@ end
 # TODO: option to list subtrees of type tree, or other symbol types
 const types_tree = Dict{String, TTNode}()
 
-for m = [Base, Core, Main]
+for m in (Base, Core, Main)
     store_all_from(m)
 end
 
-print_tree(types_tree)
+# print_tree(types_tree)
 
+end # module
\ No newline at end of file
