@@ -6,8 +6,11 @@
 #include <stdarg.h>
 #include <assert.h>
 #include "julia.h"
-#include "newobj_internal.h"
-#include "builtin_proto.h"
+#include "julia_internal.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 jl_value_t *jl_true;
 jl_value_t *jl_false;
@@ -50,6 +53,7 @@ jl_datatype_t *jl_errorexception_type=NULL;
 jl_datatype_t *jl_typeerror_type;
 jl_datatype_t *jl_methoderror_type;
 jl_datatype_t *jl_loaderror_type;
+jl_datatype_t *jl_undefvarerror_type;
 jl_datatype_t *jl_pointer_type;
 jl_datatype_t *jl_voidpointer_type;
 jl_value_t *jl_an_empty_cell=NULL;
@@ -69,7 +73,7 @@ jl_sym_t *export_sym;  jl_sym_t *import_sym;
 jl_sym_t *importall_sym; jl_sym_t *toplevel_sym;
 jl_sym_t *quote_sym;   jl_sym_t *amp_sym;
 jl_sym_t *top_sym;     jl_sym_t *colons_sym;
-jl_sym_t *line_sym;    jl_sym_t *jl_continue_sym;
+jl_sym_t *line_sym;    jl_sym_t *jl_incomplete_sym;
 // head symbols for each expression type
 jl_sym_t *goto_sym;    jl_sym_t *goto_ifnot_sym;
 jl_sym_t *label_sym;   jl_sym_t *return_sym;
@@ -87,21 +91,53 @@ jl_sym_t *compositetype_sym; jl_sym_t *type_goto_sym;
 jl_sym_t *global_sym; jl_sym_t *tuple_sym;
 jl_sym_t *dot_sym;    jl_sym_t *newvar_sym;
 jl_sym_t *boundscheck_sym; jl_sym_t *copyast_sym;
+jl_sym_t *simdloop_sym;
 
 typedef struct {
     int64_t a;
     int64_t b;
 } bits128_t;
 
-jl_value_t *jl_new_bits(jl_datatype_t *bt, void *data)
+static size_t jl_new_bits_align(jl_value_t *dt)
 {
-    if (bt == jl_uint8_type)        return jl_box_uint8(*(uint8_t*)data);
-    else if (bt == jl_int64_type)   return jl_box_int64(*(int64_t*)data);
-    else if (bt == jl_bool_type)    return (*(int8_t*)data) ? jl_true:jl_false;
-    else if (bt == jl_int32_type)   return jl_box_int32(*(int32_t*)data);
-    else if (bt == jl_float64_type) return jl_box_float64(*(double*)data);
-    
+    if (jl_is_tuple(dt)) {
+        size_t i, l = jl_tuple_len(dt), align = 0;
+        for (i = 0; i < l; i++) {
+            size_t l = jl_new_bits_align(jl_tupleref(dt,i));
+            if (l > align)
+                align = l;
+        }
+        return align;
+    }
+    return ((jl_datatype_t*)dt)->alignment;
+}
+
+static jl_value_t *jl_new_bits_internal(jl_value_t *dt, void *data, size_t *len)
+{
+    if (jl_is_tuple(dt)) {
+        jl_tuple_t *tuple = (jl_tuple_t*)dt;
+        *len = LLT_ALIGN(*len, jl_new_bits_align(dt));
+        size_t i, l = jl_tuple_len(tuple);
+        jl_value_t *v = (jl_value_t*) jl_alloc_tuple(l);
+        JL_GC_PUSH1(v);
+        for (i = 0; i < l; i++) {
+            jl_tupleset(v,i,jl_new_bits_internal(jl_tupleref(tuple,i), (char*)data, len));
+        }
+        JL_GC_POP();
+        return v;
+    }
+
+    jl_datatype_t *bt = (jl_datatype_t*)dt;
     size_t nb = jl_datatype_size(bt);
+    *len = LLT_ALIGN(*len, bt->alignment);
+    data = (char*)data + (*len);
+    *len += nb;
+    if (bt == jl_uint8_type)   return jl_box_uint8(*(uint8_t*)data);
+    if (bt == jl_int64_type)   return jl_box_int64(*(int64_t*)data);
+    if (bt == jl_bool_type)    return (*(int8_t*)data) ? jl_true:jl_false;
+    if (bt == jl_int32_type)   return jl_box_int32(*(int32_t*)data);
+    if (bt == jl_float64_type) return jl_box_float64(*(double*)data);
+
     jl_value_t *v = 
         (jl_value_t*)allocobj((NWORDS(LLT_ALIGN(nb,sizeof(void*)))+1)*
                               sizeof(void*));
@@ -117,6 +153,12 @@ jl_value_t *jl_new_bits(jl_datatype_t *bt, void *data)
     return v;
 }
 
+jl_value_t *jl_new_bits(jl_value_t *bt, void *data)
+{
+    size_t len = 0;
+    return jl_new_bits_internal(bt, data, &len);
+}
+
 void jl_assign_bits(void *dest, jl_value_t *bits)
 {
     size_t nb = jl_datatype_size(jl_typeof(bits));
@@ -150,23 +192,25 @@ jl_value_t *jl_get_nth_field(jl_value_t *v, size_t i)
     if (st->fields[i].isptr) {
         return *(jl_value_t**)((char*)v + offs);
     }
-    return jl_new_bits((jl_datatype_t*)jl_tupleref(st->types,i),
-                       (char*)v + offs);
+    return jl_new_bits(jl_tupleref(st->types,i), (char*)v + offs);
 }
 
-int jl_field_isdefined(jl_value_t *v, jl_sym_t *fld, int err)
+jl_value_t *jl_get_nth_field_checked(jl_value_t *v, size_t i)
 {
     jl_datatype_t *st = (jl_datatype_t*)jl_typeof(v);
-    int i = jl_field_index(st, fld, err);
-    if (i == -1) return 0;
+    if (i >= jl_tuple_len(st->names))
+        jl_throw(jl_bounds_exception);
     size_t offs = jl_field_offset(st,i) + sizeof(void*);
     if (st->fields[i].isptr) {
-        return *(jl_value_t**)((char*)v + offs) != NULL;
+        jl_value_t *fval = *(jl_value_t**)((char*)v + offs);
+        if (fval == NULL)
+            jl_throw(jl_undefref_exception);
+        return fval;
     }
-    return 1;
+    return jl_new_bits(jl_tupleref(st->types,i), (char*)v + offs);
 }
 
-jl_value_t *jl_set_nth_field(jl_value_t *v, size_t i, jl_value_t *rhs)
+void jl_set_nth_field(jl_value_t *v, size_t i, jl_value_t *rhs)
 {
     jl_datatype_t *st = (jl_datatype_t*)jl_typeof(v);
     size_t offs = jl_field_offset(st,i) + sizeof(void*);
@@ -176,7 +220,16 @@ jl_value_t *jl_set_nth_field(jl_value_t *v, size_t i, jl_value_t *rhs)
     else {
         jl_assign_bits((char*)v + offs, rhs);
     }
-    return rhs;
+}
+
+int jl_field_isdefined(jl_value_t *v, size_t i)
+{
+    jl_datatype_t *st = (jl_datatype_t*)jl_typeof(v);
+    size_t offs = jl_field_offset(st,i) + sizeof(void*);
+    if (st->fields[i].isptr) {
+        return *(jl_value_t**)((char*)v + offs) != NULL;
+    }
+    return 1;
 }
 
 DLLEXPORT jl_value_t *jl_new_struct(jl_datatype_t *type, ...)
@@ -219,7 +272,7 @@ DLLEXPORT jl_value_t *jl_new_struct_uninit(jl_datatype_t *type)
     return jv;
 }
 
-jl_tuple_t *jl_tuple(size_t n, ...)
+DLLEXPORT jl_tuple_t *jl_tuple(size_t n, ...)
 {
     va_list args;
     if (n == 0) return jl_null;
@@ -345,6 +398,8 @@ jl_lambda_info_t *jl_new_lambda_info(jl_value_t *ast, jl_tuple_t *sparams)
     li->roots = NULL;
     li->functionObject = NULL;
     li->cFunctionObject = NULL;
+    li->functionID = 0;
+    li->cFunctionID = 0;
     li->specTypes = NULL;
     li->inferred = 0;
     li->inInference = 0;
@@ -361,19 +416,40 @@ jl_lambda_info_t *jl_new_lambda_info(jl_value_t *ast, jl_tuple_t *sparams)
 
 static jl_sym_t *symtab = NULL;
 
+static uptrint_t hash_symbol(const char *str, size_t len)
+{
+    return memhash(str, len) ^ ~(uptrint_t)0/3*2;
+}
+
+#define SYM_POOL_SIZE 524288
+
 static jl_sym_t *mk_symbol(const char *str)
 {
+#ifndef MEMDEBUG
+    static char *sym_pool = NULL;
+    static char *pool_ptr = NULL;
+#endif
     jl_sym_t *sym;
     size_t len = strlen(str);
+    size_t nb = (sizeof(jl_sym_t)+len+1+7)&-8;
 
-    sym = (jl_sym_t*)malloc((sizeof(jl_sym_t)+len+1+7)&-8);
+    if (nb >= SYM_POOL_SIZE) {
+        jl_error("Symbol length exceeds maximum length");
+    }
+
+#ifdef MEMDEBUG
+    sym = (jl_sym_t*)malloc(nb);
+#else
+    if (sym_pool == NULL || pool_ptr+nb > sym_pool+SYM_POOL_SIZE) {
+        sym_pool = malloc(SYM_POOL_SIZE);
+        pool_ptr = sym_pool;
+    }
+    sym = (jl_sym_t*)pool_ptr;
+    pool_ptr += nb;
+#endif
     sym->type = (jl_value_t*)jl_sym_type;
     sym->left = sym->right = NULL;
-#ifdef _P64
-    sym->hash = memhash(str, len)^0xAAAAAAAAAAAAAAAAL;
-#else
-    sym->hash = memhash32(str, len)^0xAAAAAAAA;
-#endif
+    sym->hash = hash_symbol(str, len);
     strcpy(&sym->name[0], str);
     return sym;
 }
@@ -392,11 +468,16 @@ void jl_unmark_symbols(void) { unmark_symbols_(symtab); }
 static jl_sym_t **symtab_lookup(jl_sym_t **ptree, const char *str)
 {
     int x;
+    uptrint_t h = hash_symbol(str, strlen(str));
 
-    while(*ptree != NULL) {
-        x = strcmp(str, (*ptree)->name);
-        if (x == 0)
-            return ptree;
+    // Tree nodes sorted by major key of (int(hash)) and minor key o (str).
+    while (*ptree != NULL) {
+        x = (int)(h-(*ptree)->hash);
+        if (x == 0) {
+            x = strcmp(str, (*ptree)->name);
+            if (x == 0)
+                return ptree;
+        }
         if (x < 0)
             ptree = &(*ptree)->left;
         else
@@ -422,7 +503,7 @@ jl_sym_t *jl_symbol_lookup(const char *str)
 
 DLLEXPORT jl_sym_t *jl_symbol_n(const char *str, int32_t len)
 {
-    char name[len+1];
+    char *name = (char*)alloca(len+1);
     memcpy(name, str, len);
     name[len] = '\0';
     return jl_symbol(name);
@@ -444,10 +525,10 @@ DLLEXPORT jl_sym_t *jl_gensym(void)
     return jl_symbol(n);
 }
 
-DLLEXPORT jl_sym_t *jl_tagged_gensym(const char* str, int32_t len)
+DLLEXPORT jl_sym_t *jl_tagged_gensym(const char *str, int32_t len)
 {
     static char gs_name[14];
-    char name[sizeof(gs_name)+len+3];
+    char *name = (char*)alloca(sizeof(gs_name)+len+3);
     char *n;
     name[0] = '#'; name[1] = '#'; name[2+len] = '#';
     memcpy(name+2, str, len);
@@ -587,17 +668,12 @@ jl_datatype_t *jl_new_datatype(jl_sym_t *name, jl_datatype_t *super,
         else if (!strcmp(((jl_sym_t*)name)->name, "Bool"))
             t = jl_bool_type;
     }
-    if (t == NULL) {
+    if (t == NULL)
         t = jl_new_uninitialized_datatype(jl_tuple_len(fnames));
-        if (jl_is_typename(name))
-            tn = (jl_typename_t*)name;
-        else
-            tn = jl_new_typename((jl_sym_t*)name);
-        t->name = tn;
-    }
+    else
+        tn = t->name;
 
-    if (t->name->primary == NULL)
-        t->name->primary = (jl_value_t*)t;
+    // init before possibly calling jl_new_typename
     t->super = super;
     t->parameters = parameters;
     t->names = fnames;
@@ -613,6 +689,19 @@ jl_datatype_t *jl_new_datatype(jl_sym_t *name, jl_datatype_t *super,
     t->struct_decl = NULL;
     t->size = 0;
     t->alignment = 0;
+
+    if (tn == NULL) {
+        t->name = NULL;
+        if (jl_is_typename(name))
+            tn = (jl_typename_t*)name;
+        else
+            tn = jl_new_typename((jl_sym_t*)name);
+        t->name = tn;
+    }
+
+    if (t->name->primary == NULL)
+        t->name->primary = (jl_value_t*)t;
+
     if (abstract || jl_tuple_len(parameters) > 0) {
         t->uid = 0;
     }
@@ -659,15 +748,15 @@ jl_typector_t *jl_new_type_ctor(jl_tuple_t *params, jl_value_t *body)
 
 // bits constructors ----------------------------------------------------------
 
-#define BOXN_FUNC(nb,nw)                                        \
-jl_value_t *jl_box##nb(jl_datatype_t *t, int##nb##_t x)         \
-{                                                               \
-    assert(jl_is_bitstype(t));                                  \
-    assert(jl_datatype_size(t) == sizeof(x));                   \
-    jl_value_t *v = alloc_##nw##w();                            \
-    v->type = (jl_value_t*)t;                                   \
-    *(int##nb##_t*)jl_data_ptr(v) = x;                          \
-    return v;                                                   \
+#define BOXN_FUNC(nb,nw)                                       \
+jl_value_t *jl_box##nb(jl_datatype_t *t, int##nb##_t x)        \
+{                                                              \
+    assert(jl_is_bitstype(t));                                 \
+    assert(jl_datatype_size(t) == sizeof(x));                  \
+    jl_value_t *v = (jl_value_t*)alloc_##nw##w();              \
+    v->type = (jl_value_t*)t;                                  \
+    *(int##nb##_t*)jl_data_ptr(v) = x;                         \
+    return v;                                                  \
 }
 BOXN_FUNC(8,  2)
 BOXN_FUNC(16, 2)
@@ -698,13 +787,13 @@ UNBOX_FUNC(float32, float)
 UNBOX_FUNC(float64, double)
 UNBOX_FUNC(voidpointer, void*)
 
-#define BOX_FUNC(typ,c_type,pfx,nw)             \
-jl_value_t *pfx##_##typ(c_type x)               \
-{                                               \
-    jl_value_t *v = alloc_##nw##w();            \
-    v->type = (jl_value_t*)jl_##typ##_type;     \
-    *(c_type*)jl_data_ptr(v) = x;               \
-    return v;                                   \
+#define BOX_FUNC(typ,c_type,pfx,nw)               \
+jl_value_t *pfx##_##typ(c_type x)                 \
+{                                                 \
+    jl_value_t *v = (jl_value_t*)alloc_##nw##w(); \
+    v->type = (jl_value_t*)jl_##typ##_type;       \
+    *(c_type*)jl_data_ptr(v) = x;                 \
+    return v;                                     \
 }
 BOX_FUNC(float32, float,  jl_box, 2)
 BOX_FUNC(voidpointer, void*,  jl_box, 2) //2 pointers == two words on all platforms
@@ -723,21 +812,21 @@ jl_value_t *jl_box_##typ(c_type x)                      \
     c_type idx = x+NBOX_C/2;                            \
     if ((u##c_type)idx < (u##c_type)NBOX_C)             \
         return boxed_##typ##_cache[idx];                \
-    jl_value_t *v = alloc_##nw##w();                    \
+    jl_value_t *v = (jl_value_t*)alloc_##nw##w();       \
     v->type = (jl_value_t*)jl_##typ##_type;             \
     *(c_type*)jl_data_ptr(v) = x;                       \
     return v;                                           \
 }
-#define UIBOX_FUNC(typ,c_type,nw)               \
-static jl_value_t *boxed_##typ##_cache[NBOX_C]; \
-jl_value_t *jl_box_##typ(c_type x)              \
-{                                               \
-    if (x < NBOX_C)                             \
-        return boxed_##typ##_cache[x];          \
-    jl_value_t *v = alloc_##nw##w();            \
-    v->type = (jl_value_t*)jl_##typ##_type;     \
-    *(c_type*)jl_data_ptr(v) = x;               \
-    return v;                                   \
+#define UIBOX_FUNC(typ,c_type,nw)                  \
+static jl_value_t *boxed_##typ##_cache[NBOX_C];    \
+jl_value_t *jl_box_##typ(c_type x)                 \
+{                                                  \
+    if (x < NBOX_C)                                \
+        return boxed_##typ##_cache[x];             \
+    jl_value_t *v = (jl_value_t*)alloc_##nw##w();  \
+    v->type = (jl_value_t*)jl_##typ##_type;        \
+    *(c_type*)jl_data_ptr(v) = x;                  \
+    return v;                                      \
 }
 SIBOX_FUNC(int16,  int16_t, 2)
 SIBOX_FUNC(int32,  int32_t, 2)
@@ -857,3 +946,30 @@ JL_CALLABLE(jl_f_new_box)
     ((jl_value_t**)box)[1] = args[0];
     return box;
 }
+
+JL_CALLABLE(jl_f_default_ctor_1)
+{
+    if (nargs != 1)
+        jl_error("wrong number of arguments (expected 1)");
+    jl_value_t *ft = jl_t0(((jl_datatype_t*)F)->types);
+    if (!jl_subtype(args[0], ft, 1))
+        jl_type_error(((jl_datatype_t*)F)->name->name->name, ft, args[0]);
+    return jl_new_struct((jl_datatype_t*)F, args[0]);
+}
+
+JL_CALLABLE(jl_f_default_ctor_2)
+{
+    if (nargs != 2)
+        jl_error("wrong number of arguments (expected 2)");
+    jl_value_t *ft = jl_t0(((jl_datatype_t*)F)->types);
+    if (!jl_subtype(args[0], ft, 1))
+        jl_type_error(((jl_datatype_t*)F)->name->name->name, ft, args[0]);
+    ft = jl_t1(((jl_datatype_t*)F)->types);
+    if (!jl_subtype(args[1], ft, 1))
+        jl_type_error(((jl_datatype_t*)F)->name->name->name, ft, args[1]);
+    return jl_new_struct((jl_datatype_t*)F, args[0], args[1]);
+}
+
+#ifdef __cplusplus
+}
+#endif
