@@ -8,6 +8,11 @@
 #include <malloc.h>
 #endif
 #include "julia.h"
+#include "julia_internal.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 // array constructors ---------------------------------------------------------
 
@@ -21,15 +26,14 @@ int jl_array_store_unboxed(jl_value_t *el_type)
     return store_unboxed(el_type);
 }
 
-// at this size use malloc
-#define MALLOC_THRESH 1048576
-
-#ifdef _P64
+#if defined(_P64) && defined(UINT128MAX)
 typedef __uint128_t wideint_t;
 #else
 typedef uint64_t wideint_t;
 #endif
 
+size_t jl_arr_xtralloc_limit = 0;
+
 #define MAXINTVAL (((size_t)-1)>>1)
 
 static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
@@ -73,7 +77,7 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
         size_t doffs = tsz;
         tsz += tot;
         tsz = (tsz+15)&-16; // align whole object 16
-        a = allocobj(tsz);
+        a = (jl_array_t*)allocobj(tsz);
         a->type = atype;
         a->how = 0;
         data = (char*)a + doffs;
@@ -83,7 +87,7 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
     }
     else {
         tsz = (tsz+15)&-16; // align whole object size 16
-        a = allocobj(tsz);
+        a = (jl_array_t*)allocobj(tsz);
         JL_GC_PUSH1(&a);
         a->type = atype;
         // temporarily initialize to make gc-safe
@@ -143,7 +147,7 @@ jl_array_t *jl_reshape_array(jl_value_t *atype, jl_array_t *data, jl_tuple_t *di
     size_t ndims = jl_tuple_len(dims);
 
     int ndimwords = jl_array_ndimwords(ndims);
-    a = allocobj((sizeof(jl_array_t) + sizeof(void*) + ndimwords*sizeof(size_t) + 15)&-16);
+    a = (jl_array_t*)allocobj((sizeof(jl_array_t) + sizeof(void*) + ndimwords*sizeof(size_t) + 15)&-16);
     a->type = atype;
     a->ndims = ndims;
     a->offset = 0;
@@ -208,7 +212,7 @@ jl_array_t *jl_ptr_to_array_1d(jl_value_t *atype, void *data, size_t nel,
     else
         elsz = sizeof(void*);
 
-    a = allocobj((sizeof(jl_array_t)+jl_array_ndimwords(1)*sizeof(size_t)+15)&-16);
+    a = (jl_array_t*)allocobj((sizeof(jl_array_t)+jl_array_ndimwords(1)*sizeof(size_t)+15)&-16);
     a->type = atype;
     a->data = data;
 #ifdef STORE_ARRAY_LEN
@@ -256,7 +260,7 @@ jl_array_t *jl_ptr_to_array(jl_value_t *atype, void *data, jl_tuple_t *dims,
         elsz = sizeof(void*);
 
     int ndimwords = jl_array_ndimwords(ndims);
-    a = allocobj((sizeof(jl_array_t) + ndimwords*sizeof(size_t)+15)&-16);
+    a = (jl_array_t*)allocobj((sizeof(jl_array_t) + ndimwords*sizeof(size_t)+15)&-16);
     a->type = atype;
     a->data = data;
 #ifdef STORE_ARRAY_LEN
@@ -292,7 +296,7 @@ jl_array_t *jl_ptr_to_array(jl_value_t *atype, void *data, jl_tuple_t *dims,
 jl_array_t *jl_new_array(jl_value_t *atype, jl_tuple_t *dims)
 {
     size_t ndims = jl_tuple_len(dims);
-    size_t *adims = alloca(ndims*sizeof(size_t));
+    size_t *adims = (size_t*)alloca(ndims*sizeof(size_t));
     size_t i;
     for(i=0; i < ndims; i++)
         adims[i] = jl_unbox_long(jl_tupleref(dims,i));
@@ -326,9 +330,9 @@ jl_array_t *jl_pchar_to_array(const char *str, size_t len)
 jl_value_t *jl_array_to_string(jl_array_t *a)
 {
     // TODO: check type of array?
-    jl_datatype_t* string_type = u8_isvalid(a->data, jl_array_len(a)) == 1 ? // ASCII
+    jl_datatype_t *string_type = u8_isvalid((char*)a->data, jl_array_len(a)) == 1 ? // ASCII
         jl_ascii_string_type : jl_utf8_string_type;
-    jl_value_t *s = alloc_2w();
+    jl_value_t *s = (jl_value_t*)alloc_2w();
     s->type = (jl_value_t*)string_type;
     jl_set_nth_field(s, 0, (jl_value_t*)a);
     return s;
@@ -353,6 +357,15 @@ jl_array_t *jl_alloc_cell_1d(size_t n)
     return jl_alloc_array_1d(jl_array_any_type, n);
 }
 
+jl_value_t *jl_apply_array_type(jl_datatype_t *type, size_t dim)
+{
+    jl_value_t *boxed_dim = jl_box_long(dim);
+    JL_GC_PUSH1(&boxed_dim);
+    jl_value_t *ret = jl_apply_type((jl_value_t*)jl_array_type, jl_tuple2(type, boxed_dim));
+    JL_GC_POP();
+    return ret;
+}
+
 // array primitives -----------------------------------------------------------
 
 #ifndef STORE_ARRAY_LEN
@@ -403,8 +416,7 @@ jl_value_t *jl_arrayref(jl_array_t *a, size_t i)
     jl_value_t *el_type = (jl_value_t*)jl_tparam0(jl_typeof(a));
     jl_value_t *elt;
     if (!a->ptrarray) {
-        elt = jl_new_bits((jl_datatype_t*)el_type,
-                          &((char*)a->data)[i*a->elsize]);
+        elt = jl_new_bits(el_type, &((char*)a->data)[i*a->elsize]);
     }
     else {
         elt = ((jl_value_t**)a->data)[i];
@@ -510,6 +522,9 @@ void jl_arrayunset(jl_array_t *a, size_t i)
         memset(ptail, 0, a->elsize);
 }
 
+// at this size and bigger, allocate resized array data with malloc
+#define MALLOC_THRESH 1048576
+
 // allocate buffer of newlen elements, placing old data at given offset (in #elts)
 static void array_resize_buffer(jl_array_t *a, size_t newlen, size_t oldlen, size_t offs)
 {
@@ -524,21 +539,27 @@ static void array_resize_buffer(jl_array_t *a, size_t newlen, size_t oldlen, siz
     char *newdata;
     if (a->how == 2) {
         // already malloc'd - use realloc
-        newdata = jl_gc_managed_realloc((char*)a->data - oldoffsnb, nbytes,
-                                        oldnbytes+oldoffsnb, a->isaligned);
+        newdata = (char*)jl_gc_managed_realloc((char*)a->data - oldoffsnb, nbytes,
+                                               oldnbytes+oldoffsnb, a->isaligned);
         if (offs != a->offset) {
             memmove(&newdata[offsnb], &newdata[oldoffsnb], oldnbytes);
         }
     }
     else {
-        if (nbytes >= MALLOC_THRESH) {
-            newdata = jl_gc_managed_malloc(nbytes);
+        if (
+#ifdef _P64
+            nbytes >= MALLOC_THRESH
+#else
+            es > 4
+#endif
+            ) {
+            newdata = (char*)jl_gc_managed_malloc(nbytes);
             jl_gc_track_malloced_array(a);
             a->how = 2;
             a->isaligned = 1;
         }
         else {
-            newdata = allocb(nbytes);
+            newdata = (char*)allocb(nbytes);
             a->how = 1;
         }
         memcpy(newdata + offsnb, (char*)a->data, oldnbytes);
@@ -561,6 +582,18 @@ static void array_try_unshare(jl_array_t *a)
     }
 }
 
+static size_t limit_overallocation(jl_array_t *a, size_t alen, size_t newlen, size_t inc)
+{
+    // Limit overallocation to jl_arr_xtralloc_limit
+    size_t es = a->elsize;
+    size_t xtra_elems_mem = (newlen - a->offset - alen - inc) * es;
+    if (xtra_elems_mem > jl_arr_xtralloc_limit) {
+        // prune down
+        return alen + inc + a->offset + (jl_arr_xtralloc_limit / es); 
+    }
+    return newlen;
+}
+
 void jl_array_grow_end(jl_array_t *a, size_t inc)
 {
     if (a->isshared && a->how!=3) jl_error("cannot resize array with shared data");
@@ -570,6 +603,8 @@ void jl_array_grow_end(jl_array_t *a, size_t inc)
         size_t newlen = a->maxsize==0 ? (inc<4?4:inc) : a->maxsize*2;
         while ((alen + inc) > newlen - a->offset)
             newlen *= 2;
+        
+        newlen = limit_overallocation(a, alen, newlen, inc);
         array_resize_buffer(a, newlen, alen, a->offset);
     }
 #ifdef STORE_ARRAY_LEN
@@ -628,6 +663,8 @@ void jl_array_grow_beg(jl_array_t *a, size_t inc)
             size_t newlen = a->maxsize==0 ? inc*2 : a->maxsize*2;
             while (alen+2*inc > newlen-a->offset)
                 newlen *= 2;
+            
+            newlen = limit_overallocation(a, alen, newlen, 2*inc);
             size_t center = (newlen - (alen + inc))/2;
             array_resize_buffer(a, newlen, alen, center+inc);
             char *newdata = (char*)a->data - (center+inc)*es;
@@ -702,3 +739,7 @@ DLLEXPORT void jl_cell_1d_push2(jl_array_t *a, jl_value_t *b, jl_value_t *c)
     jl_cellset(a, jl_array_dim(a,0)-2, b);
     jl_cellset(a, jl_array_dim(a,0)-1, c);
 }
+
+#ifdef __cplusplus
+}
+#endif
