@@ -10,23 +10,34 @@
 #include <malloc.h>
 #endif
 #include "julia.h"
+#include "julia_internal.h"
 #include "flisp.h"
 
-static char flisp_system_image[] = {
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+// MSVC complains about "julia_flisp.boot.inc : error C4335: Mac file format
+// detected: please convert the source file to either DOS or UNIX format"
+#ifdef _MSC_VER
+#pragma warning(disable:4335)
+#endif
+
+static uint8_t flisp_system_image[] = {
 #include "julia_flisp.boot.inc"
 };
 
 extern fltype_t *iostreamtype;
-static fltype_t *jvtype;
+static fltype_t *jvtype=NULL;
+
+static value_t true_sym;
+static value_t false_sym;
+static value_t fl_error_sym;
+static value_t fl_null_sym;
 
 static jl_value_t *scm_to_julia(value_t e, int expronly);
 static value_t julia_to_scm(jl_value_t *v);
 
-DLLEXPORT void jl_lisp_prompt(void)
-{
-    fl_applyn(1, symbol_value(symbol("__start")), fl_cons(FL_NIL,FL_NIL));
-}
-
 value_t fl_defined_julia_global(value_t *args, uint32_t nargs)
 {
     // tells whether a var is defined in and *by* the current module
@@ -57,18 +68,19 @@ value_t fl_invoke_julia_macro(value_t *args, uint32_t nargs)
     int i;
     for(i=0; i < nargs; i++) margs[i] = NULL;
     for(i=1; i < nargs; i++) margs[i] = scm_to_julia(args[i], 1);
-    jl_value_t *result=NULL;
+    jl_value_t *result = NULL;
 
     JL_TRY {
         margs[0] = scm_to_julia(args[0], 1);
         f = (jl_function_t*)jl_toplevel_eval(margs[0]);
+        assert(jl_is_func(f));
         result = jl_apply(f, &margs[1], nargs-1);
     }
     JL_CATCH {
         JL_GC_POP();
         value_t opaque = cvalue(jvtype, sizeof(void*));
         *(jl_value_t**)cv_data((cvalue_t*)ptr(opaque)) = jl_exception_in_transit;
-        return fl_list2(symbol("error"), opaque);
+        return fl_list2(fl_error_sym, opaque);
     }
     // protect result from GC, otherwise it could be freed during future
     // macro expansions, since it will be referenced only from scheme and
@@ -102,15 +114,12 @@ static builtinspec_t julia_flisp_ast_ext[] = {
     { NULL, NULL }
 };
 
-static value_t true_sym;
-static value_t false_sym;
-
 DLLEXPORT void jl_init_frontend(void)
 {
     fl_init(2*512*1024);
     value_t img = cvalue(iostreamtype, sizeof(ios_t));
     ios_t *pi = value2c(ios_t*, img);
-    ios_static_buffer(pi, flisp_system_image, sizeof(flisp_system_image));
+    ios_static_buffer(pi, (char*)flisp_system_image, sizeof(flisp_system_image));
     
     if (fl_load_system_image(img)) {
         JL_PRINTF(JL_STDERR, "fatal error loading system image\n");
@@ -125,6 +134,14 @@ DLLEXPORT void jl_init_frontend(void)
     assign_global_builtins(julia_flisp_ast_ext);
     true_sym = symbol("true");
     false_sym = symbol("false");
+    fl_error_sym = symbol("error");
+    fl_null_sym = symbol("null");
+}
+
+DLLEXPORT void jl_lisp_prompt(void)
+{
+    if (jvtype==NULL) jl_init_frontend();
+    fl_applyn(1, symbol_value(symbol("__start")), fl_cons(FL_NIL,FL_NIL));
 }
 
 static jl_sym_t *scmsym_to_julia(value_t s)
@@ -223,13 +240,20 @@ static jl_value_t *scm_to_julia_(value_t e, int eo)
             }
             i64 = conv_to_int64(cp_data(cp), nt);
         }
+        if (
 #ifdef _P64
-        return (jl_value_t*)jl_box_int64(i64);
+            jl_compileropts.int_literals==32
 #else
-        if (i64 > (int64_t)S32_MAX || i64 < (int64_t)S32_MIN)
-            return (jl_value_t*)jl_box_int64(i64);
-        return (jl_value_t*)jl_box_int32((int32_t)i64);
+            jl_compileropts.int_literals!=64
 #endif
+            ) {
+            if (i64 > (int64_t)S32_MAX || i64 < (int64_t)S32_MIN)
+                return (jl_value_t*)jl_box_int64(i64);
+            return (jl_value_t*)jl_box_int32((int32_t)i64);
+        }
+        else {
+            return (jl_value_t*)jl_box_int64(i64);
+        }
     }
     if (issymbol(e)) {
         if (e == true_sym)
@@ -239,7 +263,7 @@ static jl_value_t *scm_to_julia_(value_t e, int eo)
         return (jl_value_t*)scmsym_to_julia(e);
     }
     if (fl_isstring(e)) {
-        return jl_pchar_to_string(cvalue_data(e), cvalue_len(e));
+        return jl_pchar_to_string((char*)cvalue_data(e), cvalue_len(e));
     }
     if (e == FL_F) {
         return jl_false;
@@ -320,6 +344,9 @@ static jl_value_t *scm_to_julia_(value_t e, int eo)
                 }
             }
             jl_expr_t *ex = jl_exprn(sym, n);
+            // allocate a fresh args array for empty exprs passed to macros
+            if (eo && n == 0)
+                ex->args = jl_alloc_cell_1d(0);
             for(i=0; i < n; i++) {
                 assert(iscons(e));
                 jl_cellset(ex->args, i, scm_to_julia_(car_(e),eo));
@@ -344,32 +371,45 @@ static jl_value_t *scm_to_julia_(value_t e, int eo)
     return (jl_value_t*)jl_null;
 }
 
-static value_t array_to_list(jl_array_t *a)
+static value_t julia_to_scm_(jl_value_t *v);
+
+static value_t julia_to_scm(jl_value_t *v)
+{
+    value_t temp;
+    // need try/catch to reset GC handle stack in case of error
+    FL_TRY_EXTERN {
+        temp = julia_to_scm_(v);
+    }
+    FL_CATCH_EXTERN {
+        temp = fl_list2(fl_error_sym, cvalue_static_cstring("expression too large"));
+    }
+    return temp;
+}
+
+static void array_to_list(jl_array_t *a, value_t *pv)
 {
     if (jl_array_len(a) > 300000)
-        jl_error("expression too large");
-    value_t lst=FL_NIL, temp=FL_NIL;
-    fl_gc_handle(&lst);
-    fl_gc_handle(&temp);
+        lerror(MemoryError, "expression too large");
+    value_t temp;
     for(long i=jl_array_len(a)-1; i >= 0; i--) {
-        temp = julia_to_scm(jl_cellref(a,i));
-        lst = fl_cons(temp, lst);
+        *pv = fl_cons(FL_NIL, *pv);
+        temp = julia_to_scm_(jl_cellref(a,i));
+        // note: must be separate statement
+        car_(*pv) = temp;
     }
-    fl_free_gc_handles(2);
-    return lst;
 }
 
 static value_t julia_to_list2(jl_value_t *a, jl_value_t *b)
 {
-    value_t sa = julia_to_scm(a);
+    value_t sa = julia_to_scm_(a);
     fl_gc_handle(&sa);
-    value_t sb = julia_to_scm(b);
+    value_t sb = julia_to_scm_(b);
     value_t l = fl_list2(sa, sb);
     fl_free_gc_handles(1);
     return l;
 }
 
-static value_t julia_to_scm(jl_value_t *v)
+static value_t julia_to_scm_(jl_value_t *v)
 {
     if (jl_is_symbol(v)) {
         return symbol(((jl_sym_t*)v)->name);
@@ -381,13 +421,14 @@ static value_t julia_to_scm(jl_value_t *v)
         return FL_F;
     }
     if (v == jl_nothing) {
-        return fl_cons(symbol("null"), FL_NIL);
+        return fl_cons(fl_null_sym, FL_NIL);
     }
     if (jl_is_expr(v)) {
         jl_expr_t *ex = (jl_expr_t*)v;
-        value_t args = array_to_list(ex->args);
+        value_t args = FL_NIL;
         fl_gc_handle(&args);
-        value_t hd = julia_to_scm((jl_value_t*)ex->head);
+        array_to_list(ex->args, &args);
+        value_t hd = julia_to_scm_((jl_value_t*)ex->head);
         value_t scmv = fl_cons(hd, args);
         fl_free_gc_handles(1);
         return scmv;
@@ -461,26 +502,31 @@ int jl_start_parsing_file(const char *fname)
     return 0;
 }
 
-void jl_stop_parsing()
+void jl_stop_parsing(void)
 {
     fl_applyn(0, symbol_value(symbol("jl-parser-close-stream")));
 }
 
 extern int jl_lineno;
 
-jl_value_t *jl_parse_next()
+jl_value_t *jl_parse_next(void)
 {
     value_t c = fl_applyn(0, symbol_value(symbol("jl-parser-next")));
     if (c == FL_EOF)
         return NULL;
     if (iscons(c)) {
+        if (cdr_(c) == FL_EOF)
+            return NULL;
         value_t a = car_(c);
         if (isfixnum(a)) {
             jl_lineno = numval(a);
             //jl_printf(JL_STDERR, "  on line %d\n", jl_lineno);
-            return scm_to_julia(cdr_(c),0);
+            c = cdr_(c);
         }
     }
+    // for error, get most recent line number
+    if (iscons(c) && car_(c) == fl_error_sym)
+        jl_lineno = numval(fl_applyn(0, symbol_value(symbol("jl-parser-current-lineno"))));
     return scm_to_julia(c,0);
 }
 
@@ -553,15 +599,27 @@ jl_lambda_info_t *jl_wrap_expr(jl_value_t *expr)
 // get array of formal argument expressions
 jl_array_t *jl_lam_args(jl_expr_t *l)
 {
+    assert(jl_is_expr(l));
     assert(l->head == lambda_sym);
     jl_value_t *ae = jl_exprarg(l,0);
     assert(jl_is_array(ae));
     return (jl_array_t*)ae;
 }
 
+jl_sym_t *jl_lam_argname(jl_lambda_info_t *li, int i)
+{
+    jl_expr_t *ast;
+    if (jl_is_expr(li->ast))
+        ast = (jl_expr_t*)li->ast;
+    else
+        ast = (jl_expr_t*)jl_uncompress_ast(li, li->ast);
+    return (jl_sym_t*)jl_arrayref(jl_lam_args(ast),i);
+}
+
 // get array of local var symbols
 jl_array_t *jl_lam_locals(jl_expr_t *l)
 {
+    assert(jl_is_expr(l));
     jl_value_t *le = jl_exprarg(l, 1);
     assert(jl_is_array(le));
     jl_value_t *ll = jl_cellref(le, 0);
@@ -572,6 +630,7 @@ jl_array_t *jl_lam_locals(jl_expr_t *l)
 // get array of var info records
 jl_array_t *jl_lam_vinfo(jl_expr_t *l)
 {
+    assert(jl_is_expr(l));
     jl_value_t *le = jl_exprarg(l, 1);
     assert(jl_is_array(le));
     jl_value_t *ll = jl_cellref(le, 1);
@@ -683,6 +742,10 @@ static jl_value_t *copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
             jl_exprarg(ne, 1) = copy_ast(jl_exprarg(e,1), sp, 0);
             jl_exprarg(ne, 2) = copy_ast(jl_exprarg(e,2), sp, 1);
         }
+        else if (e->head == assign_sym) {
+            jl_exprarg(ne, 0) = copy_ast(jl_exprarg(e,0), sp, 0);
+            jl_exprarg(ne, 1) = copy_ast(jl_exprarg(e,1), sp, 1);
+        }
         else {
             for(size_t i=0; i < jl_array_len(e->args); i++)
                 jl_exprarg(ne, i) = copy_ast(jl_exprarg(e,i), sp, 1);
@@ -747,6 +810,10 @@ static jl_value_t *dont_copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
             jl_exprarg(e, 1) = dont_copy_ast(jl_exprarg(e,1), sp, 0);
             jl_exprarg(e, 2) = dont_copy_ast(jl_exprarg(e,2), sp, 1);
         }
+        else if (e->head == assign_sym) {
+            jl_exprarg(e, 0) = dont_copy_ast(jl_exprarg(e,0), sp, 0);
+            jl_exprarg(e, 1) = dont_copy_ast(jl_exprarg(e,1), sp, 1);
+        }
         else {
             for(size_t i=0; i < jl_array_len(e->args); i++)
                 jl_exprarg(e, i) = dont_copy_ast(jl_exprarg(e,i), sp, 1);
@@ -757,20 +824,19 @@ static jl_value_t *dont_copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
 }
 
 // TODO: eval decl types for arguments of non-generic functions
-static void eval_decl_types(jl_array_t *vi, jl_tuple_t *spenv)
+static void eval_decl_types(jl_array_t *vi, jl_value_t *ast, jl_tuple_t *spenv)
 {
-    size_t i;
-    for(i=0; i < jl_array_len(vi); i++) {
+    size_t i, l = jl_array_len(vi);
+    for(i=0; i < l; i++) {
         jl_array_t *v = (jl_array_t*)jl_cellref(vi, i);
         assert(jl_array_len(v) > 1);
-        JL_TRY {
-            jl_value_t *ty =
-                jl_interpret_toplevel_expr_with(jl_cellref(v,1),
-                                                &jl_tupleref(spenv,0),
-                                                jl_tuple_len(spenv)/2);
+        jl_value_t *ty = jl_static_eval(jl_cellref(v,1), NULL, jl_current_module,
+                                        (jl_value_t*)spenv, (jl_expr_t*)ast, 1, 1);
+        if (ty != NULL && (jl_is_type(ty) || jl_is_typevar(ty))) {
             jl_cellref(v, 1) = ty;
         }
-        JL_CATCH {
+        else {
+            jl_cellref(v, 1) = (jl_value_t*)jl_any_type;
         }
     }
 }
@@ -810,8 +876,8 @@ jl_value_t *jl_prepare_ast(jl_lambda_info_t *li, jl_tuple_t *sparams)
     jl_module_t *last_m = jl_current_module;
     JL_TRY {
         jl_current_module = li->module;
-        eval_decl_types(jl_lam_vinfo((jl_expr_t*)ast), spenv);
-        eval_decl_types(jl_lam_capt((jl_expr_t*)ast), spenv);
+        eval_decl_types(jl_lam_vinfo((jl_expr_t*)ast), ast, spenv);
+        eval_decl_types(jl_lam_capt((jl_expr_t*)ast), ast, spenv);
     }
     JL_CATCH {
         jl_current_module = last_m;
@@ -821,3 +887,12 @@ jl_value_t *jl_prepare_ast(jl_lambda_info_t *li, jl_tuple_t *sparams)
     JL_GC_POP();
     return ast;
 }
+
+DLLEXPORT int jl_is_operator(char *sym) {
+     return fl_applyn(1, symbol_value(symbol("operator?")), symbol(sym))
+             == FL_T;
+}
+
+#ifdef __cplusplus
+}
+#endif
