@@ -25,11 +25,16 @@
 #endif
 #include <ctype.h>
 #include "julia.h"
+#include "julia_internal.h"
 #include "builtin_proto.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 // exceptions -----------------------------------------------------------------
 
-void jl_error(const char *str)
+DLLEXPORT void jl_error(const char *str)
 {
     if (jl_errorexception_type == NULL) {
         JL_PRINTF(JL_STDERR, "%s", str);
@@ -42,16 +47,17 @@ void jl_error(const char *str)
 
 void jl_errorf(const char *fmt, ...)
 {
-    char buf[1024];
     va_list args;
+    ios_t buf;
+    ios_mem(&buf, 0);
     va_start(args, fmt);
-    int nc = vsnprintf(buf, sizeof(buf), fmt, args);
+    ios_vprintf(&buf, fmt, args);
     va_end(args);
     if (jl_errorexception_type == NULL) {
-        JL_PRINTF(JL_STDERR, "%s", &buf);
+        JL_PRINTF(JL_STDERR, "%s", buf.buf);
         jl_exit(1);
     }
-    jl_value_t *msg = jl_pchar_to_string(buf, nc);
+    jl_value_t *msg = jl_takebuf_string(&buf);
     JL_GC_PUSH1(&msg);
     jl_throw(jl_new_struct(jl_errorexception_type, msg));
 }
@@ -78,11 +84,28 @@ void jl_type_error_rt(const char *fname, const char *context,
     jl_throw(ex);
 }
 
+void jl_type_error_rt_line(const char *fname, const char *context,
+                           jl_value_t *ty, jl_value_t *got, int line)
+{
+    jl_type_error_rt(fname, context, ty, got);
+}
+
 void jl_type_error(const char *fname, jl_value_t *expected, jl_value_t *got)
 {
     jl_type_error_rt(fname, "", expected, got);
 }
 
+void jl_undefined_var_error(jl_sym_t *var)
+{
+    if (var->name[0] == '#') {
+        // convention for renamed variables: #...#original_name
+        char *nxt = strchr(var->name+1, '#');
+        if (nxt)
+            var = jl_symbol(nxt+1);
+    }
+    jl_throw(jl_new_struct(jl_undefvarerror_type, var));
+}
+
 JL_CALLABLE(jl_f_throw)
 {
     JL_NARGS(throw, 1, 1);
@@ -228,6 +251,7 @@ JL_CALLABLE(jl_f_typeassert)
 }
 
 static jl_function_t *jl_append_any_func;
+extern size_t jl_page_size;
 
 JL_CALLABLE(jl_f_apply)
 {
@@ -253,6 +277,10 @@ JL_CALLABLE(jl_f_apply)
                             jl_tuple_len(args[1]));
         }
     }
+    jl_value_t *argarr = NULL;
+    JL_GC_PUSH1(&argarr);
+    jl_value_t *result;
+    jl_value_t **newargs;
     size_t n=0, i, j;
     for(i=1; i < nargs; i++) {
         if (jl_is_tuple(args[i])) {
@@ -271,15 +299,27 @@ JL_CALLABLE(jl_f_apply)
                     JL_TYPECHK(apply, tuple, args[i]);
                 }
             }
-            goto fancy_apply;
+            argarr = jl_apply(jl_append_any_func, &args[1], nargs-1);
+            assert(jl_typeis(argarr, jl_array_any_type));
+            result = jl_apply((jl_function_t*)args[0], jl_cell_data(argarr), jl_array_len(argarr));
+            JL_GC_POP();
+            return result;
         }
     }
-    jl_value_t **newargs = alloca(n * sizeof(jl_value_t*));
+    if (n > jl_page_size/sizeof(jl_value_t*)) {
+        // put arguments on the heap if there are too many
+        argarr = (jl_value_t*)jl_alloc_cell_1d(n);
+        newargs = jl_cell_data(argarr);
+    }
+    else {
+        newargs = (jl_value_t**)alloca(n * sizeof(jl_value_t*));
+    }
     n = 0;
     for(i=1; i < nargs; i++) {
         if (jl_is_tuple(args[i])) {
             jl_tuple_t *t = (jl_tuple_t*)args[i];
-            for(j=0; j < jl_tuple_len(t); j++)
+            size_t al = jl_tuple_len(t);
+            for(j=0; j < al; j++)
                 newargs[n++] = jl_tupleref(t, j);
         }
         else {
@@ -288,20 +328,11 @@ JL_CALLABLE(jl_f_apply)
                 newargs[n++] = jl_cellref(args[i], j);
         }
     }
-    return jl_apply((jl_function_t*)args[0], newargs, n);
-
- fancy_apply: ;
-    jl_value_t *argarr = jl_apply(jl_append_any_func, &args[1], nargs-1);
-    JL_GC_PUSH1(&argarr);
-    assert(jl_typeis(argarr, jl_array_any_type));
-    jl_value_t *result = jl_apply((jl_function_t*)args[0],
-                                  jl_cell_data(argarr), jl_array_len(argarr));
+    result = jl_apply((jl_function_t*)args[0], newargs, n);
     JL_GC_POP();
     return result;
 }
 
-#include "newobj_internal.h"
-
 void jl_add_constructors(jl_datatype_t *t);
 
 JL_CALLABLE(jl_f_kwcall)
@@ -329,7 +360,13 @@ JL_CALLABLE(jl_f_kwcall)
         jl_cellset(container, i+1, args[2+i+1]);
     }
 
-    return jl_apply(sorter, &args[pa-1], nargs-(pa-1));
+    assert(jl_is_gf(sorter));
+    jl_function_t *m = jl_method_lookup((jl_methtable_t*)sorter->env, &args[pa-1], nargs-(pa-1), 1);
+    if (m == jl_bottom_func) {
+        return jl_no_method_error(f, &args[pa], nargs-pa);
+    }
+
+    return jl_apply(m, &args[pa-1], nargs-(pa-1));
 }
 
 // eval -----------------------------------------------------------------------
@@ -355,23 +392,21 @@ JL_CALLABLE(jl_f_top_eval)
     }
     jl_value_t *v=NULL;
     int last_lineno = jl_lineno;
-    if (m == jl_current_module) {
-        v = jl_toplevel_eval(ex);
-        jl_lineno = last_lineno;
-        return v;
-    }
     jl_module_t *last_m = jl_current_module;
+    jl_module_t *task_last_m = jl_current_task->current_module;
     JL_TRY {
-        jl_current_module = m;
+        jl_current_task->current_module = jl_current_module = m;
         v = jl_toplevel_eval(ex);
     }
     JL_CATCH {
         jl_lineno = last_lineno;
         jl_current_module = last_m;
+        jl_current_task->current_module = task_last_m;
         jl_rethrow();
     }
     jl_lineno = last_lineno;
     jl_current_module = last_m;
+    jl_current_task->current_module = task_last_m;
     assert(v);
     return v;
 }
@@ -392,17 +427,29 @@ JL_CALLABLE(jl_f_isdefined)
         JL_NARGS(isdefined, 1, 1);
     }
     else {
-        JL_TYPECHK(isdefined, symbol, args[1]);
-        s = (jl_sym_t*)args[1];
         if (!jl_is_module(args[0])) {
-            jl_value_t *vt = (jl_value_t*)jl_typeof(args[0]);
+            jl_datatype_t *vt = (jl_datatype_t*)jl_typeof(args[0]);
             if (!jl_is_datatype(vt)) {
                 jl_type_error("isdefined", (jl_value_t*)jl_datatype_type, args[0]);
             }
-            return jl_field_isdefined(args[0], s, 0) ? jl_true : jl_false;
+            size_t idx;
+            if (jl_is_long(args[1])) {
+                idx = jl_unbox_long(args[1])-1;
+                if (idx >= jl_tuple_len(vt->names))
+                    return jl_false;
+            }
+            else {
+                JL_TYPECHK(isdefined, symbol, args[1]);
+                idx = jl_field_index(vt, (jl_sym_t*)args[1], 0);
+                if ((int)idx == -1)
+                    return jl_false;
+            }
+            return jl_field_isdefined(args[0], idx) ? jl_true : jl_false;
         }
         JL_TYPECHK(isdefined, module, args[0]);
+        JL_TYPECHK(isdefined, symbol, args[1]);
         m = (jl_module_t*)args[0];
+        s = (jl_sym_t*)args[1];
     }
     assert(s);
     return jl_boundp(m, s) ? jl_true : jl_false;
@@ -445,18 +492,27 @@ JL_CALLABLE(jl_f_tuplelen)
 JL_CALLABLE(jl_f_get_field)
 {
     JL_NARGS(getfield, 2, 2);
-    JL_TYPECHK(getfield, symbol, args[1]);
     jl_value_t *v = args[0];
     jl_value_t *vt = (jl_value_t*)jl_typeof(v);
     if (vt == (jl_value_t*)jl_module_type) {
+        JL_TYPECHK(getfield, symbol, args[1]);
         return jl_eval_global_var((jl_module_t*)v, (jl_sym_t*)args[1]);
     }
-    if (!jl_is_datatype(vt)) {
+    if (!jl_is_datatype(vt))
         jl_type_error("getfield", (jl_value_t*)jl_datatype_type, v);
-    }
     jl_datatype_t *st = (jl_datatype_t*)vt;
-    jl_sym_t *fld = (jl_sym_t*)args[1];
-    jl_value_t *fval = jl_get_nth_field(v, jl_field_index(st, fld, 1));
+    size_t idx;
+    if (jl_is_long(args[1])) {
+        idx = jl_unbox_long(args[1])-1;
+        if (idx >= jl_tuple_len(st->names))
+            jl_throw(jl_bounds_exception);
+    }
+    else {
+        JL_TYPECHK(getfield, symbol, args[1]);
+        jl_sym_t *fld = (jl_sym_t*)args[1];
+        idx = jl_field_index(st, fld, 1);
+    }
+    jl_value_t *fval = jl_get_nth_field(v, idx);
     if (fval == NULL)
         jl_throw(jl_undefref_exception);
     return fval;
@@ -464,31 +520,37 @@ JL_CALLABLE(jl_f_get_field)
 
 JL_CALLABLE(jl_f_set_field)
 {
-    JL_NARGS(setfield, 3, 3);
-    JL_TYPECHK(setfield, symbol, args[1]);
+    JL_NARGS(setfield!, 3, 3);
     jl_value_t *v = args[0];
     jl_value_t *vt = (jl_value_t*)jl_typeof(v);
     if (vt == (jl_value_t*)jl_module_type)
         jl_error("cannot assign variables in other modules");
     if (!jl_is_datatype(vt))
-        jl_type_error("setfield", (jl_value_t*)jl_datatype_type, v);
+        jl_type_error("setfield!", (jl_value_t*)jl_datatype_type, v);
     jl_datatype_t *st = (jl_datatype_t*)vt;
     if (!st->mutabl)
         jl_errorf("type %s is immutable", st->name->name->name);
-    jl_sym_t *fld = (jl_sym_t*)args[1];
-    size_t i = jl_field_index(st, fld, 1);
-    jl_value_t *ft = jl_tupleref(st->types,i);
-    if (!jl_subtype(args[2], ft, 1)) {
-        jl_type_error("setfield", ft, args[2]);
+    size_t idx;
+    if (jl_is_long(args[1])) {
+        idx = jl_unbox_long(args[1])-1;
+        if (idx >= jl_tuple_len(st->names))
+            jl_throw(jl_bounds_exception);
     }
-    jl_set_nth_field(v, i, args[2]);
+    else {
+        JL_TYPECHK(setfield!, symbol, args[1]);
+        idx = jl_field_index(st, (jl_sym_t*)args[1], 1);
+    }
+    jl_value_t *ft = jl_tupleref(st->types, idx);
+    if (!jl_subtype(args[2], ft, 1)) {
+        jl_type_error("setfield!", ft, args[2]);
+    }
+    jl_set_nth_field(v, idx, args[2]);
     return args[2];
 }
 
 JL_CALLABLE(jl_f_field_type)
 {
     JL_NARGS(fieldtype, 2, 2);
-    JL_TYPECHK(fieldtype, symbol, args[1]);
     jl_value_t *v = args[0];
     jl_value_t *vt = (jl_value_t*)jl_typeof(v);
     if (vt == (jl_value_t*)jl_module_type)
@@ -496,57 +558,21 @@ JL_CALLABLE(jl_f_field_type)
     if (!jl_is_datatype(vt))
         jl_type_error("fieldtype", (jl_value_t*)jl_datatype_type, v);
     jl_datatype_t *st = (jl_datatype_t*)vt;
-    jl_sym_t *fld = (jl_sym_t*)args[1];
-    return jl_tupleref(st->types, jl_field_index(st, fld, 1));
+    int field_index;
+    if (jl_is_long(args[1])) {
+        field_index = jl_unbox_long(args[1]) - 1;
+        if (field_index < 0 || field_index >= jl_tuple_len(st->names))
+            jl_throw(jl_bounds_exception);
+    }
+    else {
+        JL_TYPECHK(fieldtype, symbol, args[1]);
+        field_index = jl_field_index(st, (jl_sym_t*)args[1], 1);
+    }
+    return jl_tupleref(st->types, field_index);
 }
 
 // conversion -----------------------------------------------------------------
 
-static jl_value_t *convert(jl_value_t *to, jl_value_t *x, jl_function_t *conv_f)
-{
-    jl_value_t *args[2];
-    if (jl_subtype(x, (jl_value_t*)to, 1))
-        return x;
-    args[0] = (jl_value_t*)to; args[1] = x;
-    return jl_apply(conv_f, args, 2);
-}
-
-JL_CALLABLE(jl_f_convert_tuple)
-{
-    jl_tuple_t *to = (jl_tuple_t*)args[0];
-    jl_tuple_t *x = (jl_tuple_t*)args[1];
-    if (to == jl_tuple_type)
-        return (jl_value_t*)x;
-    size_t i, cl=jl_tuple_len(x), pl=jl_tuple_len(to);
-    jl_tuple_t *out = jl_alloc_tuple(cl);
-    JL_GC_PUSH1(&out);
-    jl_value_t *ce, *pe=NULL;
-    int pseq=0;
-    jl_function_t *f = (jl_function_t*)args[2];
-    for(i=0; i < cl; i++) {
-        ce = jl_tupleref(x,i);
-        if (pseq) {
-        }
-        else if (i < pl) {
-            pe = jl_tupleref(to,i);
-            if (jl_is_vararg_type(pe)) {
-                pe = jl_tparam0(pe);
-                pseq = 1;
-            }
-        }
-        else {
-            out = NULL;
-            break;
-        }
-        assert(pe != NULL);
-        jl_tupleset(out, i, convert((jl_value_t*)pe, ce, f));
-    }
-    JL_GC_POP();
-    if (out == NULL)
-        jl_error("convert: invalid tuple conversion");
-    return (jl_value_t*)out;
-}
-
 JL_CALLABLE(jl_f_convert_default)
 {
     jl_value_t *to = args[0];
@@ -590,44 +616,73 @@ DLLEXPORT int jl_substrtod(char *str, size_t offset, int len, double *out)
     char *p;
     errno = 0;
     char *bstr = str+offset;
-    *out = strtod(bstr, &p);
-    if ((p == bstr) || (p != (bstr+len)) ||
+    char *pend = bstr+len;
+    int err = 0;
+    if (!(*pend == '\0' || isspace((unsigned char)*pend) || *pend == ',')) {
+        // confusing data outside substring. must copy.
+        char *newstr = malloc(len+1);
+        memcpy(newstr, bstr, len);
+        newstr[len] = 0;
+        bstr = newstr;
+        pend = bstr+len;
+    }
+    *out = strtod_c(bstr, &p);
+    if ((p == bstr) || (p != pend) ||
         (errno==ERANGE && (*out==0 || *out==HUGE_VAL || *out==-HUGE_VAL)))
-        return 1;
-    return 0;
+        err = 1;
+    if (bstr != str+offset)
+        free(bstr);
+    return err;
 }
 
 DLLEXPORT int jl_strtod(char *str, double *out)
 {
     char *p;
     errno = 0;
-    *out = strtod(str, &p);
+    *out = strtod_c(str, &p);
     if (p == str ||
         (errno==ERANGE && (*out==0 || *out==HUGE_VAL || *out==-HUGE_VAL)))
         return 1;
     while (*p != '\0') {
-        if (!isspace(*p))
+        if (!isspace((unsigned char)*p))
             return 1;
         p++;
     }
     return 0;
 }
 
+// MSVC pre-2013 did not define HUGE_VALF
+#ifndef HUGE_VALF
+#define HUGE_VALF (1e25f * 1e25f)
+#endif
+
 DLLEXPORT int jl_substrtof(char *str, int offset, int len, float *out)
 {
     char *p;
     errno = 0;
     char *bstr = str+offset;
+    char *pend = bstr+len;
+    int err = 0;
+    if (!(*pend == '\0' || isspace((unsigned char)*pend) || *pend == ',')) {
+        // confusing data outside substring. must copy.
+        char *newstr = malloc(len+1);
+        memcpy(newstr, bstr, len);
+        newstr[len] = 0;
+        bstr = newstr;
+        pend = bstr+len;
+    }
 #if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
-    *out = (float)strtod(bstr, &p);
+    *out = (float)strtod_c(bstr, &p);
 #else
-    *out = strtof(bstr, &p);
+    *out = strtof_c(bstr, &p);
 #endif
 
-    if ((p == bstr) || (p != (bstr+len)) ||
+    if ((p == bstr) || (p != pend) ||
         (errno==ERANGE && (*out==0 || *out==HUGE_VALF || *out==-HUGE_VALF)))
-        return 1;
-    return 0;
+        err = 1;
+    if (bstr != str+offset)
+        free(bstr);
+    return err;
 }
 
 DLLEXPORT int jl_strtof(char *str, float *out)
@@ -635,15 +690,15 @@ DLLEXPORT int jl_strtof(char *str, float *out)
     char *p;
     errno = 0;
 #if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
-    *out = (float)strtod(str, &p);
+    *out = (float)strtod_c(str, &p);
 #else
-    *out = strtof(str, &p);
+    *out = strtof_c(str, &p);
 #endif
     if (p == str ||
         (errno==ERANGE && (*out==0 || *out==HUGE_VALF || *out==-HUGE_VALF)))
         return 1;
     while (*p != '\0') {
-        if (!isspace(*p))
+        if (!isspace((unsigned char)*p))
             return 1;
         p++;
     }
@@ -652,20 +707,20 @@ DLLEXPORT int jl_strtof(char *str, float *out)
 
 // showing --------------------------------------------------------------------
 
-void jl_flush_cstdio()
+void jl_flush_cstdio(void)
 {
     fflush(stdout);
     fflush(stderr);
 }
 
-jl_value_t *jl_stdout_obj()
+jl_value_t *jl_stdout_obj(void)
 {
     jl_value_t *stdout_obj = jl_get_global(jl_base_module, jl_symbol("STDOUT"));
     if (stdout_obj != NULL) return stdout_obj;
     return jl_get_global(jl_base_module, jl_symbol("OUTPUT_STREAM"));
 }
 
-jl_value_t *jl_stderr_obj()
+jl_value_t *jl_stderr_obj(void)
 {
     jl_value_t *stderr_obj = jl_get_global(jl_base_module, jl_symbol("STDERR"));
     if (stderr_obj != NULL) return stderr_obj;
@@ -691,108 +746,6 @@ void jl_show(jl_value_t *stream, jl_value_t *v)
     }
 }
 
-// comma_one prints a comma for 1 element, e.g. "(x,)"
-void jl_show_tuple(jl_value_t *st, jl_tuple_t *t, char opn, char cls, int comma_one)
-{
-    JL_STREAM *s = ((JL_STREAM**)st)[1];
-    JL_PUTC(opn, s);
-    size_t i, n=jl_tuple_len(t);
-    for(i=0; i < n; i++) {
-        jl_show(st, jl_tupleref(t, i));
-        if ((i < n-1) || (n==1 && comma_one))
-            JL_PUTC(',', s);
-    }
-    JL_PUTC(cls, s);
-}
-
-static void show_function(JL_STREAM *s, jl_value_t *v)
-{
-    if (jl_is_gf(v)) {
-        JL_PUTS(jl_gf_name(v)->name, s);
-    }
-    else {
-        JL_PUTS("# function", s);
-    }
-}
-
-static void show_type(jl_value_t *st, jl_value_t *t)
-{
-    uv_stream_t *s =((uv_stream_t**)st)[1];
-    if (jl_is_uniontype(t)) {
-        if (t == (jl_value_t*)jl_bottom_type) {
-            JL_WRITE(s, "None", 4);
-        }
-        else if (t == jl_top_type) {
-            JL_WRITE(s, "Top", 3);
-        }
-        else {
-            JL_WRITE(s, "Union", 5);
-            jl_show_tuple(st, ((jl_uniontype_t*)t)->types, '(', ')', 0);
-        }
-    }
-    else if (jl_is_vararg_type(t)) {
-        jl_show(st, jl_tparam0(t));
-        JL_WRITE(s, "...", 3);
-    }
-    else if (jl_is_typector(t)) {
-        jl_show(st, (jl_value_t*)((jl_typector_t*)t)->body);
-    }
-    else {
-        assert(jl_is_datatype(t));
-        jl_datatype_t *tt = (jl_datatype_t*)t;
-        JL_PUTS(tt->name->name->name, s);
-        jl_tuple_t *p = tt->parameters;
-        if (jl_tuple_len(p) > 0)
-            jl_show_tuple(st, p, '{', '}', 0);
-    }
-}
-
-DLLEXPORT void jl_show_any(jl_value_t *str, jl_value_t *v)
-{
-    uv_stream_t *s = ((uv_stream_t**)str)[1];
-    // fallback for printing some other builtin types
-    if (jl_is_tuple(v)) {
-        jl_show_tuple(str, (jl_tuple_t*)v, '(', ')', 1);
-    }
-    else if (jl_is_type(v)) {
-        show_type(str, v);
-    }
-    else if (jl_is_func(v)) {
-        show_function(s, v);
-    }
-    else if (jl_typeis(v,jl_intrinsic_type)) {
-        JL_PRINTF(s, "# intrinsic function %d", *(uint32_t*)jl_data_ptr(v));
-    }
-    else {
-        jl_value_t *t = (jl_value_t*)jl_typeof(v);
-        assert(jl_is_datatype(t));
-        jl_datatype_t *dt = (jl_datatype_t*)t;
-        show_type(str, t);
-        JL_PUTC('(', s);
-        if (jl_tuple_len(dt->names)>0 || dt->size==0) {
-            size_t i;
-            size_t n = jl_tuple_len(dt->names);
-            for(i=0; i < n; i++) {
-                jl_value_t *fval = jl_get_nth_field(v, i);
-                if (fval == NULL)
-                    JL_PUTS("#undef", s);
-                else
-                    jl_show(str, fval);
-                if (i < n-1)
-                    JL_PUTC(',', s);
-            }
-        }
-        else {
-            size_t nb = jl_datatype_size(dt);
-            char *data = (char*)jl_data_ptr(v);
-            JL_PUTS("0x", s);
-            for(int i=nb-1; i >= 0; --i)
-                jl_printf(s, "%02hhx", data[i]);
-        }
-        JL_PUTC(')', s);
-    }
-}
-
 // internal functions ---------------------------------------------------------
 
 extern int jl_in_inference;
@@ -818,7 +771,9 @@ JL_CALLABLE(jl_trampoline)
         }
     }
     jl_compile(f);
-    assert(f->fptr == &jl_trampoline);
+    // this assertion is probably not correct; the fptr could have been assigned
+    // by a recursive invocation from inference above.
+    //assert(f->fptr == &jl_trampoline);
     jl_generate_fptr(f);
     if (jl_boot_file_loaded && jl_is_expr(f->linfo->ast)) {
         f->linfo->ast = jl_compress_ast(f->linfo, f->linfo->ast);
@@ -851,9 +806,9 @@ JL_CALLABLE(jl_f_new_type_constructor)
 JL_CALLABLE(jl_f_typevar)
 {
     if (nargs < 1 || nargs > 3) {
-        JL_NARGS(typevar, 1, 1);
+        JL_NARGS(TypeVar, 1, 1);
     }
-    JL_TYPECHK(typevar, symbol, args[0]);
+    JL_TYPECHK(TypeVar, symbol, args[0]);
     jl_value_t *lb = (jl_value_t*)jl_bottom_type;
     jl_value_t *ub = (jl_value_t*)jl_any_type;
     int b = 0;
@@ -862,9 +817,9 @@ JL_CALLABLE(jl_f_typevar)
         nargs--;
     }
     if (nargs > 1) {
-        JL_TYPECHK(typevar, type, args[1]);
+        JL_TYPECHK(TypeVar, type, args[1]);
         if (nargs > 2) {
-            JL_TYPECHK(typevar, type, args[2]);
+            JL_TYPECHK(TypeVar, type, args[2]);
             lb = args[1];
             ub = args[2];
         }
@@ -933,7 +888,7 @@ JL_CALLABLE(jl_f_invoke)
     JL_TYPECHK(invoke, tuple, args[1]);
     jl_check_type_tuple((jl_tuple_t*)args[1], jl_gf_name(args[0]), "invoke");
     if (!jl_tuple_subtype(&args[2], nargs-2, &jl_tupleref(args[1],0),
-                          jl_tuple_len(args[1]), 1, 0))
+                          jl_tuple_len(args[1]), 1))
         jl_error("invoke: argument type error");
     return jl_gf_invoke((jl_function_t*)args[0],
                         (jl_tuple_t*)args[1], &args[2], nargs-2);
@@ -1048,11 +1003,10 @@ void jl_init_primitives(void)
     
     // functions for internal use
     add_builtin_func("convert_default", jl_f_convert_default);
-    add_builtin_func("convert_tuple", jl_f_convert_tuple);
     add_builtin_func("tupleref",  jl_f_tupleref);
     add_builtin_func("tuplelen",  jl_f_tuplelen);
     add_builtin_func("getfield",  jl_f_get_field);
-    add_builtin_func("setfield",  jl_f_set_field);
+    add_builtin_func("setfield!",  jl_f_set_field);
     add_builtin_func("fieldtype", jl_f_field_type);
 
     add_builtin_func("arraylen", jl_f_arraylen);
@@ -1090,6 +1044,7 @@ void jl_init_primitives(void)
     add_builtin("Task", (jl_value_t*)jl_task_type);
 
     add_builtin("AbstractArray", (jl_value_t*)jl_abstractarray_type);
+    add_builtin("DenseArray", (jl_value_t*)jl_densearray_type);
     add_builtin("Array", (jl_value_t*)jl_array_type);
 
     add_builtin("Expr", (jl_value_t*)jl_expr_type);
@@ -1109,14 +1064,31 @@ void jl_init_primitives(void)
     add_builtin("ANY", jl_ANY_flag);
 }
 
-// toys for debugging -----------------------------------------------------------------------
+// toys for debugging ---------------------------------------------------------
+
+// comma_one prints a comma for 1 element, e.g. "(x,)"
+static size_t jl_show_tuple(JL_STREAM *out, jl_tuple_t *t, char *opn, char *cls, int comma_one)
+{
+    size_t i, n=0, len = jl_tuple_len(t);
+    n += JL_PRINTF(out, "(");
+    for (i = 0; i < len; i++) {
+        jl_value_t *v = jl_tupleref(t,i);
+        n += jl_static_show(out, v);
+        if (len == 1)
+            n += JL_PRINTF(out, ",");
+        else if (i != len-1)
+            n += JL_PRINTF(out, ", ");
+    }
+    n += JL_PRINTF(out, ")");
+    return n;
+}
 
 DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
 {
     // mimic jl_show, but never calling a julia method
     size_t n = 0;
     if (v == NULL) {
-        n += JL_PRINTF(out, "<null>");
+        n += JL_PRINTF(out, "#<null>");
     }
     else if (jl_is_lambda_info(v)) {
         jl_lambda_info_t *li = (jl_lambda_info_t*)v;
@@ -1130,18 +1102,7 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
         }
     }
     else if (jl_is_tuple(v)) {
-        jl_tuple_t *t = (jl_tuple_t*)v;
-        size_t i, len = jl_tuple_len(t);
-        n += JL_PRINTF(out, "(");
-        for (i = 0; i < len; i++) {
-            jl_value_t *v = jl_tupleref(t,i);
-            n += jl_static_show(out, v);
-            if (len == 1)
-                n += JL_PRINTF(out, ",");
-            else if (i != len-1)
-                n += JL_PRINTF(out, ", ");
-        }
-        n += JL_PRINTF(out, ")");
+        n += jl_show_tuple(out, (jl_tuple_t*)v, "(", ")", 1);
     }
     else if (jl_is_vararg_type(v)) {
         n += jl_static_show(out, jl_tparam0(v));
@@ -1173,11 +1134,11 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
             n += JL_PRINTF(out, "%s", jl_gf_name(v)->name);
         }
         else {
-            n += JL_PRINTF(out, "<# function>");
+            n += JL_PRINTF(out, "#<function>");
         }
     }
     else if (jl_typeis(v, jl_intrinsic_type)) {
-        n += JL_PRINTF(out, "<# intrinsic function %d>", *(uint32_t*)jl_data_ptr(v));
+        n += JL_PRINTF(out, "#<intrinsic function %d>", *(uint32_t*)jl_data_ptr(v));
     }
     else if (jl_is_int64(v)) {
         n += JL_PRINTF(out, "%d", jl_unbox_int64(v));
@@ -1236,7 +1197,9 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
         n += jl_static_show(out, ((jl_typector_t*)v)->body);
     }
     else if (jl_is_typevar(v)) {
-        n += JL_PRINTF(out, "%s", ((jl_tvar_t*)v)->name->name);
+        n += jl_static_show(out, ((jl_tvar_t*)v)->lb);
+        n += JL_PRINTF(out, "<:%s<:", ((jl_tvar_t*)v)->name->name);
+        n += jl_static_show(out, ((jl_tvar_t*)v)->ub);
     }
     else if (jl_is_module(v)) {
         jl_module_t *m = (jl_module_t*)v;
@@ -1266,14 +1229,10 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
         n += JL_PRINTF(out, "goto %d", jl_gotonode_label(v));
     }
     else if (jl_is_quotenode(v)) {
-        n += JL_PRINTF(out, "quote ");
-        n += jl_static_show(out, jl_fieldref(v,0));
-        n += JL_PRINTF(out, " end");
-    }
-    else if (jl_is_newvarnode(v)) {
-        n += JL_PRINTF(out, "<newvar ");
-        n += jl_static_show(out, jl_fieldref(v,0));
-        n += JL_PRINTF(out, ">");
+        jl_value_t *qv = jl_fieldref(v,0);
+        if (!jl_is_symbol(qv)) { n += JL_PRINTF(out, "quote "); }
+        n += jl_static_show(out, qv);
+        if (!jl_is_symbol(qv)) { n += JL_PRINTF(out, " end"); }
     }
     else if (jl_is_topnode(v)) {
         n += JL_PRINTF(out, "top(");
@@ -1309,9 +1268,14 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
         n += JL_PRINTF(out, "[");
         size_t j, tlen = jl_array_len(v);
         for (j = 0; j < tlen; j++) {
-            n += jl_static_show(out, jl_arrayref((jl_array_t*)v,j));
+            jl_value_t *elt;
+            if (((jl_array_t*)v)->ptrarray)
+                elt = jl_cellref(v, j);
+            else
+                elt = jl_arrayref((jl_array_t*)v,j);
+            n += jl_static_show(out, elt);
             if (j != tlen-1)
-               n += JL_PRINTF(out, ", ");
+                n += JL_PRINTF(out, ", ");
         }
         n += JL_PRINTF(out, "]");
     }
@@ -1333,20 +1297,27 @@ DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
         jl_datatype_t *t = (jl_datatype_t*)jl_typeof(v);
         n += jl_static_show(out, (jl_value_t*)t);
         n += JL_PRINTF(out, "(");
-        size_t i, tlen = jl_tuple_len(t->names);
-        for (i = 0; i < tlen; i++) {
-            n += JL_PRINTF(out, ((jl_sym_t*)jl_tupleref(t->names, i))->name);
-            jl_fielddesc_t f = t->fields[i];
-            if (f.isptr) {
+        size_t nb = jl_datatype_size(t);
+        size_t tlen = jl_tuple_len(t->names);
+        if (nb > 0 && tlen == 0) {
+            char *data = (char*)jl_data_ptr(v);
+            n += JL_PRINTF(out, "0x");
+            for(int i=nb-1; i >= 0; --i)
+                n += JL_PRINTF(out, "%02hhx", data[i]);
+        }
+        else {
+            jl_value_t *fldval=NULL;
+            JL_GC_PUSH1(&fldval);
+            for (size_t i = 0; i < tlen; i++) {
+                n += JL_PRINTF(out, ((jl_sym_t*)jl_tupleref(t->names, i))->name);
+                //jl_fielddesc_t f = t->fields[i];
                 n += JL_PRINTF(out, "=");
-                n += jl_static_show(out, *((jl_value_t**)jl_data_ptr(v)+f.offset/sizeof(void*)));
+                fldval = jl_get_nth_field(v, i);
+                n += jl_static_show(out, fldval);
+                if (i != tlen-1)
+                    n += JL_PRINTF(out, ", ");
             }
-            else {
-                n += JL_PRINTF(out, "::");
-                n += jl_static_show(out, jl_tupleref(t->types, i));
-            }
-            if (i != tlen-1)
-                n += JL_PRINTF(out, ", ");
+            JL_GC_POP();
         }
         n += JL_PRINTF(out, ")");
     }
@@ -1372,7 +1343,11 @@ DLLEXPORT void jl_(void *jl_value)
     in_jl_--;
 }
 
-DLLEXPORT void jl_breakpoint(jl_value_t* v)
+DLLEXPORT void jl_breakpoint(jl_value_t *v)
 {
     // put a breakpoint in you debugger here
 }
+
+#ifdef __cplusplus
+}
+#endif
