@@ -3,35 +3,47 @@
 // --- library symbol lookup ---
 
 // map from "libX" to full soname "libX.so.ver"
-#if defined(__linux__)
+#if defined(__linux__) || defined(__FreeBSD__)
 static std::map<std::string, std::string> sonameMap;
 static bool got_sonames = false;
 
-extern "C" DLLEXPORT void jl_read_sonames()
+extern "C" DLLEXPORT void jl_read_sonames(void)
 {
     char *line=NULL;
     size_t sz=0;
+#if defined(__linux__)
     FILE *ldc = popen("/sbin/ldconfig -p", "r");
+#else
+    FILE *ldc = popen("/sbin/ldconfig -r", "r");
+#endif
 
     while (!feof(ldc)) {
         ssize_t n = getline(&line, &sz, ldc);
         if (n == -1)
             break;
-        if (n > 2 && isspace(line[0])) {
+        if (n > 2 && isspace((unsigned char)line[0])) {
             int i=0;
-            while (isspace(line[++i])) ;
+#ifdef __linux__
+            while (isspace((unsigned char)line[++i])) ;
             char *name = &line[i];
             char *dot = strstr(name, ".so");
+#else
+            char *name = strstr(line, ":-l");
+            if (name == NULL) continue;
+            strncpy(name, "lib", 3);
+            char *dot = strchr(name, '.');
+#endif
             i=0;
 
             if (NULL == dot)
                 continue;
 
+#ifdef __linux__
             // Detect if this entry is for the current architecture
-            while (!isspace(dot[++i])) ;
-            while (isspace(dot[++i])) ;
+            while (!isspace((unsigned char)dot[++i])) ;
+            while (isspace((unsigned char)dot[++i])) ;
             int j = i;
-            while (!isspace(dot[++j])) ;
+            while (!isspace((unsigned char)dot[++j])) ;
             char *arch = strstr(dot+i,"x86-64");
             if (arch != NULL && arch < dot + j) {
 #ifdef _P32
@@ -43,6 +55,7 @@ extern "C" DLLEXPORT void jl_read_sonames()
                 continue;
 #endif
             }
+#endif // __linux__
 
             char *abslibpath = strrchr(line, ' ');
             if (dot != NULL && abslibpath != NULL) {
@@ -58,7 +71,7 @@ extern "C" DLLEXPORT void jl_read_sonames()
     pclose(ldc);
 }
 
-extern "C" const char *jl_lookup_soname(char *pfx, size_t n)
+extern "C" DLLEXPORT const char *jl_lookup_soname(char *pfx, size_t n)
 {
     if (!got_sonames) {
         jl_read_sonames();
@@ -73,53 +86,137 @@ extern "C" const char *jl_lookup_soname(char *pfx, size_t n)
 #endif
 
 // map from user-specified lib names to handles
-static std::map<std::string, void*> libMap;
+static std::map<std::string, uv_lib_t*> libMap;
 
-extern "C" int add_library_mapping(char *lib, void *hnd)
+static uv_lib_t *get_library(char *lib)
 {
-    JL_PUTS(const_cast<char*>("WARNING: add_library_mapping is deprecated, use push!(DL_LOAD_PATH,\"/path/to/search\") instead.\n"), JL_STDERR);
-    if (libMap[lib] == NULL && hnd != NULL) {
+    uv_lib_t *hnd;
+#ifdef _OS_WINDOWS_
+    if ((intptr_t)lib == 1)
+        return jl_exe_handle;
+    if ((intptr_t)lib == 2)
+        return jl_dl_handle;
+#endif
+    if (lib == NULL)
+        return jl_RTLD_DEFAULT_handle;
+    hnd = libMap[lib];
+    if (hnd != NULL)
+        return hnd;
+    hnd = jl_load_dynamic_library(lib, JL_RTLD_DEFAULT);
+    if (hnd != NULL)
         libMap[lib] = hnd;
-        return 0;
-    }
-    else {
-        return -1;
-    }
+    return hnd;
 }
 
-static void *add_library_sym(char *name, char *lib)
+extern "C" DLLEXPORT
+void *jl_load_and_lookup(char *f_lib, char *f_name, uv_lib_t **hnd)
 {
-    void *hnd;
-    if (lib == NULL) {
-        hnd = jl_dl_handle;
-    }
-    else {
-        hnd = libMap[lib];
-        if (hnd == NULL) {
-            hnd = jl_load_dynamic_library(lib, JL_RTLD_DEFAULT);
-            if (hnd != NULL)
-                libMap[lib] = hnd;
-            else
-                return NULL;
-        }
-    }
-    // add a symbol->address mapping for the JIT
-    void *sval = jl_dlsym_e((uv_lib_t*)hnd, name);
-    if (lib != NULL && hnd != jl_dl_handle) {
-        void *exist = sys::DynamicLibrary::SearchForAddressOfSymbol(name);
-        if (exist != NULL && exist != sval &&
-            // openlibm conflicts with libm, and lots of our libraries
-            // (including LLVM) link to libm. fortunately AddSymbol() is
-            // able to resolve these in favor of openlibm, but this could
-            // be an issue in the future (TODO).
-            strcmp(lib,"libopenlibm")) {
-            ios_printf(ios_stderr, "Warning: Possible conflict in library symbol %s\n", name);
-        }
-        sys::DynamicLibrary::AddSymbol(name, sval);
-    }
-    return sval;
+    uv_lib_t *handle = *hnd;
+    if (!handle)
+        *hnd = handle = get_library(f_lib);
+    void *ptr = jl_dlsym_e(handle, f_name);
+    if (!ptr)
+        jl_errorf("symbol could not be found %s: %s\n", f_name, uv_dlerror(handle));
+    return ptr;
 }
 
+static std::map<std::string, GlobalVariable*> libMapGV;
+static std::map<std::string, GlobalVariable*> symMapGV;
+static Value *runtime_sym_lookup(PointerType *funcptype, char *f_lib, char *f_name, jl_codectx_t *ctx)
+{
+    // in pseudo-code, this function emits the following:
+    //   global uv_lib_t **libptrgv
+    //   global void **llvmgv
+    //   if (*llvmgv == NULL) {
+    //       *llvmgv = jl_load_and_lookup(f_lib, f_name, libptrgv);
+    //   }
+    //   return (*llvmgv)
+    Constant *initnul = ConstantPointerNull::get((PointerType*)T_pint8);
+
+    uv_lib_t *libsym = NULL;
+    bool runtime_lib = false;
+    GlobalVariable *libptrgv;
+#ifdef _OS_WINDOWS_
+    if ((intptr_t)f_lib == 1) {
+        libptrgv = prepare_global(jlexe_var);
+        libsym = jl_exe_handle;
+    }
+    else if ((intptr_t)f_lib == 2) {
+        libptrgv = prepare_global(jldll_var);
+        libsym = jl_dl_handle;
+    }
+    else
+#endif
+    if (f_lib == NULL) {
+        libptrgv = prepare_global(jlRTLD_DEFAULT_var);
+        libsym = jl_RTLD_DEFAULT_handle;
+    }
+    else {
+        runtime_lib = true;
+        libptrgv = libMapGV[f_lib];
+        if (libptrgv == NULL) {
+            libptrgv = new GlobalVariable(*jl_Module, T_pint8,
+               false, GlobalVariable::PrivateLinkage,
+               initnul, f_lib);
+            libMapGV[f_lib] = libptrgv;
+            libsym = get_library(f_lib);
+            assert(libsym != NULL);
+#ifdef USE_MCJIT
+            llvm_to_jl_value[libptrgv] = libsym;
+#else
+            *((uv_lib_t**)jl_ExecutionEngine->getPointerToGlobal(libptrgv)) = libsym;
+#endif
+        }
+    }
+    if (libsym == NULL) {
+#ifdef USE_MCJIT
+        libsym = (uv_lib_t*)llvm_to_jl_value[libptrgv];
+#else
+        libsym = *((uv_lib_t**)jl_ExecutionEngine->getPointerToGlobal(libptrgv));
+#endif
+    }
+
+    assert(libsym != NULL);
+
+    GlobalVariable *llvmgv = symMapGV[f_name];
+    if (llvmgv == NULL) {
+        // MCJIT forces this to have external linkage eventually, so we would clobber
+        // the symbol of the actual function.
+        std::string name = f_name;
+        name = "ccall_" + name;
+        llvmgv = new GlobalVariable(*jl_Module, T_pint8,
+           false, GlobalVariable::PrivateLinkage,
+           initnul, name);
+        symMapGV[f_name] = llvmgv;
+#ifdef USE_MCJIT
+        llvm_to_jl_value[llvmgv] = jl_dlsym_e(libsym, f_name);
+#else
+        *((void**)jl_ExecutionEngine->getPointerToGlobal(llvmgv)) = jl_dlsym_e(libsym, f_name);
+#endif
+    }
+
+    BasicBlock *dlsym_lookup = BasicBlock::Create(getGlobalContext(), "dlsym"),
+               *ccall_bb = BasicBlock::Create(getGlobalContext(), "ccall");
+    builder.CreateCondBr(builder.CreateICmpNE(builder.CreateLoad(llvmgv), initnul), ccall_bb, dlsym_lookup);
+
+    ctx->f->getBasicBlockList().push_back(dlsym_lookup);
+    builder.SetInsertPoint(dlsym_lookup);
+    Value *libname;
+    if (runtime_lib) {
+        libname = builder.CreateGlobalStringPtr(f_lib);
+    }
+    else {
+        libname = literal_static_pointer_val(f_lib, T_pint8);
+    }
+    Value *llvmf = builder.CreateCall3(prepare_call(jldlsym_func), libname, builder.CreateGlobalStringPtr(f_name), libptrgv);
+    builder.CreateStore(llvmf, llvmgv);
+    builder.CreateBr(ccall_bb);
+
+    ctx->f->getBasicBlockList().push_back(ccall_bb);
+    builder.SetInsertPoint(ccall_bb);
+    llvmf = builder.CreateLoad(llvmgv);
+    return builder.CreatePointerCast(llvmf,funcptype);
+}
 // --- argument passing and scratch space utilities ---
 
 static Function *value_to_pointer_func;
@@ -134,12 +231,12 @@ static uint32_t arg_block_n = 0;
 static Function *save_arg_area_loc_func;
 static Function *restore_arg_area_loc_func;
 
-static uint64_t save_arg_area_loc()
+extern "C" DLLEXPORT uint64_t save_arg_area_loc()
 {
     return (((uint64_t)arg_block_n)<<32) | ((uint64_t)arg_area_loc);
 }
 
-static void restore_arg_area_loc(uint64_t l)
+extern "C" DLLEXPORT void restore_arg_area_loc(uint64_t l)
 {
     arg_area_loc = l&0xffffffff;
     uint32_t ab = l>>32;
@@ -178,7 +275,7 @@ static void *alloc_temp_arg_copy(void *obj, uint32_t sz)
 
 // this is a run-time function
 // warning: cannot allocate memory except using alloc_temp_arg_space
-extern "C" void *jl_value_to_pointer(jl_value_t *jt, jl_value_t *v, int argn,
+extern "C" DLLEXPORT void *jl_value_to_pointer(jl_value_t *jt, jl_value_t *v, int argn,
                                      int addressof)
 {
     jl_value_t *jvt = (jl_value_t*)jl_typeof(v);
@@ -246,7 +343,7 @@ static Value *julia_to_native(Type *ty, jl_value_t *jt, Value *jv,
 {
     Type *vt = jv->getType();
     if (ty == jl_pvalue_llvmt) {
-        return boxed(jv);
+        return boxed(jv,ctx);
     }
     else if (ty == vt && !addressOf) {
         return jv;
@@ -272,7 +369,7 @@ static Value *julia_to_native(Type *ty, jl_value_t *jt, Value *jv,
             }
         }
         // error. box for error handling.
-        jv = boxed(jv);
+        jv = boxed(jv,ctx);
     }
     else if (jl_is_cpointer_type(jt)) {
         assert(ty->isPointerTy());
@@ -284,7 +381,7 @@ static Value *julia_to_native(Type *ty, jl_value_t *jt, Value *jv,
             return builder.CreateBitCast(emit_arrayptr(jv), ty);
         }
         if (aty == (jl_value_t*)jl_ascii_string_type || aty == (jl_value_t*)jl_utf8_string_type) {
-            return builder.CreateBitCast(emit_arrayptr(emit_nthptr(jv,1)), ty);
+            return builder.CreateBitCast(emit_arrayptr(emit_nthptr(jv,1,tbaa_const)), ty);
         }
         if (jl_is_structtype(aty) && jl_is_leaf_type(aty) && !jl_is_array_type(aty)) {
             if (!addressOf) {
@@ -294,7 +391,7 @@ static Value *julia_to_native(Type *ty, jl_value_t *jt, Value *jv,
             return builder.CreateBitCast(emit_nthptr_addr(jv, (size_t)1), ty); // skip type tag field
         }
         *mightNeedTempSpace = true;
-        Value *p = builder.CreateCall4(value_to_pointer_func,
+        Value *p = builder.CreateCall4(prepare_call(value_to_pointer_func),
                                        literal_pointer_val(jl_tparam0(jt)), jv,
                                        ConstantInt::get(T_int32, argn),
                                        ConstantInt::get(T_int32, (int)addressOf));
@@ -318,6 +415,9 @@ static Value *julia_to_native(Type *ty, jl_value_t *jt, Value *jv,
         Value *pjv = builder.CreateBitCast(emit_nthptr_addr(jv, (size_t)1), PointerType::get(ty,0));
         return builder.CreateLoad(pjv, false);
     }
+    else if (jl_is_tuple(jt)) {
+        return emit_unbox(ty,jv,jt);
+    }
     // TODO: error for & with non-pointer argument type
     assert(jl_is_bitstype(jt));
     std::stringstream msg;
@@ -356,7 +456,7 @@ static native_sym_arg_t interpret_symbol_arg(jl_value_t *arg, jl_codectx_t *ctx,
                                "cglobal: first argument not a pointer or valid constant expression",
                                ctx);
         }
-        jl_ptr = emit_unbox(T_size, T_psize, arg1);
+        jl_ptr = emit_unbox(T_size, arg1, (jl_value_t*)jl_voidpointer_type);
     }
 
     void *fptr=NULL;
@@ -371,11 +471,9 @@ static native_sym_arg_t interpret_symbol_arg(jl_value_t *arg, jl_codectx_t *ctx,
             f_name = jl_string_data(ptr);
         if (f_name != NULL) {
             // just symbol, default to JuliaDLHandle
-#ifdef _OS_WINDOWS_
-            //TODO: store the f_lib name instead of fptr
-            fptr = jl_dlsym_win32(f_name);
-#else
             // will look in process symbol table
+#ifdef _OS_WINDOWS_
+            f_lib = jl_dlfind_win32(f_name);
 #endif
         }
         else if (jl_is_cpointer_type(jl_typeof(ptr))) {
@@ -450,43 +548,32 @@ static Value *emit_cglobal(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         res = builder.CreateIntToPtr(sym.jl_ptr, lrt);
     }
     else if (sym.fptr != NULL) {
-        res = literal_pointer_val(sym.fptr, lrt);
+        res = literal_static_pointer_val(sym.fptr, lrt);
+        if (imaging_mode)
+            JL_PRINTF(JL_STDERR,"warning: literal address used in cglobal for %s; code cannot be statically compiled\n", sym.f_name);
     }
     else {
-        void *symaddr;
-        if (sym.f_lib != NULL)
-            symaddr = add_library_sym(sym.f_name, sym.f_lib);
-        else
-            symaddr = sys::DynamicLibrary::SearchForAddressOfSymbol(sym.f_name);
-        if (symaddr == NULL) {
-            std::stringstream msg;
-            msg << "cglobal: could not find symbol ";
-            msg << sym.f_name;
-            if (sym.f_lib != NULL) {
-                msg << " in library ";
-                msg << sym.f_lib;
-            }
-            emit_error(msg.str(), ctx);
-            res = literal_pointer_val(NULL, lrt);
+        if (imaging_mode) {
+            res = runtime_sym_lookup((PointerType*)lrt, sym.f_lib, sym.f_name, ctx);
         }
         else {
-            Value *nv = jl_Module->getNamedValue(sym.f_name);
-            if (nv != NULL) {
-                // if the symbol already exists, it might be a function or
-                // something else other than a GlobalVariable, so return
-                // whatever it is.
-                res = nv;
-                if (res->getType() != lrt) {
-                    // if you attempt to access a cglobal multiple times with
-                    // different types, the type of the cached global might be
-                    // wrong.
-                    res = builder.CreateBitCast(res, lrt);
+            void *symaddr = jl_dlsym_e(get_library(sym.f_lib), sym.f_name);
+            if (symaddr == NULL) {
+                std::stringstream msg;
+                msg << "cglobal: could not find symbol ";
+                msg << sym.f_name;
+                if (sym.f_lib != NULL) {
+#ifdef _OS_WINDOWS_
+                    assert((intptr_t)sym.f_lib != 1 && (intptr_t)sym.f_lib != 2);
+#endif
+                    msg << " in library ";
+                    msg << sym.f_lib;
                 }
+                emit_error(msg.str(), ctx);
             }
-            else {
-                res = jl_Module->getOrInsertGlobal(sym.f_name,
-                                                   lrt->getContainedType(0));
-            }
+            // since we aren't saving this code, there's no sense in
+            // putting anything complicated here: just JIT the address of the cglobal
+            res = literal_static_pointer_val(symaddr, lrt);
         }
     }
 
@@ -615,7 +702,7 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
             }
         }
         Type *t = julia_struct_to_llvm(tti);
-        if (t == NULL) {
+        if (t == NULL || t == T_void) {
             JL_GC_POP();
             std::stringstream msg;
             msg << "ccall: the type of argument ";
@@ -656,15 +743,14 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         jl_error("ccall: wrong number of arguments to C function");
 
     // some special functions
-    if (fptr == &jl_array_ptr ||
+    if (fptr == (void *) &jl_array_ptr ||
         (f_lib==NULL && f_name && !strcmp(f_name,"jl_array_ptr"))) {
         assert(lrt->isPointerTy());
         Value *ary = emit_expr(args[4], ctx);
         JL_GC_POP();
-        return mark_julia_type(builder.CreateBitCast(emit_arrayptr(ary),lrt),
-                               rt);
+        return mark_julia_type(builder.CreateBitCast(emit_arrayptr(ary),lrt), rt);
     }
-    if (fptr == &jl_value_ptr ||
+    if (fptr == (void *) &jl_value_ptr ||
         (f_lib==NULL && f_name && !strcmp(f_name,"jl_value_ptr"))) {
         assert(lrt->isPointerTy());
         jl_value_t *argi = args[4];
@@ -673,48 +759,23 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
             addressOf = true;
             argi = jl_exprarg(argi,0);
         }
-        Value *ary = boxed(emit_expr(argi, ctx));
+        Value *ary = boxed(emit_expr(argi, ctx),ctx);
         JL_GC_POP();
         return mark_julia_type(
-                builder.CreateBitCast(emit_nthptr_addr(ary, addressOf?1:0),lrt),
+                builder.CreateBitCast(emit_nthptr_addr(ary, addressOf?1:0), lrt),
                 rt);
     }
-
-    // make LLVM function object for the target
-    Value *llvmf;
-    FunctionType *functype = FunctionType::get(lrt, fargt_sig, isVa);
-    
-    if (jl_ptr != NULL) {
-        null_pointer_check(jl_ptr,ctx);
-        Type *funcptype = PointerType::get(functype,0);
-        llvmf = builder.CreateIntToPtr(jl_ptr, funcptype);
-    }
-    else if (fptr != NULL) {
-        Type *funcptype = PointerType::get(functype,0);
-        llvmf = literal_pointer_val(fptr, funcptype);
-    }
-    else {
-        void *symaddr;
-        if (f_lib != NULL)
-            symaddr = add_library_sym(f_name, f_lib);
-        else
-            symaddr = sys::DynamicLibrary::SearchForAddressOfSymbol(f_name);
-        if (symaddr == NULL) {
+    if (fptr == (void *) &jl_is_leaf_type ||
+        (f_lib==NULL && f_name && !strcmp(f_name, "jl_is_leaf_type"))) {
+        jl_value_t *arg = args[4];
+        jl_value_t *ty = expr_type(arg, ctx);
+        if (jl_is_type_type(ty) && !jl_is_typevar(jl_tparam0(ty))) {
+            int isleaf = jl_is_leaf_type(jl_tparam0(ty));
             JL_GC_POP();
-            std::stringstream msg;
-            msg << "ccall: could not find function ";
-            msg << f_name;
-            if (f_lib != NULL) {
-                msg << " in library ";
-                msg << f_lib;
-            }
-            emit_error(msg.str(), ctx);
-            return literal_pointer_val(jl_nothing);
+            return ConstantInt::get(T_int32, isleaf);
         }
-        llvmf = jl_Module->getOrInsertFunction(f_name, functype);
     }
 
-
     // save place before arguments, for possible insertion of temp arg
     // area saving code.
     Value *saveloc=NULL;
@@ -730,7 +791,7 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         savespot = &_savespot;
     }
 
-    if (0 && f_name != NULL && f_lib != NULL) {
+    if (0 && f_name != NULL) {
         // print the f_name before each ccall
         Value *zeros[2] = { ConstantInt::get(T_int32, 0),
                             ConstantInt::get(T_int32, 0) };
@@ -738,24 +799,24 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
             msg << "ccall: ";
             msg << f_name;
             msg << "(...)";
-            if (f_lib != NULL) {
+            if (f_lib != NULL && (intptr_t)f_lib != 1 && (intptr_t)f_lib != 2) {
                 msg << " in library ";
                 msg << f_lib;
             }
             msg << "\n";
-        builder.CreateCall2(jlputs_func,
+        builder.CreateCall2(prepare_call(jlputs_func),
                             builder.CreateGEP(stringConst(msg.str()),
                                          ArrayRef<Value*>(zeros)),
-                            literal_pointer_val(JL_STDERR));
+                            prepare_global(jlstderr_var));
     }
 
     // emit arguments
-    Value *argvals[(nargs-3)/2 + sret];
+    Value **argvals = (Value**) alloca(((nargs-3)/2 + sret)*sizeof(Value*));
     Value *result;
     if (sret) {
         assert(jl_is_structtype(rt));
         result = builder.CreateCall(
-                jlallocobj_func,
+                prepare_call(jlallocobj_func),
                 ConstantInt::get(T_size,
                     sizeof(void*)+((jl_datatype_t*)rt)->size));
         //TODO: Fill type pointer fields with C_NULL's
@@ -793,23 +854,22 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         if (largty == jl_pvalue_llvmt || largty->isStructTy()) {
             arg = emit_expr(argi, ctx, true);
             if (largty == jl_pvalue_llvmt && arg->getType() != jl_pvalue_llvmt) {
-                arg = boxed(arg);
+                arg = boxed(arg,ctx);
                 needroot = true;
             }
         }
         else {
             arg = emit_unboxed(argi, ctx);
             if (jl_is_bitstype(expr_type(argi, ctx))) {
-                Type *totype = addressOf ? largty->getContainedType(0) : largty;
-                Type *ptype  = addressOf ? largty : PointerType::get(largty,0);
                 Type *at = arg->getType();
+                Type *totype = addressOf ? largty->getContainedType(0) : largty;
                 if (at != jl_pvalue_llvmt && at != totype &&
                     !(at->isPointerTy() && jargty==(jl_value_t*)jl_voidpointer_type)) {
                     emit_type_error(arg, jargty, "ccall", ctx);
                     arg = UndefValue::get(totype);
                 }
                 else {
-                    arg = emit_unbox(totype, ptype, arg);
+                    arg = emit_unbox(totype, arg, jargty);
                 }
             }
         }
@@ -828,10 +888,59 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         needTempSpace |= mightNeed;
         needStackRestore |= nSR;
     }
+
+
+    // make LLVM function object for the target
+    // keep this close to the function call, so that the compiler can
+    // optimize the global pointer load in the common case
+    Value *llvmf;
+    FunctionType *functype = FunctionType::get(lrt, fargt_sig, isVa);
+
+    if (jl_ptr != NULL) {
+        null_pointer_check(jl_ptr,ctx);
+        Type *funcptype = PointerType::get(functype,0);
+        llvmf = builder.CreateIntToPtr(jl_ptr, funcptype);
+    }
+    else if (fptr != NULL) {
+        Type *funcptype = PointerType::get(functype,0);
+        llvmf = literal_static_pointer_val(fptr, funcptype);
+        if (imaging_mode)
+            JL_PRINTF(JL_STDERR,"warning: literal address used in ccall for %s; code cannot be statically compiled\n", f_name);
+    }
+    else {
+        assert(f_name != NULL);
+
+        PointerType *funcptype = PointerType::get(functype,0);
+        if (imaging_mode) {
+            llvmf = runtime_sym_lookup(funcptype, f_lib, f_name, ctx);
+        }
+        else {
+            void *symaddr = jl_dlsym_e(get_library(f_lib), f_name);
+            if (symaddr == NULL) {
+                JL_GC_POP();
+                std::stringstream msg;
+                msg << "ccall: could not find function ";
+                msg << f_name;
+                if (f_lib != NULL) {
+#ifdef _OS_WINDOWS_
+                    assert((intptr_t)f_lib != 1 && (intptr_t)f_lib != 2);
+#endif
+                    msg << " in library ";
+                    msg << f_lib;
+                }
+                emit_error(msg.str(), ctx);
+                return literal_pointer_val(jl_nothing);
+            }
+            // since we aren't saving this code, there's no sense in
+            // putting anything complicated here: just JIT the function address
+            llvmf = literal_static_pointer_val(symaddr, funcptype);
+        }
+    }
+
     if (needTempSpace) {
         // save temp argument area stack pointer
         // TODO: inline this
-        saveloc = CallInst::Create(save_arg_area_loc_func);
+        saveloc = CallInst::Create(prepare_call(save_arg_area_loc_func));
         if (savespot)
             instList.insertAfter(savespot, (Instruction*)saveloc);
         else
@@ -846,9 +955,10 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         else
             instList.push_front((Instruction*)stacksave);
     }
+
     // the actual call
     Value *ret = builder.CreateCall(
-            llvmf,
+            prepare_call(llvmf),
             ArrayRef<Value*>(&argvals[0],(nargs-3)/2+sret));
 
     attr_type attributes;
@@ -880,7 +990,7 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
     if (needTempSpace) {
         // restore temp argument area stack pointer
         assert(saveloc != NULL);
-        builder.CreateCall(restore_arg_area_loc_func, saveloc);
+        builder.CreateCall(prepare_call(restore_arg_area_loc_func), saveloc);
     }
     ctx->argDepth = last_depth;
     if (0) { // Enable this to turn on SSPREQ (-fstack-protector) on the function containing this ccall
@@ -898,7 +1008,7 @@ static Value *emit_ccall(jl_value_t **args, size_t nargs, jl_codectx_t *ctx)
         //fprintf(stderr, "ccall rt: %s -> %s\n", f_name, ((jl_tag_type_t*)rt)->name->name->name);
         assert(jl_is_structtype(rt));
         Value *strct =
-            builder.CreateCall(jlallocobj_func,
+            builder.CreateCall(prepare_call(jlallocobj_func),
                                ConstantInt::get(T_size,
                                     sizeof(void*)+((jl_datatype_t*)rt)->size));
         builder.CreateStore(literal_pointer_val((jl_value_t*)rt),
