@@ -15,6 +15,10 @@
 #include <dlfcn.h>
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #if defined(__APPLE__)
 static char *extensions[] = { "", ".dylib" };
 #define N_EXTENSIONS 2
@@ -32,13 +36,9 @@ static char *extensions[] = { ".so", "" };
 
 extern char *julia_home;
 
-#if defined(__linux__)
-char *jl_lookup_soname(char *pfx, size_t n);
-#endif
-
 #define JL_RTLD(flags, FLAG) (flags & JL_RTLD_ ## FLAG ? RTLD_ ## FLAG : 0)
 
-static int jl_uv_dlopen(const char* filename, uv_lib_t* lib, unsigned flags)
+DLLEXPORT int jl_uv_dlopen(const char *filename, uv_lib_t *lib, unsigned flags)
 {
 #if defined(_OS_WINDOWS_)
     needsSymRefreshModuleList = 1;
@@ -80,13 +80,13 @@ static uv_lib_t *jl_load_dynamic_library_(char *modname, unsigned flags, int thr
     char *ext;
     char path[PATHBUF];
     int i;
-    uv_lib_t *handle=malloc(sizeof(uv_lib_t));
-    handle->errmsg=NULL;
+    uv_lib_t *handle = (uv_lib_t*)malloc(sizeof(uv_lib_t));
+    handle->errmsg = NULL;
 
     if (modname == NULL) {
 #ifdef _OS_WINDOWS_
-        if (!GetModuleHandleExA(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
-                                (LPCSTR)(&jl_load_dynamic_library),
+        if (!GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
+                                (LPCWSTR)(&jl_load_dynamic_library),
                                 &handle->handle))
             jl_errorf("could not load base module", modname);
 #else
@@ -102,8 +102,8 @@ static uv_lib_t *jl_load_dynamic_library_(char *modname, unsigned flags, int thr
         error = jl_uv_dlopen(modname,handle,flags);
         if (!error) goto done;
     }
-    else {
-        jl_array_t* DL_LOAD_PATH = (jl_array_t*)jl_get_global(jl_base_module, jl_symbol("DL_LOAD_PATH"));
+    else if (jl_base_module != NULL) {
+        jl_array_t *DL_LOAD_PATH = (jl_array_t*)jl_get_global(jl_base_module, jl_symbol("DL_LOAD_PATH"));
         if (DL_LOAD_PATH != NULL) {
             size_t j;
             for (j = 0; j < jl_array_len(DL_LOAD_PATH); j++) {
@@ -134,8 +134,9 @@ static uv_lib_t *jl_load_dynamic_library_(char *modname, unsigned flags, int thr
         error = jl_uv_dlopen(path, handle, flags);
         if (!error) goto done;
     }
-#if defined(__linux__)
-    char *soname = jl_lookup_soname(modname, strlen(modname));
+
+#if defined(__linux__) || defined(__FreeBSD__)
+    const char *soname = jl_lookup_soname(modname, strlen(modname));
     error = (soname==NULL) || jl_uv_dlopen(soname, handle, flags);
     if (!error) goto done;
 #endif
@@ -161,7 +162,7 @@ uv_lib_t *jl_load_dynamic_library(char *modname, unsigned flags)
     return jl_load_dynamic_library_(modname, flags, 1);
 }
 
-DLLEXPORT void *jl_dlsym_e(uv_lib_t *handle, char *symbol)
+void *jl_dlsym_e(uv_lib_t *handle, char *symbol)
 {
     void *ptr;
     int  error=uv_dlsym(handle, symbol, &ptr);
@@ -169,7 +170,7 @@ DLLEXPORT void *jl_dlsym_e(uv_lib_t *handle, char *symbol)
     return ptr;
 }
 
-DLLEXPORT void *jl_dlsym(uv_lib_t *handle, char *symbol)
+void *jl_dlsym(uv_lib_t *handle, char *symbol)
 {
     void *ptr;
     int  error = uv_dlsym(handle, symbol, &ptr);
@@ -181,25 +182,33 @@ DLLEXPORT void *jl_dlsym(uv_lib_t *handle, char *symbol)
 
 #ifdef _OS_WINDOWS_
 //Look for symbols in win32 libraries
-void *jl_dlsym_win32(char *f_name)
+char *jl_dlfind_win32(char *f_name)
 {
-    void *fptr = jl_dlsym_e(jl_exe_handle, f_name);
-    if (!fptr) {
-        fptr = jl_dlsym_e(jl_dl_handle, f_name);
-        if (!fptr) {
-            fptr = jl_dlsym_e(jl_kernel32_handle, f_name);
-            if (!fptr) {
-                fptr = jl_dlsym_e(jl_ntdll_handle, f_name);
-                if (!fptr) {
-                    fptr = jl_dlsym_e(jl_crtdll_handle, f_name);
-                    if (!fptr) {
-                        fptr = jl_dlsym(jl_winsock_handle, f_name);
-                    }
-                }
-            }
-        }
-    }
-    return fptr;
+    if (jl_dlsym_e(jl_exe_handle, f_name))
+        return (char*)1;
+    if (jl_dlsym_e(jl_dl_handle, f_name))
+        return (char*)2;
+    if (jl_dlsym_e(jl_kernel32_handle, f_name))
+        return "kernel32";
+    if (jl_dlsym_e(jl_ntdll_handle, f_name))
+        return "ntdll";
+    if (jl_dlsym_e(jl_crtdll_handle, f_name))
+        return "msvcrt";
+    if (jl_dlsym(jl_winsock_handle, f_name))
+        return "ws2_32";
+    // additional common libraries (libc?) could be added here, but in general,
+    // it is better to specify the library explicitly in the code. This exists
+    // mainly to ease compatibility with linux, and for libraries that don't
+    // have a name (julia.exe and libjulia.dll)
+    // We could also loop over all libraries that have been used so far, but, again,
+    // explicit is preferred over implicit
+    return NULL;
+    // oops, we didn't find it. NULL defaults to searching jl_RTLD_DEFAULT_handle,
+    // which defaults to jl_dl_handle, where we won't find it, and will throw the
+    // appropriate error.
+}
+#endif
+
+#ifdef __cplusplus
 }
-
 #endif
