@@ -8,9 +8,12 @@
 #include <malloc.h>
 #endif
 #include "julia.h"
+#include "julia_internal.h"
 #include "builtin_proto.h"
-#include "newobj_internal.h"
-#include "jltypes_internal.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 static htable_t ser_tag;
 static htable_t deser_tag;
@@ -87,11 +90,178 @@ static void write_as_tag(ios_t *s, uint8_t tag)
     write_uint8(s, tag);
 }
 
-// --- serialize ---
+// --- Static Compile ---
 
 #define jl_serialize_value(s, v) jl_serialize_value_(s,(jl_value_t*)(v))
-
 static void jl_serialize_value_(ios_t *s, jl_value_t *v);
+static jl_value_t *jl_deserialize_value(ios_t *s);
+static jl_value_t *jl_deserialize_value_internal(ios_t *s);
+jl_value_t ***sysimg_gvars = NULL;
+
+extern int globalUnique;
+extern void jl_cpuid(int32_t CPUInfo[4], int32_t InfoType);
+extern const char *jl_cpu_string;
+uv_lib_t *jl_sysimg_handle = NULL;
+char *jl_sysimage_name = NULL;
+
+static void jl_load_sysimg_so(char *fname)
+{
+    // attempt to load the pre-compiled sysimg at fname
+    // if this succeeds, sysimg_gvars will be a valid array
+    // otherwise, it will be NULL
+    jl_sysimg_handle = jl_load_dynamic_library_e(fname, JL_RTLD_DEFAULT | JL_RTLD_GLOBAL);
+    if (jl_sysimg_handle != 0) {
+        sysimg_gvars = (jl_value_t***)jl_dlsym(jl_sysimg_handle, "jl_sysimg_gvars");
+        globalUnique = *(size_t*)jl_dlsym(jl_sysimg_handle, "jl_globalUnique");
+        const char *cpu_target = (const char*)jl_dlsym(jl_sysimg_handle, "jl_sysimg_cpu_target");
+        if (strcmp(cpu_target,jl_cpu_string) != 0)
+            jl_error("Julia and the system image were compiled for different architectures.\n"
+                     "Please delete or regenerate sys.{so,dll,dylib}.\n");
+        uint32_t info[4];
+        jl_cpuid((int32_t*)info, 1);
+        if (strcmp(cpu_target, "native") == 0) {
+            uint64_t saved_cpuid = *(uint64_t*)jl_dlsym(jl_sysimg_handle, "jl_sysimg_cpu_cpuid");
+            if (saved_cpuid != (((uint64_t)info[2])|(((uint64_t)info[3])<<32)))
+                jl_error("Target architecture mismatch. Please delete or regenerate sys.{so,dll,dylib}.");
+        }
+        else if (strcmp(cpu_target,"core2") == 0) {
+            int HasSSSE3 = (info[2] & 1<<9);
+            if (!HasSSSE3)
+                jl_error("The current host does not support SSSE3, but the system image was compiled for Core2.\n"
+                         "Please delete or regenerate sys.{so,dll,dylib}.");
+        }
+        jl_sysimage_name = strdup(fname);
+    }
+    else {
+        sysimg_gvars = 0;
+    }
+}
+
+static jl_value_t *jl_deserialize_gv(ios_t *s, jl_value_t *v)
+{
+    // Restore the GlobalVariable reference to this jl_value_t via the sysimg_gvars table
+    int32_t gvname_index = read_int32(s)-1;
+    if (sysimg_gvars != NULL && gvname_index >= 0) {
+        *sysimg_gvars[gvname_index] = v;
+    }
+    return v;
+}
+
+static void jl_serialize_gv(ios_t *s, jl_value_t *v)
+{
+    // write the index of the literal_pointer_val into the system image
+    write_int32(s, jl_get_llvm_gv(v));
+}
+
+static void jl_serialize_globalvals(ios_t *s)
+{
+    size_t i, len = backref_table.size;
+    void **p = backref_table.table;
+    for(i=0; i < len; i+=2) {
+        void *offs = p[i+1];
+        if (offs != HT_NOTFOUND) {
+            int32_t gv = jl_get_llvm_gv((jl_value_t*)p[i]);
+            if (gv != 0) {
+                write_int32(s, (int)(intptr_t)offs);
+                write_int32(s, gv);
+            }
+        }
+    }
+    write_int32(s, 0);
+}
+
+static void jl_deserialize_globalvals(ios_t *s)
+{
+    while (1) {
+        intptr_t key = read_int32(s);
+        if (key == 0) break;
+        jl_deserialize_gv(s, (jl_value_t*)ptrhash_get(&backref_table, (void*)key));
+    }
+}
+
+static void jl_serialize_gv_syms(ios_t *s, jl_sym_t *v)
+{
+    // ensures all symbols referenced in the code have
+    // references in the system image to their global variable
+    // since symbols are static, they might not have had a
+    // reference anywhere in the code image other than here
+    void *bp = ptrhash_get(&backref_table, v);
+    if (bp == HT_NOTFOUND) {
+        int32_t gv = jl_get_llvm_gv((jl_value_t*)v);
+        if (gv != 0) {
+            jl_serialize_value(s, v);
+            write_int32(s, gv);
+        }
+    }
+    if (v->left) jl_serialize_gv_syms(s, v->left);
+    if (v->right) jl_serialize_gv_syms(s, v->right);
+}
+
+static void jl_deserialize_gv_syms(ios_t *s)
+{
+    while (1) {
+        jl_value_t *v = jl_deserialize_value(s);
+        if (!v) break;
+        jl_deserialize_gv(s, v);
+    }
+}
+
+struct delayed_fptrs_t {
+    jl_lambda_info_t *li;
+    int32_t func;
+    int32_t cfunc;
+} *delayed_fptrs = NULL;
+static size_t delayed_fptrs_n = 0;
+static size_t delayed_fptrs_max = 0;
+
+static void jl_delayed_fptrs(jl_lambda_info_t *li, int32_t func, int32_t cfunc)
+{
+    // can't restore the fptrs until after the system image is fully restored,
+    // since it will try to decompress the function AST to determine the argument types
+    if (cfunc || func) {
+        if (delayed_fptrs_max < delayed_fptrs_n + 1) {
+            if (delayed_fptrs_max == 0)
+                // current measurements put the number of functions at 1130
+                delayed_fptrs_max = 2048;
+            else
+                delayed_fptrs_max *= 2;
+            delayed_fptrs = (struct delayed_fptrs_t*)realloc(delayed_fptrs, delayed_fptrs_max*sizeof(delayed_fptrs[0])); //assumes sizeof==alignof
+        }
+        delayed_fptrs[delayed_fptrs_n].li = li;
+        delayed_fptrs[delayed_fptrs_n].func = func;
+        delayed_fptrs[delayed_fptrs_n].cfunc = cfunc;
+        delayed_fptrs_n++;
+    }
+}
+
+static void jl_update_all_fptrs()
+{
+    //printf("delayed_fptrs_n: %d\n", delayed_fptrs_n);
+    jl_value_t ***gvars = sysimg_gvars;
+    if (gvars == 0) return;
+    // jl_fptr_to_llvm needs to decompress some ASTs, therefore this needs to be NULL
+    // to skip trying to restore GlobalVariable pointers in jl_deserialize_gv
+    sysimg_gvars = NULL;
+    size_t i;
+    for (i = 0; i < delayed_fptrs_n; i++) {
+        jl_lambda_info_t *li = delayed_fptrs[i].li;
+        int32_t func = delayed_fptrs[i].func-1;
+        if (func >= 0) {
+            jl_fptr_to_llvm((void*)gvars[func], li, 0);
+        }
+        int32_t cfunc = delayed_fptrs[i].cfunc-1;
+        if (cfunc >= 0) {
+            jl_fptr_to_llvm((void*)gvars[cfunc], li, 1);
+        }
+    }
+    delayed_fptrs_n = 0;
+    delayed_fptrs_max = 0;
+    free(delayed_fptrs);
+    delayed_fptrs = NULL;
+}
+
+
+// --- serialize ---
 
 static void jl_serialize_fptr(ios_t *s, void *fptr)
 {
@@ -122,7 +292,8 @@ static void jl_serialize_datatype(ios_t *s, jl_datatype_t *dt)
         jl_serialize_value(s, dt->names);
         jl_serialize_value(s, dt->types);
     }
-    write_uint8(s, dt->abstract | (dt->mutabl<<1) | (dt->pointerfree<<2));
+    int has_instance = !!(dt->instance != NULL);
+    write_uint8(s, dt->abstract | (dt->mutabl<<1) | (dt->pointerfree<<2) | (has_instance<<3));
     if (!dt->abstract)
         write_int32(s, dt->uid);
 
@@ -132,12 +303,14 @@ static void jl_serialize_datatype(ios_t *s, jl_datatype_t *dt)
     jl_serialize_value(s, dt->ctor_factory);
     jl_serialize_value(s, dt->env);
     jl_serialize_value(s, dt->linfo);
-    jl_serialize_fptr(s, dt->fptr);
+    if (has_instance)
+        jl_serialize_value(s, dt->instance);
+    jl_serialize_fptr(s, (void*)dt->fptr);
 }
 
 static void jl_serialize_module(ios_t *s, jl_module_t *m)
 {
-    // set on every startup; don't save
+    // set on every startup; don't save value
     jl_sym_t *jhsym = jl_symbol("JULIA_HOME");
     writetag(s, jl_module_type);
     jl_serialize_value(s, m->name);
@@ -145,15 +318,20 @@ static void jl_serialize_module(ios_t *s, jl_module_t *m)
     size_t i;
     void **table = m->bindings.table;
     for(i=1; i < m->bindings.size; i+=2) {
-        if (table[i] != HT_NOTFOUND &&
-            !(table[i-1] == jhsym && m == jl_core_module)) {
+        if (table[i] != HT_NOTFOUND) {
             jl_binding_t *b = (jl_binding_t*)table[i];
-            if (!(b->owner != m && m == jl_main_module)) {
+            if (b->owner == m || m != jl_main_module) {
                 jl_serialize_value(s, b->name);
-                jl_serialize_value(s, b->value);
+                if (table[i-1] == jhsym && m == jl_core_module) {
+                    jl_serialize_value(s, NULL);
+                }
+                else {
+                    jl_serialize_value(s, b->value);
+                }
                 jl_serialize_value(s, b->type);
                 jl_serialize_value(s, b->owner);
                 write_int8(s, (b->constp<<2) | (b->exportp<<1) | (b->imported));
+                jl_serialize_gv(s, (jl_value_t*)b);
             }
         }
     }
@@ -277,7 +455,7 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
             jl_serialize_value(s, jl_box_long(jl_array_dim(ar,i)));
         if (!ar->ptrarray) {
             size_t tot = jl_array_len(ar) * ar->elsize;
-            ios_write(s, jl_array_data(ar), tot);
+            ios_write(s, (char*)jl_array_data(ar), tot);
         }
         else {
             for(i=0; i < jl_array_len(ar); i++) {
@@ -318,10 +496,10 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
         jl_serialize_value(s, (jl_value_t*)f->linfo);
         jl_serialize_value(s, f->env);
         if (f->linfo && f->linfo->ast && f->fptr != &jl_trampoline) {
-            jl_serialize_fptr(s, &jl_trampoline);
+            jl_serialize_fptr(s, (void*)&jl_trampoline);
         }
         else {
-            jl_serialize_fptr(s, f->fptr);
+            jl_serialize_fptr(s, (void*)f->fptr);
         }
     }
     else if (jl_is_lambda_info(v)) {
@@ -345,6 +523,9 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
         jl_serialize_value(s, (jl_value_t*)li->roots);
         jl_serialize_value(s, (jl_value_t*)li->def);
         jl_serialize_value(s, (jl_value_t*)li->capt);
+        // save functionObject pointers
+        write_int32(s, li->functionID);
+        write_int32(s, li->cFunctionID);
     }
     else if (jl_typeis(v, jl_module_type)) {
         jl_serialize_module(s, (jl_module_t*)v);
@@ -359,45 +540,45 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
             *(int64_t*)data >= S32_MIN && *(int64_t*)data <= S32_MAX) {
             writetag(s, (jl_value_t*)SmallInt64_tag);
             write_int32(s, (int32_t)*(int64_t*)data);
-            return;
         }
-        if (t == jl_int32_type) {
+        else if (t == jl_int32_type) {
             writetag(s, (jl_value_t*)Int32_tag);
             write_int32(s, (int32_t)*(int32_t*)data);
-            return;
-        }
-        if ((jl_value_t*)t == jl_idtable_type)
-            writetag(s, (jl_value_t*)IdTable_tag);
-        else
-            writetag(s, (jl_value_t*)jl_datatype_type);
-        jl_serialize_value(s, t);
-        size_t nf = jl_tuple_len(t->names);
-        if (nf == 0 && jl_datatype_size(t)>0) {
-            if (t->name == jl_pointer_type->name) {
-                write_int32(s, 0);
-#ifdef _P64
-                write_int32(s, 0);
-#endif
-            }
-            else {
-                ios_write(s, data, jl_datatype_size(t));
-            }
         }
         else {
-            if ((jl_value_t*)t == jl_idtable_type) {
-                jl_array_t *data = (jl_array_t*)jl_get_nth_field(v, 0);
-                jl_value_t **d = (jl_value_t**)data->data;
-                for(size_t i=0; i < jl_array_len(data); i+=2) {
-                    if (d[i+1] != NULL) {
-                        jl_serialize_value(s, d[i+1]);
-                        jl_serialize_value(s, d[i]);
-                    }
+            if ((jl_value_t*)t == jl_idtable_type)
+                writetag(s, (jl_value_t*)IdTable_tag);
+            else
+                writetag(s, (jl_value_t*)jl_datatype_type);
+            jl_serialize_value(s, t);
+            size_t nf = jl_tuple_len(t->names);
+            if (nf == 0 && jl_datatype_size(t)>0) {
+                if (t->name == jl_pointer_type->name) {
+                    write_int32(s, 0);
+#ifdef _P64
+                    write_int32(s, 0);
+#endif
+                }
+                else {
+                    ios_write(s, (char*)data, jl_datatype_size(t));
                 }
-                jl_serialize_value(s, NULL);
             }
             else {
-                for(size_t i=0; i < nf; i++) {
-                    jl_serialize_value(s, jl_get_nth_field(v, i));
+                if ((jl_value_t*)t == jl_idtable_type) {
+                    jl_array_t *data = (jl_array_t*)jl_get_nth_field(v, 0);
+                    jl_value_t **d = (jl_value_t**)data->data;
+                    for(size_t i=0; i < jl_array_len(data); i+=2) {
+                        if (d[i+1] != NULL) {
+                            jl_serialize_value(s, d[i+1]);
+                            jl_serialize_value(s, d[i]);
+                        }
+                    }
+                    jl_serialize_value(s, NULL);
+                }
+                else {
+                    for(size_t i=0; i < nf; i++) {
+                        jl_serialize_value(s, jl_get_nth_field(v, i));
+                    }
                 }
             }
         }
@@ -406,8 +587,6 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
 
 // --- deserialize ---
 
-static jl_value_t *jl_deserialize_value(ios_t *s);
-
 static jl_fptr_t jl_deserialize_fptr(ios_t *s)
 {
     int fptr = read_uint16(s);
@@ -419,6 +598,8 @@ static jl_fptr_t jl_deserialize_fptr(ios_t *s)
     return *(jl_fptr_t*)pbp;
 }
 
+#define DTINSTANCE_PLACEHOLDER ((void*)2)
+
 static jl_value_t *jl_deserialize_datatype(ios_t *s, int pos)
 {
     int tag = read_uint8(s);
@@ -456,6 +637,7 @@ static jl_value_t *jl_deserialize_datatype(ios_t *s, int pos)
     dt->abstract = flags&1;
     dt->mutabl = (flags>>1)&1;
     dt->pointerfree = (flags>>2)&1;
+    int has_instance = (flags>>3)&1;
     if (!dt->abstract)
         dt->uid = read_int32(s);
     else
@@ -466,21 +648,30 @@ static jl_value_t *jl_deserialize_datatype(ios_t *s, int pos)
     dt->ctor_factory = jl_deserialize_value(s);
     dt->env = jl_deserialize_value(s);
     dt->linfo = (jl_lambda_info_t*)jl_deserialize_value(s);
+    if (has_instance) {
+        jl_value_t *instance = (jl_value_t*)jl_deserialize_value_internal(s);
+        if (instance == DTINSTANCE_PLACEHOLDER)
+            instance = jl_new_struct_uninit(dt);
+        dt->instance = instance;
+    }
     dt->fptr = jl_deserialize_fptr(s);
     if (dt->name == jl_array_type->name || dt->name == jl_pointer_type->name ||
         dt->name == jl_type_type->name || dt->name == jl_vararg_type->name ||
-        dt->name == jl_abstractarray_type->name) {
+        dt->name == jl_abstractarray_type->name ||
+        dt->name == jl_densearray_type->name) {
         // builtin types are not serialized, so their caches aren't
         // explicitly saved. so we reconstruct the caches of builtin
         // parametric types here.
         jl_cell_1d_push(datatype_list, (jl_value_t*)dt);
     }
+
     return (jl_value_t*)dt;
 }
 
 jl_array_t *jl_eqtable_put(jl_array_t *h, void *key, void *val);
 
-static jl_value_t *jl_deserialize_value(ios_t *s)
+// Internal jl_deserialize_value. May return the placeholder value DTINSTANCE_PLACEHOLDER, unlike jl_deserialize_value
+static jl_value_t *jl_deserialize_value_internal(ios_t *s)
 {
     int pos = ios_pos(s);
     int32_t tag = read_uint8(s);
@@ -488,7 +679,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         return NULL;
     if (tag == 0) {
         tag = read_uint8(s);
-        jl_value_t *v = ptrhash_get(&deser_tag, (void*)(ptrint_t)tag);
+        jl_value_t *v = (jl_value_t*)ptrhash_get(&deser_tag, (void*)(ptrint_t)tag);
         assert(v != HT_NOTFOUND);
         return v;
     }
@@ -504,6 +695,9 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
     if (tag >= VALUE_TAGS) {
         return vtag;
     }
+    else if (vtag == (jl_value_t*)LiteralVal_tag) {
+        return jl_cellref(tree_literal_values, read_uint16(s));
+    }
 
     int usetable = (tree_literal_values == NULL);
 
@@ -529,13 +723,13 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             len = read_uint8(s);
         else
             len = read_int32(s);
-        char *name = alloca(len+1);
+        char *name = (char*)alloca(len+1);
         ios_read(s, name, len);
         name[len] = '\0';
-        jl_value_t *s = (jl_value_t*)jl_symbol(name);
+        jl_value_t *sym = (jl_value_t*)jl_symbol(name);
         if (usetable)
-            ptrhash_put(&backref_table, (void*)(ptrint_t)pos, s);
-        return s;
+            ptrhash_put(&backref_table, (void*)(ptrint_t)pos, sym);
+        return sym;
     }
     else if (vtag == (jl_value_t*)jl_array_type ||
              vtag == (jl_value_t*)Array1d_tag) {
@@ -554,7 +748,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             elsize = elsize&0x7fff;
         }
         jl_value_t *aty = jl_deserialize_value(s);
-        size_t *dims = alloca(ndims*sizeof(size_t));
+        size_t *dims = (size_t*)alloca(ndims*sizeof(size_t));
         for(i=0; i < ndims; i++)
             dims[i] = jl_unbox_long(jl_deserialize_value(s));
         jl_array_t *a = jl_new_array_for_deserialization((jl_value_t*)aty, ndims, dims, isunboxed, elsize);
@@ -562,7 +756,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, (jl_value_t*)a);
         if (!a->ptrarray) {
             size_t tot = jl_array_len(a) * a->elsize;
-            ios_read(s, jl_array_data(a), tot);
+            ios_read(s, (char*)jl_array_data(a), tot);
         }
         else {
             for(i=0; i < jl_array_len(a); i++) {
@@ -587,9 +781,6 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         }
         return (jl_value_t*)e;
     }
-    else if (vtag == (jl_value_t*)LiteralVal_tag) {
-        return jl_cellref(tree_literal_values, read_uint16(s));
-    }
     else if (vtag == (jl_value_t*)jl_tvar_type) {
         jl_tvar_t *tv = (jl_tvar_t*)newobj((jl_value_t*)jl_tvar_type, 4);
         if (usetable)
@@ -629,13 +820,18 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         li->roots = (jl_array_t*)jl_deserialize_value(s);
         li->def = (jl_lambda_info_t*)jl_deserialize_value(s);
         li->capt = jl_deserialize_value(s);
-
         li->fptr = &jl_trampoline;
         li->functionObject = NULL;
         li->cFunctionObject = NULL;
         li->inInference = 0;
         li->inCompile = 0;
         li->unspecialized = NULL;
+        li->functionID = 0;
+        li->cFunctionID = 0;
+        int32_t cfunc_llvm, func_llvm;
+        func_llvm = read_int32(s);
+        cfunc_llvm = read_int32(s);
+        jl_delayed_fptrs(li, func_llvm, cfunc_llvm);
         return (jl_value_t*)li;
     }
     else if (vtag == (jl_value_t*)jl_module_type) {
@@ -645,10 +841,10 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, m);
         m->parent = (jl_module_t*)jl_deserialize_value(s);
         while (1) {
-            jl_value_t *name = jl_deserialize_value(s);
+            jl_sym_t *name = (jl_sym_t*)jl_deserialize_value(s);
             if (name == NULL)
                 break;
-            jl_binding_t *b = jl_get_binding_wr(m, (jl_sym_t*)name);
+            jl_binding_t *b = jl_get_binding_wr(m, name);
             b->value = jl_deserialize_value(s);
             b->type = (jl_value_t*)jl_deserialize_value(s);
             b->owner = (jl_module_t*)jl_deserialize_value(s);
@@ -656,6 +852,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             b->constp = (flags>>2) & 1;
             b->exportp = (flags>>1) & 1;
             b->imported = (flags) & 1;
+            jl_deserialize_gv(s, (jl_value_t*)b);
         }
         size_t ni = read_int32(s);
         for(size_t i=0; i < ni; i++) {
@@ -677,6 +874,14 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         return v;
     }
     else if (vtag == (jl_value_t*)jl_datatype_type || vtag == (jl_value_t*)IdTable_tag) {
+        // If the value v we are about to deserialize is some dt->instance, we have a circular 
+        // reference, because v == v->type->instance. To avoid this, we put a null value in 
+        // the backref table to reserve the space. If jl_deserialize_value encounters this,
+        // it knows to do the allocation itself. This work, because in all instances where
+        // v->type->instance != null, v->type has no fields, so there is no further serialized
+        // data stored that we would need to construct the type. 
+        if (usetable)
+            ptrhash_put(&backref_table, (void*)(ptrint_t)pos, DTINSTANCE_PLACEHOLDER);
         jl_datatype_t *dt = (jl_datatype_t*)jl_deserialize_value(s);
         if (dt == jl_datatype_type)
             return jl_deserialize_datatype(s, pos);
@@ -684,7 +889,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         jl_value_t *v;
         if (nf == 0 && jl_datatype_size(dt)>0) {
             int nby = jl_datatype_size(dt);
-            char *data = alloca(nby);
+            char *data = (char*)alloca(nby);
             ios_read(s, data, nby);
             v = NULL;
             if (dt == jl_int32_type)
@@ -695,7 +900,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
                 v = jl_box_bool(*(int8_t*)data);
             else {
                 switch (nby) {
-                case 1: v = jl_box8 (dt, *(int8_t*) data); break;
+                case 1: v = jl_box8 (dt, *(int8_t *)data); break;
                 case 2: v = jl_box16(dt, *(int16_t*)data); break;
                 case 4: v = jl_box32(dt, *(int32_t*)data); break;
                 case 8: v = jl_box64(dt, *(int64_t*)data); break;
@@ -709,6 +914,11 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
                 ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
         }
         else {
+            if (dt->instance) {
+                if (usetable)
+                    *ptrhash_bp(&backref_table, (void*)(ptrint_t)pos) = dt->instance;
+                return dt->instance;
+            }
             v = jl_new_struct_uninit(dt);
             if (usetable)
                 ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
@@ -736,8 +946,17 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
     return NULL;
 }
 
+static jl_value_t *jl_deserialize_value(ios_t *s)
+{
+    jl_value_t *v = jl_deserialize_value_internal(s);
+    assert(v != DTINSTANCE_PLACEHOLDER);
+    return v;
+}
+
 // --- entry points ---
 
+extern jl_array_t *jl_module_init_order;
+
 DLLEXPORT
 void jl_save_system_image(char *fname)
 {
@@ -760,8 +979,8 @@ void jl_save_system_image(char *fname)
     jl_array_t *spec = mt->defs->func->linfo->specializations;
     if (spec != NULL && jl_array_len(spec) > 0 &&
         ((jl_lambda_info_t*)jl_cellref(spec,0))->inferred == 0) {
-        mt->cache = JL_NULL;
-        mt->cache_arg1 = JL_NULL;
+        mt->cache = (jl_methlist_t*)JL_NULL;
+        mt->cache_arg1 = (jl_array_t*)JL_NULL;
         mt->defs->func->linfo->tfunc = (jl_value_t*)jl_null;
         mt->defs->func->linfo->specializations = NULL;
     }
@@ -772,6 +991,30 @@ void jl_save_system_image(char *fname)
 
     jl_serialize_value(&f, jl_main_module);
 
+    // deser_tag is an array indexed from 2 until HT_NOTFOUND
+    // ensure everything in there can be reassociated with its GlobalValue
+    ptrint_t i=2;
+    void *v = ptrhash_get(&deser_tag, (void*)i);
+    while (v != HT_NOTFOUND) {
+        jl_serialize_gv(&f, (jl_value_t*)v);
+        v = ptrhash_get(&deser_tag, (void*)i);
+        i += 1;
+    }
+    jl_serialize_globalvals(&f);
+    jl_serialize_gv_syms(&f, jl_get_root_symbol()); // serialize symbols with GlobalValue references
+    jl_serialize_value(&f, NULL); // signal the end of the symbols list
+
+    // save module initialization order
+    if (jl_module_init_order != NULL) {
+        for(i=0; i < jl_array_len(jl_module_init_order); i++) {
+            // NULL out any modules that weren't saved
+            jl_value_t *mod = jl_cellref(jl_module_init_order, i);
+            if (ptrhash_get(&backref_table, mod) == HT_NOTFOUND)
+                jl_cellset(jl_module_init_order, i, NULL);
+        }
+    }
+    jl_serialize_value(&f, jl_module_init_order);
+
     write_int32(&f, jl_get_t_uid_ctr());
     write_int32(&f, jl_get_gs_ctr());
     htable_reset(&backref_table, 0);
@@ -784,7 +1027,7 @@ extern jl_function_t *jl_typeinf_func;
 extern int jl_boot_file_loaded;
 extern void jl_get_builtin_hooks(void);
 extern void jl_get_system_hooks(void);
-extern void jl_get_uv_hooks(void);
+extern void jl_get_uv_hooks();
 
 DLLEXPORT
 void jl_restore_system_image(char *fname)
@@ -795,6 +1038,19 @@ void jl_restore_system_image(char *fname)
         JL_PRINTF(JL_STDERR, "System image file \"%s\" not found\n", fname);
         exit(1);
     }
+    int build_mode = (jl_compileropts.build_path != NULL);
+#ifdef _OS_WINDOWS_
+    //XXX: the windows linker forces our system image to be
+    //     linked against only one dll, I picked libjulia-release
+    if (jl_is_debugbuild()) build_mode = 1;
+#endif
+    if (!build_mode) {
+        char *fname_shlib = (char*)alloca(strlen(fname));
+        strcpy(fname_shlib, fname);
+        char *fname_shlib_dot = strrchr(fname_shlib, '.');
+        if (fname_shlib_dot != NULL) *fname_shlib_dot = 0;
+        jl_load_sysimg_so(fname_shlib);
+    }
 #ifdef JL_GC_MARKSWEEP
     int en = jl_gc_is_enabled();
     jl_gc_disable();
@@ -805,12 +1061,27 @@ void jl_restore_system_image(char *fname)
     jl_array_type->env = jl_deserialize_value(&f);
     
     jl_main_module = (jl_module_t*)jl_deserialize_value(&f);
+    jl_internal_main_module = jl_main_module;
     jl_core_module = (jl_module_t*)jl_get_global(jl_main_module,
                                                  jl_symbol("Core"));
     jl_base_module = (jl_module_t*)jl_get_global(jl_main_module,
                                                  jl_symbol("Base"));
     jl_current_module = jl_base_module; // run start_image in Base
 
+    // deser_tag is an array indexed from 2 until HT_NOTFOUND
+    // ensure everything in there is reassociated with its GlobalValue
+    ptrint_t i=2;
+    void *v = ptrhash_get(&deser_tag, (void*)i);
+    while (v != HT_NOTFOUND) {
+        jl_deserialize_gv(&f, (jl_value_t*)v);
+        v = ptrhash_get(&deser_tag, (void*)i);
+        i += 1;
+    }
+    jl_deserialize_globalvals(&f);
+    jl_deserialize_gv_syms(&f);
+
+    jl_module_init_order = (jl_array_t*)jl_deserialize_value(&f);
+
     // cache builtin parametric types
     for(int i=0; i < jl_array_len(datatype_list); i++) {
         jl_value_t *v = jl_cellref(datatype_list, i);
@@ -837,6 +1108,26 @@ void jl_restore_system_image(char *fname)
 #ifdef JL_GC_MARKSWEEP
     if (en) jl_gc_enable();
 #endif
+    // restore the value of our "magic" JULIA_HOME variable/constant
+    jl_get_binding_wr(jl_core_module, jl_symbol("JULIA_HOME"))->value =
+        jl_cstr_to_string(julia_home);
+    jl_update_all_fptrs();
+}
+
+void jl_init_restored_modules()
+{
+    if (jl_module_init_order != NULL) {
+        jl_array_t *temp = jl_module_init_order;
+        jl_module_init_order = NULL;
+        JL_GC_PUSH1(&temp);
+        int i;
+        for(i=0; i < jl_array_len(temp); i++) {
+            jl_value_t *mod = jl_cellref(temp, i);
+            jl_module_run_initializer((jl_module_t*)mod);
+        }
+        jl_module_init_order = NULL;
+        JL_GC_POP();
+    }
 }
 
 DLLEXPORT
@@ -848,7 +1139,7 @@ jl_value_t *jl_ast_rettype(jl_lambda_info_t *li, jl_value_t *ast)
     ios_t src;
     jl_array_t *bytes = (jl_array_t*)ast;
     ios_mem(&src, 0);
-    ios_setbuf(&src, bytes->data, jl_array_len(bytes), 0);
+    ios_setbuf(&src, (char*)bytes->data, jl_array_len(bytes), 0);
     src.size = jl_array_len(bytes);
     int en = jl_gc_is_enabled();
     jl_gc_disable();
@@ -896,7 +1187,7 @@ jl_value_t *jl_uncompress_ast(jl_lambda_info_t *li, jl_value_t *data)
     tree_literal_values = li->module->constant_table;
     ios_t src;
     ios_mem(&src, 0);
-    ios_setbuf(&src, bytes->data, jl_array_len(bytes), 0);
+    ios_setbuf(&src, (char*)bytes->data, jl_array_len(bytes), 0);
     src.size = jl_array_len(bytes);
     int en = jl_gc_is_enabled();
     jl_gc_disable();
@@ -957,7 +1248,7 @@ void jl_init_serializer(void)
                      jl_symbol("arrayset"), jl_symbol("arrayref"),
                      jl_symbol("arraylen"), jl_symbol("boundscheck"),
                      jl_symbol("convert"), jl_symbol("typeassert"),
-                     jl_symbol("getfield"), jl_symbol("setfield"),
+                     jl_symbol("getfield"), jl_symbol("setfield!"),
                      jl_symbol("tupleref"), jl_symbol("tuplelen"),
                      jl_symbol("apply_type"), tuple_sym,
 
@@ -982,8 +1273,7 @@ void jl_init_serializer(void)
                      jl_box_int32(51), jl_box_int32(52), jl_box_int32(53),
                      jl_box_int32(54), jl_box_int32(55), jl_box_int32(56),
                      jl_box_int32(57), jl_box_int32(58), jl_box_int32(59),
-                     jl_box_int32(60), jl_box_int32(61), jl_box_int32(62),
-                     jl_box_int32(63), jl_box_int32(64),
+                     jl_box_int32(60), jl_box_int32(61),
 #endif
                      jl_box_int64(0), jl_box_int64(1), jl_box_int64(2),
                      jl_box_int64(3), jl_box_int64(4), jl_box_int64(5),
@@ -1006,30 +1296,28 @@ void jl_init_serializer(void)
                      jl_box_int64(51), jl_box_int64(52), jl_box_int64(53),
                      jl_box_int64(54), jl_box_int64(55), jl_box_int64(56),
                      jl_box_int64(57), jl_box_int64(58), jl_box_int64(59),
-                     jl_box_int64(60), jl_box_int64(61), jl_box_int64(62),
-                     jl_box_int64(63), jl_box_int64(64),
+                     jl_box_int64(60), jl_box_int64(61),
 #endif
                      jl_labelnode_type, jl_linenumbernode_type,
                      jl_gotonode_type, jl_quotenode_type, jl_topnode_type,
                      jl_type_type, jl_bottom_type, jl_pointer_type,
                      jl_vararg_type, jl_ntuple_type, jl_abstractarray_type,
-                     jl_box_type, jl_typector_type, jl_undef_type, jl_top_type,
-                     jl_typename_type, jl_task_type, jl_uniontype_type,
-                     jl_typetype_type, jl_typetype_tvar, jl_ANY_flag,
-                     jl_array_any_type, jl_intrinsic_type, jl_method_type,
+                     jl_densearray_type, jl_box_type,
+                     jl_typector_type, jl_undef_type, jl_top_type, jl_typename_type,
+                     jl_task_type, jl_uniontype_type, jl_typetype_type, jl_typetype_tvar,
+                     jl_ANY_flag, jl_array_any_type, jl_intrinsic_type, jl_method_type,
                      jl_methtable_type, jl_voidpointer_type, jl_newvarnode_type,
                      jl_array_symbol_type, jl_tupleref(jl_tuple_type,0),
 
-                     jl_symbol_type->name, jl_pointer_type->name,
-                     jl_datatype_type->name, jl_uniontype_type->name,
-                     jl_array_type->name, jl_expr_type->name,
+                     jl_symbol_type->name, jl_pointer_type->name, jl_datatype_type->name,
+                     jl_uniontype_type->name, jl_array_type->name, jl_expr_type->name,
                      jl_typename_type->name, jl_type_type->name, jl_methtable_type->name,
-                     jl_method_type->name, jl_tvar_type->name,
-                     jl_vararg_type->name, jl_ntuple_type->name, jl_abstractarray_type->name,
-                     jl_lambda_info_type->name, jl_module_type->name,
-                     jl_box_type->name, jl_function_type->name,
-                     jl_typector_type->name, jl_intrinsic_type->name, jl_undef_type->name,
-                     jl_task_type->name,
+                     jl_method_type->name, jl_tvar_type->name, jl_vararg_type->name,
+                     jl_ntuple_type->name, jl_abstractarray_type->name,
+                     jl_densearray_type->name,
+                     jl_lambda_info_type->name, jl_module_type->name, jl_box_type->name,
+                     jl_function_type->name, jl_typector_type->name,
+                     jl_intrinsic_type->name, jl_undef_type->name, jl_task_type->name,
                      jl_labelnode_type->name, jl_linenumbernode_type->name,
                      jl_gotonode_type->name, jl_quotenode_type->name,
                      jl_topnode_type->name,
@@ -1046,31 +1334,35 @@ void jl_init_serializer(void)
     assert(i <= Null_tag);
     VALUE_TAGS = (ptrint_t)ptrhash_get(&ser_tag, jl_null);
 
-    void *fptrs[] = { jl_f_new_expr, jl_f_new_box,
-                      jl_f_throw, jl_f_is, 
-                      jl_f_no_function, jl_f_typeof, 
-                      jl_f_subtype, jl_f_isa, 
-                      jl_f_typeassert, jl_f_apply, 
-                      jl_f_top_eval, jl_f_isdefined, 
-                      jl_f_tuple, jl_f_tupleref, 
-                      jl_f_tuplelen, jl_f_get_field, 
-                      jl_f_set_field, jl_f_field_type, 
-                      jl_f_arraylen, jl_f_arrayref, 
-                      jl_f_arrayset, jl_f_arraysize, 
-                      jl_f_instantiate_type, jl_f_kwcall,
-                      jl_f_convert_default, jl_f_convert_tuple,
-                      jl_trampoline, jl_f_new_type_constructor, 
-                      jl_f_typevar, jl_f_union, 
-                      jl_f_methodexists, jl_f_applicable, 
-                      jl_f_invoke, jl_apply_generic, 
-                      jl_unprotect_stack, jl_f_task, 
-                      jl_f_yieldto, jl_f_ctor_trampoline,
-                      jl_f_new_module,
-                      NULL };
+    jl_fptr_t fptrs[] = { jl_f_new_expr, jl_f_new_box,
+                          jl_f_throw, jl_f_is,
+                          jl_f_no_function, jl_f_typeof,
+                          jl_f_subtype, jl_f_isa,
+                          jl_f_typeassert, jl_f_apply,
+                          jl_f_top_eval, jl_f_isdefined,
+                          jl_f_tuple, jl_f_tupleref,
+                          jl_f_tuplelen, jl_f_get_field,
+                          jl_f_set_field, jl_f_field_type,
+                          jl_f_arraylen, jl_f_arrayref,
+                          jl_f_arrayset, jl_f_arraysize,
+                          jl_f_instantiate_type, jl_f_kwcall,
+                          jl_f_convert_default,
+                          jl_trampoline, jl_f_new_type_constructor,
+                          jl_f_typevar, jl_f_union,
+                          jl_f_methodexists, jl_f_applicable,
+                          jl_f_invoke, jl_apply_generic,
+                          jl_unprotect_stack, jl_f_task,
+                          jl_f_yieldto, jl_f_ctor_trampoline,
+                          jl_f_new_module,
+                          NULL };
     i=2;
     while (fptrs[i-2] != NULL) {
-        ptrhash_put(&fptr_to_id, fptrs[i-2], (void*)i);
-        ptrhash_put(&id_to_fptr, (void*)i, fptrs[i-2]);
+        ptrhash_put(&fptr_to_id, (void*)fptrs[i-2], (void*)i);
+        ptrhash_put(&id_to_fptr, (void*)i, (void*)fptrs[i-2]);
         i += 1;
     }
 }
+
+#ifdef __cplusplus
+}
+#endif
