@@ -19,6 +19,10 @@
 #include <sys/time.h>
 #endif /* !_OS_WINDOWS_ */
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 size_t llength(value_t v)
 {
     size_t n = 0;
@@ -131,7 +135,7 @@ static value_t fl_symbol(value_t *args, u_int32_t nargs)
     argcount("symbol", nargs, 1);
     if (!fl_isstring(args[0]))
         type_error("symbol", "string", args[0]);
-    return symbol(cvalue_data(args[0]));
+    return symbol((char*)cvalue_data(args[0]));
 }
 
 static value_t fl_keywordp(value_t *args, u_int32_t nargs)
@@ -313,9 +317,10 @@ static value_t fl_path_cwd(value_t *args, uint32_t nargs)
         argcount("path.cwd", nargs, 1);
     if (nargs == 0) {
         char buf[1024];
-        err = uv_cwd(buf, sizeof(buf));
+        size_t len = sizeof(buf);
+        err = uv_cwd(buf, &len);
         if (err != 0)
-          lerrorf(IOError, "path.cwd: could not get cwd: %s", uv_strerror(err));
+            lerrorf(IOError, "path.cwd: could not get cwd: %s", uv_strerror(err));
         return string_from_cstr(buf);
     }
     char *ptr = tostring(args[0], "path.cwd");
@@ -418,3 +423,7 @@ void builtins_init(void)
     table_init();
     iostream_init();
 }
+
+#ifdef __cplusplus
+}
+#endif
