@@ -8,6 +8,9 @@
 #include "libsupport.h"
 #include "../../deps/libuv/include/uv.h"
 
+//#define MEMDEBUG
+//#define MEMDEBUG2
+
 typedef uptrint_t value_t;
 typedef int_t fixnum_t;
 #if NBITS==64
@@ -17,6 +20,10 @@ typedef int_t fixnum_t;
 #define T_FIXNUM T_INT32
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 typedef struct {
     value_t car;
     value_t cdr;
@@ -107,8 +114,12 @@ typedef struct {
 #define isconstant(s) ((s)->flags&0x1)
 #define iskeyword(s) ((s)->flags&0x2)
 #define symbol_value(s) (((symbol_t*)ptr(s))->binding)
+#ifdef MEMDEBUG2
+#define ismanaged(v) (!issymbol(v) && !isfixnum(v) && ((v)>(N_OPCODES<<3)) && !iscbuiltin(v))
+#else
 #define ismanaged(v) ((((unsigned char*)ptr(v)) >= fromspace) && \
                       (((unsigned char*)ptr(v)) < fromspace+heapsize))
+#endif
 #define isgensym(x)  (issymbol(x) && ismanaged(x))
 
 #define isfunction(x) (tag(x) == TAG_FUNCTION && (x) > (N_BUILTINS<<3))
@@ -147,7 +158,7 @@ value_t fl_list2(value_t a, value_t b);
 value_t fl_listn(size_t n, ...);
 value_t symbol(char *str);
 char *symbol_name(value_t v);
-int fl_is_keyword_name(char *str, size_t len);
+int fl_is_keyword_name(const char *str, size_t len);
 value_t alloc_vector(size_t n, int init);
 size_t llength(value_t v);
 value_t fl_compare(value_t a, value_t b);  // -1, 0, or 1
@@ -376,4 +387,12 @@ value_t cvalue_wchar(value_t *args, uint32_t nargs);
 void fl_init(size_t initial_heapsize);
 int fl_load_system_image(value_t ios);
 
+/* julia extensions */
+DLLEXPORT int jl_id_char(uint32_t wc);
+DLLEXPORT int jl_id_start_char(uint32_t wc);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif
