@@ -7,6 +7,10 @@
 #include <setjmp.h>
 #include "flisp.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 static value_t iostreamsym, rdsym, wrsym, apsym, crsym, truncsym;
 static value_t instrsym, outstrsym;
 fltype_t *iostreamtype;
@@ -27,11 +31,8 @@ void relocate_iostream(value_t oldv, value_t newv)
 {
     ios_t *olds = value2c(ios_t*, oldv);
     ios_t *news = value2c(ios_t*, newv);
-    cvalue_t *cv = (cvalue_t*)ptr(oldv);
-    if (isinlined(cv)) {
-        if (olds->buf == &olds->local[0]) {
-            news->buf = &news->local[0];
-        }
+    if (news->buf == &olds->local[0]) {
+        news->buf = &news->local[0];
     }
 }
 
@@ -217,7 +218,7 @@ value_t fl_ioseek(value_t *args, u_int32_t nargs)
     ios_t *s = toiostream(args[0], "io.seek");
     size_t pos = tosize(args[1], "io.seek");
     off_t res = ios_seek(s, (off_t)pos);
-    if (res == -1)
+    if (res < 0)
         return FL_F;
     return FL_T;
 }
@@ -265,7 +266,7 @@ value_t fl_ioread(value_t *args, u_int32_t nargs)
     }
     value_t cv = cvalue(ft, n);
     char *data;
-    if (iscvalue(cv)) data = cv_data((cvalue_t*)ptr(cv));
+    if (iscvalue(cv)) data = (char*)cv_data((cvalue_t*)ptr(cv));
     else data = cp_data((cprim_t*)ptr(cv));
     size_t got = ios_read(value2c(ios_t*,args[0]), data, n);
     if (got < n)
@@ -329,7 +330,7 @@ value_t fl_ioreaduntil(value_t *args, u_int32_t nargs)
     argcount("io.readuntil", nargs, 2);
     value_t str = cvalue_string(80);
     cvalue_t *cv = (cvalue_t*)ptr(str);
-    char *data = cv_data(cv);
+    char *data = (char*)cv_data(cv);
     ios_t dest;
     ios_mem(&dest, 0);
     ios_setbuf(&dest, data, 80, 0);
@@ -450,3 +451,7 @@ void iostream_init(void)
     setc(symbol("*stdin*" ), cvalue_from_ref(iostreamtype, ios_stdin,
                                              sizeof(ios_t), FL_NIL));
 }
+
+#ifdef __cplusplus
+}
+#endif
