@@ -39,7 +39,7 @@ int isnumtok_base(char *tok, value_t *pval, int base)
         return 0;
     if (!((tok[0]=='0' && tok[1]=='x') || (base >= 15)) &&
         strpbrk(tok, ".eEpP")) {
-        d = strtod(tok, &end);
+        d = strtod_c(tok, &end);
         if (*end == '\0') {
             if (pval) *pval = mk_double(d);
             return 1;
@@ -55,7 +55,7 @@ int isnumtok_base(char *tok, value_t *pval, int base)
     // hexadecimal float literals
     else if (((tok[0]=='0' && tok[1]=='x') || (base == 16)) &&
         strpbrk(tok, "pP")) {
-        d = strtod(tok, &end);
+        d = strtod_c(tok, &end);
         if (*end == '\0') {
             if (pval) *pval = mk_double(d);
             return 1;
@@ -149,7 +149,7 @@ static char nextchar(void)
             } while ((char)ch != '\n');
             c = (char)ch;
         }
-    } while (c==' ' || isspace(c));
+    } while (c==' ' || isspace((unsigned char)c));
     return c;
 }
 
@@ -463,11 +463,11 @@ static value_t read_string(void)
     value_t s;
     u_int32_t wc=0;
 
-    buf = malloc(sz);
+    buf = (char*)malloc(sz);
     while (1) {
         if (i >= sz-4) {  // -4: leaves room for longest utf8 sequence
             sz *= 2;
-            temp = realloc(buf, sz);
+            temp = (char*)realloc(buf, sz);
             if (temp == NULL) {
                 free(buf);
                 lerror(ParseError, "read: out of memory reading string");
@@ -607,10 +607,14 @@ static value_t do_read_sexpr(value_t label)
     case TOK_QUOTE:
         head = &QUOTE;
     listwith:
+#ifdef MEMDEBUG2
+        v = fl_list2(*head, NIL);
+#else
         v = cons_reserve(2);
         car_(v) = *head;
         cdr_(v) = tagptr(((cons_t*)ptr(v))+1, TAG_CONS);
         car_(cdr_(v)) = cdr_(cdr_(v)) = NIL;
+#endif
         PUSH(v);
         if (label != UNBOUND)
             ptrhash_put(&readstate->backrefs, (void*)label, (void*)v);
