@@ -3,31 +3,36 @@
   . non-moving, precise mark and sweep collector
   . pool-allocates small objects, keeps big objects on a simple list
 */
+// use mmap instead of malloc to allocate pages. default = off.
+//#define USE_MMAP
+
+// free pages as soon as they are empty. if not defined, then we
+// will wait for the next GC, to allow the space to be reused more
+// efficiently. default = on.
+#define FREE_PAGES_EAGER
+
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
+#ifdef USE_MMAP
+# include <sys/mman.h>
+# include <malloc.h>
+#endif
 #include "julia.h"
-
-// with MEMDEBUG, every object is allocated explicitly with malloc, and
-// filled with 0xbb before being freed.
-//#define MEMDEBUG
-
-// MEMPROFILE prints pool summary statistics after every GC
-//#define MEMPROFILE
-
-// GCTIME prints time taken by each phase of GC
-//#define GCTIME
-
-// GC_FINAL_STATS prints total GC stats at exit
-// set in julia.h
-
-// OBJPROFILE counts objects by type
-//#define OBJPROFILE
+#include "julia_internal.h"
 
 #ifdef _P64
-#define GC_PAGE_SZ (1536*sizeof(void*))//bytes
+# ifdef USE_MMAP
+#  define GC_PAGE_SZ 16384//bytes
+# else
+#  define GC_PAGE_SZ 12288//bytes
+# endif
 #else
-#define GC_PAGE_SZ (2048*sizeof(void*))//bytes
+#define GC_PAGE_SZ  8192//bytes
+#endif
+
+#ifdef __cplusplus
+extern "C" {
 #endif
 
 typedef struct _gcpage_t {
@@ -75,14 +80,17 @@ typedef struct _bigval_t {
 // GC knobs and self-measurement variables
 static size_t allocd_bytes = 0;
 static int64_t total_allocd_bytes = 0;
+static int64_t last_gc_total_bytes = 0;
 static size_t freed_bytes = 0;
-#define default_collect_interval (3200*1024*sizeof(void*))
-static size_t collect_interval = default_collect_interval;
+static uint64_t total_gc_time=0;
 #ifdef _P64
+#define default_collect_interval (5600*1024*sizeof(void*))
 static size_t max_collect_interval = 1250000000UL;
 #else
-static size_t max_collect_interval = 500000000UL;
+#define default_collect_interval (3200*1024*sizeof(void*))
+static size_t max_collect_interval =  500000000UL;
 #endif
+static size_t collect_interval = default_collect_interval;
 int jl_in_gc; // referenced from switchto task.c
 
 #ifdef OBJPROFILE
@@ -90,7 +98,6 @@ static htable_t obj_counts;
 #endif
 
 #ifdef GC_FINAL_STATS
-static double total_gc_time=0;
 static size_t total_freed_bytes=0;
 #endif
 
@@ -286,7 +293,7 @@ static void run_finalizer(jl_value_t *o, jl_value_t *ff)
         }
         JL_CATCH {
             JL_PRINTF(JL_STDERR, "error in running finalizer: ");
-            jl_show(jl_stderr_obj(), jl_exception_in_transit);
+            jl_static_show(JL_STDERR, jl_exception_in_transit);
             JL_PUTC('\n',JL_STDERR);
         }
         ff = jl_t1(ff);
@@ -298,7 +305,7 @@ static void run_finalizer(jl_value_t *o, jl_value_t *ff)
     }
     JL_CATCH {
         JL_PRINTF(JL_STDERR, "error in running finalizer: ");
-        jl_show(jl_stderr_obj(), jl_exception_in_transit);
+        jl_static_show(JL_STDERR, jl_exception_in_transit);
         JL_PUTC('\n',JL_STDERR);
     }
 }
@@ -313,15 +320,16 @@ static void run_finalizers(void)
         ff = (jl_value_t*)ptrhash_get(&finalizer_table, o);
         assert(ff != HT_NOTFOUND);
         ptrhash_remove(&finalizer_table, o);
-        run_finalizer(o, ff);
+        run_finalizer((jl_value_t*)o, ff);
     }
     JL_GC_POP();
 }
 
-void jl_gc_run_all_finalizers()
+void jl_gc_run_all_finalizers(void)
 {
     for(size_t i=0; i < finalizer_table.size; i+=2) {
-        if (finalizer_table.table[i+1] != HT_NOTFOUND) {
+        jl_value_t *f = (jl_value_t*)finalizer_table.table[i+1];
+        if (f != HT_NOTFOUND && !jl_is_cpointer(f)) {
             schedule_finalization(finalizer_table.table[i]);
         }
     }
@@ -401,7 +409,7 @@ void jl_gc_track_malloced_array(jl_array_t *a)
 {
     mallocarray_t *ma;
     if (mafreelist == NULL) {
-        ma = malloc(sizeof(mallocarray_t));
+        ma = (mallocarray_t*)malloc(sizeof(mallocarray_t));
     }
     else {
         ma = mafreelist;
@@ -461,7 +469,12 @@ static pool_t *pools = &norm_pools[0];
 
 static void add_page(pool_t *p)
 {
-    gcpage_t *pg = malloc_a16(sizeof(gcpage_t));
+#ifdef USE_MMAP
+    gcpage_t *pg = (gcpage_t*)mmap(NULL, sizeof(gcpage_t), PROT_READ|PROT_WRITE,
+                                   MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);
+#else
+    gcpage_t *pg = (gcpage_t*)malloc_a16(sizeof(gcpage_t));
+#endif
     if (pg == NULL)
         jl_throw(jl_memory_exception);
     gcval_t *v = (gcval_t*)&pg->data[0];
@@ -514,8 +527,11 @@ static int szclass(size_t sz)
 
 static void sweep_pool(pool_t *p)
 {
-    //int empty;
+#ifdef FREE_PAGES_EAGER
     int freedall;
+#else
+    int empty;
+#endif
     gcval_t **prev_pfl;
     gcval_t *v;
     gcpage_t *pg = p->pages;
@@ -534,18 +550,32 @@ static void sweep_pool(pool_t *p)
     while (pg != NULL) {
         v = (gcval_t*)&pg->data[0];
         char *lim = (char*)v + GC_PAGE_SZ - osize;
-        //empty = 1;
+#ifdef FREE_PAGES_EAGER
         freedall = 1;
+#else
+        empty = 1;
+#endif
         prev_pfl = pfl;
         while ((char*)v <= lim) {
             if (!v->marked) {
+#ifndef FREE_PAGES_EAGER
+                // check that all but last object points to its next object,
+                // which is a heuristic check for being on the freelist.
+                if ((char*)v->next != (char*)v + osize && v->next != NULL &&
+                    (char*)v+osize <= lim)
+                    empty = 0;
+#endif
                 *pfl = v;
                 pfl = &v->next;
                 nfreed++;
             }
             else {
                 v->marked = 0;
+#ifdef FREE_PAGES_EAGER
                 freedall = 0;
+#else
+                empty = 0;
+#endif
             }
             v = (gcval_t*)((char*)v + osize);
         }
@@ -553,13 +583,23 @@ static void sweep_pool(pool_t *p)
         // lazy version: (empty) if the whole page was already unused, free it
         // eager version: (freedall) free page as soon as possible
         // the eager one uses less memory.
-        if (freedall) {
+        if (
+#ifdef FREE_PAGES_EAGER
+            freedall
+#else
+            empty
+#endif
+            ) {
             pfl = prev_pfl;
             *ppg = nextpg;
 #ifdef MEMDEBUG
             memset(pg, 0xbb, sizeof(gcpage_t));
 #endif
+#ifdef USE_MMAP
+            munmap(pg, sizeof(gcpage_t));
+#else
             free_a16(pg);
+#endif
             //freed_bytes += GC_PAGE_SZ;
         }
         else {
@@ -595,7 +635,7 @@ static size_t mark_sp = 0;
 
 static void push_root(jl_value_t *v, int d);
 
-#define gc_push_root(v,d) if (!gc_marked(v)) push_root((jl_value_t*)(v),d);
+#define gc_push_root(v,d) do {  assert(v != NULL); if (!gc_marked(v)) { push_root((jl_value_t*)(v),d); } } while (0)
 
 void jl_gc_setmark(jl_value_t *v)
 {
@@ -639,6 +679,13 @@ static void gc_mark_module(jl_module_t *m, int d)
                 gc_push_root(b->type, d);
         }
     }
+    // this is only necessary because bindings for "using" modules
+    // are added only when accessed. therefore if a module is replaced
+    // after "using" it but before accessing it, this array might
+    // contain the only reference.
+    for(i=0; i < m->usings.len; i++) {
+        gc_push_root(m->usings.items[i], d);
+    }
     if (m->constant_table)
         gc_push_root(m->constant_table, d);
 }
@@ -646,7 +693,7 @@ static void gc_mark_module(jl_module_t *m, int d)
 static void gc_mark_task(jl_task_t *ta, int d)
 {
     if (ta->parent) gc_push_root(ta->parent, d);
-    gc_push_root(ta->last, d);
+    if (ta->last) gc_push_root(ta->last, d);
     gc_push_root(ta->tls, d);
     gc_push_root(ta->consumers, d);
     gc_push_root(ta->donenotify, d);
@@ -678,7 +725,7 @@ static jl_value_t *lookforme = NULL;
 DLLEXPORT void jl_gc_lookfor(jl_value_t *v) { lookforme = v; }
 */
 
-#define MAX_MARK_DEPTH 1000
+#define MAX_MARK_DEPTH 400
 
 static void push_root(jl_value_t *v, int d)
 {
@@ -786,28 +833,9 @@ extern jl_value_t * volatile jl_task_arg_in_transit;
 double clock_now(void);
 #endif
 
-static void gc_mark_uv_handle(uv_handle_t *handle, void *arg)
-{
-    if (handle->data) {
-        gc_push_root((jl_value_t*)(handle->data), 0);
-    }
-}
-
-#include "../deps/libuv/src/queue.h"
-
-static void gc_mark_uv_state(uv_loop_t *loop)
-{
-    QUEUE *q;
-    uv_walk(loop,gc_mark_uv_handle,0);
-    QUEUE_FOREACH(q,&loop->active_reqs)
-    {
-        uv_req_t *req = QUEUE_DATA(q,uv_req_t,active_queue);
-        if (req->data)
-            gc_push_root((jl_value_t*)(req->data), 0);
-    }
-}
-
 extern jl_module_t *jl_old_base_module;
+extern jl_array_t *typeToTypeId;
+extern jl_array_t *jl_module_init_order;
 
 static void gc_mark(void)
 {
@@ -819,6 +847,7 @@ static void gc_mark(void)
 
     // modules
     gc_push_root(jl_main_module, 0);
+    gc_push_root(jl_internal_main_module, 0);
     gc_push_root(jl_current_module, 0);
     if (jl_old_base_module) gc_push_root(jl_old_base_module, 0);
 
@@ -830,15 +859,15 @@ static void gc_mark(void)
     gc_push_root(jl_bottom_func, 0);
     gc_push_root(jl_typetype_type, 0);
     gc_push_root(jl_tupletype_type, 0);
+    gc_push_root(typeToTypeId, 0);
+    if (jl_module_init_order != NULL)
+        gc_push_root(jl_module_init_order, 0);
 
     // constants
     gc_push_root(jl_null, 0);
     gc_push_root(jl_true, 0);
     gc_push_root(jl_false, 0);
 
-    // libuv loops
-    gc_mark_uv_state(jl_global_event_loop());
-
     jl_mark_box_caches();
 
     size_t i;
@@ -859,8 +888,16 @@ static void gc_mark(void)
     // this must happen last.
     for(i=0; i < finalizer_table.size; i+=2) {
         if (finalizer_table.table[i+1] != HT_NOTFOUND) {
-            jl_value_t *v = finalizer_table.table[i];
+            jl_value_t *v = (jl_value_t*)finalizer_table.table[i];
             if (!gc_marked(v)) {
+                jl_value_t *fin = (jl_value_t*)finalizer_table.table[i+1];
+                if (gc_typeof(fin) == (jl_value_t*)jl_voidpointer_type) {
+                    void *p = ((void**)fin)[1];
+                    if (p)
+                        ((void (*)(void*))p)(jl_data_ptr(v));
+                    finalizer_table.table[i+1] = HT_NOTFOUND;
+                    continue;
+                }
                 gc_push_root(v, 0);
                 schedule_finalization(v);
             }
@@ -879,6 +916,16 @@ DLLEXPORT void jl_gc_disable(void)   { is_gc_enabled = 0; }
 DLLEXPORT int jl_gc_is_enabled(void) { return is_gc_enabled; }
 
 DLLEXPORT int64_t jl_gc_total_bytes(void) { return total_allocd_bytes + allocd_bytes; }
+DLLEXPORT uint64_t jl_gc_total_hrtime(void) { return total_gc_time; }
+
+int64_t diff_gc_total_bytes(void)
+{
+    int64_t oldtb = last_gc_total_bytes;
+    int64_t newtb = jl_gc_total_bytes();
+    last_gc_total_bytes = newtb;
+    return newtb - oldtb;
+}
+void sync_gc_total_bytes(void) {last_gc_total_bytes = jl_gc_total_bytes();}
 
 void jl_gc_ephemeral_on(void)  { pools = &ephe_pools[0]; }
 void jl_gc_ephemeral_off(void) { pools = &norm_pools[0]; }
@@ -909,31 +956,29 @@ void jl_gc_collect(void)
     if (is_gc_enabled) {
         JL_SIGATOMIC_BEGIN();
         jl_in_gc = 1;
-#if defined(GCTIME) || defined(GC_FINAL_STATS)
-        double t0 = clock_now();
-#endif
+        uint64_t t0 = jl_hrtime();
         gc_mark();
 #ifdef GCTIME
-        JL_PRINTF(JL_STDERR, "mark time %.3f ms\n", (clock_now()-t0)*1000);
+        JL_PRINTF(JL_STDERR, "mark time %.3f ms\n", (jl_hrtime()-t0)*1.0e6);
 #endif
 #if defined(MEMPROFILE)
         all_pool_stats();
         big_obj_stats();
 #endif
 #ifdef GCTIME
-        t0 = clock_now();
+        uint64_t t1 = jl_hrtime();
 #endif
         sweep_weak_refs();
         gc_sweep();
 #ifdef GCTIME
-        JL_PRINTF(JL_STDERR, "sweep time %.3f ms\n", (clock_now()-t0)*1000);
+        JL_PRINTF(JL_STDERR, "sweep time %.3f ms\n", (jl_hrtime()-t1)*1.0e6);
 #endif
         int nfinal = to_finalize.len;
         run_finalizers();
         jl_in_gc = 0;
         JL_SIGATOMIC_END();
+        total_gc_time += (jl_hrtime()-t0);
 #if defined(GC_FINAL_STATS)
-        total_gc_time += (clock_now()-t0);
         total_freed_bytes += freed_bytes;
 #endif
 #ifdef OBJPROFILE
@@ -981,7 +1026,7 @@ void *allocb(size_t sz)
     return (void*)((void**)b + 1);
 }
 
-void *allocobj(size_t sz)
+DLLEXPORT void *allocobj(size_t sz)
 {
 #ifdef MEMDEBUG
     return alloc_big(sz);
@@ -991,7 +1036,7 @@ void *allocobj(size_t sz)
     return pool_alloc(&pools[szclass(sz)]);
 }
 
-void *alloc_2w(void)
+DLLEXPORT void *alloc_2w(void)
 {
 #ifdef MEMDEBUG
     return alloc_big(2*sizeof(void*));
@@ -1003,7 +1048,7 @@ void *alloc_2w(void)
 #endif
 }
 
-void *alloc_3w(void)
+DLLEXPORT void *alloc_3w(void)
 {
 #ifdef MEMDEBUG
     return alloc_big(3*sizeof(void*));
@@ -1015,7 +1060,7 @@ void *alloc_3w(void)
 #endif
 }
 
-void *alloc_4w(void)
+DLLEXPORT void *alloc_4w(void)
 {
 #ifdef MEMDEBUG
     return alloc_big(4*sizeof(void*));
@@ -1032,16 +1077,15 @@ static double process_t0;
 #include <malloc.h>
 void jl_print_gc_stats(JL_STREAM *s)
 {
+    double gct = total_gc_time/1e9;
     malloc_stats();
     double ptime = clock_now()-process_t0;
     jl_printf(s, "exec time\t%.5f sec\n", ptime);
-    jl_printf(s, "gc time  \t%.5f sec (%2.1f%%)\n", total_gc_time,
-              (total_gc_time/ptime)*100);
+    jl_printf(s, "gc time  \t%.5f sec (%2.1f%%)\n", gct, (gct/ptime)*100);
     struct mallinfo mi = mallinfo();
     jl_printf(s, "malloc size\t%d MB\n", mi.uordblks/1024/1024);
     jl_printf(s, "total freed\t%llu b\n", total_freed_bytes);
-    jl_printf(s, "free rate\t%.1f MB/sec\n",
-              (total_freed_bytes/total_gc_time)/1024/1024);
+    jl_printf(s, "free rate\t%.1f MB/sec\n", (total_freed_bytes/gct)/1024/1024);
 }
 #endif
 
@@ -1171,3 +1215,7 @@ static void big_obj_stats(void)
     JL_PRINTF(JL_STDOUT, "%d bytes in %d large objects\n", nbytes, nused);
 }
 #endif //MEMPROFILE
+
+#ifdef __cplusplus
+}
+#endif
