@@ -13,24 +13,21 @@
 #include <malloc.h>
 #endif
 #include "julia.h"
-#include "builtin_proto.h"
+#include "julia_internal.h"
 
-#define ENABLE_INFERENCE
-
-// debugging options
-//#define TRACE_INFERENCE
-//#define JL_TRACE
-//#define JL_GF_PROFILE
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 static jl_methtable_t *new_method_table(jl_sym_t *name)
 {
     jl_methtable_t *mt = (jl_methtable_t*)allocobj(sizeof(jl_methtable_t));
     mt->type = (jl_value_t*)jl_methtable_type;
     mt->name = name;
-    mt->defs = JL_NULL;
-    mt->cache = JL_NULL;
-    mt->cache_arg1 = JL_NULL;
-    mt->cache_targ = JL_NULL;
+    mt->defs = (jl_methlist_t*)JL_NULL;
+    mt->cache = (jl_methlist_t*)JL_NULL;
+    mt->cache_arg1 = (jl_array_t*)JL_NULL;
+    mt->cache_targ = (jl_array_t*)JL_NULL;
     mt->max_args = 0;
     mt->kwsorter = NULL;
 #ifdef JL_GF_PROFILE
@@ -122,8 +119,7 @@ static inline int cache_match(jl_value_t **args, size_t n, jl_tuple_t *sig,
             // signatures for tuples of every length
             if (!jl_is_tuple(a) || //!jl_subtype(a, decl, 1))
                 !jl_tuple_subtype(((jl_tuple_t*)a)->data, jl_tuple_len(a),
-                                  ((jl_tuple_t*)decl)->data, jl_tuple_len(decl),
-                                  1, 0))
+                                  ((jl_tuple_t*)decl)->data, jl_tuple_len(decl), 1))
                 return 0;
         }
         else if (jl_is_type_type(decl) &&
@@ -158,7 +154,7 @@ jl_methlist_t *mtcache_hash_lookup(jl_array_t *a, jl_value_t *ty, int tparam)
         if (t == ty)
             return ml;
     }
-    return JL_NULL;
+    return (jl_methlist_t*)JL_NULL;
 }
 
 static void mtcache_rehash(jl_array_t **pa)
@@ -189,7 +185,7 @@ static jl_methlist_t **mtcache_hash_bp(jl_array_t **pa, jl_value_t *ty,
         while (1) {
             jl_methlist_t **pml = (jl_methlist_t**)&jl_cellref(*pa, uid & ((*pa)->nrows-1));
             if (*pml == NULL || *pml == JL_NULL) {
-                *pml = JL_NULL;
+                *pml = (jl_methlist_t*)JL_NULL;
                 return pml;
             }
             jl_value_t *t = jl_tupleref((*pml)->sig,0);
@@ -211,7 +207,7 @@ static jl_methlist_t **mtcache_hash_bp(jl_array_t **pa, jl_value_t *ty,
 static jl_function_t *jl_method_table_assoc_exact_by_type(jl_methtable_t *mt,
                                                           jl_tuple_t *types)
 {
-    jl_methlist_t *ml = JL_NULL;
+    jl_methlist_t *ml = (jl_methlist_t*)JL_NULL;
     if (jl_tuple_len(types) > 0) {
         jl_value_t *ty = jl_t0(types);
         if (jl_is_type_type(ty)) {
@@ -243,7 +239,7 @@ static jl_function_t *jl_method_table_assoc_exact(jl_methtable_t *mt,
                                                   jl_value_t **args, size_t n)
 {
     // NOTE: This function is a huge performance hot spot!!
-    jl_methlist_t *ml = JL_NULL;
+    jl_methlist_t *ml = (jl_methlist_t*)JL_NULL;
     if (n > 0) {
         jl_value_t *a0 = args[0];
         jl_value_t *ty = (jl_value_t*)jl_typeof(a0);
@@ -260,14 +256,16 @@ static jl_function_t *jl_method_table_assoc_exact(jl_methtable_t *mt,
                 if (n==2) {
                     // some manually-unrolled common special cases
                     jl_value_t *a1 = args[1];
-                    jl_methlist_t *mn = ml;
-                    if (jl_tuple_len(mn->sig)==2 &&
-                        jl_tupleref(mn->sig,1)==(jl_value_t*)jl_typeof(a1))
-                        return mn->func;
-                    mn = mn->next;
-                    if (mn!=JL_NULL && jl_tuple_len(mn->sig)==2 &&
-                        jl_tupleref(mn->sig,1)==(jl_value_t*)jl_typeof(a1))
-                        return mn->func;
+                    if (!jl_is_tuple(a1)) {  // issue #6426
+                        jl_methlist_t *mn = ml;
+                        if (jl_tuple_len(mn->sig)==2 &&
+                            jl_tupleref(mn->sig,1)==(jl_value_t*)jl_typeof(a1))
+                            return mn->func;
+                        mn = mn->next;
+                        if (mn!=JL_NULL && jl_tuple_len(mn->sig)==2 &&
+                            jl_tupleref(mn->sig,1)==(jl_value_t*)jl_typeof(a1))
+                            return mn->func;
+                    }
                 }
             }
         }
@@ -306,8 +304,6 @@ jl_lambda_info_t *jl_add_static_parameters(jl_lambda_info_t *l, jl_tuple_t *sp)
     return nli;
 }
 
-JL_CALLABLE(jl_trampoline);
-
 jl_function_t *jl_instantiate_method(jl_function_t *f, jl_tuple_t *sp)
 {
     if (f->linfo == NULL)
@@ -441,6 +437,19 @@ static int is_kind(jl_value_t *v)
             v==(jl_value_t*)jl_typector_type);
 }
 
+static int jl_is_specializable_tuple(jl_tuple_t *t)
+{
+    if (t == jl_null) return 1;
+    jl_value_t *e0 = jl_tupleref(t,0);
+    if (jl_is_tuple(e0) || e0 == (jl_value_t*)jl_datatype_type) return 0;
+    size_t i, l=jl_tuple_len(t);
+    // allow specialization on homogeneous tuples
+    for(i=1; i < l; i++) {
+        if (jl_tupleref(t,i) != e0) return 0;
+    }
+    return 1;
+}
+
 static jl_value_t *ml_matches(jl_methlist_t *ml, jl_value_t *type,
                               jl_sym_t *name, int lim);
 
@@ -458,6 +467,14 @@ static jl_function_t *cache_method(jl_methtable_t *mt, jl_tuple_t *type,
         jl_value_t *elt = jl_tupleref(type,i);
         jl_value_t *decl_i = nth_slot_type(decl,i);
         if (jl_is_type_type(elt) && jl_is_tuple(jl_tparam0(elt)) &&
+            /*
+              NOTE: without this, () is sometimes specialized as () and
+              sometimes as Type{()}. In #6624, this caused a
+                TypeError(func=:tuplelen, context="", expected=(Any...,), got=Type{()}())
+              inside ==, inside isstructtype. Not quite clear why, however.
+            */
+            jl_tparam0(elt) != (jl_value_t*)jl_null &&
+
             !jl_is_type_type(decl_i)) {
             jl_methlist_t *curr = mt->defs;
             int ok=1;
@@ -507,7 +524,7 @@ static jl_function_t *cache_method(jl_methtable_t *mt, jl_tuple_t *type,
         }
         if (set_to_any) {
         }
-        else if (jl_is_tuple(elt)) {
+        else if (jl_is_tuple(elt) && !jl_is_specializable_tuple((jl_tuple_t*)elt)) {
             /*
               don't cache tuple type exactly; just remember that it was
               a tuple, unless the declaration asks for something more
@@ -567,7 +584,9 @@ static jl_function_t *cache_method(jl_methtable_t *mt, jl_tuple_t *type,
                 while (curr != JL_NULL && curr->func!=method) {
                     jl_tuple_t *sig = curr->sig;
                     if (jl_tuple_len(sig) > i &&
-                        jl_is_tuple(jl_tupleref(sig,i))) {
+                        (jl_is_tuple(jl_tupleref(sig,i)) ||
+                         // tuples can also be Types (issue #5577)
+                         jl_subtype(jl_tupleref(sig,i), (jl_value_t*)jl_type_type, 0))) {
                         need_guard_entries = 1;
                         break;
                     }
@@ -832,8 +851,9 @@ static jl_value_t *lookup_match(jl_value_t *a, jl_value_t *b, jl_tuple_t **penv,
     jl_value_t *ti = jl_type_intersection_matching(a, b, penv, tvars);
     if (ti == (jl_value_t*)jl_bottom_type)
         return ti;
+    JL_GC_PUSH1(&ti);
     assert(jl_is_tuple(*penv));
-    jl_value_t **ee = alloca(sizeof(void*) * jl_tuple_len(*penv));
+    jl_value_t **ee = (jl_value_t**)alloca(sizeof(void*) * jl_tuple_len(*penv));
     int n=0;
     // only keep vars in tvars list
     jl_value_t **tvs;
@@ -846,21 +866,37 @@ static jl_value_t *lookup_match(jl_value_t *a, jl_value_t *b, jl_tuple_t **penv,
         tvs = &jl_t0(tvars);
         tvarslen = jl_tuple_len(tvars);
     }
-    for(int i=0; i < jl_tuple_len(*penv); i+=2) {
+    int l = jl_tuple_len(*penv);
+    for(int i=0; i < l; i+=2) {
         jl_value_t *v = jl_tupleref(*penv,i);
         jl_value_t *val = jl_tupleref(*penv,i+1);
         for(int j=0; j < tvarslen; j++) {
             if (v == tvs[j]) {
                 ee[n++] = v;
                 ee[n++] = val;
+                /*
+                  since "a" is a concrete type, we assume that
+                  (a∩b != None) => a<:b. However if a static parameter is
+                  forced to equal None, then part of "b" might become None,
+                  and therefore a subtype of "a". For example
+                  (Type{None},Int) ∩ (Type{T},T)
+                  issue #5254
+                */
+                if (val == (jl_value_t*)jl_bottom_type) {
+                    if (!jl_subtype(a, ti, 0)) {
+                        JL_GC_POP();
+                        return (jl_value_t*)jl_bottom_type;
+                    }
+                }
             }
         }
     }
-    if (n != jl_tuple_len(*penv)) {
+    if (n != l) {
         jl_tuple_t *en = jl_alloc_tuple_uninit(n);
         memcpy(en->data, ee, n*sizeof(void*));
         *penv = en;
     }
+    JL_GC_POP();
     return ti;
 }
 
@@ -900,7 +936,7 @@ static jl_function_t *jl_mt_assoc_by_type(jl_methtable_t *mt, jl_tuple_t *tt, in
         }
         else if (jl_tuple_subtype(&jl_tupleref(tt,0), nargs,
                                   &jl_tupleref(m->sig,0),
-                                  jl_tuple_len(m->sig), 0, 0)) {
+                                  jl_tuple_len(m->sig), 0)) {
             break;
         }
         m = m->next;
@@ -952,7 +988,7 @@ static int sigs_eq(jl_value_t *a, jl_value_t *b, int useenv)
 
 int jl_args_morespecific(jl_value_t *a, jl_value_t *b)
 {
-    int msp = jl_type_morespecific(a,b,0);
+    int msp = jl_type_morespecific(a,b);
     int btv = jl_has_typevars(b);
     if (btv) {
         if (jl_type_match_morespecific(a,b) == (jl_value_t*)jl_false) {
@@ -967,12 +1003,12 @@ int jl_args_morespecific(jl_value_t *a, jl_value_t *b)
             if (jl_type_match(b,a) == (jl_value_t*)jl_false)
                 return 1;
         }
-        int nmsp = jl_type_morespecific(b,a,0);
+        int nmsp = jl_type_morespecific(b,a);
         if (nmsp == msp)
             return 0;
     }
     if (jl_has_typevars((jl_value_t*)a)) {
-        int nmsp = jl_type_morespecific(b,a,0);
+        int nmsp = jl_type_morespecific(b,a);
         if (nmsp && msp)
             return 1;
         if (!btv && jl_types_equal(a,b))
@@ -1022,18 +1058,27 @@ static void check_ambiguous(jl_methlist_t *ml, jl_tuple_t *type,
         !jl_args_morespecific((jl_value_t*)sig, (jl_value_t*)type)) {
         jl_value_t *isect = jl_type_intersection((jl_value_t*)type,
                                                  (jl_value_t*)sig);
-        if (isect == (jl_value_t*)jl_bottom_type)
+        if (isect == (jl_value_t*)jl_bottom_type ||
+            // we're ok if the new definition is actually the one we just
+            // inferred to be required (see issue #3609). ideally this would
+            // never happen, since if New ⊓ Old == New then we should have
+            // considered New more specific, but jl_args_morespecific is not
+            // perfect, so this is a useful fallback.
+            sigs_eq(isect, (jl_value_t*)type, 1))
             return;
         JL_GC_PUSH1(&isect);
         jl_methlist_t *l = ml;
+        char *n;
+        jl_value_t *errstream;
+        JL_STREAM *s;
         while (l != JL_NULL) {
             if (sigs_eq(isect, (jl_value_t*)l->sig, 0))
                 goto done_chk_amb;  // ok, intersection is covered
             l = l->next;
         }
-        char *n = fname->name;
-        jl_value_t *errstream = jl_stderr_obj();
-        JL_STREAM *s = JL_STDERR;
+        n = fname->name;
+        errstream = jl_stderr_obj();
+        s = JL_STDERR;
         JL_PRINTF(s, "Warning: New definition \n    %s", n);
         jl_show(errstream, (jl_value_t*)type);
         print_func_loc(s, linfo);
@@ -1095,7 +1140,7 @@ jl_methlist_t *jl_method_list_insert(jl_methlist_t **pml, jl_tuple_t *type,
             l->va = (jl_tuple_len(type) > 0 &&
                      jl_is_vararg_type(jl_tupleref(type,jl_tuple_len(type)-1))) ?
                 1 : 0;
-            l->invokes = JL_NULL;
+            l->invokes = (struct _jl_methtable_t *)JL_NULL;
             l->func = method;
             JL_SIGATOMIC_END();
             return l;
@@ -1108,7 +1153,7 @@ jl_methlist_t *jl_method_list_insert(jl_methlist_t **pml, jl_tuple_t *type,
         if (jl_args_morespecific((jl_value_t*)type, (jl_value_t*)l->sig))
             break;
         if (check_amb) {
-            check_ambiguous(*pml, (jl_tuple_t*)type, l,
+            check_ambiguous(*pml, type, l,
                             method->linfo ? method->linfo->name :
                             anonymous_sym, method->linfo);
         }
@@ -1123,7 +1168,7 @@ jl_methlist_t *jl_method_list_insert(jl_methlist_t **pml, jl_tuple_t *type,
                   jl_is_vararg_type(jl_tupleref(type,jl_tuple_len(type)-1))) ?
         1 : 0;
     newrec->func = method;
-    newrec->invokes = JL_NULL;
+    newrec->invokes = (struct _jl_methtable_t*)JL_NULL;
     newrec->next = l;
     JL_SIGATOMIC_BEGIN();
     *pl = newrec;
@@ -1263,12 +1308,33 @@ jl_function_t *jl_method_lookup(jl_methtable_t *mt, jl_value_t **args, size_t na
     return sf;
 }
 
+void jl_add_constructors(jl_datatype_t *t);
+DLLEXPORT jl_value_t *jl_matching_methods(jl_function_t *gf, jl_value_t *type, int lim);
+
 // compile-time method lookup
 jl_function_t *jl_get_specialization(jl_function_t *f, jl_tuple_t *types)
 {
-    assert(jl_is_gf(f));
     if (!jl_is_leaf_type((jl_value_t*)types))
         return NULL;
+    if (f->fptr == jl_f_ctor_trampoline)
+        jl_add_constructors((jl_datatype_t*)f);
+    assert(jl_is_gf(f));
+
+    // make sure exactly 1 method matches (issue #7302).
+    int i;
+    for(i=0; i < jl_tuple_len(types); i++) {
+        jl_value_t *ti = jl_tupleref(types, i);
+        // if one argument type is DataType, multiple Type{} definitions
+        // might match. also be conservative with tuples rather than trying
+        // to analyze them in detail.
+        if (ti == (jl_value_t*)jl_datatype_type || jl_is_tuple(ti)) {
+            jl_value_t *matches = jl_matching_methods(f, (jl_value_t*)types, 1);
+            if (matches == jl_false)
+                return NULL;
+            break;
+        }
+    }
+
     jl_methtable_t *mt = jl_gf_mtable(f);
     jl_function_t *sf = jl_method_lookup_by_type(mt, types, 1, 1);
     if (sf == jl_bottom_func) {
@@ -1313,9 +1379,8 @@ JL_CALLABLE(jl_apply_generic)
     mt->ncalls++;
 #endif
 #ifdef JL_TRACE
-    if (trace_en) {
+    if (trace_en)
         show_call(F, args, nargs);
-    }
 #endif
     /*
       search order:
@@ -1327,6 +1392,7 @@ JL_CALLABLE(jl_apply_generic)
       otherwise instantiate the generic method and use it
     */
     jl_function_t *mfunc = jl_method_table_assoc_exact(mt, args, nargs);
+
     if (mfunc != jl_bottom_func) {
         if (mfunc->linfo != NULL && 
             (mfunc->linfo->inInference || mfunc->linfo->inCompile)) {
@@ -1337,26 +1403,31 @@ JL_CALLABLE(jl_apply_generic)
                 li->unspecialized = jl_instantiate_method(mfunc, li->sparams);
             }
             mfunc = li->unspecialized;
+            assert(mfunc != jl_bottom_func);
         }
+        assert(!mfunc->linfo || !mfunc->linfo->inInference);
+        return jl_apply(mfunc, args, nargs);
     }
-    else {
-        jl_tuple_t *tt = arg_type_tuple(args, nargs);
-        JL_GC_PUSH1(&tt);
-        mfunc = jl_mt_assoc_by_type(mt, tt, 1, 0);
-        JL_GC_POP();
-    }
+
+    // cache miss case
+    jl_tuple_t *tt = arg_type_tuple(args, nargs);
+    // if running inference overwrites this particular method, it becomes
+    // unreachable from the method table, so root mfunc.
+    JL_GC_PUSH2(&tt, &mfunc);
+    mfunc = jl_mt_assoc_by_type(mt, tt, 1, 0);
 
     if (mfunc == jl_bottom_func) {
 #ifdef JL_TRACE
-        if (error_en) {
+        if (error_en)
             show_call(F, args, nargs);
-        }
 #endif
+        JL_GC_POP();
         return jl_no_method_error((jl_function_t*)F, args, nargs);
     }
     assert(!mfunc->linfo || !mfunc->linfo->inInference);
-
-    return jl_apply(mfunc, args, nargs);
+    jl_value_t *res = jl_apply(mfunc, args, nargs);
+    JL_GC_POP();
+    return res;
 }
 
 // invoke()
@@ -1386,7 +1457,7 @@ jl_value_t *jl_gf_invoke(jl_function_t *gf, jl_tuple_t *types,
         }
         else if (jl_tuple_subtype(&jl_tupleref(types,0), typelen,
                                   &jl_tupleref(m->sig,0),
-                                  jl_tuple_len(m->sig), 0, 0)) {
+                                  jl_tuple_len(m->sig), 0)) {
             break;
         }
         m = m->next;
@@ -1465,45 +1536,6 @@ void print_func_loc(JL_STREAM *s, jl_lambda_info_t *li)
     }
 }
 
-static void print_methlist(jl_value_t *outstr, char *name, jl_methlist_t *ml)
-{
-    JL_STREAM *s = (JL_STREAM*)jl_iostr_data(outstr);
-    while (ml != JL_NULL) {
-        JL_PRINTF(s, "%s", name);
-        if (ml->tvars != jl_null) {
-            if (jl_is_typevar(ml->tvars)) {
-                JL_PUTC('{', s); jl_show(outstr, (jl_value_t*)ml->tvars);
-                JL_PUTC('}', s);
-            }
-            else {
-                jl_show_tuple(outstr, ml->tvars, '{', '}', 0);
-            }
-        }
-        jl_show(outstr, (jl_value_t*)ml->sig);
-        if (ml->func == jl_bottom_func)  {
-            // mark guard cache entries
-            JL_PRINTF(s, " *");
-        }
-        else {
-            jl_lambda_info_t *li = ml->func->linfo;
-            assert(li);
-            print_func_loc(s, li);
-        }
-        if (ml->next != JL_NULL)
-            JL_PRINTF(s, "\n");
-        ml = ml->next;
-    }
-}
-
-void jl_show_method_table(jl_value_t *outstr, jl_function_t *gf)
-{
-    char *name = jl_gf_name(gf)->name;
-    jl_methtable_t *mt = jl_gf_mtable(gf);
-    print_methlist(outstr, name, mt->defs);
-    //JL_PRINTF(JL_STDOUT, "\ncache:\n");
-    //print_methlist(outstr, name, mt->cache);
-}
-
 void jl_initialize_generic_function(jl_function_t *f, jl_sym_t *name)
 {
     f->fptr = jl_apply_generic;
@@ -1548,6 +1580,28 @@ DLLEXPORT jl_tuple_t *jl_match_method(jl_value_t *type, jl_value_t *sig,
     return result;
 }
 
+// Determine whether a typevar exists inside at most one DataType.
+// These are the typevars that will always be matched by any matching
+// arguments.
+static int tvar_exists_at_top_level(jl_value_t *tv, jl_tuple_t *sig, int attop)
+{
+    int i, l=jl_tuple_len(sig);
+    for(i=0; i < l; i++) {
+        jl_value_t *a = jl_tupleref(sig, i);
+        if (jl_is_vararg_type(a))
+            a = jl_tparam0(a);
+        if (a == tv)
+            return 1;
+        if (jl_is_tuple(a) && tvar_exists_at_top_level(tv, (jl_tuple_t*)a, attop))
+            return 1;
+        if (attop && jl_is_datatype(a)) {
+            if (tvar_exists_at_top_level(tv, ((jl_datatype_t*)a)->parameters, 0))
+                return 1;
+        }
+    }
+    return 0;
+}
+
 // returns a match as (argtypes, static_params, Method)
 static jl_value_t *ml_matches(jl_methlist_t *ml, jl_value_t *type,
                               jl_sym_t *name, int lim)
@@ -1584,6 +1638,13 @@ static jl_value_t *ml_matches(jl_methlist_t *ml, jl_value_t *type,
                         break;
                     }
                 }
+                // don't analyze slots declared with ANY
+                l = jl_tuple_len(ml->sig);
+                size_t m = jl_tuple_len(ti);
+                for(i=0; i < l && i < m; i++) {
+                    if (jl_tupleref(ml->sig, i) == jl_ANY_flag)
+                        jl_tupleset(ti, i, jl_any_type);
+                }
             }
             if (!skip) {
                 len++;
@@ -1602,7 +1663,10 @@ static jl_value_t *ml_matches(jl_methlist_t *ml, jl_value_t *type,
                 int matched_all_typevars = 1;
                 size_t l = jl_tuple_len(env);
                 for(i=1; i < l; i+=2) {
-                    if (jl_is_typevar(jl_tupleref(env,i))) {
+                    if (jl_is_typevar(jl_tupleref(env,i)) &&
+                        // if tvar is at the top level it will definitely be matched.
+                        // see issue #5575
+                        !tvar_exists_at_top_level(jl_tupleref(env,i-1), ml->sig, 1)) {
                         matched_all_typevars = 0;
                         break;
                     }
@@ -1630,9 +1694,6 @@ static jl_value_t *ml_matches(jl_methlist_t *ml, jl_value_t *type,
     return (jl_value_t*)t;
 }
 
-void jl_add_constructors(jl_datatype_t *t);
-JL_CALLABLE(jl_f_ctor_trampoline);
-
 // return a cell array of tuples, each describing a method match:
 // {(t, spvals, li, cenv), ...}
 // t is the intersection of the type argument and the method signature,
@@ -1655,3 +1716,7 @@ jl_value_t *jl_matching_methods(jl_function_t *gf, jl_value_t *type, int lim)
     jl_methtable_t *mt = jl_gf_mtable(gf);
     return ml_matches(mt->defs, type, jl_gf_name(gf), lim);
 }
+
+#ifdef __cplusplus
+}
+#endif
