@@ -17,6 +17,16 @@
 #include <unistd.h>
 #endif
 
+#if defined(__APPLE__)
+#include <AvailabilityMacros.h>
+#define __need_ucontext64_t
+#ifdef MAC_OS_X_VERSION_10_10
+#include <sys/_types/_ucontext64.h>
+#else
+#include <machine/_structs.h>
+#endif
+#endif
+
 #include <errno.h>
 #include <signal.h>
 
@@ -25,11 +35,16 @@
 #endif
 
 #include "julia.h"
+#include "julia_internal.h"
 #include <stdio.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #ifdef _OS_WINDOWS_
 #define WIN32_LEAN_AND_MEAN
-// Copied from MINGW_FLOAT_H which may not be found due to a colision with the builtin gcc float.h
+// Copied from MINGW_FLOAT_H which may not be found due to a collision with the builtin gcc float.h
 // eventually we can probably integrate this into OpenLibm.
 #if defined(_COMPILER_MINGW_)
 void __cdecl __MINGW_NOTHROW _fpreset (void);
@@ -52,14 +67,27 @@ void __cdecl fpreset (void);
 #include <windows.h>
 #include <dbghelp.h>
 extern int needsSymRefreshModuleList;
-extern WINBOOL WINAPI (*hSymRefreshModuleList)(HANDLE);
+extern BOOL (WINAPI *hSymRefreshModuleList)(HANDLE);
 #endif
 #if defined(__linux__)
 //#define _GNU_SOURCE
 #include <sched.h>   // for setting CPU affinity
 #endif
 
+DLLEXPORT void jlbacktrace();
+DLLEXPORT void gdbbacktrace();
+DLLEXPORT void gdblookup(ptrint_t ip);
+
+char *julia_home = NULL;
+jl_compileropts_t jl_compileropts = { NULL, // build_path
+                                      0,    // code_coverage
+                                      0,    // malloc_log
+                                      JL_COMPILEROPT_CHECK_BOUNDS_DEFAULT,
+                                      0     // int32_literals
+};
+
 int jl_boot_file_loaded = 0;
+int exit_on_sigint = 0;
 
 char *jl_stack_lo;
 char *jl_stack_hi;
@@ -80,18 +108,9 @@ static void jl_find_stack_bottom(void)
 }
 
 #ifdef _OS_WINDOWS_
-void __cdecl fpe_handler(int arg,int num)
-#else
-void fpe_handler(int arg)
-#endif
+void __cdecl fpe_handler(int arg, int num)
 {
     (void)arg;
-#ifndef _OS_WINDOWS_
-    sigset_t sset;
-    sigemptyset(&sset);
-    sigaddset(&sset, SIGFPE);
-    sigprocmask(SIG_UNBLOCK, &sset, NULL);
-#else
     fpreset();
     signal(SIGFPE, (void (__cdecl *)(int))fpe_handler);
     switch(num) {
@@ -102,44 +121,84 @@ void fpe_handler(int arg)
         jl_errorf("Unexpected FPE Error 0x%X", num);
         break;
     case _FPE_ZERODIVIDE:
-#endif
         jl_throw(jl_diverror_exception);
-#ifdef _OS_WINDOWS_
         break;
     }
+}
+#else
+void fpe_handler(int arg)
+{
+    (void)arg;
+    sigset_t sset;
+    sigemptyset(&sset);
+    sigaddset(&sset, SIGFPE);
+    sigprocmask(SIG_UNBLOCK, &sset, NULL);
+
+    jl_throw(jl_diverror_exception);
+}
+#endif
+
+#ifndef _OS_WINDOWS_
+static int is_addr_on_stack(void *addr)
+{
+#ifdef COPY_STACKS
+    return ((char*)addr > (char*)jl_stack_lo-3000000 &&
+            (char*)addr < (char*)jl_stack_hi);
+#else
+    return ((char*)addr > (char*)jl_current_task->stack-8192 &&
+            (char*)addr < (char*)jl_current_task->stack+jl_current_task->ssize);
 #endif
 }
 
+#ifndef SIGINFO
+#define SIGINFO SIGUSR1
+#endif
+
+void sigdie_handler(int sig, siginfo_t *info, void *context)
+{
+    if (sig != SIGINFO) {
+        sigset_t sset;
+        uv_tty_reset_mode();
+        sigfillset(&sset);
+        sigprocmask(SIG_UNBLOCK, &sset, NULL);
+        signal(sig, SIG_DFL);
+    }
+    ios_printf(ios_stderr,"\nsignal (%d): %s\n", sig, strsignal(sig));
+#ifdef __APPLE__
+    bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, (bt_context_t)&((ucontext64_t*)context)->uc_mcontext64->__ss);
+#else
+    bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, (ucontext_t*)context);
+#endif
+    jlbacktrace();
+    if (sig != SIGSEGV &&
+        sig != SIGBUS &&
+        sig != SIGILL &&
+        sig != SIGINFO) {
+        raise(sig);
+    }
+}
+#endif
+
 #if defined(__linux__) || defined(__FreeBSD__)
 extern int in_jl_;
 void segv_handler(int sig, siginfo_t *info, void *context)
 {
     sigset_t sset;
 
-    if ( in_jl_ || (
-#ifdef COPY_STACKS
-        (char*)info->si_addr > (char*)jl_stack_lo-3000000 &&
-        (char*)info->si_addr < (char*)jl_stack_hi
-#else
-        (char*)info->si_addr > (char*)jl_current_task->stack-8192 &&
-        (char*)info->si_addr <
-        (char*)jl_current_task->stack+jl_current_task->ssize
-#endif
-        )) {
+    if (sig == SIGSEGV && (in_jl_ || is_addr_on_stack(info->si_addr))) { // stack overflow
         sigemptyset(&sset);
         sigaddset(&sset, SIGSEGV);
         sigprocmask(SIG_UNBLOCK, &sset, NULL);
         jl_throw(jl_stackovf_exception);
     }
-    else {
-        uv_tty_reset_mode();
-        sigfillset(&sset);
+    else if (info->si_code == SEGV_ACCERR) {  // writing to read-only memory (e.g., mmap)
+        sigemptyset(&sset);
+        sigaddset(&sset, SIGSEGV);
         sigprocmask(SIG_UNBLOCK, &sset, NULL);
-        signal(sig, SIG_DFL);
-        if (sig != SIGSEGV &&
-            sig != SIGBUS &&
-            sig != SIGILL)
-            raise(sig);
+        jl_throw(jl_memory_exception);
+    }
+    else {
+        sigdie_handler(sig, info, context);
     }
 }
 #endif
@@ -148,11 +207,14 @@ volatile sig_atomic_t jl_signal_pending = 0;
 volatile sig_atomic_t jl_defer_signal = 0;
 
 #ifdef _OS_WINDOWS_
-void restore_signals()
+void restore_signals(void)
 {
-    SetConsoleCtrlHandler(NULL,0); //turn on ctrl-c handler
+    SetConsoleCtrlHandler(NULL, 0); //turn on ctrl-c handler
 }
-void jl_throw_in_ctx(jl_value_t* excpt, CONTEXT *ctxThread, int bt) {
+
+void jl_throw_in_ctx(jl_value_t *excpt, CONTEXT *ctxThread, int bt)
+{
+    assert(excpt != NULL);
     bt_size = bt ? rec_backtrace_ctx(bt_data, MAX_BT_SIZE, ctxThread) : 0;
     jl_exception_in_transit = excpt;
 #if defined(_CPU_X86_64_)
@@ -167,11 +229,12 @@ void jl_throw_in_ctx(jl_value_t* excpt, CONTEXT *ctxThread, int bt) {
 #error WIN16 not supported :P
 #endif
 }
+
 volatile HANDLE hMainThread = NULL;
-DLLEXPORT void jlbacktrace();
-DLLEXPORT void gdblookup(ptrint_t ip);
+
 static BOOL WINAPI sigint_handler(DWORD wsig) //This needs winapi types to guarantee __stdcall
 {
+    if (exit_on_sigint) jl_exit(0);
     int sig;
     //windows signals use different numbers from unix
     switch(wsig) {
@@ -213,76 +276,114 @@ static BOOL WINAPI sigint_handler(DWORD wsig) //This needs winapi types to guara
     }
     return 1;
 }
-static LONG WINAPI exception_handler(struct _EXCEPTION_POINTERS *ExceptionInfo) {
+
+static LONG WINAPI _exception_handler(struct _EXCEPTION_POINTERS *ExceptionInfo, int in_ctx)
+{
     if (ExceptionInfo->ExceptionRecord->ExceptionFlags == 0) {
         switch (ExceptionInfo->ExceptionRecord->ExceptionCode) {
-        case EXCEPTION_STACK_OVERFLOW:
-            jl_throw_in_ctx(jl_stackovf_exception, ExceptionInfo->ContextRecord, 0);
-            return EXCEPTION_CONTINUE_EXECUTION;
-        default:
-            ios_puts("Please submit a bug report with steps to reproduce this fault, and any error messages that follow (in their entirety). Thanks.\nException: ", ios_stderr);
-            switch (ExceptionInfo->ExceptionRecord->ExceptionCode) {
-                case EXCEPTION_ACCESS_VIOLATION:
-                    ios_puts("EXCEPTION_ACCESS_VIOLATION", ios_stderr); break;
-                case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
-                    ios_puts("EXCEPTION_ARRAY_BOUNDS_EXCEEDED", ios_stderr); break;
-                case EXCEPTION_BREAKPOINT:
-                    ios_puts("EXCEPTION_BREAKPOINT", ios_stderr); break;
-                case EXCEPTION_DATATYPE_MISALIGNMENT:
-                    ios_puts("EXCEPTION_DATATYPE_MISALIGNMENT", ios_stderr); break;
-                case EXCEPTION_FLT_DENORMAL_OPERAND:
-                    ios_puts("EXCEPTION_FLT_DENORMAL_OPERAND", ios_stderr); break;
-                case EXCEPTION_FLT_DIVIDE_BY_ZERO:
-                    ios_puts("EXCEPTION_FLT_DIVIDE_BY_ZERO", ios_stderr); break;
-                case EXCEPTION_FLT_INEXACT_RESULT:
-                    ios_puts("EXCEPTION_FLT_INEXACT_RESULT", ios_stderr); break;
-                case EXCEPTION_FLT_INVALID_OPERATION:
-                    ios_puts("EXCEPTION_FLT_INVALID_OPERATION", ios_stderr); break;
-                case EXCEPTION_FLT_OVERFLOW:
-                    ios_puts("EXCEPTION_FLT_OVERFLOW", ios_stderr); break;
-                case EXCEPTION_FLT_STACK_CHECK:
-                    ios_puts("EXCEPTION_FLT_STACK_CHECK", ios_stderr); break;
-                case EXCEPTION_FLT_UNDERFLOW:
-                    ios_puts("EXCEPTION_FLT_UNDERFLOW", ios_stderr); break;
-                case EXCEPTION_ILLEGAL_INSTRUCTION:
-                    ios_puts("EXCEPTION_ILLEGAL_INSTRUCTION", ios_stderr); break;
-                case EXCEPTION_IN_PAGE_ERROR:
-                    ios_puts("EXCEPTION_IN_PAGE_ERROR", ios_stderr); break;
-                case EXCEPTION_INT_DIVIDE_BY_ZERO:
-                    ios_puts("EXCEPTION_INT_DIVIDE_BY_ZERO", ios_stderr); break;
-                case EXCEPTION_INT_OVERFLOW:
-                    ios_puts("EXCEPTION_INT_OVERFLOW", ios_stderr); break;
-                case EXCEPTION_INVALID_DISPOSITION:
-                    ios_puts("EXCEPTION_INVALID_DISPOSITION", ios_stderr); break;
-                case EXCEPTION_NONCONTINUABLE_EXCEPTION:
-                    ios_puts("EXCEPTION_NONCONTINUABLE_EXCEPTION", ios_stderr); break;
-                case EXCEPTION_PRIV_INSTRUCTION:
-                    ios_puts("EXCEPTION_PRIV_INSTRUCTION", ios_stderr); break;
-                case EXCEPTION_SINGLE_STEP:
-                    ios_puts("EXCEPTION_SINGLE_STEP", ios_stderr); break;
-                case EXCEPTION_STACK_OVERFLOW:
-                    ios_puts("EXCEPTION_STACK_OVERFLOW", ios_stderr); break;
-                default:
-                    ios_puts("UNKNOWN", ios_stderr); break;
-            }
-            ios_printf(ios_stderr," at 0x%Ix -- ", (size_t)ExceptionInfo->ExceptionRecord->ExceptionAddress);
-            gdblookup((ptrint_t)ExceptionInfo->ExceptionRecord->ExceptionAddress);
-            bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, ExceptionInfo->ContextRecord);
-            jlbacktrace();
-            break;
+            case EXCEPTION_INT_DIVIDE_BY_ZERO:
+                fpreset();
+                jl_throw_in_ctx(jl_diverror_exception, ExceptionInfo->ContextRecord,in_ctx);
+                return EXCEPTION_CONTINUE_EXECUTION;
+            case EXCEPTION_STACK_OVERFLOW:
+                bt_size = 0;
+                jl_throw_in_ctx(jl_stackovf_exception, ExceptionInfo->ContextRecord,0);
+                return EXCEPTION_CONTINUE_EXECUTION;
         }
+        ios_puts("\nPlease submit a bug report with steps to reproduce this fault, and any error messages that follow (in their entirety). Thanks.\nException: ", ios_stderr);
+        switch (ExceptionInfo->ExceptionRecord->ExceptionCode) {
+            case EXCEPTION_ACCESS_VIOLATION:
+                ios_puts("EXCEPTION_ACCESS_VIOLATION", ios_stderr); break;
+            case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
+                ios_puts("EXCEPTION_ARRAY_BOUNDS_EXCEEDED", ios_stderr); break;
+            case EXCEPTION_BREAKPOINT:
+                ios_puts("EXCEPTION_BREAKPOINT", ios_stderr); break;
+            case EXCEPTION_DATATYPE_MISALIGNMENT:
+                ios_puts("EXCEPTION_DATATYPE_MISALIGNMENT", ios_stderr); break;
+            case EXCEPTION_FLT_DENORMAL_OPERAND:
+                ios_puts("EXCEPTION_FLT_DENORMAL_OPERAND", ios_stderr); break;
+            case EXCEPTION_FLT_DIVIDE_BY_ZERO:
+                ios_puts("EXCEPTION_FLT_DIVIDE_BY_ZERO", ios_stderr); break;
+            case EXCEPTION_FLT_INEXACT_RESULT:
+                ios_puts("EXCEPTION_FLT_INEXACT_RESULT", ios_stderr); break;
+            case EXCEPTION_FLT_INVALID_OPERATION:
+                ios_puts("EXCEPTION_FLT_INVALID_OPERATION", ios_stderr); break;
+            case EXCEPTION_FLT_OVERFLOW:
+                ios_puts("EXCEPTION_FLT_OVERFLOW", ios_stderr); break;
+            case EXCEPTION_FLT_STACK_CHECK:
+                ios_puts("EXCEPTION_FLT_STACK_CHECK", ios_stderr); break;
+            case EXCEPTION_FLT_UNDERFLOW:
+                ios_puts("EXCEPTION_FLT_UNDERFLOW", ios_stderr); break;
+            case EXCEPTION_ILLEGAL_INSTRUCTION:
+                ios_puts("EXCEPTION_ILLEGAL_INSTRUCTION", ios_stderr); break;
+            case EXCEPTION_IN_PAGE_ERROR:
+                ios_puts("EXCEPTION_IN_PAGE_ERROR", ios_stderr); break;
+            case EXCEPTION_INT_DIVIDE_BY_ZERO:
+                ios_puts("EXCEPTION_INT_DIVIDE_BY_ZERO", ios_stderr); break;
+            case EXCEPTION_INT_OVERFLOW:
+                ios_puts("EXCEPTION_INT_OVERFLOW", ios_stderr); break;
+            case EXCEPTION_INVALID_DISPOSITION:
+                ios_puts("EXCEPTION_INVALID_DISPOSITION", ios_stderr); break;
+            case EXCEPTION_NONCONTINUABLE_EXCEPTION:
+                ios_puts("EXCEPTION_NONCONTINUABLE_EXCEPTION", ios_stderr); break;
+            case EXCEPTION_PRIV_INSTRUCTION:
+                ios_puts("EXCEPTION_PRIV_INSTRUCTION", ios_stderr); break;
+            case EXCEPTION_SINGLE_STEP:
+                ios_puts("EXCEPTION_SINGLE_STEP", ios_stderr); break;
+            case EXCEPTION_STACK_OVERFLOW:
+                ios_puts("EXCEPTION_STACK_OVERFLOW", ios_stderr); break;
+            default:
+                ios_puts("UNKNOWN", ios_stderr); break;
+        }
+        ios_printf(ios_stderr," at 0x%Ix -- ", (size_t)ExceptionInfo->ExceptionRecord->ExceptionAddress);
+        gdblookup((ptrint_t)ExceptionInfo->ExceptionRecord->ExceptionAddress);
+        bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, ExceptionInfo->ContextRecord);
+        jlbacktrace();
     }
     return EXCEPTION_CONTINUE_SEARCH;
 }
-#else
-void restore_signals()
+
+static LONG WINAPI exception_handler(struct _EXCEPTION_POINTERS *ExceptionInfo)
+{
+    return _exception_handler(ExceptionInfo,1);
+}
+
+#if defined(_CPU_X86_64_)
+EXCEPTION_DISPOSITION _seh_exception_handler(PEXCEPTION_RECORD ExceptionRecord, void *EstablisherFrame, PCONTEXT ContextRecord, void *DispatcherContext)
+{
+    EXCEPTION_POINTERS ExceptionInfo;
+    ExceptionInfo.ExceptionRecord = ExceptionRecord;
+    ExceptionInfo.ContextRecord = ContextRecord;
+    
+    EXCEPTION_DISPOSITION rval;
+    switch (_exception_handler(&ExceptionInfo,1)) {
+        case EXCEPTION_CONTINUE_EXECUTION:
+            rval = ExceptionContinueExecution; break;
+        case EXCEPTION_CONTINUE_SEARCH:
+            rval = ExceptionContinueSearch; break;
+#ifndef _MSC_VER
+        case EXCEPTION_EXECUTE_HANDLER:
+            rval = ExceptionExecuteHandler; break;
+#endif
+    }
+
+    return rval;
+} 
+void* CALLBACK jl_getUnwindInfo(HANDLE hProcess, ULONG64 AddrBase, ULONG64 UserContext);
+#endif
+
+#else // #ifdef _OS_WINDOWS_
+
+void restore_signals(void)
 {
     sigset_t sset;
     sigemptyset(&sset);
     sigprocmask(SIG_SETMASK, &sset, 0);
 }
+
 void sigint_handler(int sig, siginfo_t *info, void *context)
 {
+    if (exit_on_sigint) jl_exit(0);
     if (jl_defer_signal) {
         jl_signal_pending = sig;
     }
@@ -300,24 +401,34 @@ void sigint_handler(int sig, siginfo_t *info, void *context)
 struct uv_shutdown_queue_item { uv_handle_t *h; struct uv_shutdown_queue_item *next; };
 struct uv_shutdown_queue { struct uv_shutdown_queue_item *first; struct uv_shutdown_queue_item *last; };
 
-static void jl_uv_exitcleanup_add(uv_handle_t* handle, struct uv_shutdown_queue *queue)
+static void jl_uv_exitcleanup_add(uv_handle_t *handle, struct uv_shutdown_queue *queue)
 {
-    struct uv_shutdown_queue_item *item = malloc(sizeof(struct uv_shutdown_queue_item));
+    struct uv_shutdown_queue_item *item = (struct uv_shutdown_queue_item*)malloc(sizeof(struct uv_shutdown_queue_item));
     item->h = handle;
     item->next = NULL;
     if (queue->last) queue->last->next = item;
     if (!queue->first) queue->first = item;
     queue->last = item;
 }
-static void jl_uv_exitcleanup_walk(uv_handle_t* handle, void *arg) {
+
+static void jl_uv_exitcleanup_walk(uv_handle_t *handle, void *arg)
+{
     if (handle != (uv_handle_t*)jl_uv_stdout && handle != (uv_handle_t*)jl_uv_stderr)
-        jl_uv_exitcleanup_add(handle, arg);
+        jl_uv_exitcleanup_add(handle, (struct uv_shutdown_queue*)arg);
 }
+
+void jl_write_coverage_data(void);
+void jl_write_malloc_log(void);
+
 DLLEXPORT void uv_atexit_hook()
 {
 #if defined(JL_GC_MARKSWEEP) && defined(GC_FINAL_STATS)
     jl_print_gc_stats(JL_STDERR);
 #endif
+    if (jl_compileropts.code_coverage)
+        jl_write_coverage_data();
+    if (jl_compileropts.malloc_log)
+        jl_write_malloc_log();
     if (jl_base_module) {
         jl_value_t *f = jl_get_global(jl_base_module, jl_symbol("_atexit"));
         if (f!=NULL && jl_is_function(f)) {
@@ -333,13 +444,15 @@ DLLEXPORT void uv_atexit_hook()
 
     jl_gc_run_all_finalizers();
 
-    uv_loop_t* loop = jl_global_event_loop();
+    uv_loop_t *loop = jl_global_event_loop();
     struct uv_shutdown_queue queue = {NULL, NULL};
     uv_walk(loop, jl_uv_exitcleanup_walk, &queue);
     // close stdout and stderr last, since we like being
     // able to show stuff (incl. printf's)
     jl_uv_exitcleanup_add((uv_handle_t*)jl_uv_stdout, &queue);
     jl_uv_exitcleanup_add((uv_handle_t*)jl_uv_stderr, &queue);
+    //uv_unref((uv_handle_t*)jl_uv_stdout);
+    //uv_unref((uv_handle_t*)jl_uv_stderr);
     struct uv_shutdown_queue_item *item = queue.first;
     while (item) {
         JL_TRY {
@@ -394,6 +507,8 @@ DLLEXPORT void uv_atexit_hook()
 void jl_get_builtin_hooks(void);
 
 uv_lib_t *jl_dl_handle;
+uv_lib_t _jl_RTLD_DEFAULT_handle;
+uv_lib_t *jl_RTLD_DEFAULT_handle=&_jl_RTLD_DEFAULT_handle;
 #ifdef _OS_WINDOWS_
 uv_lib_t _jl_ntdll_handle;
 uv_lib_t _jl_exe_handle;
@@ -409,30 +524,20 @@ uv_lib_t *jl_winsock_handle=&_jl_winsock_handle;
 #endif
 uv_loop_t *jl_io_loop;
 
-#ifdef COPY_STACKS
-void jl_switch_stack(jl_task_t *t, jl_jmp_buf *where);
-extern jl_jmp_buf * volatile jl_jmp_target;
-#endif
-
 void *init_stdio_handle(uv_file fd,int readable)
 {
     void *handle;
     uv_handle_type type = uv_guess_handle(fd);
     jl_uv_file_t *file;
 #ifndef _OS_WINDOWS_    
-    // Duplicate the file descritor so we can later dup it over if we want to redirect
+    // Duplicate the file descriptor so we can later dup it over if we want to redirect
     // STDIO without having to worry about closing the associated libuv object.
     // On windows however, libuv objects remember streams by their HANDLE, so this is
-    // unnessecary.
+    // unnecessary.
     fd = dup(fd);
-#else
-    // Duplicate the file descritor so libuv can later close it without having to worry
-    // about needing to keep its handle open
-    fd = _dup(fd);
 #endif
     //printf("%d: %d -- %d\n", fd, type, 0);
-    switch(type)
-    {
+    switch(type) {
         case UV_TTY:
             handle = malloc(sizeof(uv_tty_t));
             if (uv_tty_init(jl_io_loop,(uv_tty_t*)handle,fd,readable)) {
@@ -440,10 +545,10 @@ void *init_stdio_handle(uv_file fd,int readable)
                 abort();
             }
             ((uv_tty_t*)handle)->data=0;
-            uv_tty_set_mode((void*)handle,0); //cooked stdio
+            uv_tty_set_mode((uv_tty_t*)handle,0); //cooked stdio
             break;
         case UV_FILE: 
-            file = malloc(sizeof(jl_uv_file_t));
+            file = (jl_uv_file_t*)malloc(sizeof(jl_uv_file_t));
             file->loop = jl_io_loop;
             file->type = UV_FILE;
             file->file = fd;
@@ -485,11 +590,21 @@ void *init_stdio_handle(uv_file fd,int readable)
 
 void init_stdio()
 {   //order must be 2,1,0
-    JL_STDERR = init_stdio_handle(2,0);
-    JL_STDOUT = init_stdio_handle(1,0);
-    JL_STDIN = init_stdio_handle(0,1);
+    JL_STDERR = (uv_stream_t*)init_stdio_handle(2,0);
+    JL_STDOUT = (uv_stream_t*)init_stdio_handle(1,0);
+    JL_STDIN = (uv_stream_t*)init_stdio_handle(0,1);
 }
 
+#ifdef JL_USE_INTEL_JITEVENTS
+char jl_using_intel_jitevents; // Non-zero if running under Intel VTune Amplifier
+#endif
+
+#if defined(JL_USE_INTEL_JITEVENTS) && defined(__linux__)
+unsigned sig_stack_size = SIGSTKSZ; 
+#else
+#define sig_stack_size SIGSTKSZ
+#endif
+
 #ifndef _OS_WINDOWS_
 static void *signal_stack;
 #endif
@@ -502,9 +617,9 @@ static mach_port_t segv_port = 0;
 
 extern boolean_t exc_server(mach_msg_header_t *, mach_msg_header_t *);
 
-void * mach_segv_listener(void *arg)
+void *mach_segv_listener(void *arg)
 {
-    (void) arg;
+    (void)arg;
     while (1) {
         int ret = mach_msg_server(exc_server,2048,segv_port,MACH_MSG_TIMEOUT_NONE);
         printf("mach_msg_server: %s\n", mach_error_string(ret));
@@ -519,6 +634,13 @@ void darwin_stack_overflow_handler(unw_context_t *uc)
     jl_rethrow();
 }
 
+void darwin_accerr_handler(unw_context_t *uc)
+{
+    bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, uc);
+    jl_exception_in_transit = jl_memory_exception;
+    jl_rethrow();
+}
+
 #define HANDLE_MACH_ERROR(msg, retval) \
     if (retval!=KERN_SUCCESS) { mach_error(msg ":", (retval)); jl_exit(1); }
 
@@ -527,14 +649,20 @@ extern kern_return_t profiler_segv_handler(mach_port_t,mach_port_t,mach_port_t,e
 extern volatile mach_port_t mach_profiler_thread;
 #endif
 
+enum x86_trap_flags {
+    USER_MODE = 0x4,
+    WRITE_FAULT = 0x2,
+    PAGE_PRESENT = 0x1
+};
+
 //exc_server uses dlsym to find symbol
-DLLEXPORT kern_return_t catch_exception_raise
-                (mach_port_t                          exception_port,
-                 mach_port_t                                  thread,
-                 mach_port_t                                    task,
-                 exception_type_t                          exception,
-                 exception_data_t                               code,
-                 mach_msg_type_number_t                   code_count)
+DLLEXPORT
+kern_return_t catch_exception_raise(mach_port_t            exception_port,
+                                    mach_port_t            thread,
+                                    mach_port_t            task,
+                                    exception_type_t       exception,
+                                    exception_data_t       code,
+                                    mach_msg_type_number_t code_count)
 {
     unsigned int count = MACHINE_THREAD_STATE_COUNT;
     unsigned int exc_count = X86_EXCEPTION_STATE64_COUNT;
@@ -544,37 +672,27 @@ DLLEXPORT kern_return_t catch_exception_raise
     //memset(&state,0,sizeof(x86_thread_state64_t));
     //memset(&exc_state,0,sizeof(x86_exception_state64_t));
 #ifdef LIBOSXUNWIND
-    if (thread == mach_profiler_thread)
-    {
+    if (thread == mach_profiler_thread) {
         return profiler_segv_handler(exception_port,thread,task,exception,code,code_count);
     }
 #endif
     ret = thread_get_state(thread,x86_EXCEPTION_STATE64,(thread_state_t)&exc_state,&exc_count);
     HANDLE_MACH_ERROR("thread_get_state(1)",ret);
     uint64_t fault_addr = exc_state.__faultvaddr;
-    if (
-#ifdef COPY_STACKS
-        (char*)fault_addr > (char*)jl_stack_lo-3000000 &&
-        (char*)fault_addr < (char*)jl_stack_hi
-#else
-        (char*)fault_addr > (char*)jl_current_task->stack-8192 &&
-        (char*)fault_addr <
-        (char*)jl_current_task->stack+jl_current_task->ssize
-#endif
-        ) 
-    {
+    if (is_addr_on_stack((void*)fault_addr) ||
+        ((exc_state.__err & PAGE_PRESENT) == PAGE_PRESENT)) {
         ret = thread_get_state(thread,x86_THREAD_STATE64,(thread_state_t)&state,&count);
         HANDLE_MACH_ERROR("thread_get_state(2)",ret);
         old_state = state;
         // memset(&state,0,sizeof(x86_thread_state64_t));
         // Setup libunwind information
-        state.__rsp = (uint64_t)signal_stack + SIGSTKSZ;
+        state.__rsp = (uint64_t)signal_stack + sig_stack_size;
         state.__rsp -= sizeof(unw_context_t);
         state.__rsp &= -16;
         unw_context_t *uc = (unw_context_t*)state.__rsp;
         state.__rsp -= 512;
         // This is for alignment. In particular note that the sizeof(void*) is necessary
-        // since it would usually specify the return address (i.e. we are aligning the call
+        // since it would usually specify the return address (i.e., we are aligning the call
         // frame to a 16 byte boundary as required by the abi, but the stack pointer
         // to point to the byte beyond that. Not doing this leads to funny behavior on
         // the first access to an external function will fail due to stack misalignment
@@ -583,7 +701,10 @@ DLLEXPORT kern_return_t catch_exception_raise
         memset(uc,0,sizeof(unw_context_t));
         memcpy(uc,&old_state,sizeof(x86_thread_state64_t));
         state.__rdi = (uint64_t)uc;
-        state.__rip = (uint64_t)darwin_stack_overflow_handler;
+        if ((exc_state.__err & PAGE_PRESENT) == PAGE_PRESENT)
+            state.__rip = (uint64_t)darwin_accerr_handler;
+        else
+            state.__rip = (uint64_t)darwin_stack_overflow_handler;
 
         state.__rbp = state.__rsp;
         ret = thread_set_state(thread,x86_THREAD_STATE64,(thread_state_t)&state,count);
@@ -591,6 +712,11 @@ DLLEXPORT kern_return_t catch_exception_raise
         return KERN_SUCCESS;
     }
     else {
+        ret = thread_get_state(thread,x86_THREAD_STATE64,(thread_state_t)&state,&count);
+        HANDLE_MACH_ERROR("thread_get_state(3)",ret);
+        ios_printf(ios_stderr,"\nsignal (%d): %s\n", SIGSEGV, strsignal(SIGSEGV));
+        bt_size = rec_backtrace_ctx(bt_data, MAX_BT_SIZE, (unw_context_t*)&state);
+        jlbacktrace();
         return KERN_INVALID_ARGUMENT;
     }
 }
@@ -599,30 +725,55 @@ DLLEXPORT kern_return_t catch_exception_raise
 
 void julia_init(char *imageFile)
 {
+    jl_io_loop = uv_default_loop(); // this loop will internal events (spawning process etc.),
+                                    // best to call this first, since it also initializes libuv
     jl_page_size = jl_getpagesize();
+    jl_arr_xtralloc_limit = uv_get_total_memory() / 100;  // Extra allocation limited to 1% of total RAM 
     jl_find_stack_bottom();
     jl_dl_handle = jl_load_dynamic_library(NULL, JL_RTLD_DEFAULT);
+#ifdef RTLD_DEFAULT
+    jl_RTLD_DEFAULT_handle->handle = RTLD_DEFAULT;
+#else
+    jl_RTLD_DEFAULT_handle->handle = jl_dl_handle->handle;
+#endif
 #ifdef _OS_WINDOWS_
-    uv_dlopen("ntdll.dll",jl_ntdll_handle); //bypass julia's pathchecking for system dlls
-    uv_dlopen("kernel32.dll",jl_kernel32_handle);
-    uv_dlopen("msvcrt.dll",jl_crtdll_handle);
-    uv_dlopen("ws2_32.dll",jl_winsock_handle);
+    uv_dlopen("ntdll.dll", jl_ntdll_handle); // bypass julia's pathchecking for system dlls
+    uv_dlopen("kernel32.dll", jl_kernel32_handle);
+    uv_dlopen("msvcrt.dll", jl_crtdll_handle);
+    uv_dlopen("ws2_32.dll", jl_winsock_handle);
     _jl_exe_handle.handle = GetModuleHandleA(NULL);
-    if (!DuplicateHandle( GetCurrentProcess(), GetCurrentThread(),
-        GetCurrentProcess(), (PHANDLE)&hMainThread, 0,
-        TRUE, DUPLICATE_SAME_ACCESS )) {
-        JL_PRINTF(JL_STDERR, "Couldn't access handle to main thread\n");
+    if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
+                         GetCurrentProcess(), (PHANDLE)&hMainThread, 0,
+                         TRUE, DUPLICATE_SAME_ACCESS)) {
+        JL_PRINTF(JL_STDERR, "WARNING: failed to access handle to main thread\n");
     }
     SymSetOptions(SYMOPT_UNDNAME | SYMOPT_DEFERRED_LOADS | SYMOPT_LOAD_LINES);
-    SymInitialize(GetCurrentProcess(), NULL, 1);
+    if (!SymInitialize(GetCurrentProcess(), NULL, 1)) {
+        JL_PRINTF(JL_STDERR, "WARNING: failed to initalize stack walk info\n");
+    }
+#if defined(_CPU_X86_64_)
+    if (!SymRegisterFunctionEntryCallback64(GetCurrentProcess(), jl_getUnwindInfo, 0)) JL_PRINTF(JL_STDERR, "WARNING: failed to install backtrace info callback\n");
+#endif
     needsSymRefreshModuleList = 0;
     uv_lib_t jl_dbghelp;
     uv_dlopen("dbghelp.dll",&jl_dbghelp);
-    if (uv_dlsym(&jl_dbghelp,"SymRefreshModuleList",(void**)&hSymRefreshModuleList)) hSymRefreshModuleList = 0;
+    if (uv_dlsym(&jl_dbghelp, "SymRefreshModuleList", (void**)&hSymRefreshModuleList))
+        hSymRefreshModuleList = 0;
 #endif
-    jl_io_loop = uv_default_loop(); //this loop will internal events (spawining process etc.)
     init_stdio();
 
+#if defined(JL_USE_INTEL_JITEVENTS)
+    const char *jit_profiling = getenv("ENABLE_JITPROFILING");
+    if (jit_profiling && atoi(jit_profiling)) {
+        jl_using_intel_jitevents = 1;
+#if defined(__linux__)
+        // Intel VTune Amplifier needs at least 64k for alternate stack.
+        if (SIGSTKSZ < 1<<16) 
+            sig_stack_size = 1<<16; 
+#endif
+    }
+#endif
+
 #if defined(__linux__)
     int ncores = jl_cpu_cores();
     if (ncores > 1) {
@@ -648,20 +799,24 @@ void julia_init(char *imageFile)
     jl_init_serializer();
 
     if (!imageFile) {
-        jl_main_module = jl_new_module(jl_symbol("Main"));
-        jl_main_module->parent = jl_main_module;
         jl_core_module = jl_new_module(jl_symbol("Core"));
-        jl_core_module->parent = jl_main_module;
-        jl_set_const(jl_main_module, jl_symbol("Core"),
-                     (jl_value_t*)jl_core_module);
-        jl_module_using(jl_main_module, jl_core_module);
+        jl_new_main_module();
+        jl_internal_main_module = jl_main_module;
+
         jl_current_module = jl_core_module;
+        jl_root_task->current_module = jl_current_module;
+
         jl_init_intrinsic_functions();
         jl_init_primitives();
         jl_load("boot.jl");
         jl_get_builtin_hooks();
         jl_boot_file_loaded = 1;
         jl_init_box_caches();
+        // Core.JULIA_HOME is a "magic" constant, we set it at runtime here
+        // since its value gets excluded from the system image
+        jl_set_const(jl_core_module, jl_symbol("JULIA_HOME"),
+                     jl_cstr_to_string(julia_home));
+        jl_module_export(jl_core_module, jl_symbol("JULIA_HOME"));
     }
 
     if (imageFile) {
@@ -698,10 +853,11 @@ void julia_init(char *imageFile)
     // eval() uses Main by default, so Main.eval === Core.eval
     jl_module_import(jl_main_module, jl_core_module, jl_symbol("eval"));
     jl_current_module = jl_main_module;
+    jl_root_task->current_module = jl_current_module;
 
 
 #ifndef _OS_WINDOWS_
-    signal_stack = malloc(SIGSTKSZ);
+    signal_stack = malloc(sig_stack_size);
     struct sigaction actf;
     memset(&actf, 0, sizeof(struct sigaction));
     sigemptyset(&actf.sa_mask);
@@ -726,25 +882,23 @@ void julia_init(char *imageFile)
     // Alright, create a thread to serve as the listener for exceptions
     pthread_t thread;
     pthread_attr_t attr;
-    if (pthread_attr_init(&attr) != 0)
-    {
+    if (pthread_attr_init(&attr) != 0) {
         JL_PRINTF(JL_STDERR, "pthread_attr_init failed");
         jl_exit(1);  
     }
     pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
-    if (pthread_create(&thread,&attr,mach_segv_listener,NULL) != 0)
-    {
+    if (pthread_create(&thread,&attr,mach_segv_listener,NULL) != 0) {
         JL_PRINTF(JL_STDERR, "pthread_create failed");
         jl_exit(1);  
     }     
     pthread_attr_destroy(&attr);
 
-    ret = task_set_exception_ports(self,EXC_MASK_BAD_ACCESS,segv_port,EXCEPTION_DEFAULT,MACHINE_THREAD_STATE);
-    HANDLE_MACH_ERROR("task_set_exception_ports",ret);
+    ret = thread_set_exception_ports(mach_thread_self(),EXC_MASK_BAD_ACCESS,segv_port,EXCEPTION_DEFAULT,MACHINE_THREAD_STATE);
+    HANDLE_MACH_ERROR("thread_set_exception_ports",ret);
 #else // defined(_OS_DARWIN_)
     stack_t ss;
     ss.ss_flags = 0;
-    ss.ss_size = SIGSTKSZ;
+    ss.ss_size = sig_stack_size;
     ss.ss_sp = signal_stack;
     if (sigaltstack(&ss, NULL) < 0) {
         JL_PRINTF(JL_STDERR, "sigaltstack: %s\n", strerror(errno));
@@ -761,6 +915,43 @@ void julia_init(char *imageFile)
         jl_exit(1);
     }
 #endif // defined(_OS_DARWIN_)
+    struct sigaction act_die;
+    memset(&act_die, 0, sizeof(struct sigaction));
+    sigemptyset(&act_die.sa_mask);
+    act_die.sa_sigaction = sigdie_handler;
+    act_die.sa_flags = SA_SIGINFO;
+    if (sigaction(SIGINFO, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGBUS, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGILL, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGTERM, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGABRT, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGQUIT, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGSYS, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
+    if (sigaction(SIGPIPE, &act_die, NULL) < 0) {
+        JL_PRINTF(JL_STDERR, "sigaction: %s\n", strerror(errno));
+        jl_exit(1);
+    }
 #else // defined(_OS_WINDOWS_)
     if (signal(SIGFPE, (void (__cdecl *)(int))fpe_handler) == SIG_ERR) {
         JL_PRINTF(JL_STDERR, "Couldn't set SIGFPE\n");
@@ -768,10 +959,14 @@ void julia_init(char *imageFile)
     }
 #endif
 
-
 #ifdef JL_GC_MARKSWEEP
     jl_gc_enable();
 #endif
+
+    if (imageFile)
+        jl_init_restored_modules();
+
+    jl_install_sigint_handler();
 }
 
 DLLEXPORT void jl_install_sigint_handler()
@@ -792,15 +987,15 @@ DLLEXPORT void jl_install_sigint_handler()
     //printf("sigint installed\n");
 }
 
-
-extern void * __stack_chk_guard;
+extern int asprintf(char **str, const char *fmt, ...);
+extern void *__stack_chk_guard;
 
 DLLEXPORT int julia_trampoline(int argc, char **argv, int (*pmain)(int ac,char *av[]))
 {
-#if defined(_OS_WINDOWS_) //&& !defined(_WIN64)
+#if defined(_OS_WINDOWS_)
     SetUnhandledExceptionFilter(exception_handler);
 #endif
-    unsigned char * p = (unsigned char *) &__stack_chk_guard;
+    unsigned char *p = (unsigned char *)&__stack_chk_guard;
     char a = p[sizeof(__stack_chk_guard)-1];
     char b = p[sizeof(__stack_chk_guard)-2];
     char c = p[0];
@@ -808,14 +1003,27 @@ DLLEXPORT int julia_trampoline(int argc, char **argv, int (*pmain)(int ac,char *
     p[sizeof(__stack_chk_guard)-1] = 255;
     p[sizeof(__stack_chk_guard)-2] = '\n';
     p[0] = 0;
-#ifdef COPY_STACKS
-    // initialize base context of root task
-    jl_root_task->stackbase = (char*)&argc;
-    if (jl_setjmp(jl_root_task->base_ctx, 0)) {
-        jl_switch_stack(jl_current_task, jl_jmp_target);
-    }
-#endif
+    JL_SET_STACK_BASE;
     int ret = pmain(argc, argv);
+    char *build_path = jl_compileropts.build_path;
+    if (build_path) {
+        char *build_ji;
+        if (asprintf(&build_ji, "%s.ji",build_path) > 0) {
+            jl_save_system_image(build_ji);
+            free(build_ji);
+            char *build_o;
+            if (asprintf(&build_o, "%s.o",build_path) > 0) {
+                jl_dump_objfile(build_o,0);
+                free(build_o);
+            }
+            else {
+                ios_printf(ios_stderr,"\nFATAL: failed to create string for .o build path\n");
+            }
+        }
+        else {
+            ios_printf(ios_stderr,"\nFATAL: failed to create string for .ji build path\n");
+        }
+    }
     p[sizeof(__stack_chk_guard)-1] = a;
     p[sizeof(__stack_chk_guard)-2] = b;
     p[0] = c;
@@ -826,7 +1034,6 @@ jl_function_t *jl_typeinf_func=NULL;
 
 DLLEXPORT void jl_enable_inference(void)
 {
-    if (jl_typeinf_func != NULL) return;
     jl_typeinf_func = (jl_function_t*)jl_get_global(jl_base_module,
                                                     jl_symbol("typeinf_ext"));
 }
@@ -862,25 +1069,18 @@ void jl_get_builtin_hooks(void)
     jl_float32_type = (jl_datatype_t*)core("Float32");
     jl_float64_type = (jl_datatype_t*)core("Float64");
     jl_floatingpoint_type = (jl_datatype_t*)core("FloatingPoint");
+    jl_number_type = (jl_datatype_t*)core("Number");
 
-    jl_stackovf_exception =
-        jl_apply((jl_function_t*)core("StackOverflowError"), NULL, 0);
-    jl_diverror_exception =
-        jl_apply((jl_function_t*)core("DivideError"), NULL, 0);
-    jl_domain_exception =
-        jl_apply((jl_function_t*)core("DomainError"), NULL, 0);
-    jl_overflow_exception =
-        jl_apply((jl_function_t*)core("OverflowError"), NULL, 0);
-    jl_inexact_exception =
-        jl_apply((jl_function_t*)core("InexactError"), NULL, 0);
-    jl_undefref_exception =
-        jl_apply((jl_function_t*)core("UndefRefError"),NULL,0);
-    jl_interrupt_exception =
-        jl_apply((jl_function_t*)core("InterruptException"),NULL,0);
-    jl_bounds_exception =
-        jl_apply((jl_function_t*)core("BoundsError"),NULL,0);
-    jl_memory_exception =
-        jl_apply((jl_function_t*)core("MemoryError"),NULL,0);
+    jl_stackovf_exception  = jl_new_struct((jl_datatype_t*)core("StackOverflowError"));
+    jl_diverror_exception  = jl_new_struct((jl_datatype_t*)core("DivideError"));
+    jl_domain_exception    = jl_new_struct((jl_datatype_t*)core("DomainError"));
+    jl_overflow_exception  = jl_new_struct((jl_datatype_t*)core("OverflowError"));
+    jl_inexact_exception   = jl_new_struct((jl_datatype_t*)core("InexactError"));
+    jl_undefref_exception  = jl_new_struct((jl_datatype_t*)core("UndefRefError"));
+    jl_undefvarerror_type  = (jl_datatype_t*)core("UndefVarError");
+    jl_interrupt_exception = jl_new_struct((jl_datatype_t*)core("InterruptException"));
+    jl_bounds_exception    = jl_new_struct((jl_datatype_t*)core("BoundsError"));
+    jl_memory_exception    = jl_new_struct((jl_datatype_t*)core("MemoryError"));
 
     jl_ascii_string_type = (jl_datatype_t*)core("ASCIIString");
     jl_utf8_string_type = (jl_datatype_t*)core("UTF8String");
@@ -902,3 +1102,9 @@ DLLEXPORT void jl_get_system_hooks(void)
     jl_loaderror_type = (jl_datatype_t*)basemod("LoadError");
     jl_weakref_type = (jl_datatype_t*)basemod("WeakRef");
 }
+
+DLLEXPORT void jl_exit_on_sigint(int on) {exit_on_sigint = on;}
+
+#ifdef __cplusplus
+}
+#endif
