@@ -5,8 +5,13 @@
 #include <malloc.h>
 #endif
 #include "julia.h"
+#include "julia_internal.h"
 #include "builtin_proto.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 extern int jl_lineno;
 
 static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl);
@@ -26,20 +31,23 @@ jl_value_t *jl_interpret_toplevel_expr_with(jl_value_t *e,
     return eval(e, locals, nl);
 }
 
-jl_value_t *jl_interpret_toplevel_expr_in(jl_module_t *m, jl_value_t *e,
-                                          jl_value_t **locals, size_t nl)
+DLLEXPORT jl_value_t *jl_interpret_toplevel_expr_in(jl_module_t *m, jl_value_t *e,
+                                                    jl_value_t **locals, size_t nl)
 {
     jl_value_t *v=NULL;
     jl_module_t *last_m = jl_current_module;
+    jl_module_t *task_last_m = jl_current_task->current_module;
     JL_TRY {
-        jl_current_module = m;
+        jl_current_task->current_module = jl_current_module = m;
         v = eval(e, locals, nl);
     }
     JL_CATCH {
         jl_current_module = last_m;
+        jl_current_task->current_module = task_last_m;
         jl_rethrow();
     }
     jl_current_module = last_m;
+    jl_current_task->current_module = task_last_m;
     assert(v);
     return v;
 }
@@ -63,15 +71,39 @@ jl_value_t *jl_eval_global_var(jl_module_t *m, jl_sym_t *e)
 {
     jl_value_t *v = jl_get_global(m, e);
     if (v == NULL)
-        jl_errorf("%s not defined", e->name);
+        jl_undefined_var_error(e);
     return v;
 }
 
+void jl_check_static_parameter_conflicts(jl_lambda_info_t *li, jl_tuple_t *t, jl_sym_t *fname);
+
 int jl_has_intrinsics(jl_expr_t *e);
 
 extern int jl_boot_file_loaded;
 extern int inside_typedef;
 
+// this is a heuristic for allowing "redefining" a type to something identical
+static int equiv_type(jl_datatype_t *dta, jl_datatype_t *dtb)
+{
+    return (jl_typeof(dta) == jl_typeof(dtb) &&
+            // TODO: can't yet handle parametric types due to how constructors work
+            dta->parameters == jl_null &&
+            dta->name->name == dtb->name->name &&
+            jl_egal((jl_value_t*)dta->types, (jl_value_t*)dtb->types) &&
+            dta->abstract == dtb->abstract &&
+            dta->mutabl == dtb->mutabl &&
+            dta->size == dtb->size &&
+            jl_egal((jl_value_t*)dta->super, (jl_value_t*)dtb->super) &&
+            jl_egal((jl_value_t*)dta->names, (jl_value_t*)dtb->names) &&
+            jl_egal((jl_value_t*)dta->parameters, (jl_value_t*)dtb->parameters));
+}
+
+static void check_can_assign_type(jl_binding_t *b)
+{
+    if (b->constp && b->value != NULL && !jl_is_datatype(b->value))
+        jl_errorf("invalid redefinition of constant %s", b->name->name);
+}
+
 static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
 {
     if (jl_is_symbol(e)) {
@@ -87,7 +119,7 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
             v = jl_get_global(jl_current_module, (jl_sym_t*)e);
         }
         if (v == NULL) {
-            jl_errorf("%s not defined", ((jl_sym_t*)e)->name);
+            jl_undefined_var_error((jl_sym_t*)e);
         }
         return v;
     }
@@ -101,7 +133,7 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_sym_t *s = (jl_sym_t*)jl_fieldref(e,0);
         jl_value_t *v = jl_get_global(jl_base_relative_to(jl_current_module),s);
         if (v == NULL)
-            jl_errorf("%s not defined", s->name);
+            jl_undefined_var_error(s);
         return v;
     }
     if (!jl_is_expr(e)) {
@@ -176,6 +208,7 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
     }
     else if (ex->head == assign_sym) {
         jl_value_t *sym = args[0];
+        assert(jl_is_symbol(sym));
         size_t i;
         for (i=0; i < nl; i++) {
             if (locals[i*2] == sym) {
@@ -248,6 +281,9 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_value_t *atypes=NULL, *meth=NULL;
         JL_GC_PUSH2(&atypes, &meth);
         atypes = eval(args[1], locals, nl);
+        if (jl_is_lambda_info(args[2])) {
+            jl_check_static_parameter_conflicts((jl_lambda_info_t*)args[2], (jl_tuple_t*)jl_t1(atypes), fname);
+        }
         meth = eval(args[2], locals, nl);
         jl_method_def(fname, bp, b, (jl_tuple_t*)atypes, (jl_function_t*)meth);
         JL_GC_POP();
@@ -258,6 +294,7 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
     }
     else if (ex->head == const_sym) {
         jl_value_t *sym = args[0];
+        assert(jl_is_symbol(sym));
         for (size_t i=0; i < nl; i++) {
             if (locals[i*2] == sym) {
                 return (jl_value_t*)jl_nothing;
@@ -280,21 +317,32 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_value_t *name = args[0];
         jl_value_t *para = eval(args[1], locals, nl);
         jl_value_t *super = NULL;
-        JL_GC_PUSH2(&para, &super);
+        jl_value_t *temp = NULL;
+        JL_GC_PUSH3(&para, &super, &temp);
+        assert(jl_is_tuple(para));
+        assert(jl_is_symbol(name));
         jl_datatype_t *dt =
             jl_new_abstracttype(name, jl_any_type, (jl_tuple_t*)para);
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)dt);
+        temp = b->value;
+        check_can_assign_type(b);
+        b->value = (jl_value_t*)dt;
         super = eval(args[2], locals, nl);
         jl_set_datatype_super(dt, super);
+        b->value = temp;
+        if (temp==NULL || !equiv_type(dt, (jl_datatype_t*)temp)) {
+            jl_checked_assignment(b, (jl_value_t*)dt);
+        }
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
     else if (ex->head == bitstype_sym) {
         jl_value_t *name = args[0];
-        jl_value_t *super = NULL, *para = NULL, *vnb = NULL;
-        JL_GC_PUSH3(&para, &super, &vnb);
+        jl_value_t *super = NULL, *para = NULL, *vnb = NULL, *temp = NULL;
+        JL_GC_PUSH3(&para, &super, &temp);
+        assert(jl_is_symbol(name));
         para = eval(args[1], locals, nl);
+        assert(jl_is_tuple(para));
         vnb  = eval(args[2], locals, nl);
         if (!jl_is_long(vnb))
             jl_errorf("invalid declaration of bits type %s", ((jl_sym_t*)name)->name);
@@ -305,39 +353,69 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_datatype_t *dt =
             jl_new_bitstype(name, jl_any_type, (jl_tuple_t*)para, nb);
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)dt);
+        temp = b->value;
+        check_can_assign_type(b);
+        b->value = (jl_value_t*)dt;
         super = eval(args[3], locals, nl);
         jl_set_datatype_super(dt, super);
+        b->value = temp;
+        if (temp==NULL || !equiv_type(dt, (jl_datatype_t*)temp)) {
+            jl_checked_assignment(b, (jl_value_t*)dt);
+        }
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
     else if (ex->head == compositetype_sym) {
         void jl_add_constructors(jl_datatype_t *t);
         jl_value_t *name = args[0];
+        assert(jl_is_symbol(name));
         jl_value_t *para = eval(args[1], locals, nl);
-        jl_value_t *fnames = NULL;
+        assert(jl_is_tuple(para));
+        jl_value_t *temp = NULL;
         jl_value_t *super = NULL;
         jl_datatype_t *dt = NULL;
-        JL_GC_PUSH4(&para, &super, &fnames, &dt);
-        fnames = eval(args[2], locals, nl);
+        JL_GC_PUSH4(&para, &super, &temp, &dt);
+        temp = eval(args[2], locals, nl);  // field names
         dt = jl_new_datatype((jl_sym_t*)name, jl_any_type, (jl_tuple_t*)para,
-                             (jl_tuple_t*)fnames, NULL,
+                             (jl_tuple_t*)temp, NULL,
                              0, args[6]==jl_true ? 1 : 0);
         dt->fptr = jl_f_ctor_trampoline;
         dt->ctor_factory = eval(args[3], locals, nl);
+
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)dt);
-        inside_typedef = 1;
-        dt->types = (jl_tuple_t*)eval(args[5], locals, nl);
-        inside_typedef = 0;
-        jl_check_type_tuple(dt->types, dt->name->name, "type definition");
-        super = eval(args[4], locals, nl);
-        jl_set_datatype_super(dt, super);
+        temp = b->value;  // save old value
+        // temporarily assign so binding is available for field types
+        check_can_assign_type(b);
+        b->value = (jl_value_t*)dt;
+
+        JL_TRY {
+            // operations that can fail
+            inside_typedef = 1;
+            dt->types = (jl_tuple_t*)eval(args[5], locals, nl);
+            inside_typedef = 0;
+            jl_check_type_tuple(dt->types, dt->name->name, "type definition");
+            super = eval(args[4], locals, nl);
+            jl_set_datatype_super(dt, super);
+        }
+        JL_CATCH {
+            b->value = temp;
+            jl_rethrow();
+        }
         for(size_t i=0; i < jl_tuple_len(para); i++) {
             ((jl_tvar_t*)jl_tupleref(para,i))->bound = 0;
         }
         jl_compute_field_offsets(dt);
-        jl_add_constructors(dt);
+
+        b->value = temp;
+        if (temp==NULL || !equiv_type(dt, (jl_datatype_t*)temp)) {
+            jl_checked_assignment(b, (jl_value_t*)dt);
+
+            jl_add_constructors(dt);
+        }
+        else {
+            // TODO: remove all old ctors and set temp->ctor_factory = dt->ctor_factory
+        }
+
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
@@ -362,7 +440,9 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
     else if (ex->head == module_sym) {
         return jl_eval_module_expr(ex);
     }
-    else if (ex->head == error_sym || ex->head == jl_continue_sym) {
+    else if (ex->head == error_sym || ex->head == jl_incomplete_sym) {
+        if (nargs == 0)
+            jl_error("malformed \"error\" expression");
         if (jl_is_byte_string(args[0]))
             jl_errorf("syntax: %s", jl_string_data(args[0]));
         jl_throw(args[0]);
@@ -370,6 +450,9 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
     else if (ex->head == boundscheck_sym) {
         return (jl_value_t*)jl_nothing;
     }
+    else if (ex->head == simdloop_sym) {
+        return (jl_value_t*)jl_nothing;
+    }
     jl_errorf("unsupported or misplaced expression %s", ex->head->name);
     return (jl_value_t*)jl_nothing;
 }
@@ -430,6 +513,10 @@ static jl_value_t *eval_body(jl_array_t *stmts, jl_value_t **locals, size_t nl,
                     return eval_body(stmts, locals, nl, i+1, toplevel);
                 }
                 else {
+#ifdef _OS_WINDOWS_
+                    if (jl_exception_in_transit == jl_stackovf_exception)
+                        _resetstkoflw();
+#endif
                     i = label_idx(jl_exprarg(stmt,0), stmts);
                     continue;
                 }
@@ -487,3 +574,7 @@ jl_value_t *jl_interpret_toplevel_thunk(jl_lambda_info_t *lam)
 {
     return jl_interpret_toplevel_thunk_with(lam, NULL, 0);
 }
+
+#ifdef __cplusplus
+}
+#endif
