@@ -25,6 +25,7 @@
 #endif
 
 #include "julia.h"
+#include "julia_internal.h"
 #include "support/ios.h"
 #include "uv.h"
 
@@ -35,6 +36,15 @@ extern "C" {
 
 /** libuv callbacks */
 
+enum CALLBACK_TYPE { CB_PTR, CB_INT32, CB_UINT32, CB_INT64, CB_UINT64 };
+#ifdef _P64
+#define CB_INT CB_INT64
+#define CB_UINT CB_UINT64
+#else
+#define CB_INT CB_INT32
+#define CB_UINT CB_UINT32
+#endif
+
 /*
  * Notes for adding new callbacks
  * - Make sure to type annotate the callback, so we'll get the one in the new
@@ -44,20 +54,22 @@ extern "C" {
 
 //These callbacks are implemented in stream.jl
 #define JL_CB_TYPES(XX) \
-	XX(close) \
-	XX(return_spawn) \
-	XX(readcb) \
-	XX(alloc_buf) \
-	XX(connectcb) \
-	XX(connectioncb) \
-	XX(asynccb) \
+    XX(close) \
+    XX(return_spawn) \
+    XX(readcb) \
+    XX(alloc_buf) \
+    XX(connectcb) \
+    XX(connectioncb) \
+    XX(asynccb) \
     XX(getaddrinfo) \
     XX(pollcb) \
     XX(fspollcb) \
     XX(isopen) \
     XX(fseventscb) \
     XX(writecb) \
-    XX(writecb_task)
+    XX(writecb_task) \
+    XX(recv) \
+    XX(send)
 //TODO add UDP and other missing callbacks
 
 #define JULIA_HOOK_(m,hook)  ((jl_function_t*)jl_get_global(m, jl_symbol("_uv_hook_" #hook)))
@@ -74,35 +86,23 @@ DLLEXPORT void jl_get_uv_hooks()
 }
 #undef JL_CB_TYPES
 
-int base_module_conflict = 0; //set to 1 if Base is getting redefined since it means there are two place to try the callbacks
-// warning: this is defined without the standard do {...} while (0) wrapper, since I wanted ret to escape
-// warning: during bootstrapping, callbacks will be called twice if a MethodError occured at ANY time during callback call
+extern jl_module_t *jl_old_base_module;
 // Use:  JULIA_CB(hook, arg1, numberOfAdditionalArgs, arg2Type, arg2, ..., argNType, argN)
-#define JULIA_CB(hook, ...) \
-    jl_value_t *ret; \
-    if (!base_module_conflict) { \
-        ret = jl_callback_call(JULIA_HOOK(hook),__VA_ARGS__); \
-    } else { \
-        JL_TRY { \
-            ret = jl_callback_call(JULIA_HOOK(hook),__VA_ARGS__); \
-            /* jl_puts(#hook " original succeeded\n",jl_uv_stderr); */ \
-        } \
-        JL_CATCH { \
-            if (jl_typeof(jl_exception_in_transit) == (jl_value_t*)jl_methoderror_type) { \
-                /* jl_puts("\n" #hook " being retried with new Base bindings --> ",jl_uv_stderr); */ \
-                jl_function_t *cb_func = JULIA_HOOK_((jl_module_t*)jl_get_global(jl_main_module, jl_symbol("Base")), hook); \
-                ret = jl_callback_call(cb_func,__VA_ARGS__); \
-                /* jl_puts(#hook " succeeded\n",jl_uv_stderr); */ \
-            } else { \
-                jl_rethrow(); \
-            } \
-        } \
-    }
+#define JULIA_CB(hook,val, ...) \
+    (!jl_old_base_module ? ( \
+        jl_callback_call(JULIA_HOOK(hook),(jl_value_t*)val,__VA_ARGS__) \
+    ) : ( \
+        jl_callback_call( \
+            JULIA_HOOK_(jl_base_relative_to(((jl_datatype_t*)jl_typeof(val))->name->module), hook), \
+            (jl_value_t*)val,__VA_ARGS__) \
+    ))
 
 jl_value_t *jl_callback_call(jl_function_t *f,jl_value_t *val,int count,...)
 {
     if (val != 0)
         count += 1;
+    else
+        return NULL;
     jl_value_t **argv;
     JL_GC_PUSHARGS(argv,count);
     memset(argv, 0, count*sizeof(jl_value_t*));
@@ -120,28 +120,35 @@ jl_value_t *jl_callback_call(jl_function_t *f,jl_value_t *val,int count,...)
         case CB_INT32:
             argv[i] = jl_box_int32(va_arg(argp,int32_t));
             break;
+        case CB_UINT32:
+            argv[i] = jl_box_uint32(va_arg(argp,uint32_t));
+            break;
         case CB_INT64:
             argv[i] = jl_box_int64(va_arg(argp,int64_t));
             break;
+        case CB_UINT64:
+            argv[i] = jl_box_uint64(va_arg(argp,uint64_t));
+            break;
         default: jl_error("callback: only Ints and Pointers are supported at this time");
             //excecution never reaches here
             break;
         }
     }
+    va_end(argp);
     v = jl_apply(f,(jl_value_t**)argv,count);
     JL_GC_POP();
     return v;
 }
 
-DLLEXPORT void jl_uv_closeHandle(uv_handle_t* handle)
+DLLEXPORT void jl_uv_closeHandle(uv_handle_t *handle)
 {
     if (handle->data) {
-        JULIA_CB(close,handle->data,0); (void)ret;
+        JULIA_CB(close,handle->data,0);
     }
     free(handle);
 }
 
-DLLEXPORT void jl_uv_shutdownCallback(uv_shutdown_t* req, int status)
+DLLEXPORT void jl_uv_shutdownCallback(uv_shutdown_t *req, int status)
 {
     /*
      * This happens if the remote machine closes the connecition while we're
@@ -149,77 +156,86 @@ DLLEXPORT void jl_uv_shutdownCallback(uv_shutdown_t* req, int status)
      * request.
      */
     if (status != UV__ECANCELED && !uv_is_closing((uv_handle_t*)req->handle)) {
-        uv_close((uv_handle_t*) req->handle, &jl_uv_closeHandle);
+        uv_close((uv_handle_t*)req->handle, &jl_uv_closeHandle);
     }
     free(req);
 }
 
-DLLEXPORT void jl_uv_return_spawn(uv_process_t *p, int exit_status, int term_signal)
+DLLEXPORT void jl_uv_return_spawn(uv_process_t *p, int64_t exit_status, int term_signal)
 {
-    JULIA_CB(return_spawn,p->data,2,CB_INT32,exit_status,CB_INT32,term_signal);
-    (void)ret;
+    JULIA_CB(return_spawn,p->data,2,CB_INT64,exit_status,CB_INT32,term_signal);
 }
 
-DLLEXPORT void jl_uv_readcb(uv_stream_t *handle, ssize_t nread, uv_buf_t buf)
+DLLEXPORT void jl_uv_readcb(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf)
 {
-    JULIA_CB(readcb,handle->data,3,CB_INT,nread,CB_PTR,(buf.base),CB_INT32,buf.len);
-    (void)ret;
+    JULIA_CB(readcb,handle->data,3,CB_INT,nread,CB_PTR,(buf->base),CB_UINT,buf->len);
 }
 
-DLLEXPORT uv_buf_t jl_uv_alloc_buf(uv_handle_t *handle, size_t suggested_size)
+DLLEXPORT void jl_uv_alloc_buf(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
 {
-    uv_buf_t buf;
-    JULIA_CB(alloc_buf,handle->data,1,CB_INT32,suggested_size);
-    if (!jl_is_tuple(ret) || !jl_is_pointer(jl_t0(ret)) || !jl_is_int32(jl_t1(ret))) {
-        jl_error("jl_alloc_buf: Julia function returned invalid value for buffer allocation callback");
+    if (handle->data) {
+        jl_value_t *ret = JULIA_CB(alloc_buf,handle->data,1,CB_UINT,suggested_size);
+        assert(jl_is_tuple(ret) && jl_is_pointer(jl_t0(ret)));
+        buf->base = (char*)jl_unbox_voidpointer(jl_t0(ret));
+#ifdef _P64
+        assert(jl_is_uint64(jl_t1(ret)));
+        buf->len = jl_unbox_uint64(jl_t1(ret));
+#else
+        assert(jl_is_uint32(jl_t1(ret)));
+        buf->len = jl_unbox_uint32(jl_t1(ret));
+#endif
+    }
+    else {
+        buf->len = 0;
     }
-    buf.base = jl_unbox_voidpointer(jl_t0(ret));
-    buf.len = jl_unbox_int32(jl_t1(ret));
-    return buf;
 }
 
 DLLEXPORT void jl_uv_connectcb(uv_connect_t *connect, int status)
 {
     JULIA_CB(connectcb,connect->handle->data,1,CB_INT32,status);
     free(connect);
-    (void)ret;
 }
 
 DLLEXPORT void jl_uv_connectioncb(uv_stream_t *stream, int status)
 {
     JULIA_CB(connectioncb,stream->data,1,CB_INT32,status);
-    (void)ret;
 }
 
 DLLEXPORT void jl_uv_getaddrinfocb(uv_getaddrinfo_t *req,int status, struct addrinfo *addr)
 {
     JULIA_CB(getaddrinfo,req->data,2,CB_PTR,addr,CB_INT32,status);
-    (void)ret;
 }
 
-DLLEXPORT void jl_uv_asynccb(uv_handle_t *handle, int status)
+DLLEXPORT void jl_uv_asynccb(uv_handle_t *handle)
 {
-    JULIA_CB(asynccb,handle->data,1,CB_INT32,status);
-    (void)ret;
+    JULIA_CB(asynccb,handle->data,0);
 }
 
 DLLEXPORT void jl_uv_pollcb(uv_poll_t *handle, int status, int events)
 {
-    JULIA_CB(pollcb,handle->data,2,CB_INT32,status,CB_INT32,events)
-    (void)ret;
+    JULIA_CB(pollcb,handle->data,2,CB_INT32,status,CB_INT32,events);
 }
 
-DLLEXPORT void jl_uv_fspollcb(uv_fs_poll_t* handle, int status, const uv_stat_t* prev, const uv_stat_t* curr)
+DLLEXPORT void jl_uv_fspollcb(uv_fs_poll_t *handle, int status, const uv_stat_t *prev, const uv_stat_t *curr)
 {
-    JULIA_CB(fspollcb,handle->data,3,CB_INT32,status,CB_PTR,prev,CB_PTR,curr)
-    (void)ret;
+    JULIA_CB(fspollcb,handle->data,3,CB_INT32,status,CB_PTR,prev,CB_PTR,curr);
 }
 
 
-DLLEXPORT void jl_uv_fseventscb(uv_fs_event_t* handle, const char* filename, int events, int status)
+DLLEXPORT void jl_uv_fseventscb(uv_fs_event_t *handle, const char *filename, int events, int status)
 {
-    JULIA_CB(fseventscb,handle->data,3,CB_PTR,filename,CB_INT32,events,CB_INT32,status)
-    (void)ret;
+    JULIA_CB(fseventscb,handle->data,3,CB_PTR,filename,CB_INT32,events,CB_INT32,status);
+}
+
+DLLEXPORT void jl_uv_recvcb(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, struct sockaddr* addr, unsigned flags)
+{
+    JULIA_CB(recv,handle->data,5,CB_INT,nread,CB_PTR,(buf->base),CB_UINT,buf->len,CB_PTR,addr,CB_INT32,flags);
+}
+
+DLLEXPORT void jl_uv_sendcb(uv_udp_send_t *handle, int status)
+{
+    JULIA_CB(send,handle->data,1,CB_INT32,status);
+    free(handle);
 }
 
 /** This file contains wrappers for most of libuv's stream functionailty. Once we can allocate structs in Julia, this file will be removed */
@@ -250,7 +266,7 @@ DLLEXPORT int jl_process_events(uv_loop_t *loop)
     else return 0;
 }
 
-DLLEXPORT int jl_init_pipe(uv_pipe_t *pipe, int writable, int readable, int julia_only, jl_value_t *julia_struct)
+DLLEXPORT int jl_init_pipe(uv_pipe_t *pipe, int writable, int readable, int julia_only)
 {
      int flags = 0;
      flags |= writable ? UV_PIPE_WRITABLE : 0;
@@ -258,7 +274,6 @@ DLLEXPORT int jl_init_pipe(uv_pipe_t *pipe, int writable, int readable, int juli
      if (!julia_only)
          flags |= UV_PIPE_SPAWN_SAFE;
      int err = uv_pipe_init(jl_io_loop, pipe, flags);
-     pipe->data = julia_struct;//will be initilized on io
      return err;
 }
 
@@ -267,18 +282,9 @@ DLLEXPORT void jl_close_uv(uv_handle_t *handle)
     if (handle->type==UV_TTY)
         uv_tty_set_mode((uv_tty_t*)handle,0);
 
-    if ( (handle->type == UV_NAMED_PIPE || handle->type == UV_TCP) && uv_is_writable( (uv_stream_t *) handle)) { 
-        // Make sure that the stream has not already been marked closed in Julia.
-        // A double shutdown would cause the process to hang on exit.
-        JULIA_CB(isopen, handle->data, 0);
-        if (!jl_is_int32(ret)) {
-            jl_error("jl_close_uv: _uv_hook_isopen must return an int32.");
-        }
-        if (!jl_unbox_int32(ret)){
-            return;
-        }
-
-        uv_shutdown_t *req = malloc(sizeof(uv_shutdown_t));
+    if ((handle->type == UV_NAMED_PIPE || handle->type == UV_TCP) &&
+        uv_is_writable((uv_stream_t*)handle)) {
+        uv_shutdown_t *req = (uv_shutdown_t*)malloc(sizeof(uv_shutdown_t));
         req->data = 0;
         /*
          * We are explicity ignoring the error here for the following reason:
@@ -319,7 +325,7 @@ DLLEXPORT void jl_uv_disassociate_julia_struct(uv_handle_t *handle)
     handle->data = NULL;
 }
 
-DLLEXPORT int jl_listen(uv_stream_t* stream, int backlog)
+DLLEXPORT int jl_listen(uv_stream_t *stream, int backlog)
 {
     return uv_listen(stream,backlog,&jl_uv_connectioncb);
 }
@@ -329,7 +335,7 @@ DLLEXPORT int jl_spawn(char *name, char **argv, uv_loop_t *loop,
                        uv_handle_type stdin_type, uv_pipe_t *stdin_pipe,
                        uv_handle_type stdout_type, uv_pipe_t *stdout_pipe,
                        uv_handle_type stderr_type, uv_pipe_t *stderr_pipe, 
-                       int detach, char **env)
+                       int detach, char **env, char *cwd)
 {
     uv_process_options_t opts;
     uv_stdio_container_t stdio[3];
@@ -341,7 +347,7 @@ DLLEXPORT int jl_spawn(char *name, char **argv, uv_loop_t *loop,
 #else
     opts.flags = UV_PROCESS_RESET_SIGPIPE;
 #endif
-    opts.cwd = NULL;
+    opts.cwd = cwd;
     opts.args = argv;
     if (detach)
         opts.flags |= UV_PROCESS_DETACHED;
@@ -354,7 +360,7 @@ DLLEXPORT int jl_spawn(char *name, char **argv, uv_loop_t *loop,
     stdio[2].type = stderr_type;
     stdio[2].data.stream = (uv_stream_t*)(stderr_pipe);
     opts.exit_cb = &jl_uv_return_spawn;
-    error = uv_spawn(loop,proc,opts);
+    error = uv_spawn(loop,proc,&opts);
     return error;
 }
 
@@ -362,7 +368,7 @@ DLLEXPORT int jl_spawn(char *name, char **argv, uv_loop_t *loop,
 #include <time.h>
 DLLEXPORT struct tm* localtime_r(const time_t *t, struct tm *tm)
 {
-    auto struct tm *tmp = localtime(t); //localtime is reentrant on windows
+    struct tm *tmp = localtime(t); //localtime is reentrant on windows
     if (tmp)
         *tm = *tmp;
     return tmp;
@@ -374,39 +380,75 @@ DLLEXPORT struct tm* localtime_r(const time_t *t, struct tm *tm)
     return uv_loop_new();
 }*/
 
-DLLEXPORT uv_loop_t *jl_global_event_loop()
+DLLEXPORT uv_loop_t *jl_global_event_loop(void)
 {
     return jl_io_loop;
 }
 
-DLLEXPORT int jl_poll_start(uv_poll_t* handle, int32_t events)
+DLLEXPORT int jl_poll_start(uv_poll_t *handle, int32_t events)
 {
     return uv_poll_start(handle, events, &jl_uv_pollcb);
 }
 
-DLLEXPORT int jl_fs_poll_start(uv_fs_poll_t* handle, char *file, uint32_t interval)
+DLLEXPORT int jl_fs_poll_start(uv_fs_poll_t *handle, char *file, uint32_t interval)
 {
     return uv_fs_poll_start(handle,&jl_uv_fspollcb,file,interval);
 }
 
-DLLEXPORT int jl_fs_event_init(uv_loop_t* loop, uv_fs_event_t* handle,
-    const char* filename, int flags)
+DLLEXPORT int jl_fs_event_init(uv_loop_t *loop, uv_fs_event_t *handle,
+                               const char *filename, int flags)
 {
-    return uv_fs_event_init(loop,handle,filename,&jl_uv_fseventscb,flags);
+    uv_fs_event_init(loop,handle); 
+    return uv_fs_event_start(handle,&jl_uv_fseventscb,filename,flags);
 }
 
 DLLEXPORT int jl_fs_unlink(char *path)
 {
     uv_fs_t req;
+    JL_SIGATOMIC_BEGIN();
     int ret = uv_fs_unlink(jl_io_loop, &req, path, NULL);
     uv_fs_req_cleanup(&req);
+    JL_SIGATOMIC_END();
+    return ret;
+}
+
+DLLEXPORT int jl_fs_rename(char *src_path, char *dst_path)
+{
+    uv_fs_t req;
+    JL_SIGATOMIC_BEGIN();
+    int ret = uv_fs_rename(jl_io_loop, &req, src_path, dst_path, NULL);
+    uv_fs_req_cleanup(&req);
+    JL_SIGATOMIC_END();
+    return ret;
+}
+
+DLLEXPORT int jl_fs_sendfile(int src_fd, int dst_fd,
+                             int64_t in_offset, size_t len)
+{
+    uv_fs_t req;
+    JL_SIGATOMIC_BEGIN();
+    int ret = uv_fs_sendfile(jl_io_loop, &req, dst_fd, src_fd,
+                             in_offset, len, NULL);
+    uv_fs_req_cleanup(&req);
+    JL_SIGATOMIC_END();
+    return ret;
+}
+
+DLLEXPORT int jl_fs_symlink(char *path, char *new_path, int flags)
+{
+    uv_fs_t req;
+    int ret = uv_fs_symlink(jl_io_loop, &req, path, new_path, flags, NULL);
+    uv_fs_req_cleanup(&req);
     return ret;
 }
 
-DLLEXPORT int jl_fs_write(int handle, char *buf, size_t len, size_t offset)
+DLLEXPORT int jl_fs_write(int handle, char *data, size_t len, size_t offset)
 {
     uv_fs_t req;
-    int ret = uv_fs_write(jl_io_loop, &req, handle, buf, len, offset, NULL);
+    uv_buf_t buf[1];
+    buf[0].base = data;
+    buf[0].len = len;
+    int ret = uv_fs_write(jl_io_loop, &req, handle, buf, 1, offset, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
@@ -414,15 +456,21 @@ DLLEXPORT int jl_fs_write(int handle, char *buf, size_t len, size_t offset)
 DLLEXPORT int jl_fs_write_byte(int handle, char c)
 {
     uv_fs_t req;
-    int ret = uv_fs_write(jl_io_loop, &req, handle, &c, 1, -1, NULL);
+    uv_buf_t buf[1];
+    buf[0].base = &c;
+    buf[0].len = 1;
+    int ret = uv_fs_write(jl_io_loop, &req, handle, buf, 1, -1, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
 
-DLLEXPORT int jl_fs_read(int handle, char *buf, size_t len)
+DLLEXPORT int jl_fs_read(int handle, char *data, size_t len)
 {
     uv_fs_t req;
-    int ret = uv_fs_read(jl_io_loop, &req, handle, buf, len, -1, NULL);
+    uv_buf_t buf[1];
+    buf[0].base = data;
+    buf[0].len = len;
+    int ret = uv_fs_read(jl_io_loop, &req, handle, buf, 1, -1, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
@@ -430,12 +478,15 @@ DLLEXPORT int jl_fs_read(int handle, char *buf, size_t len)
 DLLEXPORT int jl_fs_read_byte(int handle)
 {
     uv_fs_t req;
-    char buf;
-    int ret = uv_fs_read(jl_io_loop, &req, handle, &buf, 1, -1, NULL);
+    char c;
+    uv_buf_t buf[1];
+    buf[0].base = &c;
+    buf[0].len = 1;
+    int ret = uv_fs_read(jl_io_loop, &req, handle, buf, 1, -1, NULL);
     uv_fs_req_cleanup(&req);
     if (ret == -1)
         return ret;
-    return (int)buf;
+    return (int)c;
 }
 
 DLLEXPORT int jl_fs_close(int handle)
@@ -449,21 +500,22 @@ DLLEXPORT int jl_fs_close(int handle)
 //units are in ms
 DLLEXPORT int jl_puts(char *str, uv_stream_t *stream)
 {
+    if (!stream) return 0;
     return jl_write(stream,str,strlen(str));
 }
 
-DLLEXPORT void jl_uv_writecb(uv_write_t* req, int status)
+DLLEXPORT void jl_uv_writecb(uv_write_t *req, int status)
 {
-    JULIA_CB(writecb, req->handle->data, 2, CB_PTR, req, CB_INT32, status)
+    if (req->data) {
+        JULIA_CB(writecb, req->data, 2, CB_PTR, req, CB_INT32, status);
+    }
     free(req);
-    (void)ret;
 }
 
-DLLEXPORT void jl_uv_writecb_task(uv_write_t* req, int status)
+DLLEXPORT void jl_uv_writecb_task(uv_write_t *req, int status)
 {
-    JULIA_CB(writecb_task, req->handle->data, 2, CB_PTR, req, CB_INT32, status)
+    JULIA_CB(writecb_task, req->handle->data, 2, CB_PTR, req, CB_INT32, status);
     free(req);
-    (void)ret;
 }
 
 DLLEXPORT int jl_write_copy(uv_stream_t *stream, const char *str, size_t n, uv_write_t *uvw, void *writecb)
@@ -471,14 +523,16 @@ DLLEXPORT int jl_write_copy(uv_stream_t *stream, const char *str, size_t n, uv_w
     JL_SIGATOMIC_BEGIN();
     char *data = (char*)(uvw+1);
     memcpy(data,str,n);
-    uv_buf_t buf[]  = {{.base = data,.len=n}};
+    uv_buf_t buf[1];
+    buf[0].base = data;
+    buf[0].len = n;
     uvw->data = NULL;
-    int err = uv_write(uvw,stream,buf,1,writecb);  
+    int err = uv_write(uvw,stream,buf,1,(uv_write_cb)writecb);
     JL_SIGATOMIC_END();
     return err;
 }
 
-DLLEXPORT int jl_putc(unsigned char c, uv_stream_t *stream)
+DLLEXPORT int jl_putc(char c, uv_stream_t *stream)
 {
     int err;
     if (stream!=0) {
@@ -488,13 +542,16 @@ DLLEXPORT int jl_putc(unsigned char c, uv_stream_t *stream)
                 jl_uv_file_t *file = (jl_uv_file_t *)stream;
                 // Do a blocking write for now
                 uv_fs_t req;
-                err = uv_fs_write(file->loop, &req, file->file, &c, 1, -1, NULL);
+                uv_buf_t buf[1];
+                buf[0].base = &c;
+                buf[0].len = 1;
+                err = uv_fs_write(file->loop, &req, file->file, buf, 1, -1, NULL);
                 JL_SIGATOMIC_END();
                 return err ? 0 : 1;
             }
             else {
-                uv_write_t *uvw = malloc(sizeof(uv_write_t)+1);
-                err = jl_write_copy(stream,(char*)&c,1,uvw,&jl_uv_writecb);
+                uv_write_t *uvw = (uv_write_t*)malloc(sizeof(uv_write_t)+1);
+                err = jl_write_copy(stream,(char*)&c,1,uvw, (void*)&jl_uv_writecb);
                 if (err < 0) {
                     free(uvw);
                     return 0;
@@ -512,9 +569,11 @@ DLLEXPORT int jl_putc(unsigned char c, uv_stream_t *stream)
 
 DLLEXPORT int jl_write_no_copy(uv_stream_t *stream, char *data, size_t n, uv_write_t *uvw, void *writecb)
 {
-    uv_buf_t buf[]  = {{.base = data,.len=n}};
+    uv_buf_t buf[1];
+    buf[0].base = data;
+    buf[0].len = n;
     JL_SIGATOMIC_BEGIN();
-    int err = uv_write(uvw,stream,buf,1,writecb);
+    int err = uv_write(uvw,stream,buf,1,(uv_write_cb)writecb);
     uvw->data = NULL;
     JL_SIGATOMIC_END();
     return err;
@@ -522,7 +581,7 @@ DLLEXPORT int jl_write_no_copy(uv_stream_t *stream, char *data, size_t n, uv_wri
 
 DLLEXPORT int jl_putc_copy(unsigned char c, uv_stream_t *stream, void *uvw, void *writecb)
 {
-    return jl_write_copy(stream,(char *)&c,1,uvw,writecb);
+    return jl_write_copy(stream,(char *)&c,1,(uv_write_t*)uvw,writecb);
 }
 
 DLLEXPORT int jl_pututf8(uv_stream_t *s, uint32_t wchar )
@@ -540,7 +599,7 @@ DLLEXPORT int jl_pututf8_copy(uv_stream_t *s, uint32_t wchar, void *uvw, void *w
     if (wchar < 0x80)
         return jl_putc_copy((int)wchar, s, uvw, writecb);
     size_t n = u8_toutf8(buf, 8, &wchar, 1);
-    return jl_write_copy(s, buf, n, uvw, writecb);
+    return jl_write_copy(s, buf, n, (uv_write_t*)uvw, writecb);
 }
 
 DLLEXPORT size_t jl_write(uv_stream_t *stream, const char *str, size_t n)
@@ -555,13 +614,16 @@ DLLEXPORT size_t jl_write(uv_stream_t *stream, const char *str, size_t n)
             jl_uv_file_t *file = (jl_uv_file_t *)stream;
             // Do a blocking write for now
             uv_fs_t req;
-            err = uv_fs_write(file->loop, &req, file->file, (void*)str, n, -1, NULL);
+            uv_buf_t buf[1];
+            buf[0].base = (char*)str;
+            buf[0].len = n;
+            err = uv_fs_write(file->loop, &req, file->file, buf, 1, -1, NULL);
             JL_SIGATOMIC_END();
             return err ? 0 : n;
         }
         else {
-            uv_write_t *uvw = malloc(sizeof(uv_write_t)+n);
-            err = jl_write_copy(stream,str,n,uvw,&jl_uv_writecb);
+            uv_write_t *uvw = (uv_write_t*)malloc(sizeof(uv_write_t)+n);
+            err = jl_write_copy(stream,str,n,uvw, (void*)&jl_uv_writecb);
             if (err < 0) {
                 free(uvw);
                 return 0;
@@ -625,11 +687,6 @@ DLLEXPORT void jl_exit(int exitcode)
     exit(exitcode);
 }
 
-DLLEXPORT int jl_cwd(char *buffer, size_t size)
-{
-    return uv_cwd(buffer,size);
-}
-
 DLLEXPORT int jl_getpid()
 {
 #ifdef _OS_WINDOWS_
@@ -640,23 +697,72 @@ DLLEXPORT int jl_getpid()
 }
 
 //NOTE: These function expects port/host to be in network byte-order (Big Endian)
-DLLEXPORT int jl_tcp_bind(uv_tcp_t* handle, uint16_t port, uint32_t host)
+DLLEXPORT int jl_tcp_bind(uv_tcp_t *handle, uint16_t port, uint32_t host, unsigned int flags)
 {
     struct sockaddr_in addr;
     memset(&addr, 0, sizeof(struct sockaddr_in));
     addr.sin_port = port;
     addr.sin_addr.s_addr = host;
     addr.sin_family = AF_INET;
-    return uv_tcp_bind(handle, addr);
+    return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
 }
-DLLEXPORT int jl_tcp_bind6(uv_tcp_t* handle, uint16_t port, void *host)
+DLLEXPORT int jl_tcp_bind6(uv_tcp_t *handle, uint16_t port, void *host, unsigned int flags)
 {
     struct sockaddr_in6 addr;
     memset(&addr, 0, sizeof(struct sockaddr_in6));
     addr.sin6_port = port;
     memcpy(&addr.sin6_addr, host, 16);
     addr.sin6_family = AF_INET6;
-    return uv_tcp_bind6(handle, addr);
+    return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
+}
+
+DLLEXPORT int jl_udp_bind(uv_udp_t *handle, uint16_t port, uint32_t host, uint32_t flags)
+{
+    struct sockaddr_in addr;
+    memset(&addr, 0, sizeof(struct sockaddr_in));
+    addr.sin_port = port;
+    addr.sin_addr.s_addr = host;
+    addr.sin_family = AF_INET;
+    return uv_udp_bind(handle, (struct sockaddr*)&addr, flags);
+}
+DLLEXPORT int jl_udp_bind6(uv_udp_t *handle, uint16_t port, void *host, uint32_t flags)
+{
+    struct sockaddr_in6 addr;
+    memset(&addr, 0, sizeof(struct sockaddr_in6));
+    addr.sin6_port = port;
+    memcpy(&addr.sin6_addr, host, 16);
+    addr.sin6_family = AF_INET6;
+    return uv_udp_bind(handle, (struct sockaddr*)&addr, flags);
+}
+
+DLLEXPORT int jl_udp_send(uv_udp_t *handle, uint16_t port, uint32_t host, void *data, uint32_t size)
+{
+    struct sockaddr_in addr;
+    memset(&addr, 0, sizeof(struct sockaddr_in));
+    addr.sin_port = port;
+    addr.sin_addr.s_addr = host;
+    addr.sin_family = AF_INET;
+    uv_buf_t buf[1];
+    buf[0].base = (char *) data;
+    buf[0].len = size;
+    uv_udp_send_t *req = (uv_udp_send_t*)malloc(sizeof(uv_udp_send_t));
+    req->data = handle->data;
+    return uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, &jl_uv_sendcb);
+}
+
+DLLEXPORT int jl_udp_send6(uv_udp_t *handle, uint16_t port, void *host, void *data, uint32_t size)
+{
+    struct sockaddr_in6 addr;
+    memset(&addr, 0, sizeof(struct sockaddr_in6));
+    addr.sin6_port = port;
+    memcpy(&addr.sin6_addr, host, 16);
+    addr.sin6_family = AF_INET6;
+    uv_buf_t buf[1];
+    buf[0].base = (char *) data;
+    buf[0].len = size;
+    uv_udp_send_t *req = (uv_udp_send_t *) malloc(sizeof(uv_udp_send_t));
+    req->data = handle->data;
+    return uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, &jl_uv_sendcb);
 }
 
 DLLEXPORT int jl_uv_sizeof_interface_address()
@@ -681,7 +787,7 @@ DLLEXPORT struct sockaddr_in *jl_uv_interface_address_sockaddr(uv_interface_addr
 
 DLLEXPORT int jl_getaddrinfo(uv_loop_t *loop, const char *host, const char *service, jl_function_t *cb)
 {
-    uv_getaddrinfo_t *req = malloc(sizeof(uv_getaddrinfo_t));
+    uv_getaddrinfo_t *req = (uv_getaddrinfo_t*)malloc(sizeof(uv_getaddrinfo_t));
     struct addrinfo hints;
 
     memset (&hints, 0, sizeof (hints));
@@ -747,41 +853,47 @@ DLLEXPORT void jl_sockaddr_set_port(struct sockaddr_storage *addr,uint16_t port)
 DLLEXPORT int jl_tcp4_connect(uv_tcp_t *handle,uint32_t host, uint16_t port)
 {
     struct sockaddr_in addr;
-    uv_connect_t *req = malloc(sizeof(uv_connect_t));
+    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
     req->data = 0;
     memset(&addr, 0, sizeof(struct sockaddr_in));
     addr.sin_family = AF_INET;
     addr.sin_addr.s_addr = host;
     addr.sin_port = port;
-    return uv_tcp_connect(req,handle,addr,&jl_uv_connectcb);
+    return uv_tcp_connect(req,handle,(struct sockaddr*)&addr,&jl_uv_connectcb);
 }
 
 DLLEXPORT int jl_tcp6_connect(uv_tcp_t *handle, void *host, uint16_t port)
 {
     struct sockaddr_in6 addr;
-    uv_connect_t *req = malloc(sizeof(uv_connect_t));
+    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
     req->data = 0;
     memset(&addr, 0, sizeof(struct sockaddr_in6));
     addr.sin6_family = AF_INET6;
     memcpy(&addr.sin6_addr, host, 16);
     addr.sin6_port = port;
-    return uv_tcp_connect6(req,handle,addr,&jl_uv_connectcb);
+    return uv_tcp_connect(req,handle,(struct sockaddr*)&addr,&jl_uv_connectcb);
 }
 
 DLLEXPORT int jl_connect_raw(uv_tcp_t *handle,struct sockaddr_storage *addr)
 {
-    uv_connect_t *req = malloc(sizeof(uv_connect_t));
+    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
     req->data = 0;
-    if (addr->ss_family==AF_INET) {
-        return uv_tcp_connect(req,handle,*((struct sockaddr_in*)addr),&jl_uv_connectcb);
-    }
-    else {
-        return uv_tcp_connect6(req,handle,*((struct sockaddr_in6*)addr),&jl_uv_connectcb);
-    }
-    free(req);
-    return -2; //error! Only IPv4 and IPv6 are implemented atm
+    return uv_tcp_connect(req,handle,(struct sockaddr*)addr,&jl_uv_connectcb);
 }
 
+#ifdef _OS_LINUX_
+DLLEXPORT int jl_tcp_quickack(uv_tcp_t *handle, int on)
+{
+    int fd = (handle)->io_watcher.fd;
+    if (fd != -1) {
+        if (setsockopt(fd, IPPROTO_TCP, TCP_QUICKACK, &on, sizeof(on))) {
+            return -1;
+        }
+    }
+    return 0;
+}
+#endif
+
 DLLEXPORT char *jl_ios_buf_base(ios_t *ios)
 {
     return ios->buf;
@@ -789,8 +901,12 @@ DLLEXPORT char *jl_ios_buf_base(ios_t *ios)
 
 DLLEXPORT uv_lib_t *jl_wrap_raw_dl_handle(void *handle)
 {
-    uv_lib_t *lib = malloc(sizeof(uv_lib_t));
+    uv_lib_t *lib = (uv_lib_t*)malloc(sizeof(uv_lib_t));
+    #ifdef _OS_WINDOWS_
+    lib->handle=(HMODULE)handle;
+    #else
     lib->handle=handle;
+    #endif
     lib->errmsg=NULL;
     return lib;
 }
@@ -810,11 +926,60 @@ DLLEXPORT int jl_uv_unix_fd_is_watched(int fd, uv_poll_t *handle, uv_loop_t *loo
 
 #endif
 
+#ifdef _OS_WINDOWS_
+static inline int ishexchar(char c)
+{
+   if (c >= '0' && c <= '9') return 1;
+   if (c >= 'a' && c <= 'z') return 1;
+   return 0;
+}
+
+DLLEXPORT int jl_ispty(uv_pipe_t *pipe)
+{
+    if (pipe->type != UV_NAMED_PIPE) return 0;
+    size_t len = 0;
+    if (uv_pipe_getsockname(pipe, NULL, &len) != UV_ENOBUFS) return 0;
+    char *name = (char *) alloca(len);
+    if (uv_pipe_getsockname(pipe, name, &len)) return 0;
+    // return true if name matches regex:
+    // ^\\\\?\\pipe\\(msys|cygwin)-[0-9a-z]{16}-[pt]ty[1-9][0-9]*-
+    //JL_PRINTF(JL_STDERR,"pipe_name: %s\n", name);
+    int n = 0;
+    if (!strncmp(name,"\\\\?\\pipe\\msys-",14))
+        n = 14;
+    else if (!strncmp(name,"\\\\?\\pipe\\cygwin-",16)) 
+        n = 16;
+    else
+        return 0;
+    //JL_PRINTF(JL_STDERR,"prefix pass\n");
+    name += n;
+    for (int n = 0; n < 16; n++)
+        if (!ishexchar(*name++)) return 0;
+    //JL_PRINTF(JL_STDERR,"hex pass\n");
+    if ((*name++)!='-') return 0;
+    if (*name != 'p' && *name != 't') return 0;
+    name++;
+    if (*name++ != 't' || *name++ != 'y') return 0;
+    //JL_PRINTF(JL_STDERR,"tty pass\n");
+    return 1;
+}
+#endif
+ 
 DLLEXPORT uv_handle_type jl_uv_handle_type(uv_handle_t *handle)
 {
+#ifdef _OS_WINDOWS_
+    if (jl_ispty((uv_pipe_t*)handle))
+        return UV_TTY;
+#endif
     return handle->type;
 }
 
+DLLEXPORT int jl_tty_set_mode(uv_tty_t *handle, int mode)
+{
+    if (handle->type != UV_TTY) return 0;
+    return uv_tty_set_mode(handle, mode);
+}
+
 DLLEXPORT uv_file jl_uv_file_handle(jl_uv_file_t *f)
 {
     return f->file;
@@ -838,19 +1003,26 @@ DLLEXPORT void *jl_uv_handle_data(uv_handle_t *handle)
 
 #ifndef _OS_WINDOWS_
 #if defined(__APPLE__)
-int uv___stream_fd(uv_stream_t* handle);
-#define uv__stream_fd(handle) (uv___stream_fd((uv_stream_t*) (handle)))
+int uv___stream_fd(uv_stream_t *handle);
+#define uv__stream_fd(handle) (uv___stream_fd((uv_stream_t*)(handle)))
 #else
 #define uv__stream_fd(handle) ((handle)->io_watcher.fd)
 #endif /* defined(__APPLE__) */
-DLLEXPORT int jl_uv_pipe_fd(uv_pipe_t *handle)
+DLLEXPORT int jl_uv_handle(uv_stream_t *handle)
 {
-    return uv__stream_fd((uv_stream_t*)handle);
+    return uv__stream_fd(handle);
 }
 #else
-DLLEXPORT HANDLE jl_uv_pipe_handle(uv_pipe_t *handle)
+DLLEXPORT HANDLE jl_uv_handle(uv_stream_t *handle)
 {
-    return handle->handle;
+    switch (handle->type) {
+    case UV_TTY:
+        return ((uv_tty_t*)handle)->handle;
+    case UV_NAMED_PIPE:
+        return ((uv_pipe_t*)handle)->handle;
+    default:
+        return INVALID_HANDLE_VALUE;
+    }
 }
 #endif
 #ifdef __cplusplus
