@@ -3,6 +3,11 @@
 */
 #include <assert.h>
 #include "julia.h"
+#include "julia_internal.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 jl_module_t *jl_main_module=NULL;
 jl_module_t *jl_core_module=NULL;
@@ -12,7 +17,9 @@ jl_module_t *jl_current_module=NULL;
 jl_module_t *jl_new_module(jl_sym_t *name)
 {
     jl_module_t *m = (jl_module_t*)allocobj(sizeof(jl_module_t));
+    JL_GC_PUSH1(&m);
     m->type = (jl_value_t*)jl_module_type;
+    assert(jl_is_symbol(name));
     m->name = name;
     m->constant_table = NULL;
     htable_new(&m->bindings, 0);
@@ -23,6 +30,7 @@ jl_module_t *jl_new_module(jl_sym_t *name)
     // export own name, so "using Foo" makes "Foo" itself visible
     jl_set_const(m, name, (jl_value_t*)m);
     jl_module_export(m, name);
+    JL_GC_POP();
     return m;
 }
 
@@ -112,7 +120,7 @@ jl_binding_t *jl_get_binding_for_method_def(jl_module_t *m, jl_sym_t *var)
 }
 
 static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
-                           int explicit);
+                           int explici);
 
 typedef struct _modstack_t {
     jl_module_t *m;
@@ -139,7 +147,8 @@ static jl_binding_t *jl_get_binding_(jl_module_t *m, jl_sym_t *var, modstack_t *
             if (b != HT_NOTFOUND && b->exportp) {
                 b = jl_get_binding_(imp, var, &top);
                 if (b == NULL || b->owner == NULL)
-                    return NULL;
+                    // couldn't resolve; try next using (see issue #6105)
+                    continue;
                 // do a full import to prevent the result of this lookup
                 // from changing, for example if this var is assigned to
                 // later.
@@ -167,8 +176,9 @@ static int eq_bindings(jl_binding_t *a, jl_binding_t *b)
     return 0;
 }
 
+// NOTE: we use explici since explicit is a C++ keyword
 static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
-                           int explicit)
+                           int explici)
 {
     if (to == from)
         return;
@@ -187,14 +197,14 @@ static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
             }
             else if (bto->owner == b->owner) {
                 // already imported
-                bto->imported = (explicit!=0);
+                bto->imported = (explici!=0);
             }
             else if (bto->owner != to && bto->owner != NULL) {
                 // already imported from somewhere else
                 jl_binding_t *bval = jl_get_binding(to, s);
                 if (bval->constp && bval->value && b->constp && b->value == bval->value) {
                     // equivalent binding
-                    bto->imported = (explicit!=0);
+                    bto->imported = (explici!=0);
                     return;
                 }
                 jl_printf(JL_STDERR,
@@ -214,13 +224,13 @@ static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
             }
             else {
                 bto->owner = b->owner;
-                bto->imported = (explicit!=0);
+                bto->imported = (explici!=0);
             }
         }
         else {
             jl_binding_t *nb = new_binding(s);
             nb->owner = b->owner;
-            nb->imported = (explicit!=0);
+            nb->imported = (explici!=0);
             *bp = nb;
         }
     }
@@ -346,20 +356,21 @@ DLLEXPORT int jl_is_const(jl_module_t *m, jl_sym_t *var)
     return b && b->constp;
 }
 
-void jl_checked_assignment(jl_binding_t *b, jl_value_t *rhs)
+DLLEXPORT void jl_checked_assignment(jl_binding_t *b, jl_value_t *rhs)
 {
     if (b->constp && b->value != NULL) {
-        if (!jl_egal(rhs, b->value) &&
-            (jl_typeof(rhs) != jl_typeof(b->value) ||
-             jl_is_type(rhs) || jl_is_function(rhs) || jl_is_module(rhs))) {
-            jl_errorf("invalid redefinition of constant %s", b->name->name);
+        if (!jl_egal(rhs, b->value)) {
+            if (jl_typeof(rhs) != jl_typeof(b->value) ||
+                jl_is_type(rhs) || jl_is_function(rhs) || jl_is_module(rhs)) {
+                jl_errorf("invalid redefinition of constant %s", b->name->name);
+            }
+            JL_PRINTF(JL_STDERR,"Warning: redefining constant %s\n",b->name->name);
         }
-        JL_PRINTF(JL_STDERR,"Warning: redefining constant %s\n",b->name->name);
     }
     b->value = rhs;
 }
 
-void jl_declare_constant(jl_binding_t *b)
+DLLEXPORT void jl_declare_constant(jl_binding_t *b)
 {
     if (b->value != NULL && !b->constp) {
         jl_errorf("cannot declare %s constant; it already has a value",
@@ -414,3 +425,27 @@ DLLEXPORT jl_value_t *jl_module_names(jl_module_t *m, int all, int imported)
 
 DLLEXPORT jl_sym_t *jl_module_name(jl_module_t *m) { return m->name; }
 DLLEXPORT jl_module_t *jl_module_parent(jl_module_t *m) { return m->parent; }
+
+int jl_module_has_initializer(jl_module_t *m)
+{
+    return jl_get_global(m, jl_symbol("__init__")) != NULL;
+}
+
+void jl_module_run_initializer(jl_module_t *m)
+{
+    jl_value_t *f = jl_get_global(m, jl_symbol("__init__"));
+    if (f == NULL || !jl_is_function(f))
+        return;
+    JL_TRY {
+        jl_apply((jl_function_t*)f, NULL, 0);
+    }
+    JL_CATCH {
+        JL_PRINTF(JL_STDERR, "Warning: error initializing module %s:\n", m->name->name);
+        jl_static_show(JL_STDERR, jl_exception_in_transit);
+        JL_PRINTF(JL_STDERR, "\n");
+    }
+}
+
+#ifdef __cplusplus
+}
+#endif
