@@ -1,13 +1,16 @@
 #ifndef DTYPES_H
 #define DTYPES_H
 
-#include <inttypes.h>
 #include <stddef.h>
 #include <stddef.h> // double include of stddef.h fixes #3421
 #include <stdint.h>
 
 #include "platform.h"
 
+#if !defined(_OS_WINDOWS_)
+#include <inttypes.h>
+#endif
+
 #if defined(_OS_WINDOWS_)
 
 #include <stdio.h>
@@ -53,10 +56,10 @@
 
 #ifdef _OS_WINDOWS_
 #define STDCALL __stdcall
-# ifdef IMPORT_EXPORTS
-#  define DLLEXPORT __declspec(dllimport)
-# else
+# ifdef LIBRARY_EXPORTS
 #  define DLLEXPORT __declspec(dllexport)
+# else
+#  define DLLEXPORT __declspec(dllimport)
 # endif
 #else
 #define STDCALL
@@ -112,9 +115,12 @@
 #define LLT_REALLOC(p,n) realloc((p),(n))
 #define LLT_FREE(x) free(x)
 
-#if defined(_COMPILER_INTEL_) && defined(_OS_WINDOWS_)
-# define STATIC_INLINE static
-# define INLINE
+#if defined(_OS_WINDOWS_) && defined(_COMPILER_INTEL_)
+#  define STATIC_INLINE static
+#  define INLINE
+#elif defined(_OS_WINDOWS_) && defined(_COMPILER_MICROSOFT_)
+#  define STATIC_INLINE static __inline
+#  define INLINE __inline
 #else
 # define STATIC_INLINE static inline
 # define INLINE inline
@@ -167,20 +173,6 @@ typedef uptrint_t u_ptrint_t;
 #define S32_MIN    (-S32_MAX - 1L)
 #define BIT31      0x80000000
 
-#define DBL_EPSILON      2.2204460492503131e-16
-#define FLT_EPSILON      1.192092896e-7
-#define DBL_MAX          1.7976931348623157e+308
-#define DBL_MIN          2.2250738585072014e-308
-#define FLT_MAX          3.402823466e+38
-#define FLT_MIN          1.175494351e-38
-#define LOG2_10          3.3219280948873626
-#define rel_zero(a, b) (fabs((a)/(b)) < DBL_EPSILON)
-#define sign_bit(r) ((*(int64_t*)&(r)) & BIT63)
-#define LABS(n) (((n)^((n)>>(NBITS-1))) - ((n)>>(NBITS-1)))
-#define NBABS(n,nb) (((n)^((n)>>((nb)-1))) - ((n)>>((nb)-1)))
-#define DFINITE(d) (((*(int64_t*)&(d))&0x7ff0000000000000LL)!=0x7ff0000000000000LL)
-#define DNAN(d) ((d)!=(d))
-
 extern double D_PNAN;
 extern double D_NNAN;
 extern double D_PINF;
@@ -204,4 +196,3 @@ typedef enum { T_INT8, T_UINT8, T_INT16, T_UINT16, T_INT32, T_UINT32,
 #endif
 
 #endif /* DTYPES_H */
-
