@@ -8,6 +8,10 @@
 #include "timefuncs.h"
 #include "ios.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 uint_t nextipow2(uint_t i)
 {
     if (i==0) return 1;
@@ -59,7 +63,7 @@ u_int32_t int64to32hash(u_int64_t key)
 
 #define _MHASH_SEED_ 0xcafe8881
 
-uint64_t memhash(const char* buf, size_t n)
+uint64_t memhash(const char *buf, size_t n)
 {
     uint64_t out[2];
 
@@ -72,7 +76,7 @@ uint64_t memhash(const char* buf, size_t n)
     return out[1];
 }
 
-uint64_t memhash_seed(const char* buf, size_t n, uint32_t seed)
+uint64_t memhash_seed(const char *buf, size_t n, uint32_t seed)
 {
     uint64_t out[2];
 
@@ -85,7 +89,7 @@ uint64_t memhash_seed(const char* buf, size_t n, uint32_t seed)
     return out[1];
 }
 
-uint32_t memhash32(const char* buf, size_t n)
+uint32_t memhash32(const char *buf, size_t n)
 {
     uint32_t out;
 
@@ -93,10 +97,14 @@ uint32_t memhash32(const char* buf, size_t n)
     return out;
 }
 
-uint32_t memhash32_seed(const char* buf, size_t n, uint32_t seed)
+uint32_t memhash32_seed(const char *buf, size_t n, uint32_t seed)
 {
     uint32_t out;
 
     MurmurHash3_x86_32(buf, n, seed, &out);
     return out;
 }
+
+#ifdef __cplusplus
+}
+#endif
