@@ -2,6 +2,10 @@
 #include "dtypes.h"
 #include "utils.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 char *uint2str(char *dest, size_t len, uint64_t num, uint32_t base)
 {
     int i = len-1;
@@ -57,3 +61,7 @@ int str2int(char *str, size_t len, int64_t *res, uint32_t base)
     return 0;
 }
 */
+
+#ifdef __cplusplus
+}
+#endif
