@@ -4,6 +4,10 @@
 #include <stdarg.h>
 #include "uv.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 // this flag controls when data actually moves out to the underlying I/O
 // channel. memory streams are a special case of this where the data
 // never moves out.
@@ -20,10 +24,13 @@ typedef struct {
     // to the buffer. reading: at the end. writing: at the beginning.
     // in general, you can do any operation in any state.
     char *buf;        // start of buffer
-    bufmode_t bm;
 
     int errcode;
 
+#ifdef _P64
+    int _pad_bm;      // put bm at same offset as type field of uv_stream_s
+#endif
+    bufmode_t bm;     //
     bufstate_t state;
 
     off_t maxsize;    // space allocated to buffer
@@ -38,7 +45,8 @@ typedef struct {
     // to be a pointer
     long fd;
 
-    unsigned char readonly:1;
+    unsigned char readable:1;
+    unsigned char writable:1;
     unsigned char ownbuf:1;
     unsigned char ownfd:1;
     unsigned char _eof:1;
@@ -77,7 +85,8 @@ DLLEXPORT char *ios_takebuf(ios_t *s, size_t *psize);  // release buffer to call
 // set buffer space to use
 DLLEXPORT int ios_setbuf(ios_t *s, char *buf, size_t size, int own);
 DLLEXPORT int ios_bufmode(ios_t *s, bufmode_t mode);
-DLLEXPORT int ios_get_readonly(ios_t *s);
+DLLEXPORT int ios_get_readable(ios_t *s);
+DLLEXPORT int ios_get_writable(ios_t *s);
 DLLEXPORT void ios_set_readonly(ios_t *s);
 DLLEXPORT size_t ios_copy(ios_t *to, ios_t *from, size_t nbytes);
 DLLEXPORT size_t ios_copyall(ios_t *to, ios_t *from);
@@ -207,4 +216,8 @@ int ios_ungetc(int c, ios_t *s);
   do no reading if the data will only be overwritten.
 */
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
