@@ -34,6 +34,10 @@
 
 #include "utf8.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 static const uint32_t offsetsFromUTF8[6] = {
     0x00000000UL, 0x00003080UL, 0x000E2080UL,
     0x03C82080UL, 0xFA082080UL, 0x82082080UL
@@ -257,8 +261,10 @@ size_t u8_strlen(const char *s)
     return count;
 }
 
-#if defined(_OS_WINDOWS_) || defined(_OS_LINUX_)
+#if defined(_OS_WINDOWS_)
 extern int wcwidth(uint32_t ch);
+#elif defined(_OS_LINUX_)
+extern int wcwidth(wchar_t ch);
 #endif
 
 size_t u8_strwidth(const char *s)
@@ -298,13 +304,13 @@ size_t u8_strwidth(const char *s)
 uint32_t u8_nextchar(const char *s, size_t *i)
 {
     uint32_t ch = 0;
-    size_t sz = 0;
+    size_t sz, j;
 
-    do {
+    sz = u8_seqlen(&s[*i]);
+    for (j = sz; j > 0; j--) {
         ch <<= 6;
-        ch += (unsigned char)s[(*i)];
-        sz++;
-    } while (s[*i] && (++(*i)) && !isutf(s[*i]));
+        ch += (unsigned char)s[(*i)++];
+    }
     ch -= offsetsFromUTF8[sz-1];
 
     return ch;
@@ -313,17 +319,7 @@ uint32_t u8_nextchar(const char *s, size_t *i)
 /* next character without NUL character terminator */
 uint32_t u8_nextmemchar(const char *s, size_t *i)
 {
-    uint32_t ch = 0;
-    size_t sz = 0;
-
-    do {
-        ch <<= 6;
-        ch += (unsigned char)s[(*i)++];
-        sz++;
-    } while (!isutf(s[*i]));
-    ch -= offsetsFromUTF8[sz-1];
-
-    return ch;
+    return u8_nextchar(s,i);
 }
 
 void u8_inc(const char *s, size_t *i)
@@ -588,12 +584,12 @@ int u8_is_locale_utf8(const char *locale)
     if (locale == NULL) return 0;
 
     /* this code based on libutf8 */
-    const char* cp = locale;
+    const char *cp = locale;
 
-    for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++) {
+    for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ',' && *cp != ';'; cp++) {
         if (*cp == '.') {
-            const char* encoding = ++cp;
-            for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++)
+            const char *encoding = ++cp;
+            for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ',' && *cp != ';'; cp++)
                 ;
             if ((cp-encoding == 5 && !strncmp(encoding, "UTF-8", 5))
                 || (cp-encoding == 4 && !strncmp(encoding, "utf8", 4)))
@@ -747,3 +743,7 @@ int u8_reverse(char *dest, char *src, size_t len)
     }
     return 0;
 }
+
+#ifdef __cplusplus
+}
+#endif
