@@ -1,7 +1,9 @@
 #ifndef UTF8_H
 #define UTF8_H
 
-extern DLLEXPORT int locale_is_utf8;
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /* is c the start of a utf8 sequence? */
 #define isutf(c) (((c)&0xC0)!=0x80)
@@ -18,10 +20,10 @@ size_t u8_toutf8(char *dest, size_t sz, const uint32_t *src, size_t srcsz);
 size_t u8_wc_toutf8(char *dest, uint32_t ch);
 
 /* character number to byte offset */
-size_t u8_offset(const char *str, size_t charnum);
+DLLEXPORT size_t u8_offset(const char *str, size_t charnum);
 
 /* byte offset to character number */
-size_t u8_charnum(const char *s, size_t offset);
+DLLEXPORT size_t u8_charnum(const char *s, size_t offset);
 
 /* return next character, updating an index variable */
 uint32_t u8_nextchar(const char *s, size_t *i);
@@ -112,4 +114,8 @@ DLLEXPORT int u8_isvalid(const char *str, size_t length);
    be allocated to at least len+1 bytes. returns 1 for error, 0 otherwise */
 DLLEXPORT int u8_reverse(char *dest, char *src, size_t len);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
