@@ -1,6 +1,10 @@
 #ifndef UTILS_H
 #define UTILS_H
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 DLLEXPORT char *uint2str(char *dest, size_t len, uint64_t num, uint32_t base);
 int str2int(char *str, size_t len, int64_t *res, uint32_t base);
 int isdigit_base(char c, int base);
@@ -29,7 +33,7 @@ int cmp_eq(void *a, numerictype_t atag, void *b, numerictype_t btag,
 #  define LEGACY_REGS "=q"
 #endif
 
-#if !defined(__INTEL_COMPILER) && (defined(__i386__) || defined(__x86_64__))
+#if (!defined(__INTEL_COMPILER) || defined(__clang__)) && (defined(__i386__) || defined(__x86_64__))
 STATIC_INLINE u_int16_t ByteSwap16(u_int16_t x)
 {
   __asm("xchgb %b0,%h0" :
@@ -71,7 +75,7 @@ STATIC_INLINE u_int64_t ByteSwap64(u_int64_t x)
 
 #define bswap_16(x) (((x) & 0x00ff) << 8 | ((x) & 0xff00) >> 8)
 
-#ifdef __INTEL_COMPILER
+#if defined(__INTEL_COMPILER) && !defined(__clang__)
 #define bswap_32(x) _bswap(x)
 #else
 #define bswap_32(x) \
@@ -94,4 +98,8 @@ STATIC_INLINE u_int64_t ByteSwap64(u_int64_t x)
 
 #endif
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
