@@ -11,19 +11,23 @@
 #ifndef _OS_WINDOWS_
 #include <sys/sysctl.h>
 #include <sys/wait.h>
+#include <sys/ptrace.h>
 #include <unistd.h>
 #include <sys/mman.h>
+#include <dlfcn.h>
 #endif
 #include <errno.h>
 #include <signal.h>
-#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
-char * basename(char *);
-char * dirname(char *);
-#else
-#include <libgen.h>
-#endif
 #include <fcntl.h>
 
+#ifdef __APPLE__
+#include <mach-o/dyld.h>
+#include <mach-o/nlist.h>
+#include <sys/types.h> // for jl_raise_debugger
+#elif !defined(_OS_WINDOWS_)
+#include <link.h>
+#endif
+
 #define __STDC_CONSTANT_MACROS
 #define __STDC_LIMIT_MACROS
 #include <llvm-c/Target.h>
@@ -32,6 +36,22 @@ char * dirname(char *);
 #include <xmmintrin.h>
 #endif
 
+#if defined _MSC_VER
+#include <io.h>
+#include <intrin.h>
+#endif
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
+DLLEXPORT char *basename(char *);
+DLLEXPORT char *dirname(char *);
+#else
+#include <libgen.h>
+#endif
+
 DLLEXPORT uint32_t jl_getutf8(ios_t *s)
 {
     uint32_t wc=0;
@@ -44,11 +64,13 @@ DLLEXPORT size_t jl_ios_size(ios_t *s) { return s->size; }
 DLLEXPORT int jl_sizeof_off_t(void) { return sizeof(off_t); }
 #ifndef _OS_WINDOWS_
 DLLEXPORT off_t jl_lseek(int fd, off_t offset, int whence) { return lseek(fd, offset, whence); }
-DLLEXPORT ssize_t jl_pwrite(int fd, const void *buf, size_t count, off_t offset) {
+DLLEXPORT ssize_t jl_pwrite(int fd, const void *buf, size_t count, off_t offset)
+{
     return pwrite(fd, buf, count, offset);
 }
 DLLEXPORT void *jl_mmap(void *addr, size_t length, int prot, int flags,
-                        int fd, off_t offset) {
+                        int fd, off_t offset)
+{
     return mmap(addr, length, prot, flags, fd, offset);
 }
 #else
@@ -75,25 +97,25 @@ DLLEXPORT int jl_ios_eof(ios_t *s)
 // --- dir/file stuff ---
 
 DLLEXPORT int jl_sizeof_uv_fs_t(void) { return sizeof(uv_fs_t); }
-DLLEXPORT void jl_uv_fs_req_cleanup(uv_fs_t* req)
+DLLEXPORT void jl_uv_fs_req_cleanup(uv_fs_t *req)
 {
     uv_fs_req_cleanup(req);
 }
 
-DLLEXPORT int jl_readdir(const char* path, uv_fs_t* readdir_req)
+DLLEXPORT int jl_readdir(const char *path, uv_fs_t *readdir_req)
 {
     // Note that the flags field is mostly ignored by libuv
     return uv_fs_readdir(uv_default_loop(), readdir_req, path, 0 /*flags*/, NULL);
 }
 
-DLLEXPORT char* jl_uv_fs_t_ptr(uv_fs_t* req) {return req->ptr; }
-DLLEXPORT char* jl_uv_fs_t_ptr_offset(uv_fs_t* req, int offset) {return (char *)req->ptr + offset; }
+DLLEXPORT char *jl_uv_fs_t_ptr(uv_fs_t *req) { return (char*)req->ptr; }
+DLLEXPORT char *jl_uv_fs_t_ptr_offset(uv_fs_t *req, int offset) { return (char*)req->ptr + offset; }
 DLLEXPORT int jl_uv_fs_result(uv_fs_t *f) { return f->result; }
 
 // --- stat ---
 DLLEXPORT int jl_sizeof_stat(void) { return sizeof(uv_stat_t); }
 
-DLLEXPORT int32_t jl_stat(const char* path, char* statbuf)
+DLLEXPORT int32_t jl_stat(const char *path, char *statbuf)
 {
     uv_fs_t req;
     int ret;
@@ -107,7 +129,7 @@ DLLEXPORT int32_t jl_stat(const char* path, char* statbuf)
     return ret;
 }
 
-DLLEXPORT int32_t jl_lstat(const char* path, char* statbuf)
+DLLEXPORT int32_t jl_lstat(const char *path, char *statbuf)
 {
     uv_fs_t req;
     int ret;
@@ -133,52 +155,52 @@ DLLEXPORT int32_t jl_fstat(int fd, char *statbuf)
 
 DLLEXPORT unsigned int jl_stat_dev(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_dev;
+    return ((uv_stat_t*)statbuf)->st_dev;
 }
 
 DLLEXPORT unsigned int jl_stat_ino(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_ino;
+    return ((uv_stat_t*)statbuf)->st_ino;
 }
 
 DLLEXPORT unsigned int jl_stat_mode(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_mode;
+    return ((uv_stat_t*)statbuf)->st_mode;
 }
 
 DLLEXPORT unsigned int jl_stat_nlink(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_nlink;
+    return ((uv_stat_t*)statbuf)->st_nlink;
 }
 
 DLLEXPORT unsigned int jl_stat_uid(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_uid;
+    return ((uv_stat_t*)statbuf)->st_uid;
 }
 
 DLLEXPORT unsigned int jl_stat_gid(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_gid;
+    return ((uv_stat_t*)statbuf)->st_gid;
 }
 
 DLLEXPORT unsigned int jl_stat_rdev(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_rdev;
+    return ((uv_stat_t*)statbuf)->st_rdev;
 }
 
 DLLEXPORT uint64_t jl_stat_size(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_size;
+    return ((uv_stat_t*)statbuf)->st_size;
 }
 
 DLLEXPORT uint64_t jl_stat_blksize(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_blksize;
+    return ((uv_stat_t*)statbuf)->st_blksize;
 }
 
 DLLEXPORT uint64_t jl_stat_blocks(char *statbuf)
 {
-    return ((uv_stat_t*) statbuf)->st_blocks;
+    return ((uv_stat_t*)statbuf)->st_blocks;
 }
 
 /*
@@ -186,7 +208,7 @@ DLLEXPORT uint64_t jl_stat_blocks(char *statbuf)
 DLLEXPORT double jl_stat_atime(char *statbuf)
 {
   uv_stat_t *s;
-  s = (uv_stat_t*) statbuf;
+  s = (uv_stat_t*)statbuf;
   return (double)s->st_atim.tv_sec + (double)s->st_atim.tv_nsec * 1e-9;
 }
 */
@@ -194,14 +216,14 @@ DLLEXPORT double jl_stat_atime(char *statbuf)
 DLLEXPORT double jl_stat_mtime(char *statbuf)
 {
     uv_stat_t *s;
-    s = (uv_stat_t*) statbuf;
+    s = (uv_stat_t*)statbuf;
     return (double)s->st_mtim.tv_sec + (double)s->st_mtim.tv_nsec * 1e-9;
 }
 
 DLLEXPORT double jl_stat_ctime(char *statbuf)
 {
     uv_stat_t *s;
-    s = (uv_stat_t*) stat;
+    s = (uv_stat_t*)statbuf;
     return (double)s->st_ctim.tv_sec + (double)s->st_ctim.tv_nsec * 1e-9;
 }
 
@@ -257,7 +279,7 @@ jl_value_t *jl_readuntil(ios_t *s, uint8_t delim)
         a = jl_alloc_array_1d(jl_array_uint8_type, 80);
         ios_t dest;
         ios_mem(&dest, 0);
-        ios_setbuf(&dest, a->data, 80, 0);
+        ios_setbuf(&dest, (char*)a->data, 80, 0);
         size_t n = ios_copyuntil(&dest, s, delim);
         if (dest.buf != a->data) {
             a = jl_takebuf_array(&dest);
@@ -281,6 +303,7 @@ void jl_free2(void *p, void *hint)
 // -- syscall utilities --
 
 int jl_errno(void) { return errno; }
+void jl_set_errno(int e) { errno = e; }
 
 // -- get the number of CPU cores --
 
@@ -351,12 +374,6 @@ jl_value_t *jl_environ(int i)
     char *env = environ[i];
     return env ? jl_pchar_to_string(env, strlen(env)) : jl_nothing;
 }
-#ifdef _OS_WINDOWS_
-jl_value_t *jl_env_done(char *pos)
-{
-    return (*pos==0)?jl_true:jl_false;
-}
-#endif
 
 // -- child process status --
 
@@ -392,16 +409,13 @@ JL_STREAM *jl_stdin_stream(void)  { return (JL_STREAM*)JL_STDIN; }
 JL_STREAM *jl_stdout_stream(void) { return (JL_STREAM*)JL_STDOUT; }
 JL_STREAM *jl_stderr_stream(void) { return (JL_STREAM*)JL_STDERR; }
 
-// -- set/clear the FZ/DAZ flags on x86 & x86-64 --
+// CPUID
 
-#ifdef __SSE__
-
-#ifdef _OS_WINDOWS_
-#define cpuid    __cpuid
-#else
-
-void cpuid(int32_t CPUInfo[4], int32_t InfoType)
+DLLEXPORT void jl_cpuid(int32_t CPUInfo[4], int32_t InfoType)
 {
+#if defined _MSC_VER
+    __cpuid(CPUInfo, InfoType);
+#else
     __asm__ __volatile__ (
         #if defined(__i386__) && defined(__PIC__)
         "xchg %%ebx, %%esi;"
@@ -417,18 +431,20 @@ void cpuid(int32_t CPUInfo[4], int32_t InfoType)
         "=d" (CPUInfo[3]) :
         "a" (InfoType)
     );
+#endif
 }
 
-#endif
+// -- set/clear the FZ/DAZ flags on x86 & x86-64 --
+#ifdef __SSE__
 
 DLLEXPORT uint8_t jl_zero_subnormals(uint8_t isZero)
 {
     uint32_t flags = 0x00000000;
     int32_t info[4];
 
-    cpuid(info, 0);
+    jl_cpuid(info, 0);
     if (info[0] >= 1) {
-        cpuid(info, 0x00000001);
+        jl_cpuid(info, 0x00000001);
         if ((info[3] & ((int)1 << 26)) != 0) {
             // SSE2 supports both FZ and DAZ
             flags = 0x00008040;
@@ -462,7 +478,8 @@ DLLEXPORT uint8_t jl_zero_subnormals(uint8_t isZero)
 
 // -- processor native alignment information --
 
-DLLEXPORT void jl_native_alignment(uint_t* int8align, uint_t* int16align, uint_t* int32align, uint_t* int64align, uint_t* float32align, uint_t* float64align)
+DLLEXPORT void jl_native_alignment(uint_t *int8align, uint_t *int16align, uint_t *int32align,
+                                   uint_t *int64align, uint_t *float32align, uint_t *float64align)
 {
     LLVMTargetDataRef tgtdata = LLVMCreateTargetData("");
     *int8align = LLVMPreferredAlignmentOfType(tgtdata, LLVMInt8Type());
@@ -507,6 +524,24 @@ long jl_getpagesize(void)
 }
 #endif
 
+#ifdef _OS_WINDOWS_
+static long cachedAllocationGranularity = 0;
+long jl_getallocationgranularity(void)
+{
+    if (!cachedAllocationGranularity) {
+        SYSTEM_INFO systemInfo;
+        GetSystemInfo (&systemInfo);
+        cachedAllocationGranularity = systemInfo.dwAllocationGranularity;
+    }
+    return cachedAllocationGranularity;
+}
+#else
+long jl_getallocationgranularity(void)
+{
+    return jl_getpagesize();
+}
+#endif
+
 DLLEXPORT long jl_SC_CLK_TCK(void)
 {
 #ifndef _OS_WINDOWS_
@@ -515,3 +550,152 @@ DLLEXPORT long jl_SC_CLK_TCK(void)
     return 0;
 #endif
 }
+
+// Dynamic Library interrogation
+
+#ifdef __APPLE__
+// This code gratefully lifted from: http://stackoverflow.com/questions/20481058
+#ifdef __LP64__
+typedef struct mach_header_64 mach_header_t;
+typedef struct segment_command_64 segment_command_t;
+typedef struct nlist_64 nlist_t;
+#else
+typedef struct mach_header mach_header_t;
+typedef struct segment_command segment_command_t;
+typedef struct nlist nlist_t;
+#endif
+
+static const char *first_external_symbol_for_image(const mach_header_t *header)
+{
+    Dl_info info;
+    if (dladdr(header, &info) == 0)
+        return NULL;
+
+    segment_command_t *seg_linkedit = NULL;
+    segment_command_t *seg_text = NULL;
+    struct symtab_command *symtab = NULL;
+
+    struct load_command *cmd = (struct load_command *)((intptr_t)header + sizeof(mach_header_t));
+    for (uint32_t i = 0; i < header->ncmds; i++, cmd = (struct load_command *)((intptr_t)cmd + cmd->cmdsize)) {
+        switch(cmd->cmd) {
+        case LC_SEGMENT:
+        case LC_SEGMENT_64:
+            if (!strcmp(((segment_command_t *)cmd)->segname, SEG_TEXT))
+                seg_text = (segment_command_t *)cmd;
+            else if (!strcmp(((segment_command_t *)cmd)->segname, SEG_LINKEDIT))
+                seg_linkedit = (segment_command_t *)cmd;
+            break;
+
+        case LC_SYMTAB:
+            symtab = (struct symtab_command *)cmd;
+            break;
+        }
+    }
+
+    if ((seg_text == NULL) || (seg_linkedit == NULL) || (symtab == NULL))
+        return NULL;
+
+    intptr_t file_slide = ((intptr_t)seg_linkedit->vmaddr - (intptr_t)seg_text->vmaddr) - seg_linkedit->fileoff;
+    intptr_t strings = (intptr_t)header + (symtab->stroff + file_slide);
+    nlist_t *sym = (nlist_t *)((intptr_t)header + (symtab->symoff + file_slide));
+
+    for (uint32_t i = 0; i < symtab->nsyms; i++, sym++) {
+        if ((sym->n_type & N_EXT) != N_EXT || !sym->n_value)
+            continue;
+
+        return (const char*)strings + sym->n_un.n_strx;
+    }
+
+    return NULL;
+}
+#endif
+
+// Takes a handle (as returned from dlopen()) and returns the absolute path to the image loaded
+DLLEXPORT const char *jl_pathname_for_handle(uv_lib_t *uv_lib)
+{
+    if (!uv_lib)
+        return NULL;
+    
+    void *handle = uv_lib->handle;
+#ifdef __APPLE__
+    for (int32_t i = _dyld_image_count(); i >= 0 ; i--) {
+        const char *first_symbol = first_external_symbol_for_image((const mach_header_t *)_dyld_get_image_header(i));
+        if (first_symbol && strlen(first_symbol) > 1) {
+            handle = (void*)((intptr_t)handle | 1); // in order to trigger findExportedSymbol instead of findExportedSymbolInImageOrDependentImages. See `dlsym` implementation at http://opensource.apple.com/source/dyld/dyld-239.3/src/dyldAPIs.cpp
+            first_symbol++; // in order to remove the leading underscore
+            void *address = dlsym(handle, first_symbol);
+            Dl_info info;
+            if (dladdr(address, &info))
+                return info.dli_fname;
+        }
+    }
+
+#elif defined(_OS_WINDOWS_)
+
+    wchar_t *pth16 = (wchar_t*)malloc(32768); // max long path length
+    DWORD n16 = GetModuleFileNameW(handle,pth16,32768);
+    if (n16 <= 0) {
+        free(pth16);
+        return NULL;
+    }
+    pth16[n16] = L'\0';
+    DWORD n8 = WideCharToMultiByte(CP_UTF8, 0, pth16, -1, NULL, 0, NULL, NULL);
+    if (n8 == 0) {
+        free(pth16);
+        return NULL;
+    }
+    char *filepath = (char*)malloc(++n8);
+    if (!WideCharToMultiByte(CP_UTF8, 0, pth16, -1, filepath, n8, NULL, NULL)) {
+        free(pth16);
+        free(filepath);
+        return NULL;
+    }
+    free(pth16);
+    return filepath;
+
+#else // Linux, FreeBSD, ...
+
+    struct link_map *map;
+    dlinfo(handle, RTLD_DI_LINKMAP, &map);
+    if (map)
+        return map->l_name;
+
+#endif
+    return NULL;
+}
+
+#ifdef _OS_WINDOWS_
+#include <dbghelp.h>
+static BOOL CALLBACK jl_EnumerateLoadedModulesProc64(
+  _In_      PCTSTR ModuleName,
+  _In_      DWORD64 ModuleBase,
+  _In_      ULONG ModuleSize,
+  _In_opt_  PVOID a
+)
+{
+    jl_array_grow_end(a, 1);
+    //XXX: change to jl_arrayset if array storage allocation for Array{String,1} changes:
+    jl_value_t *v = jl_cstr_to_string(ModuleName);
+    jl_cellset(a, jl_array_dim0(a)-1, v);
+    return TRUE;
+}
+// Takes a handle (as returned from dlopen()) and returns the absolute path to the image loaded
+DLLEXPORT int jl_dllist(jl_array_t *list)
+{
+    return EnumerateLoadedModules64(GetCurrentProcess(), jl_EnumerateLoadedModulesProc64, list);
+}
+#endif
+
+DLLEXPORT void jl_raise_debugger(void)
+{
+#if defined(_OS_WINDOWS_)
+    if (IsDebuggerPresent() == 1)
+        DebugBreak();
+#else
+    raise(SIGINT);
+#endif // _OS_WINDOWS_
+}
+
+#ifdef __cplusplus
+}
+#endif
