@@ -11,12 +11,17 @@
 #include <signal.h>
 #include <errno.h>
 #include "julia.h"
-#include "builtin_proto.h"
+#include "julia_internal.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #if defined(_OS_WINDOWS_)
 #include <winbase.h>
 #include <malloc.h>
 #include <dbghelp.h>
-static volatile int in_stackwalk = 0;
+volatile int jl_in_stackwalk = 0;
 #else
 #include <unistd.h>
 // This gives unwind only local unwinding options ==> faster code
@@ -135,12 +140,15 @@ static void _probe_arch(void)
   - stack growth
 */
 
+static jl_sym_t *done_sym;
+static jl_sym_t *failed_sym;
+static jl_sym_t *runnable_sym;
+
 extern size_t jl_page_size;
 jl_datatype_t *jl_task_type;
 DLLEXPORT jl_task_t * volatile jl_current_task;
 jl_task_t *jl_root_task;
 jl_value_t * volatile jl_task_arg_in_transit;
-static volatile int n_args_in_transit;
 jl_value_t *jl_exception_in_transit;
 #ifdef JL_GC_MARKSWEEP
 jl_gcframe_t *jl_pgcstack = NULL;
@@ -153,18 +161,18 @@ jl_jmp_buf * volatile jl_jmp_target;
 
 static void save_stack(jl_task_t *t)
 {
-    if (t->done)
+    if (t->state == done_sym || t->state == failed_sym)
         return;
     volatile int _x;
     size_t nb = (char*)t->stackbase - (char*)&_x;
     char *buf;
     if (t->stkbuf == NULL || t->bufsz < nb) {
-        buf = allocb(nb);
+        buf = (char*)allocb(nb);
         t->stkbuf = buf;
         t->bufsz = nb;
     }
     else {
-        buf = t->stkbuf;
+        buf = (char*)t->stkbuf;
     }
     t->ssize = nb;
     memcpy(buf, (char*)&_x, nb);
@@ -176,11 +184,11 @@ void __declspec(noinline) restore_stack(jl_task_t *t, jl_jmp_buf *where, char *p
 void __attribute__((noinline)) restore_stack(jl_task_t *t, jl_jmp_buf *where, char *p)
 #endif
 {
-    char* _x = (char*)t->stackbase - t->ssize;
+    char *_x = (char*)t->stackbase - t->ssize;
     if (!p) {
         p = _x;
         if ((char*)&_x > _x) {
-            p = alloca((char*)&_x - _x);
+            p = (char*)alloca((char*)&_x - _x);
     	}
         restore_stack(t, where, p);
     }
@@ -238,10 +246,18 @@ static void ctx_switch(jl_task_t *t, jl_jmp_buf *where)
         jl_current_task->gcstack = jl_pgcstack;
         jl_pgcstack = t->gcstack;
 #endif
+
+        // restore task's current module, looking at parent tasks
+        // if it hasn't set one.
+        jl_task_t *last = t;
+        while (last->current_module == NULL && last != jl_root_task) {
+            last = last->parent;
+        }
+        if (last->current_module != NULL) {
+            jl_current_module = last->current_module;
+        }
+
         t->last = jl_current_task;
-        // by default, parent is first task to switch to this one
-        if (t->parent == NULL)
-            t->parent = jl_current_task;
         jl_current_task = t;
 
 #ifdef COPY_STACKS
@@ -257,8 +273,10 @@ static void ctx_switch(jl_task_t *t, jl_jmp_buf *where)
 extern int jl_in_gc;
 static jl_value_t *switchto(jl_task_t *t)
 {
-    if (t->done) {
+    if (t->state == done_sym || t->state == failed_sym) {
         jl_task_arg_in_transit = (jl_value_t*)jl_null;
+        if (t->exception != jl_nothing)
+            jl_throw(t->exception);
         return t->result;
     }
     if (jl_in_gc) {
@@ -360,7 +378,6 @@ static void rebase_state(jl_jmp_buf *ctx, intptr_t local_sp, intptr_t new_sp)
 jl_value_t *jl_switchto(jl_task_t *t, jl_value_t *arg)
 {
     jl_task_arg_in_transit = arg;
-    n_args_in_transit = 1;
     return switchto(t);
 }
 
@@ -368,23 +385,23 @@ static jl_function_t *task_done_hook_func=NULL;
 
 static void finish_task(jl_task_t *t, jl_value_t *resultval)
 {
-    assert(t->done==0);
-    t->done = 1;
-    t->runnable = 0;
+    if (t->exception != jl_nothing)
+        t->state = failed_sym;
+    else
+        t->state = done_sym;
     t->result = resultval;
     // TODO: early free of t->stkbuf
 #ifdef COPY_STACKS
     t->stkbuf = NULL;
 #endif
-    if (t->donenotify && t->donenotify != jl_nothing) {
-        if (task_done_hook_func == NULL) {
-            task_done_hook_func = (jl_function_t*)jl_get_global(jl_base_module,
-                                                                jl_symbol("task_done_hook"));
-        }
-        if (task_done_hook_func != NULL) {
-            jl_apply(task_done_hook_func, (jl_value_t**)&t, 1);
-        }
+    if (task_done_hook_func == NULL) {
+        task_done_hook_func = (jl_function_t*)jl_get_global(jl_base_module,
+                                                            jl_symbol("task_done_hook"));
     }
+    if (task_done_hook_func != NULL) {
+        jl_apply(task_done_hook_func, (jl_value_t**)&t, 1);
+    }
+    assert(0);
 }
 
 static void start_task(jl_task_t *t)
@@ -406,27 +423,17 @@ static void start_task(jl_task_t *t)
         switch_stack(jl_current_task, jl_jmp_target);
     }
 #endif
-    if (n_args_in_transit == 0) {
-        res = jl_apply(t->start, NULL, 0);
-    }
-    else if (n_args_in_transit == 1) {
-        res = jl_apply(t->start, &arg, 1);
-    }
-    else {
-        assert(jl_is_tuple(jl_task_arg_in_transit));
-        res = jl_apply(t->start, &jl_tupleref(jl_task_arg_in_transit,0),
-                       n_args_in_transit);
-    }
+    res = jl_apply(t->start, NULL, 0);
     JL_GC_POP();
     finish_task(t, res);
-    jl_task_t *cont = t->parent;
-    // if parent task has exited, try its parent, and so on
-    while (cont->done)
-        cont = cont->parent;
-    jl_switchto(cont, t->result);
     assert(0);
 }
 
+DLLEXPORT void jl_handle_stack_switch()
+{
+    jl_switch_stack(jl_current_task, jl_jmp_target);
+}
+
 #ifndef COPY_STACKS
 static void init_task(jl_task_t *t)
 {
@@ -449,35 +456,51 @@ static void init_task(jl_task_t *t)
 ptrint_t bt_data[MAX_BT_SIZE+1];
 size_t bt_size = 0;
 
-void getFunctionInfo(const char **name, int *line, const char **filename, size_t pointer);
+void jl_getFunctionInfo(
+    const char **name,
+    size_t *line,
+    const char **filename,
+    size_t pointer,
+    int *fromC,
+    int skipC);
 
-static const char* name_unknown = "???";
-static int frame_info_from_ip(const char **func_name, int *line_num, const char **file_name, size_t ip, int doCframes)
+static int frame_info_from_ip(
+    const char **func_name,
+    size_t *line_num,
+    const char **file_name,
+    size_t ip,
+    int skipC)
 {
+    static const char *name_unknown = "???";
+#if defined(_OS_WINDOWS_)
+    static char frame_info_func[
+        sizeof(SYMBOL_INFO) +
+        MAX_SYM_NAME * sizeof(TCHAR)];
+    static IMAGEHLP_LINE64 frame_info_line;
+#endif
     int fromC = 0;
 
-    getFunctionInfo(func_name, line_num, file_name, ip);
-    if (*func_name == NULL && doCframes) {
-        fromC = 1;
+    jl_getFunctionInfo(func_name, line_num, file_name, ip, &fromC, skipC);
+    if (*func_name == NULL) {
 #if defined(_OS_WINDOWS_)
-        if (in_stackwalk) {
+        fromC = 1;
+        if (jl_in_stackwalk) {
             *func_name = name_unknown;
             *file_name = name_unknown;
             *line_num = ip;
         }
         else {
-            in_stackwalk = 1;
+            jl_in_stackwalk = 1;
             DWORD64 dwDisplacement64 = 0;
             DWORD64 dwAddress = ip;
 
-            char buffer[sizeof(SYMBOL_INFO) + MAX_SYM_NAME * sizeof(TCHAR)];
-            PSYMBOL_INFO pSymbol = (PSYMBOL_INFO)buffer;
+            PSYMBOL_INFO pSymbol = (PSYMBOL_INFO)frame_info_func;
             pSymbol->SizeOfStruct = sizeof(SYMBOL_INFO);
             pSymbol->MaxNameLen = MAX_SYM_NAME;
 
             if (SymFromAddr(GetCurrentProcess(), dwAddress, &dwDisplacement64, pSymbol)) {
                 // SymFromAddr returned success
-                *func_name = strdup(pSymbol->Name);
+                *func_name = pSymbol->Name;
             }
             else {
                 *func_name = name_unknown;
@@ -486,14 +509,13 @@ static int frame_info_from_ip(const char **func_name, int *line_num, const char
                 //printf("SymFromAddr returned error : %d\n", error);
             }
 
-            IMAGEHLP_LINE64 line;
             DWORD dwDisplacement = 0;
-            line.SizeOfStruct = sizeof(IMAGEHLP_LINE64);
+            frame_info_line.SizeOfStruct = sizeof(IMAGEHLP_LINE64);
 
-            if (SymGetLineFromAddr64(GetCurrentProcess(), dwAddress, &dwDisplacement, &line)) {
+            if (SymGetLineFromAddr64(GetCurrentProcess(), dwAddress, &dwDisplacement, &frame_info_line)) {
                 // SymGetLineFromAddr64 returned success
-                *file_name = strdup(line.FileName);
-                *line_num = line.LineNumber;
+                *file_name = frame_info_line.FileName;
+                *line_num = frame_info_line.LineNumber;
             }
             else {
                 *file_name = name_unknown;
@@ -502,27 +524,12 @@ static int frame_info_from_ip(const char **func_name, int *line_num, const char
                 //DWORD error = GetLastError();
                 //printf("SymGetLineFromAddr64 returned error : %d\n", error);
             }
-            in_stackwalk = 0;
+            jl_in_stackwalk = 0;
         }
 #else
-        Dl_info dlinfo;
-        if (dladdr((void*) ip, &dlinfo) != 0) {
-            *file_name = (dlinfo.dli_fname != NULL) ? dlinfo.dli_fname : name_unknown;
-            if (dlinfo.dli_sname != NULL) {
-                *func_name = dlinfo.dli_sname;
-                // line number in C looks tricky. addr2line and libbfd seem promising. For now, punt and just return address offset.
-                *line_num = ip-(size_t)dlinfo.dli_saddr;
-            }
-            else {
-                *func_name = name_unknown;
-                *line_num = 0;
-            }
-        }
-        else {
-            *func_name = name_unknown;
-            *file_name = name_unknown;
-            *line_num = ip;
-        }
+        *func_name = name_unknown;
+        *file_name = name_unknown;
+        *line_num = ip;
 #endif
     }
     return fromC;
@@ -530,26 +537,28 @@ static int frame_info_from_ip(const char **func_name, int *line_num, const char
 
 #if defined(_OS_WINDOWS_)
 int needsSymRefreshModuleList;
-WINBOOL WINAPI (*hSymRefreshModuleList)(HANDLE);
-DLLEXPORT size_t rec_backtrace(ptrint_t *data, size_t maxsize) {
+BOOL (WINAPI *hSymRefreshModuleList)(HANDLE);
+DLLEXPORT size_t rec_backtrace(ptrint_t *data, size_t maxsize)
+{
     CONTEXT Context;
     memset(&Context, 0, sizeof(Context));
-    in_stackwalk = 1;
+    jl_in_stackwalk = 1;
     RtlCaptureContext(&Context);
-    in_stackwalk = 0;
+    jl_in_stackwalk = 0;
     return rec_backtrace_ctx(data, maxsize, &Context);
 }
-DLLEXPORT size_t rec_backtrace_ctx(ptrint_t *data, size_t maxsize, CONTEXT *Context) {
-    if (in_stackwalk) {
+DLLEXPORT size_t rec_backtrace_ctx(ptrint_t *data, size_t maxsize, CONTEXT *Context)
+{
+    if (jl_in_stackwalk) {
         return 0;
     }
     STACKFRAME64 stk;
     memset(&stk, 0, sizeof(stk));
 
     if (needsSymRefreshModuleList && hSymRefreshModuleList != 0) {
-        in_stackwalk = 1;
+        jl_in_stackwalk = 1;
         hSymRefreshModuleList(GetCurrentProcess());
-        in_stackwalk = 0;
+        jl_in_stackwalk = 0;
         needsSymRefreshModuleList = 0;
     }
 #if defined(_CPU_X86_64_) 
@@ -572,10 +581,10 @@ DLLEXPORT size_t rec_backtrace_ctx(ptrint_t *data, size_t maxsize, CONTEXT *Cont
     size_t n = 0;
     intptr_t lastsp = stk.AddrStack.Offset;
     while (n < maxsize) {
-        in_stackwalk = 1;
+        jl_in_stackwalk = 1;
         BOOL result = StackWalk64(MachineType, GetCurrentProcess(), hMainThread,
             &stk, Context, NULL, SymFunctionTableAccess64, SymGetModuleBase64, NULL);
-        in_stackwalk = 0;
+        jl_in_stackwalk = 0;
         data[n++] = (intptr_t)stk.AddrPC.Offset;
         intptr_t sp = (intptr_t)stk.AddrStack.Offset;
         if (!result || sp == 0 || 
@@ -645,31 +654,26 @@ DLLEXPORT jl_value_t *jl_backtrace_from_here(void)
                                             jl_tuple2(jl_voidpointer_type,
                                                       jl_box_long(1)));
     jl_array_t *bt = jl_alloc_array_1d(array_ptr_void_type, MAX_BT_SIZE);
-    size_t n = rec_backtrace(jl_array_data(bt), MAX_BT_SIZE);
+    size_t n = rec_backtrace((ptrint_t*)jl_array_data(bt), MAX_BT_SIZE);
     if (n < MAX_BT_SIZE)
         jl_array_del_end(bt, MAX_BT_SIZE-n);
     return (jl_value_t*)bt;
 }
 
-DLLEXPORT jl_value_t *jl_lookup_code_address(void *ip, int doCframes)
+DLLEXPORT jl_value_t *jl_lookup_code_address(void *ip, int skipC)
 {
     const char *func_name;
-    int line_num;
+    size_t line_num;
     const char *file_name;
-#ifdef _OS_WINDOWS_
-    int fromC =
-#endif
-        frame_info_from_ip(&func_name, &line_num, &file_name, (size_t)ip, doCframes);
+    int fromC = frame_info_from_ip(&func_name, &line_num, &file_name, (size_t)ip, skipC);
     if (func_name != NULL) {
-        jl_value_t *r = (jl_value_t*)jl_alloc_tuple(3);
+        jl_value_t *r = (jl_value_t*)jl_alloc_tuple(5);
         JL_GC_PUSH1(&r);
         jl_tupleset(r, 0, jl_symbol(func_name));
         jl_tupleset(r, 1, jl_symbol(file_name));
         jl_tupleset(r, 2, jl_box_long(line_num));
-#ifdef _OS_WINDOWS_
-        if (fromC && func_name != name_unknown) free((void*)func_name);
-        if (fromC && file_name != name_unknown) free((void*)file_name);
-#endif
+        jl_tupleset(r, 3, jl_box_bool(fromC));
+        jl_tupleset(r, 4, jl_box_long((intptr_t)ip));
         JL_GC_POP();
         return r;
     }
@@ -691,18 +695,16 @@ DLLEXPORT jl_value_t *jl_get_backtrace(void)
 DLLEXPORT void gdblookup(ptrint_t ip)
 {
     const char *func_name;
-    int line_num;
+    size_t line_num;
     const char *file_name;
-    int fromC = frame_info_from_ip(&func_name, &line_num, &file_name, ip, 1);
+    frame_info_from_ip(&func_name, &line_num, &file_name, ip, 0);
     if (func_name != NULL) {
-        if (fromC)
-            ios_printf(ios_stderr, "%s at %s: offset %x\n", func_name, file_name, line_num);
+        if (line_num == ip)
+            ios_printf(ios_stderr, "unknown function (ip: %d)\n", line_num);
+        else if (line_num == -1)
+            ios_printf(ios_stderr, "%s at %s (unknown line)\n", func_name, file_name, line_num);
         else
             ios_printf(ios_stderr, "%s at %s:%d\n", func_name, file_name, line_num);
-#ifdef _OS_WINDOWS_
-        if (fromC && func_name != name_unknown) free((void*)func_name);
-        if (fromC && file_name != name_unknown) free((void*)file_name);
-#endif
     }
 }
 
@@ -722,6 +724,7 @@ DLLEXPORT void gdbbacktrace()
 // yield to exception handler
 void NORETURN throw_internal(jl_value_t *e)
 {
+    assert(e != NULL);
     jl_exception_in_transit = e;
     if (jl_current_task->eh != NULL) {
         jl_longjmp(jl_current_task->eh->eh_ctx, 1);
@@ -733,15 +736,9 @@ void NORETURN throw_internal(jl_value_t *e)
             JL_PRINTF(JL_STDERR, "\n");
             exit(1);
         }
-        jl_task_t *cont = jl_current_task->parent;
-        while (cont->done || cont->eh == NULL)
-            cont = cont->parent;
-        // for now, exit the task
         jl_current_task->exception = e;
         finish_task(jl_current_task, e);
-        jl_current_task->exception = jl_nothing;
-        ctx_switch(cont, &cont->eh->eh_ctx);
-        // TODO: continued exception
+        assert(0);
     }
     jl_exit(1);
 }
@@ -749,11 +746,12 @@ void NORETURN throw_internal(jl_value_t *e)
 // record backtrace and raise an error
 DLLEXPORT void jl_throw(jl_value_t *e)
 {
+    assert(e != NULL);
     record_backtrace();
     throw_internal(e);
 }
 
-DLLEXPORT void jl_rethrow()
+DLLEXPORT void jl_rethrow(void)
 {
     throw_internal(jl_exception_in_transit);
 }
@@ -775,12 +773,12 @@ jl_task_t *jl_new_task(jl_function_t *start, size_t ssize)
     t->type = (jl_value_t*)jl_task_type;
     ssize = LLT_ALIGN(ssize, pagesz);
     t->ssize = ssize;
-    t->parent = NULL;
-    t->last = jl_current_task;
+    t->current_module = NULL;
+    t->parent = jl_current_task;
+    t->last = NULL;
     t->tls = jl_nothing;
     t->consumers = jl_nothing;
-    t->done = 0;
-    t->runnable = 1;
+    t->state = runnable_sym;
     t->start = start;
     t->result = jl_nothing;
     t->donenotify = jl_nothing;
@@ -850,12 +848,11 @@ JL_CALLABLE(jl_f_yieldto)
 {
     JL_NARGSV(yieldto, 1);
     JL_TYPECHK(yieldto, task, args[0]);
-    n_args_in_transit = nargs-1;
     if (nargs == 2) {
         jl_task_arg_in_transit = args[1];
     }
     else if (nargs > 2) {
-        jl_task_arg_in_transit = jl_f_tuple(NULL, &args[1], n_args_in_transit);
+        jl_task_arg_in_transit = jl_f_tuple(NULL, &args[1], nargs-1);
     }
     else {
         jl_task_arg_in_transit = (jl_value_t*)jl_null;
@@ -876,31 +873,33 @@ void jl_init_tasks(void *stack, size_t ssize)
     jl_task_type = jl_new_datatype(jl_symbol("Task"),
                                    jl_any_type,
                                    jl_null,
-                                   jl_tuple(10,
+                                   jl_tuple(9,
                                             jl_symbol("parent"),
                                             jl_symbol("last"),
                                             jl_symbol("storage"),
+                                            jl_symbol("state"),
                                             jl_symbol("consumers"),
-                                            jl_symbol("done"),
-                                            jl_symbol("runnable"),
-                                            jl_symbol("result"),
                                             jl_symbol("donenotify"),
+                                            jl_symbol("result"),
                                             jl_symbol("exception"),
                                             jl_symbol("code")),
-                                   jl_tuple(10,
+                                   jl_tuple(9,
                                             jl_any_type, jl_any_type,
+                                            jl_any_type, jl_sym_type,
                                             jl_any_type, jl_any_type,
-                                            jl_bool_type, jl_bool_type,
-                                            jl_any_type, jl_any_type,
-                                            jl_any_type, jl_function_type),
+                                            jl_any_type, jl_any_type, jl_function_type),
                                    0, 1);
     jl_tupleset(jl_task_type->types, 0, (jl_value_t*)jl_task_type);
     jl_task_type->fptr = jl_f_task;
 
+    done_sym = jl_symbol("done");
+    failed_sym = jl_symbol("failed");
+    runnable_sym = jl_symbol("runnable");
+
     jl_current_task = (jl_task_t*)allocobj(sizeof(jl_task_t));
     jl_current_task->type = (jl_value_t*)jl_task_type;
 #ifdef COPY_STACKS
-    jl_current_task->stackbase = (char *)stack + ssize;
+    jl_current_task->stackbase = NULL;
     jl_current_task->ssize = 0;  // size of saved piece
     jl_current_task->bufsz = 0;
 #else
@@ -909,11 +908,11 @@ void jl_init_tasks(void *stack, size_t ssize)
 #endif
     jl_current_task->stkbuf = NULL;
     jl_current_task->parent = jl_current_task;
+    jl_current_task->current_module = jl_current_module;
     jl_current_task->last = jl_current_task;
     jl_current_task->tls = NULL;
     jl_current_task->consumers = NULL;
-    jl_current_task->done = 0;
-    jl_current_task->runnable = 1;
+    jl_current_task->state = runnable_sym;
     jl_current_task->start = NULL;
     jl_current_task->result = NULL;
     jl_current_task->donenotify = NULL;
@@ -929,3 +928,7 @@ void jl_init_tasks(void *stack, size_t ssize)
     jl_task_arg_in_transit = (jl_value_t*)jl_null;
     jl_unprotect_stack_func = jl_new_closure(jl_unprotect_stack, (jl_value_t*)jl_null, NULL);
 }
+
+#ifdef __cplusplus
+}
+#endif
