@@ -15,18 +15,25 @@
 #include <unistd.h>
 #endif
 #include "julia.h"
-#include "builtin_proto.h"
+#include "julia_internal.h"
+#include "uv.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
-DLLEXPORT char *julia_home = NULL;
 // current line number in a file
 int jl_lineno = 0;
 
 jl_module_t *jl_old_base_module = NULL;
+// the Main we started with, in case it is switched
+jl_module_t *jl_internal_main_module = NULL;
 
 jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast);
 
 void jl_add_standard_imports(jl_module_t *m)
 {
+    assert(jl_base_module != NULL);
     // using Base
     jl_module_using(m, jl_base_module);
     // importall Base.Operators
@@ -34,9 +41,55 @@ void jl_add_standard_imports(jl_module_t *m)
                                                        jl_symbol("Operators")));
 }
 
-extern int base_module_conflict;
+jl_module_t *jl_new_main_module(void)
+{
+    // switch to a new top-level module
+    if (jl_current_module != jl_main_module && jl_current_module != NULL)
+        jl_error("Main can only be replaced from the top level");
+
+    jl_module_t *old_main = jl_main_module;
+
+    jl_main_module = jl_new_module(jl_symbol("Main"));
+    jl_main_module->parent = jl_main_module;
+    jl_current_module = jl_main_module;
+
+    jl_core_module->parent = jl_main_module;
+    jl_set_const(jl_main_module, jl_symbol("Core"),
+                 (jl_value_t*)jl_core_module);
+    jl_set_global(jl_core_module, jl_symbol("Main"),
+                  (jl_value_t*)jl_main_module);
+    jl_current_task->current_module = jl_main_module;
+
+    jl_module_import(jl_main_module, jl_core_module, jl_symbol("eval"));
+
+    return old_main;
+}
+
+jl_array_t *jl_module_init_order = NULL;
+
+// load time init procedure: in build mode, only record order
+void jl_module_load_time_initialize(jl_module_t *m)
+{
+    int build_mode = (jl_compileropts.build_path != NULL);
+    if (build_mode) {
+        if (jl_module_init_order == NULL)
+            jl_module_init_order = jl_alloc_cell_1d(0);
+        jl_cell_1d_push(jl_module_init_order, (jl_value_t*)m);
+    }
+    else {
+        jl_module_run_initializer(m);
+    }
+}
+
+extern void jl_get_system_hooks(void);
 jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
 {
+    static arraylist_t module_stack;
+    static int initialized=0;
+    if (!initialized) {
+        arraylist_new(&module_stack, 0);
+        initialized = 1;
+    }
     assert(ex->head == module_sym);
     jl_module_t *last_module = jl_current_module;
     if (jl_array_len(ex->args) != 3 || !jl_is_expr(jl_exprarg(ex,2))) {
@@ -57,10 +110,17 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
     newm->parent = parent_module;
     b->value = (jl_value_t*)newm;
     if (parent_module == jl_main_module && name == jl_symbol("Base")) {
-        base_module_conflict = (jl_base_module != NULL);
-        jl_old_base_module = jl_base_module;
         // pick up Base module during bootstrap
+        jl_old_base_module = jl_base_module;
         jl_base_module = newm;
+        // reinitialize global variables
+        // to pick up new types from Base
+        jl_errorexception_type = NULL;
+        jl_typeerror_type = NULL;
+        jl_methoderror_type = NULL;
+        jl_loaderror_type = NULL;
+        jl_weakref_type = NULL;
+        jl_current_task->tls = jl_nothing;
     }
     // export all modules from Main
     if (parent_module == jl_main_module)
@@ -74,7 +134,8 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
     }
 
     JL_GC_PUSH1(&last_module);
-    jl_current_module = newm;
+    jl_module_t *task_last_m = jl_current_task->current_module;
+    jl_current_task->current_module = jl_current_module = newm;
 
     jl_array_t *exprs = ((jl_expr_t*)jl_exprarg(ex, 2))->args;
     JL_TRY {
@@ -86,10 +147,12 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
     }
     JL_CATCH {
         jl_current_module = last_module;
+        jl_current_task->current_module = task_last_m;
         jl_rethrow();
     }
     JL_GC_POP();
     jl_current_module = last_module;
+    jl_current_task->current_module = task_last_m;
 
 #if 0
     // some optional post-processing steps
@@ -110,6 +173,15 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
         }
     }
 #endif
+
+    arraylist_push(&module_stack, newm);
+
+    if (jl_current_module == jl_main_module) {
+        while (module_stack.len > 0) {
+            jl_module_load_time_initialize((jl_module_t *) arraylist_pop(&module_stack));
+        }
+    }
+
     return jl_nothing;
 }
 
@@ -164,7 +236,7 @@ int jl_eval_with_compiler_p(jl_expr_t *expr, int compileloops)
             }
         }
         size_t sz = (maxlabl+1+7)/8;
-        char *labls = alloca(sz); memset(labls,0,sz);
+        char *labls = (char*)alloca(sz); memset(labls,0,sz);
         for(i=0; i < jl_array_len(body); i++) {
             jl_value_t *stmt = jl_cellref(body,i);
             if (jl_is_labelnode(stmt)) {
@@ -218,6 +290,7 @@ static jl_module_t *eval_import_path_(jl_array_t *args, int retrying)
         m = jl_current_module;
         while (1) {
             var = (jl_sym_t*)jl_cellref(args,i);
+            assert(jl_is_symbol(var));
             i++;
             if (var != dot_sym) {
                 if (i == jl_array_len(args))
@@ -232,12 +305,20 @@ static jl_module_t *eval_import_path_(jl_array_t *args, int retrying)
     while (1) {
         if (jl_binding_resolved_p(m, var)) {
             jl_binding_t *mb = jl_get_binding(m, var);
+            jl_module_t *m0 = m;
             assert(mb != NULL);
-            if (mb->owner == m || mb->imported) {
+            if (mb->owner == m0 || mb->imported) {
                 m = (jl_module_t*)mb->value;
-                if (m == NULL || !jl_is_module(m))
+                if ((mb->owner == m0 && m != NULL && !jl_is_module(m)) ||
+                    (mb->imported && (m == NULL || !jl_is_module(m))))
                     jl_errorf("invalid module path (%s does not name a module)", var->name);
-                break;
+                // If the binding has been resolved but is (1) undefined, and (2) owned
+                // by the module we're importing into, then allow the import into the
+                // undefined variable (by setting m back to m0).
+                if (m == NULL)
+                    m = m0;
+                else
+                    break;
             }
         }
         if (m == jl_main_module) {
@@ -371,12 +452,14 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
     int ewc = 0;
     JL_GC_PUSH3(&thunk, &thk, &ex);
 
-    if (ex->head != body_sym && ex->head != thunk_sym) {
+    if (ex->head != body_sym && ex->head != thunk_sym && ex->head != return_sym &&
+        ex->head != method_sym) {
         // not yet expanded
         ex = (jl_expr_t*)jl_expand(e);
     }
+    jl_sym_t *head = jl_is_expr(ex) ? ex->head : NULL;
 
-    if (jl_is_expr(ex) && ex->head == toplevel_sym) {
+    if (head == toplevel_sym) {
         int i=0; jl_value_t *res=jl_nothing;
         for(i=0; i < jl_array_len(ex->args); i++) {
             res = jl_toplevel_eval_flex(jl_cellref(ex->args, i), fast);
@@ -385,9 +468,10 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
         return res;
     }
 
-    if (jl_is_expr(ex) && ex->head == thunk_sym) {
+    if (head == thunk_sym) {
         thk = (jl_lambda_info_t*)jl_exprarg(ex,0);
         assert(jl_is_lambda_info(thk));
+        assert(jl_is_expr(thk->ast));
         ewc = jl_eval_with_compiler_p(jl_lam_body((jl_expr_t*)thk->ast), fast);
         if (!ewc) {
             if (jl_lam_vars_captured((jl_expr_t*)thk->ast)) {
@@ -397,12 +481,12 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
         }
     }
     else {
-        if (jl_is_expr(ex) && jl_eval_with_compiler_p((jl_expr_t*)ex, fast)) {
+        if (head && jl_eval_with_compiler_p((jl_expr_t*)ex, fast)) {
             thk = jl_wrap_expr((jl_value_t*)ex);
             ewc = 1;
         }
         else {
-            if (ex->head == body_sym) {
+            if (head == body_sym) {
                 result = jl_toplevel_eval_body(ex->args);
             }
             else if (jl_is_toplevel_only_expr((jl_value_t*)ex)) {
@@ -450,7 +534,7 @@ jl_value_t *jl_parse_eval_all(char *fname)
             if (form == NULL)
                 break;
             if (jl_is_expr(form)) {
-                if (((jl_expr_t*)form)->head == jl_continue_sym) {
+                if (((jl_expr_t*)form)->head == jl_incomplete_sym) {
                     jl_errorf("syntax: %s", jl_string_data(jl_exprarg(form,0)));
                 }
                 if (((jl_expr_t*)form)->head == error_sym) {
@@ -474,13 +558,14 @@ jl_value_t *jl_parse_eval_all(char *fname)
     return result;
 }
 
-int asprintf(char **strp, const char *fmt, ...);
-
 jl_value_t *jl_load(const char *fname)
 {
     if (jl_current_module == jl_base_module) {
         //This deliberatly uses ios, because stdio initialization has been moved to Julia
-        jl_printf(JL_STDOUT, "%s\n", fname);
+        jl_printf(JL_STDOUT, "%s\r\n", fname);
+#ifdef _OS_WINDOWS_        
+        uv_run(uv_default_loop(), 1);
+#endif
     }
     char *fpath = (char*)fname;
     uv_stat_t stbuf;
@@ -492,9 +577,6 @@ jl_value_t *jl_load(const char *fname)
     }
     jl_value_t *result = jl_parse_eval_all(fpath);
     if (fpath != fname) free(fpath);
-    if (jl_current_module == jl_base_module) {
-        jl_printf(JL_STDOUT, "\x1B[1F\x1B[2K");
-    }
     return result;
 }
 
@@ -522,6 +604,7 @@ void jl_set_datatype_super(jl_datatype_t *tt, jl_value_t *super)
 {
     if (!jl_is_datatype(super) || super == (jl_value_t*)jl_undef_type ||
         !jl_is_abstracttype(super) ||
+        tt->name == ((jl_datatype_t*)super)->name ||
         jl_subtype(super,(jl_value_t*)jl_vararg_type,0) ||
         jl_subtype(super,(jl_value_t*)jl_type_type,0)) {
         jl_errorf("invalid subtyping in definition of %s",tt->name->name->name);
@@ -558,26 +641,20 @@ static int type_contains(jl_value_t *ty, jl_value_t *x)
 
 void print_func_loc(JL_STREAM *s, jl_lambda_info_t *li);
 
-jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
-                          jl_tuple_t *argtypes, jl_function_t *f)
+DLLEXPORT jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
+                                    jl_tuple_t *argtypes, jl_function_t *f)
 {
     // argtypes is a tuple ((types...), (typevars...))
     jl_tuple_t *t = (jl_tuple_t*)jl_t1(argtypes);
     argtypes = (jl_tuple_t*)jl_t0(argtypes);
-    jl_value_t *gf;
-    if (bnd) {
-        //jl_declare_constant(bnd);
-        if (bnd->value != NULL && !bnd->constp) {
-            jl_errorf("cannot define function %s; it already has a value",
-                      bnd->name->name);
-        }
-        bnd->constp = 1;
+    jl_value_t *gf=NULL;
+
+    if (bnd && bnd->value != NULL && !bnd->constp) {
+        jl_errorf("cannot define function %s; it already has a value",
+                  bnd->name->name);
     }
-    if (*bp == NULL) {
-        gf = (jl_value_t*)jl_new_generic_function(name);
-        *bp = gf;
-    }
-    else {
+
+    if (*bp != NULL) {
         gf = *bp;
         if (!jl_is_gf(gf)) {
             if (jl_is_datatype(gf) &&
@@ -589,20 +666,14 @@ jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
             }
         }
     }
-    JL_GC_PUSH1(&gf);
-    assert(jl_is_function(f));
-    assert(jl_is_tuple(argtypes));
-    assert(jl_is_tuple(t));
 
-    for(size_t i=0; i < jl_tuple_len(argtypes); i++) {
+    size_t na = jl_tuple_len(argtypes);
+    for(size_t i=0; i < na; i++) {
         jl_value_t *elt = jl_tupleref(argtypes,i);
         if (!jl_is_type(elt) && !jl_is_typevar(elt)) {
             jl_lambda_info_t *li = f->linfo;
             jl_errorf("invalid type for argument %s in method definition for %s at %s:%d",
-                      jl_is_expr(li->ast) ?
-                      ((jl_sym_t*)jl_arrayref(jl_lam_args((jl_expr_t*)li->ast),i))->name :
-                      "?",
-                      name->name, li->file->name, li->line);
+                      jl_lam_argname(li,i)->name, name->name, li->file->name, li->line);
         }
     }
 
@@ -618,6 +689,19 @@ jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
             JL_PRINTF(JL_STDERR, ".\nThe method will not be callable.\n");
         }
     }
+
+    if (bnd) {
+        bnd->constp = 1;
+    }
+    if (*bp == NULL) {
+        gf = (jl_value_t*)jl_new_generic_function(name);
+        *bp = gf;
+    }
+    JL_GC_PUSH1(&gf);
+    assert(jl_is_function(f));
+    assert(jl_is_tuple(argtypes));
+    assert(jl_is_tuple(t));
+
     jl_add_method((jl_function_t*)gf, argtypes, f, t);
     if (jl_boot_file_loaded &&
         f->linfo && f->linfo->ast && jl_is_expr(f->linfo->ast)) {
@@ -627,3 +711,33 @@ jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
     JL_GC_POP();
     return gf;
 }
+
+void jl_check_static_parameter_conflicts(jl_lambda_info_t *li, jl_tuple_t *t, jl_sym_t *fname)
+{
+    jl_array_t *vinfo;
+    size_t nvars;
+
+    if (li->ast && jl_is_expr(li->ast)) {
+        vinfo = jl_lam_vinfo((jl_expr_t*)li->ast);
+        nvars = jl_array_len(vinfo);
+        for(size_t i=0; i < jl_tuple_len(t); i++) {
+            for(size_t j=0; j < nvars; j++) {
+                jl_value_t *tv = jl_tupleref(t,i);
+                if (jl_is_typevar(tv)) {
+                    if ((jl_sym_t*)jl_arrayref((jl_array_t*)jl_arrayref(vinfo,j),0) ==
+                        ((jl_tvar_t*)tv)->name) {
+                        JL_PRINTF(JL_STDERR,
+                                  "Warning: local variable %s conflicts with a static parameter in %s",
+                                  ((jl_tvar_t*)tv)->name->name, fname->name);
+                        print_func_loc(JL_STDERR, li);
+                        JL_PRINTF(JL_STDERR, ".\n");
+                    }
+                }
+            }
+        }
+    }
+}
+
+#ifdef __cplusplus
+}
+#endif
