@@ -2,6 +2,9 @@
 
 ## basics
 
+@test length([1, 2, 3]) == 3
+@test countnz([1, 2, 3]) == 3
+
 a = ones(4)
 b = a+a
 @test b[1]==2. && b[2]==2. && b[3]==2. && b[4]==2.
@@ -9,14 +12,14 @@ b = a+a
 @test length((1,)) == 1
 @test length((1,2)) == 2
 
-@test isequal(1+[1,2,3], [2,3,4])
-@test isequal([1,2,3]+1, [2,3,4])
-@test isequal(1-[1,2,3], [0,-1,-2])
-@test isequal([1,2,3]-1, [0,1,2])
+@test isequal(1.+[1,2,3], [2,3,4])
+@test isequal([1,2,3].+1, [2,3,4])
+@test isequal(1.-[1,2,3], [0,-1,-2])
+@test isequal([1,2,3].-1, [0,1,2])
 
 @test isequal(5*[1,2,3], [5,10,15])
 @test isequal([1,2,3]*5, [5,10,15])
-@test isequal(1/[1,2,5], [1.0,0.5,0.2])
+@test isequal(1./[1,2,5], [1.0,0.5,0.2])
 @test isequal([1,2,3]/5, [0.2,0.4,0.6])
 
 a = ones(2,2)
@@ -50,12 +53,10 @@ a = reshape(b, (2, 2, 2, 2, 2))
 
 sz = (5,8,7)
 A = reshape(1:prod(sz),sz...)
-tmp = A[2:6]
-@test tmp == [2:6]
+@test A[2:6] == [2:6]
 tmp = A[1:3,2,2:4]
 @test tmp == cat(3,46:48,86:88,126:128)
-tmp = A[:,7:-3:1,5]
-@test tmp == [191 176 161; 192 177 162; 193 178 163; 194 179 164; 195 180 165]
+@test A[:,7:-3:1,5] == [191 176 161; 192 177 162; 193 178 163; 194 179 164; 195 180 165]
 tmp = A[:,3:9]
 @test tmp == reshape(11:45,5,7)
 rng = (2,2:3,2:2:5)
@@ -90,15 +91,18 @@ b = [4, 6, 2, -7, 1]
 ind = findin(a, b)
 @test ind == [3,4]
 
+rt = Base.return_types(setindex!, (Array{Int32, 3}, Uint8, Vector{Int}, Float64, Range1{Int}))
+@test length(rt) == 1 && rt[1] == Array{Int32, 3}
+
 # sub
 A = reshape(1:120, 3, 5, 8)
 sA = sub(A, 2, 1:5, :)
 @test parent(sA) == A
 @test parentindexes(sA) == (2:2, 1:5, 1:8)
-@test Base.parentdims(sA) == 1:3
+@test Base.parentdims(sA) == [1:3]
 @test size(sA) == (1, 5, 8)
-@test_throws sA[2, 1:8]
-@test sA[1, 2, 1:8][:] == 5:15:120
+@test_throws BoundsError sA[2, 1:8]
+@test sA[1, 2, 1:8][:] == [5:15:120]
 sA[2:5:end] = -1
 @test all(sA[2:5:end] .== -1)
 @test all(A[5:15:120] .== -1)
@@ -106,38 +110,45 @@ sA[2:5:end] = -1
 @test stride(sA,3) == 15
 @test stride(sA,4) == 120
 sA = sub(A, 1:3, 1:5, 5)
-@test Base.parentdims(sA) == 1:2
+@test Base.parentdims(sA) == [1:2]
 sA[1:3,1:5] = -2
 @test all(A[:,:,5] .== -2)
 sA[:] = -3
 @test all(A[:,:,5] .== -3)
 @test strides(sA) == (1,3)
 sA = sub(A, 1:3, 3, 2:5)
-@test Base.parentdims(sA) == 1:3
+@test Base.parentdims(sA) == [1:3]
 @test size(sA) == (3,1,4)
 @test sA == A[1:3,3,2:5]
 @test sA[:] == A[1:3,3,2:5][:]
 sA = sub(A, 1:2:3, 1:3:5, 1:2:8)
-@test Base.parentdims(sA) == 1:3
+@test Base.parentdims(sA) == [1:3]
 @test strides(sA) == (2,9,30)
 @test sA[:] == A[1:2:3, 1:3:5, 1:2:8][:]
 
+# sub logical indexing #4763
+A = sub([1:10], 5:8)
+@test A[A.<7] == [5, 6]
+B = reshape(1:16, 4, 4)
+sB = sub(B, 2:3, 2:3)
+@test sB[sB.>8] == [10, 11]
+
 # slice
 A = reshape(1:120, 3, 5, 8)
 sA = slice(A, 2, :, 1:8)
 @test parent(sA) == A
 @test parentindexes(sA) == (2, 1:5, 1:8)
-@test Base.parentdims(sA) == 2:3
+@test Base.parentdims(sA) == [2:3]
 @test size(sA) == (5, 8)
 @test strides(sA) == (3,15)
-@test sA[2, 1:8][:] == 5:15:120
-@test sA[:,1] == 2:3:14
-@test sA[2:5:end] == 5:15:110
+@test sA[2, 1:8][:] == [5:15:120]
+@test sA[:,1] == [2:3:14]
+@test sA[2:5:end] == [5:15:110]
 sA[2:5:end] = -1
 @test all(sA[2:5:end] .== -1)
 @test all(A[5:15:120] .== -1)
 sA = slice(A, 1:3, 1:5, 5)
-@test Base.parentdims(sA) == 1:2
+@test Base.parentdims(sA) == [1:2]
 @test size(sA) == (3,5)
 @test strides(sA) == (1,3)
 sA = slice(A, 1:2:3, 3, 1:2:8)
@@ -150,29 +161,37 @@ a = [5:8]
 @test parent(a) == a
 @test parentindexes(a) == (1:4,)
 
-# 4335
-@test_throws slice(A, 1:2)
-@test_throws slice(A, 1:2, 3:4)
-@test_throws slice(A, 1:2, 3:4, 5:6, 7:8)
+# issue #4335
+@test_throws BoundsError slice(A, 1:2)
+@test_throws BoundsError slice(A, 1:2, 3:4)
+@test_throws BoundsError slice(A, 1:2, 3:4, 5:6, 7:8)
 
 # Out-of-bounds construction. See #4044
 A = rand(7,7)
 rng = 1:4
 sA = sub(A, 2, rng-1)
-@test_throws sA[1,1]
+@test_throws BoundsError sA[1,1]
 @test sA[1,2] == A[2,1]
 sA = sub(A, 2, rng)
 B = sub(sA, 1, rng-1)
 C = sub(B, 1, rng+1)
 @test C == sA
 sA = slice(A, 2, rng-1)
-@test_throws sA[1]
+@test_throws BoundsError sA[1]
 @test sA[2] == A[2,1]
 sA = slice(A, 2, rng)
 B = slice(sA, rng-1)
 C = sub(B, rng+1)
 @test C == sA
 
+# issue #6218 - logical indexing
+A = rand(2, 2, 3)
+msk = ones(Bool, 2, 2)
+msk[2,1] = false
+sA = sub(A, :, :, 1)
+sA[msk] = 1.0
+@test sA[msk] == ones(countnz(msk))
+
 # get
 let
     A = reshape(1:24, 3, 8)
@@ -187,7 +206,7 @@ let
     X = get(A, -5:5, nan(Float32))
     @test eltype(X) == Float32
     @test isnan(X) == [trues(6),falses(5)]
-    @test X[7:11] == 1:5
+    @test X[7:11] == [1:5]
     X = get(A, (2:4, 9:-2:-13), 0)
     Xv = zeros(Int, 3, 12)
     Xv[1:2, 2:5] = A[2:3, 7:-2:1]
@@ -197,14 +216,28 @@ let
 end
 
 ## arrays as dequeues
-l = {1,2,3}
-push!(l,8)
+l = {1}
+push!(l,2,3,8)
 @test l[1]==1 && l[2]==2 && l[3]==3 && l[4]==8
 v = pop!(l)
 @test v == 8
 v = pop!(l)
 @test v == 3
 @test length(l)==2
+unshift!(l,4,7,5)
+@test l[1]==4 && l[2]==7 && l[3]==5 && l[4]==1 && l[5]==2
+v = shift!(l)
+@test v == 4
+@test length(l)==4
+
+v = [3, 7, 6]
+@test_throws BoundsError insert!(v, 0, 5)
+for i = 1:4
+    vc = copy(v)
+    @test insert!(vc, i, 5) === vc
+    @test vc == [v[1:(i-1)], 5, v[i:end]]
+end
+@test_throws BoundsError insert!(v, 5, 5)
 
 # concatenation
 @test isequal([ones(2,2)  2*ones(2,1)], [1. 1 2; 1 1 2])
@@ -296,6 +329,7 @@ for i = 1:3
 end
 
 #permutes correctly
+@test isequal(z,permutedims(y,[3,1,2]))
 @test isequal(z,permutedims(y,(3,1,2)))
 
 # of a subarray
@@ -314,59 +348,40 @@ for i = tensors
     @test isequal(i,permutedims(ipermutedims(i,perm),perm))
 end
 
+## unique across dim ##
 
-## reduce ##
+# All rows and columns unique
+A = ones(10, 10)
+A[diagind(A)] = shuffle!([1:10])
+@test unique(A, 1) == A
+@test unique(A, 2) == A
 
-z = zeros(2,2,2,2)
-for i=1:16
-    z[i] = i
+# 10 repeats of each row
+B = A[shuffle!(repmat(1:10, 10)), :]
+C = unique(B, 1)
+@test sortrows(C) == sortrows(A)
+@test unique(B, 2) == B
+@test unique(B.', 2).' == C
+
+# Along third dimension
+D = cat(3, B, B)
+@test unique(D, 1) == cat(3, C, C)
+@test unique(D, 3) == cat(3, B)
+
+# With hash collisions
+immutable HashCollision
+    x::Float64
 end
-
-@test sum(z) == sum(z,(1,2,3,4))[1] == 136
-
-v = cell(2,2,1,1)
-v[1,1,1,1] = 28.0
-v[1,2,1,1] = 36.0
-v[2,1,1,1] = 32.0
-v[2,2,1,1] = 40.0
-
-@test isequal(v,sum(z,(3,4)))
-
-z = rand(10^6)
-let es = sum_kbn(z), es2 = sum_kbn(z[1:10^5])
-    @test (es - sum(z)) < es * 1e-13
-    cs = cumsum(z)
-    @test (es - cs[end]) < es * 1e-13
-    @test (es2 - cs[10^5]) < es2 * 1e-13
-end
-@test sum(sin(z)) == sum(sin, z)
-
-@test any([true false; false false], 2) == [true false]'
-@test any([true false; false false], 1) == [true false]
-
-@test all([true true; false true], 2) == [true false]'
-@test all([true false; false true], 1) == [false false]
+Base.hash(::HashCollision, h::Uint) = h
+@test map(x->x.x, unique(map(HashCollision, B), 1)) == C
 
 ## large matrices transpose ##
 
 for i = 1 : 3
     a = rand(200, 300)
-
-    @test isequal(a', permutedims(a, (2, 1)))
+    @test isequal(a', permutedims(a, [2, 1]))
 end
 
-## cumsum, cummin, cummax
-
-@test isequal(cummin([1, 2, 5, -1, 3, -2]), [1, 1, 1, -1, -1, -2])
-@test isequal(cummax([1, 2, 5, -1, 3, -2]), [1, 2, 5, 5, 5, 5])
-
-@test isequal(cummax([1 0; 0 1], 1), [1 0; 1 1])
-@test isequal(cummax([1 0; 0 1], 2), [1 1; 0 1])
-@test isequal(cummin([1 0; 0 1], 1), [1 0; 0 0])
-@test isequal(cummin([1 0; 0 1], 2), [1 0; 0 0])
-
-@test sum_kbn([1,1e100,1,-1e100]) == 2
-
 begin
     local A, A1, A2, A3, v, v2, cv, cv2, c, R, T
     A = ones(Int,2,3,4)
@@ -398,7 +413,7 @@ begin
     @test R == [1 1; 2 2]
 
     R = repeat([1 2;
-                3 4], inner = [1, 1], outer = [1, 1]) 
+                3 4], inner = [1, 1], outer = [1, 1])
     @test R == [1 2;
                   3 4]
     R = repeat([1 2;
@@ -562,6 +577,34 @@ begin
 end
 
 @test (1:5)[[true,false,true,false,true]] == [1,3,5]
+@test [1:5][[true,false,true,false,true]] == [1,3,5]
+@test_throws BoundsError (1:5)[[true,false,true,false]]
+@test_throws BoundsError (1:5)[[true,false,true,false,true,false]]
+@test_throws BoundsError [1:5][[true,false,true,false]]
+@test_throws BoundsError [1:5][[true,false,true,false,true,false]]
+a = [1:5]
+a[[true,false,true,false,true]] = 6
+@test a == [6,2,6,4,6]
+a[[true,false,true,false,true]] = [7,8,9]
+@test a == [7,2,8,4,9]
+@test_throws DimensionMismatch (a[[true,false,true,false,true]] = [7,8,9,10])
+A = reshape(1:15, 3, 5)
+@test A[[true, false, true], [false, false, true, true, false]] == [7 10; 9 12]
+@test_throws BoundsError A[[true, false], [false, false, true, true, false]]
+@test_throws BoundsError A[[true, false, true], [false, true, true, false]]
+@test_throws BoundsError A[[true, false, true, true], [false, false, true, true, false]]
+@test_throws BoundsError A[[true, false, true], [false, false, true, true, false, true]]
+A = ones(Int, 3, 5)
+@test_throws DimensionMismatch A[2,[true, false, true, true, false]] = 2:5
+A[2,[true, false, true, true, false]] = 2:4
+@test A == [1 1 1 1 1; 2 1 3 4 1; 1 1 1 1 1]
+@test_throws DimensionMismatch A[[true,false,true], 5] = [19]
+@test_throws DimensionMismatch A[[true,false,true], 5] = 19:21
+A[[true,false,true], 5] = 7
+@test A == [1 1 1 1 7; 2 1 3 4 1; 1 1 1 1 7]
+
+B = cat(3, 1, 2, 3)
+@test B[:,:,[true, false, true]] == reshape([1,3], 1, 1, 2)  # issue #5454
 
 # issue #2342
 @test isequal(cumsum([1 2 3]), [1 2 3])
@@ -616,6 +659,32 @@ begin
     b = mapslices(sum, ones(2,3,4), [1,2])
     @test size(b) === (1,1,4)
     @test all(b.==6)
+
+    # issue #5141
+    ## Update Removed the version that removes the dimensions when dims==1:ndims(A)
+    c1 = mapslices(x-> maximum(-x), a, [])
+    @test c1 == -a
+
+    # other types than Number
+    @test mapslices(prod,["1" "2"; "3" "4"],1) == ["13" "24"]
+    @test mapslices(prod,["1"],1) == ["1"]
+
+    # issue #5177
+
+    c = ones(2,3,4)
+    m1 = mapslices(x-> ones(2,3), c, [1,2])
+    m2 = mapslices(x-> ones(2,4), c, [1,3])
+    m3 = mapslices(x-> ones(3,4), c, [2,3])
+    @test size(m1) == size(m2) == size(m3) == size(c)
+
+    n1 = mapslices(x-> ones(6), c, [1,2])
+    n2 = mapslices(x-> ones(6), c, [1,3])
+    n3 = mapslices(x-> ones(6), c, [2,3])
+    n1a = mapslices(x-> ones(1,6), c, [1,2])
+    n2a = mapslices(x-> ones(1,6), c, [1,3])
+    n3a = mapslices(x-> ones(1,6), c, [2,3])
+    @test size(n1a) == (1,6,4) && size(n2a) == (1,3,6)  && size(n3a) == (2,1,6)
+    @test size(n1) == (6,1,4) && size(n2) == (6,3,1)  && size(n3) == (2,6,1)
 end
 
 
@@ -640,6 +709,17 @@ b = [3,1,2]
 a[b] = a
 @test a == [3,5,1]
 
+# lexicographic comparison
+@test lexcmp([1.0], [1]) == 0
+@test lexcmp([1], [1.0]) == 0
+@test lexcmp([1, 1], [1, 1]) == 0
+@test lexcmp([1, 1], [2, 1]) == -1
+@test lexcmp([2, 1], [1, 1]) == 1
+@test lexcmp([1, 1], [1, 2]) == -1
+@test lexcmp([1, 2], [1, 1]) == 1
+@test lexcmp([1], [1, 1]) == -1
+@test lexcmp([1, 1], [1]) == 1
+
 # sort on arrays
 begin
     local a = rand(3,3)
@@ -673,6 +753,14 @@ end
 
 # fill
 @test fill!(Array(Float64,1),-0.0)[1] === -0.0
+A = ones(3,3)
+S = sub(A, 2, 1:3)
+fill!(S, 2)
+S = sub(A, 1:2, 3)
+fill!(S, 3)
+@test A == [1 1 3; 2 2 3; 1 1 1]
+rt = Base.return_types(fill!, (Array{Int32, 3}, Uint8))
+@test length(rt) == 1 && rt[1] == Array{Int32, 3}
 
 # splice!
 for idx in {1, 2, 5, 9, 10, 1:0, 2:1, 1:1, 2:2, 1:2, 2:4, 9:8, 10:9, 9:9, 10:10,
@@ -684,6 +772,16 @@ for idx in {1, 2, 5, 9, 10, 1:0, 2:1, 1:1, 2:2, 1:2, 2:4, 9:8, 10:9, 9:9, 10:10,
     end
 end
 
+# deleteat!
+for idx in {1, 2, 5, 9, 10, 1:0, 2:1, 1:1, 2:2, 1:2, 2:4, 9:8, 10:9, 9:9, 10:10,
+            8:9, 9:10, 6:9, 7:10}
+    a = [1:10]; acopy = copy(a)
+    @test deleteat!(a, idx) == [acopy[1:(first(idx)-1)], acopy[(last(idx)+1):end]]
+end
+a = [1:10]
+@test deleteat!(a, [1,3,5,7:10...]) == [2,4,6]
+
+
 # comprehensions
 X = [ i+2j for i=1:5, j=1:5 ]
 @test X[2,3] == 8
@@ -699,7 +797,7 @@ end
 @test isequal([1,2,3], [a for (a,b) in enumerate(2:4)])
 @test isequal([2,3,4], [b for (a,b) in enumerate(2:4)])
 
-@test_throws (10.^[-1])[1] == 0.1
+@test_throws DomainError (10.^[-1])[1] == 0.1
 @test (10.^[-1.])[1] == 0.1
 
 # reverse
@@ -719,10 +817,11 @@ end
 @test isequal(flipdim([2,3,1], 2), [2,3,1])
 @test isequal(flipdim([2 3 1], 1), [2 3 1])
 @test isequal(flipdim([2 3 1], 2), [1 3 2])
-@test_throws flipdim([2,3,1] -1)
+@test_throws ErrorException flipdim([2,3,1], -1)
 @test isequal(flipdim(1:10, 1), 10:-1:1)
 @test isequal(flipdim(1:10, 2), 1:10)
-@test_throws flipdim(1:10, -1)
+@test_throws ErrorException flipdim(1:10, -1)
+@test isequal(flipdim(Array(Int,0,0),1), Array(Int,0,0))  # issue #5872
 
 # issue 4228
 A = [[i i; i i] for i=1:2]
@@ -740,7 +839,7 @@ B = reshape(A, 4)
 @test push!(B,5) == [1,2,3,4,5]
 @test pop!(B) == 5
 C = reshape(B, 1, 4)
-@test_throws push!(C, 5)
+@test_throws MethodError push!(C, 5)
 
 A = [NaN]; B = [NaN]
 @test !(A==A)
@@ -749,3 +848,49 @@ A = [NaN]; B = [NaN]
 @test !(A==B)
 @test isequal(A,B)
 @test A!==B
+
+# complete testsuite for reducedim
+
+# Inferred types
+Nmax = 3 # TODO: go up to CARTESIAN_DIMS+2 (currently this exposes problems)
+for N = 1:Nmax
+    #indexing with (Range1, Range1, Range1)
+    args = ntuple(N, d->Range1{Int})
+    @test Base.return_types(getindex, tuple(Array{Float32, N}, args...)) == {Array{Float32, N}}
+    @test Base.return_types(getindex, tuple(BitArray{N}, args...)) == {BitArray{N}}
+    @test Base.return_types(setindex!, tuple(Array{Float32, N}, Array{Int, 1}, args...)) == {Array{Float32, N}}
+    # Indexing with (Range1, Range1, Float64)
+    args = ntuple(N, d->d<N ? Range1{Int} : Float64)
+    N > 1 && @test Base.return_types(getindex, tuple(Array{Float32, N}, args...)) == {Array{Float32, N-1}}
+    N > 1 && @test Base.return_types(getindex, tuple(BitArray{N}, args...)) == {BitArray{N-1}}
+    N > 1 && @test Base.return_types(setindex!, tuple(Array{Float32, N}, Array{Int, 1}, args...)) == {Array{Float32, N}}
+end
+
+# issue #6645 (32-bit)
+let
+    x = Float64[]
+    for i=1:5; push!(x, 1.0); end
+    @test dot(zeros(5),x) == 0.0
+end
+
+# issue #6977
+@test []' == Array(None,1,0)
+
+# issue #6996
+@test { 1 2; 3 4 }' == { 1 2; 3 4 }.'
+
+# map with promotion (issue #6541)
+@test map(join, ["z", "я"]) == ["z", "я"]
+
+# Handle block matrices
+A = [randn(2,2) for i = 1:2, j = 1:2]
+@test issym(A.'A)
+A = [complex(randn(2,2), randn(2,2)) for i = 1:2, j = 1:2]
+@test ishermitian(A'A)
+
+# issue #7197
+function i7197()
+    S = [1 2 3; 4 5 6; 7 8 9]
+    ind2sub(size(S), 5)
+end
+@test i7197() == (2,2)
