@@ -253,3 +253,26 @@ s = string(n)
 @test length(s) == 208988
 @test endswith(s, "359244926937501")
 @test beginswith(s, "316047687386689")
+
+# serialization (#5133)
+let
+    n = BigInt("359334085968622831041960188598043661065388726959079837")
+    b = IOBuffer()
+    serialize(b,n)
+    seek(b,0)
+    @test deserialize(b) == n
+end
+
+# issue #5873
+@test ndigits(big(90)) == 2
+@test ndigits(big(99)) == 2
+ndigits_mismatch(n) = ndigits(n) != ndigits(BigInt(n))
+@test !any(ndigits_mismatch, 8:9)
+@test !any(ndigits_mismatch, 64:99)
+@test !any(ndigits_mismatch, 512:999)
+@test !any(ndigits_mismatch, 8192:9999)
+
+# conversion from float
+@test BigInt(2.0) == BigInt(2.0f0) == BigInt(big(2.0)) == 2
+@test_throws InexactError convert(BigInt, 2.1)
+@test_throws InexactError convert(BigInt, big(2.1))
