@@ -103,6 +103,76 @@
 @test isequal(exp(complex( NaN, 5.0)), complex( NaN, NaN))
 @test isequal(exp(complex( NaN, NaN)), complex( NaN, NaN))
 
+# expm1:
+#  expm1(conj(z)) = conj(expm1(z))
+
+@test isequal(expm1(complex( 0.0, 0.0)), complex(0.0, 0.0))
+@test isequal(expm1(complex( 0.0,-0.0)), complex(0.0,-0.0))
+@test isequal(expm1(complex( 0.0, Inf)), complex(NaN, NaN))
+@test isequal(expm1(complex( 0.0,-Inf)), complex(NaN, NaN))
+@test isequal(expm1(complex( 0.0, NaN)), complex(NaN, NaN))
+
+@test isequal(expm1(complex(-0.0, 0.0)), complex(-0.0, 0.0))
+@test isequal(expm1(complex(-0.0,-0.0)), complex(-0.0,-0.0))
+
+@test isequal(expm1(complex( 5.0, Inf)), complex(NaN, NaN))
+
+@test isequal(expm1(complex( Inf, 0.0)), complex(Inf, 0.0))
+@test isequal(expm1(complex( Inf,-0.0)), complex(Inf,-0.0))
+@test isequal(expm1(complex( Inf, 5.0)), complex(cos(5.0)*Inf,sin(5.0)* Inf))
+@test isequal(expm1(complex( Inf,-5.0)), complex(cos(5.0)*Inf,sin(5.0)*-Inf))
+@test isequal(expm1(complex( Inf, NaN)), complex(-Inf, NaN))
+@test isequal(expm1(complex( Inf, Inf)), complex(-Inf, NaN))
+@test isequal(expm1(complex( Inf,-Inf)), complex(-Inf, NaN))
+
+@test isequal(expm1(complex(-Inf, 0.0)), complex(-1.0, 0.0))
+@test isequal(expm1(complex(-Inf,-0.0)), complex(-1.0,-0.0))
+@test isequal(expm1(complex(-Inf, 5.0)), complex(-1.0,sin(5.0)* 0.0))
+@test isequal(expm1(complex(-Inf,-5.0)), complex(-1.0,sin(5.0)*-0.0))
+@test isequal(expm1(complex(-Inf, Inf)), complex(-1.0, 0.0))
+@test isequal(expm1(complex(-Inf,-Inf)), complex(-1.0,-0.0))
+@test isequal(expm1(complex(-Inf, NaN)), complex(-1.0, 0.0))
+
+@test isequal(expm1(complex( NaN, 0.0)), complex( NaN, 0.0))
+@test isequal(expm1(complex( NaN,-0.0)), complex( NaN,-0.0))
+@test isequal(expm1(complex( NaN, 5.0)), complex( NaN, NaN))
+@test isequal(expm1(complex( NaN, NaN)), complex( NaN, NaN))
+
+@test isequal(expm1(complex( 1e-20, 0.0)), complex(expm1( 1e-20), 0.0))
+@test isequal(expm1(complex(-1e-20, 0.0)), complex(expm1(-1e-20), 0.0))
+
+@test_approx_eq expm1(complex( 1e-20, 1e-10)) complex( 5e-21, 1e-10)
+@test_approx_eq expm1(complex( 1e-20,-1e-10)) complex( 5e-21,-1e-10)
+@test_approx_eq expm1(complex(-1e-20, 1e-10)) complex(-1.5e-20, 1e-10)
+@test_approx_eq expm1(complex(-1e-20,-1e-10)) complex(-1.5e-20,-1e-10)
+
+@test_approx_eq expm1(complex( 10.0, 10.0)) exp(complex( 10.0, 10.0))-1
+@test_approx_eq expm1(complex( 10.0,-10.0)) exp(complex( 10.0,-10.0))-1
+@test_approx_eq expm1(complex(-10.0, 10.0)) exp(complex(-10.0, 10.0))-1
+@test_approx_eq expm1(complex(-10.0,-10.0)) exp(complex(-10.0,-10.0))-1
+
+# log1p:
+@test isequal(log1p(complex(Inf, 3)), complex(Inf, +0.))
+@test isequal(log1p(complex(Inf, -3)), complex(Inf, -0.))
+@test isequal(log1p(complex(-Inf, 3)), complex(Inf, +pi))
+@test isequal(log1p(complex(-Inf, -3)), complex(Inf, -pi))
+@test isequal(log1p(complex(Inf, NaN)), complex(Inf, NaN))
+@test isequal(log1p(complex(NaN, 0)), complex(NaN, NaN))
+@test isequal(log1p(complex(0, NaN)), complex(NaN, NaN))
+@test isequal(log1p(complex(-1, +0.)), complex(-Inf, +0.))
+@test isequal(log1p(complex(-1, -0.)), complex(-Inf, -0.))
+@test isequal(log1p(complex(-2, 1e-10)), log(1 + complex(-2, 1e-10)))
+@test isequal(log1p(complex(1, Inf)), complex(Inf, pi/2))
+@test isequal(log1p(complex(1, -Inf)), complex(Inf, -pi/2))
+import Base.Math.@horner
+for z in (1e-10+1e-9im, 1e-10-1e-9im, -1e-10+1e-9im, -1e-10-1e-9im)
+    @test_approx_eq log1p(z) @horner(z, 0, 1, -0.5, 1/3, -0.25, 0.2)
+end
+for z in (15+4im, 0.2+3im, 0.08-0.9im)
+    @test_approx_eq log1p(z) log(1+z)
+end
+
+
 # ^ (cpow)
 #  equivalent to exp(y*log(x))
 #    except for 0^0?
@@ -291,14 +361,14 @@
 #  tanh(-z) = -tanh(z)
 @test isequal(tanh(complex( 0.0, 0.0)),complex(0.0,0.0))
 @test isequal(tanh(complex( 0.0,-0.0)),complex(0.0,-0.0))
-@test_throws   tanh(complex( 0.0, Inf)) #complex(NaN,0.0)
-@test_throws   tanh(complex( 0.0,-Inf)) #complex(NaN,-0.0)
+@test_throws DomainError  tanh(complex( 0.0, Inf))
+@test_throws DomainError  tanh(complex( 0.0,-Inf))
 @test isequal(tanh(complex( 0.0, NaN)),complex(NaN,NaN))
 
 @test isequal(tanh(complex(-0.0, 0.0)),complex(-0.0,0.0))
 @test isequal(tanh(complex(-0.0,-0.0)),complex(-0.0,-0.0))
 
-@test_throws   tanh(complex( 5.0, Inf)) #complex(NaN,NaN)
+@test_throws DomainError  tanh(complex( 5.0, Inf))
 @test isequal(tanh(complex( 5.0, NaN)),complex(NaN,NaN))
 
 @test isequal(tanh(complex( Inf, 0.0)),complex(1.0, 0.0))
@@ -340,7 +410,7 @@
 @test isequal(tan(complex(-5.0,-Inf)),complex(sin(2*5.0)*-0.0,-1.0))
 @test isequal(tan(complex(-5.0, NaN)),complex( NaN, NaN))
 
-@test_throws   tan(complex( Inf, 5.0)) #complex(NaN,NaN)
+@test_throws DomainError  tan(complex( Inf, 5.0))
 @test isequal(tan(complex( Inf, Inf)),complex( 0.0, 1.0))
 @test isequal(tan(complex( Inf,-Inf)),complex (0.0,-1.0))
 @test isequal(tan(complex(-Inf, Inf)),complex(-0.0, 1.0))
@@ -592,9 +662,65 @@
 @test isequal(atan(complex( NaN, NaN)),complex( NaN, NaN))
 
 
+# lexcmp
+@test lexcmp(1.0-1.0im, 1.0+0.0im) == -1
+@test lexcmp(0.0+0.0im, 0.0+0.0im) == 0
+@test lexcmp(1.0-1.0im, 0.0+0.0im) == 1
+
+
 # misc.
 
 @test complex(1//2,1//3)^2 === complex(5//36, 1//3)
 @test complex(2,2)^2 === complex(0,8)
-@test_throws complex(2,2)^(-2)
+@test_throws DomainError complex(2,2)^(-2)
 @test complex(2.0,2.0)^(-2) === complex(0.0, -0.125)
+
+# robust division of Float64
+# hard complex divisions from Fig 6 of arxiv.1210.4539
+z7 = Complex{Float64}(3.898125604559113300e289, 8.174961907852353577e295)
+z9 = Complex{Float64}(0.001953125, -0.001953125)
+z10 = Complex{Float64}( 1.02951151789360578e-84, 6.97145987515076231e-220)
+harddivs = ((1.0+im*1.0, 1.0+im*2^1023.0, 2^-1023.0-im*2^-1023.0), #1
+      (1.0+im*1.0, 2^-1023.0+im*2^-1023.0, 2^1023.0+im*0.0), #2
+      (2^1023.0+im*2^-1023.0, 2^677.0+im*2^-677.0, 2^346.0-im*2^-1008.0), #3
+      (2^1023.0+im*2^1023.0, 1.0+im*1.0, 2^1023.0+im*0.0), #4
+      (2^1020.0+im*2^-844., 2^656.0+im*2^-780.0, 2^364.0-im*2^-1072.0), #5
+      (2^-71.0+im*2^1021., 2^1001.0+im*2^-323.0, 2^-1072.0+im*2^20.0), #6
+      (2^-347.0+im*2^-54., 2^-1037.0+im*2^-1058.0, z7), #7
+      (2^-1074.0+im*2^-1074., 2^-1073.0+im*2^-1074., 0.6+im*0.2), #8
+      (2^1015.0+im*2^-989., 2^1023.0+im*2^1023.0, z9), #9
+      (2^-622.0+im*2^-1071., 2^-343.0+im*2^-798.0, z10)#10
+      )
+
+# calculate "accurate bits" in range 0:53 by algorithm given in arxiv.1210.4539
+function sb_accuracy(x,expected)
+  min(logacc(real(x),real(expected)),
+    logacc(imag(x),imag(expected)) )
+end
+relacc(x,expected) = abs(x-expected)/abs(expected)
+function logacc(x::Float64,expected::Float64)
+  x == expected && (return 53)
+  expected == 0 && (return 0)
+  (x == Inf || x == -Inf) && (return 0)
+  isnan(x) && (return 0)
+  ra = relacc(BigFloat(x),BigFloat(expected))
+  max(ifloor(-log2(ra)),0)
+end
+# the robust division algorithm should have 53 or 52
+# bits accuracy for each of the hard divisions
+@test 52 <= minimum([sb_accuracy(h[1]/h[2],h[3]) for h in harddivs])
+
+# division of non-Float64
+function cdiv_test(a,b)
+  c=convert(Complex{Float64},a)/convert(Complex{Float64},b)
+  50 <= sb_accuracy(c,convert(Complex{Float64},a/b))
+end
+@test cdiv_test(complex(1//2, 3//4), complex(17//13, 4//5))
+@test cdiv_test(complex(1,2), complex(8997,2432))
+
+# inv
+@test inv(1e300+0im) == 1e-300 - 0.0im
+
+# issue #7904
+@test log10(10+0im) === 1.0 + 0.0im
+@test log2(2+0im) === 1.0 + 0.0im
