@@ -9,7 +9,7 @@ function matmultest(n, iter)
     b
 end
 
-problemsizes = [(2, 10^6, "tiny"), (2^4, 10^5, "small"), (2^6, 10^4, "medium"), (2^8, 10^2, "large")]
-testdata = [(gemvtest, "matmul", "matrix-matrix multiplication", problemsizes)]
+problemsizes = [(2^2, 10^6, "tiny"), (2^4, 10^5, "small"), (2^6, 10^4, "medium"), (2^8, 10^2, "large")]
+testdata = [(matmultest, "matmul", "matrix-matrix multiplication", problemsizes)]
 include("../perfgeneric.jl")
 
