@@ -19,7 +19,7 @@ function bench_eu_devec(numPaths)
         end
     end
 
-    V = mean( exp(-r*T)*max(K-S,0) )
+    V = mean( exp(-r*T)*max(K.-S,0) )
 end
 
 function bench_eu_vec(numPaths)
@@ -35,8 +35,8 @@ function bench_eu_vec(numPaths)
     t1 = (r-0.5*sigma.^2)*dt
     t2 = sigma*sqrt(dt)
     for i=1:steps
-        S = S .* exp(t1+t2*randn(numPaths))
+        S = S .* exp(t1.+t2*randn(numPaths))
     end
 
-    V = mean( exp(-r*T)*max(K-S,0) )
+    V = mean( exp(-r*T)*max(K.-S,0) )
 end
