@@ -155,6 +155,8 @@ end
 @test parseint("-10") == -10
 @test parseint(Int64,"3830974272") == 3830974272
 @test parseint(Int64,"-3830974272") == -3830974272
+@test parseint('3') == 3
+@test parseint('3', 8) == 3
 
 parsebin(s) = parseint(s,2)
 parseoct(s) = parseint(s,8)
@@ -214,9 +216,33 @@ parsehex(s) = parseint(s,16)
 @test parseint(" 2") == 2
 @test parseint("+2\n") == 2
 @test parseint("-2") == -2
-@test_throws parseint("   2 \n 0")
-@test_throws parseint("2x")
-@test_throws parseint("-")
+@test_throws ErrorException parseint("   2 \n 0")
+@test_throws ErrorException parseint("2x")
+@test_throws ErrorException parseint("-")
+
+@test parseint("1234") == 1234
+@test parseint("0x1234") == 0x1234
+@test parseint("0o1234") == 0o1234
+@test parseint("0b1011") == 0b1011
+@test parseint("-1234") == -1234
+@test parseint("-0x1234") == -int(0x1234)
+@test parseint("-0o1234") == -int(0o1234)
+@test parseint("-0b1011") == -int(0b1011)
+
+## FIXME: #4905, do these tests for Int128/UInt128!
+for T in (Int8, Int16, Int32, Int64)
+    @test parseint(T,string(typemin(T))) == typemin(T)
+    @test parseint(T,string(typemax(T))) == typemax(T)
+    @test_throws OverflowError parseint(T,string(big(typemin(T))-1))
+    @test_throws OverflowError parseint(T,string(big(typemax(T))+1))
+end
+
+for T in (Uint8,Uint16,Uint32,Uint64)
+    @test parseint(T,string(typemin(T))) == typemin(T)
+    @test parseint(T,string(typemax(T))) == typemax(T)
+    @test_throws ErrorException parseint(T,string(big(typemin(T))-1))
+    @test_throws OverflowError parseint(T,string(big(typemax(T))+1))
+end
 
 # string manipulation
 @test strip("\t  hi   \n") == "hi"
@@ -666,6 +692,26 @@ for i1 = 1:length(u8str2)
     end
 end
 
+str="tempus fugit"              #length(str)==12
+ss=SubString(str,1,length(str)) #match source string
+@test length(ss)==length(str)
+
+ss=SubString(str,1,0)    #empty SubString
+@test length(ss)==0
+
+ss=SubString(str,14,20)  #start indexed beyond source string length
+@test length(ss)==0
+
+ss=SubString(str,10,16)  #end indexed beyond source string length
+@test length(ss)==3
+
+str2=""
+ss=SubString(str2,1,4)  #empty source string
+@test length(ss)==0
+
+ss=SubString(str2,1,1)  #empty source string, identical start and end index
+@test length(ss)==0
+
 str = "aa\u2200\u2222bb"
 u = SubString(str, 3, 6)
 @test length(u)==2
@@ -689,6 +735,13 @@ write(b, u)
 
 @test replace("\u2202", '*', '\0') == "\u2202"
 
+# search and SubString (issue #5679)
+str = "Hello, world!"
+u = SubString(str, 1, 5)
+@test rsearch(u, "World") == 0:-1
+@test rsearch(u, 'z') == 0
+@test rsearch(u, "ll") == 3:4
+
 # quotes + interpolation (issue #455)
 @test "$("string")" == "string"
 arr = ["a","b","c"]
@@ -753,6 +806,23 @@ s = "   p"
 @test """
        $s
       """ == " $s$(nl)"
+@test """\t""" == "\t"
+@test """
+      \t""" == ""
+@test """
+      foo
+      \tbar""" == "foo$(nl)\tbar"
+@test """
+      foo
+      \tbar
+      """ == "foo$(nl)\tbar$(nl)"
+@test """
+      foo
+      bar\t""" == "foo$(nl)bar\t"
+@test """
+      foo
+      \tbar
+       """ == "foo$(nl)       bar$(nl)"
 
 # bytes2hex and hex2bytes
 hex_str = "d7a8fbb307d7809469ca9abcb0082e4f8d5651e46d3cdb762d02d0bf37c9e592"
@@ -771,10 +841,10 @@ bin_val = hex2bytes("07bf")
 @test "0123456789abcdefabcdef" == bytes2hex(hex2bytes("0123456789abcdefABCDEF"))
 
 # odd size
-@test_throws hex2bytes("0123456789abcdefABCDEF0")
+@test_throws ErrorException hex2bytes("0123456789abcdefABCDEF0")
 
 #non-hex characters
-@test_throws hex2bytes("0123456789abcdefABCDEFGH")
+@test_throws ErrorException hex2bytes("0123456789abcdefABCDEFGH")
 
 # sizeof
 @test sizeof("abc") == 3
@@ -783,7 +853,7 @@ bin_val = hex2bytes("07bf")
 @test sizeof(RopeString("abc","def")) == 6
 
 # issue #3597
-@test string(CharString(['T', 'e', 's', 't'])[1:1], "X") == "TX"
+@test string(utf32(['T', 'e', 's', 't'])[1:1], "X") == "TX"
 
 # issue #3710
 @test prevind(SubString("{var}",2,4),4) == 3
@@ -832,4 +902,198 @@ bin_val = hex2bytes("07bf")
 
 # issue #4586
 @test rsplit(RevString("ailuj"),'l') == {"ju","ia"}
-@test_throws float64(RevString("64"))
+@test float64(RevString("64")) === 46.0
+
+# issue #6772
+@test float(SubString("10",1,1)) === 1.0
+@test float(SubString("1 0",1,1)) === 1.0
+@test float32(SubString("10",1,1)) === 1.0f0
+
+for T = (Uint8,Int8,Uint16,Int16,Uint32,Int32,Uint64,Int64,Uint128,Int128,BigInt),
+    b = 2:62, _ = 1:10
+    n = T != BigInt ? rand(T) : BigInt(rand(Int128))
+    @test parseint(T,base(b,n),b) == n
+end
+
+# normalize_string (Unicode normalization etc.):
+@test normalize_string("\u006e\u0303", :NFC) == "\u00f1"
+@test "\u006e\u0303" == normalize_string("\u00f1", :NFD)
+@test normalize_string("\ufb00", :NFC) != "ff"
+@test normalize_string("\ufb00", :NFKC) == "ff"
+@test normalize_string("\u006e\u0303\ufb00", :NFKC) == "\u00f1"*"ff"
+@test normalize_string("\u00f1\ufb00", :NFKD) == "\u006e\u0303"*"ff"
+@test normalize_string("\u006e\u0303", compose=true) == "\u00f1"
+@test "\u006e\u0303" == normalize_string("\u00f1", decompose=true)
+@test normalize_string("\u006e\u0303\u00b5",compat=true) == "\u00f1\u03bc"
+@test normalize_string("Σσς",casefold=true) == "σσσ"
+@test normalize_string("∕⁄", lump=true) == "//"
+@test normalize_string("\ua\n\r\r\ua", newline2lf=true) == "\ua\ua\ua\ua"
+@test normalize_string("\ua\n\r\r\ua", newline2ls=true) == "\u2028\u2028\u2028\u2028"
+@test normalize_string("\ua\n\r\r\ua", newline2ps=true) == "\u2029\u2029\u2029\u2029"
+@test normalize_string("\u00f1", stripmark=true) == "n"
+@test isempty(normalize_string("\u00ad", stripignore=true))
+@test normalize_string("\t\r", stripcc=true) == "  "
+@test normalize_string("\t\r", stripcc=true, newline2ls=true) == " \u2028"
+
+#Tests from Unicode SA#15, "Unicode normalization forms"
+#http://www.unicode.org/reports/tr15/
+
+#1. Canonical equivalence
+let ==(a::Array{Char},b::Array{Char}) = normalize_string(string(a...), :NFC)==normalize_string(string(b...), :NFC)
+    ==(a,b) = Base.(:(==))(a,b)
+    @test ['C', '̧'] == ['Ç']
+    @test ['q', '̇', '̣'] == ['q', '̣', '̇']
+    @test ['가'] == ['ᄀ', 'ᅡ']
+    @test ['Ω'] == ['Ω']
+end
+
+#2. Compatibility Equivalence
+let ==(a::Array{Char},b::Array{Char}) = normalize_string(string(a...), :NFKC)==normalize_string(string(b...), :NFKC)
+    ==(a,b) = Base.(:(==))(a,b)
+    @test ['ℌ'] == ['ℍ'] == ['H']
+    @test ['ﻨ'] == ['ﻧ'] == ['ﻦ'] == ['ﻥ']
+    @test ['①'] == ['1']
+    @test ['ｶ'] == ['カ']
+    @test ['︷'] == ['{']
+    @test ['⁹'] == ['₉']
+    @test ['㌀'] == ['ア', 'パ', 'ー', 'ト']
+    @test ['¼'] == ['1', '⁄', '4']
+    @test ['ǆ'] == ['d', 'ž']
+end
+
+#3. Singletons
+@test normalize_string("\U212b", :NFD) == "A\U030a"
+@test normalize_string("\U212b", :NFC) == "\U00c5"
+@test normalize_string("\U2126", :NFC) == normalize_string("\U2126", :NFD) == "\U03a9"
+
+#4. Canonical Composites
+@test normalize_string("\U00c5", :NFC) == "\U00c5"
+@test normalize_string("\U00c5", :NFD) == "A\U030a"
+@test normalize_string("\U00f4", :NFC) == "\U00f4"
+@test normalize_string("\U00f4", :NFD) == "o\U0302"
+
+#5. Multiple Combining Marks
+@test normalize_string("\U1e69", :NFD) == "s\U0323\U0307"
+@test normalize_string("\U1e69", :NFC) == "\U1e69"
+@test normalize_string("\U1e0b\U0323", :NFD) == "d\U0323\U0307"
+@test normalize_string("\U1e0b\U0323", :NFC) == "\U1e0d\U0307"
+@test normalize_string("q\U0307\U0323", :NFC) == "q\U0323\U0307"
+@test normalize_string("q\U0307\U0323", :NFD) == "q\U0323\U0307"
+
+#6. Compatibility Composites
+@test normalize_string("\Ufb01", :NFD) == normalize_string("\Ufb01", :NFC) == "\Ufb01"
+@test normalize_string("\Ufb01", :NFKD) == normalize_string("\Ufb01", :NFKC) == "fi"
+@test normalize_string("2\U2075", :NFD) == normalize_string("2\U2075", :NFC) == "2\U2075"
+@test normalize_string("2\U2075", :NFKD) == normalize_string("2\U2075", :NFKC) == "25"
+@test normalize_string("\U1e9b\U0323", :NFD) == "\U017f\U0323\U0307"
+@test normalize_string("\U1e9b\U0323", :NFC) == "\U1e9b\U0323"
+@test normalize_string("\U1e9b\U0323", :NFKD) == "s\U0323\U0307"
+@test normalize_string("\U1e9b\U0323", :NFKC) == "\U1e69"
+
+# issue #5870
+@test !ismatch(Regex("aa"), SubString("",1,0))
+@test ismatch(Regex(""), SubString("",1,0))
+
+# issue #6027
+let
+    # make symbol with invalid char
+    sym = symbol(char(0xdcdb))
+    @test string(sym) == "\udcdb"
+    @test expand(sym) === sym
+    @test parse("\udcdb = 1",1,raise=false)[1] == Expr(:error, "invalid character \"\udcdb\"")
+end
+
+# issue #6949
+let f =IOBuffer(),
+    x = split("1 2 3")
+    @test write(f, x) == 3
+    @test takebuf_string(f) == "123"
+    @test invoke(write, (IO, AbstractArray), f, x) == 3
+    @test takebuf_string(f) == "123"
+end
+
+# issue #7248
+@test_throws BoundsError ind2chr("hello", -1)
+@test_throws BoundsError chr2ind("hello", -1)
+@test_throws BoundsError ind2chr("hellø", -1)
+@test_throws BoundsError chr2ind("hellø", -1)
+@test_throws BoundsError ind2chr("hello", 10)
+@test_throws BoundsError chr2ind("hello", 10)
+@test_throws BoundsError ind2chr("hellø", 10)
+@test_throws BoundsError chr2ind("hellø", 10)
+@test_throws BoundsError checkbounds("hello", 0)
+@test_throws BoundsError checkbounds("hello", 6)
+@test_throws BoundsError checkbounds("hello", 0:3)
+@test_throws BoundsError checkbounds("hello", 4:6)
+@test_throws BoundsError checkbounds("hello", [0:3])
+@test_throws BoundsError checkbounds("hello", [4:6])
+@test checkbounds("hello", 2)
+@test checkbounds("hello", 1:5)
+@test checkbounds("hello", [1:5])
+
+
+# isvalid(), chr2ind() and ind2chr() for SubString{DirectIndexString}
+let s="lorem ipsum",
+    sdict=[SubString(s,1,11)=>s, 
+        SubString(s,1,6)=>"lorem ",
+        SubString(s,1,0)=>"", 
+        SubString(s,2,4)=>"ore", 
+        SubString(s,2,16)=>"orem ipsum", 
+        SubString(s,12,14)=>""
+    ]
+    for (ss,s) in sdict
+        for i in -1:12
+            @test isvalid(ss,i)==isvalid(s,i)
+        end
+    end
+    for (ss,s) in sdict
+        for i in 1:length(ss)
+            @test ind2chr(ss,i)==ind2chr(s,i)
+        end
+    end
+    for (ss,s) in sdict
+        for i in 1:length(ss)
+            @test chr2ind(ss,i)==chr2ind(s,i)
+        end
+    end
+end #let
+
+#for isvalid(SubString{UTF8String})
+let s = utf8("Σx + βz - 2")
+  for i in -1:length(s)+2
+      ss=SubString(s,1,i)
+      @test isvalid(ss,i)==isvalid(s,i)
+  end
+end
+
+ss=SubString("hello",1,5)
+@test_throws BoundsError ind2chr(ss, -1)
+@test_throws BoundsError chr2ind(ss, -1)
+@test_throws BoundsError chr2ind(ss, 10)
+@test_throws BoundsError ind2chr(ss, 10)
+
+# length(SubString{UTF8String}) performance specialization
+let s = "|η(α)-ϕ(κ)| < ε"
+    @test length(SubString(s,1,0))==length(s[1:0])
+    @test length(SubString(s,4,4))==length(s[4:4])
+    @test length(SubString(s,1,7))==length(s[1:7])
+    @test length(SubString(s,4,11))==length(s[4:11])
+end
+
+# issue #7764
+let
+    srep = RepString("Σβ",2)
+    s="Σβ"
+    ss=SubString(s,1,endof(s))
+
+    @test ss^2 == "ΣβΣβ"
+    @test RepString(ss,2) == "ΣβΣβ"
+
+    @test endof(srep) == 7
+
+    @test next(srep, 3) == ('β',5)
+    @test next(srep, 7) == ('β',9)
+
+    @test srep[7] == 'β'
+    @test_throws BoundsError srep[8]
+end
